# ----------------------------------------------------------------------------
# - scmn-utils                                                               -
# - silkan python library: universal utility functions                       -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       scmn_utils
   :platform:     Unix, Windows
   :synopsis:     Silkan Python Library - universal utility functions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from collections import Iterable
import re

def nmap(f, x):
    """
    :type f: (T) -> U
    :type x: T | None
    :rtype: U | None
    """
    return f(x) if x is not None else None


def smap(f, s, d=None):
    """
    :type f: (T) -> U
    :type s: T | None
    :type d: U | None
    :rtype: U | None
    """
    if s is None:
        return None
    elif s:
        return f(s)
    else:
        return d


def nth(n, itms, default=None):
    """
    Return the nth element of the given iterable/iterator (starting at 0).

    :param n: Number of element to return
    :type n: int
    :param itms: Elements
    :type itms: Iterable[T]
    :param default: Default value if nth element cannot be found
    :type default: T | None
    :return: Nth object of the given iterable/iterator, or defalut-value
    :rtype: T | None
    """
    if isinstance(itms, list) or isinstance(itms, tuple):
        if 0 <= n < len(itms):
            return itms[n]
    elif isinstance(itms, Iterable):
        itms = iter(itms)

    if hasattr(itms, '__next__'):
        if n:
            for x in range(n):
                if not next(itms, None):
                    return default
        return next(itms, default)

    return default


def first(itms, default=None):
    """
    Return the first element of the given iterable/iterator.

    :param itms: Elements
    :type itms: Iterable[T]
    :param default: Default value if first element cannot be found
    :type default: T | None
    :return: First object of the given iterable/iterator, or default-value
    :rtype: T | None
    """
    return nth(0, itms, default)


def last(itms, default=None):
    """
    Return the last element of the given iterable/iterator.

    :param itms: Elements
    :type itms: Iterable[T]
    :param default: Default value if last element cannot be found
    :type default: T | None
    :return: Last object of the given iterable/iterator, or default-value
    :rtype: T | None
    """
    if isinstance(itms, list) or isinstance(itms, tuple):
        return itms[-1]
    elif isinstance(itms, Iterable):
        itms = iter(itms)

    if hasattr(itms, '__next__'):
        elem = default
        for elem in itms:
            pass
        return elem

    return default


def concatWithoutDuplicates(*seqs):
    """
    Combine the given sequences into one while preserving element order and
    eliminating all duplicate items.

    :param seqs: Sequences to concatenate
    :type seqs: list[Iterable[T]]
    :return: Concatenation of given iterables, without duplicates
    :rtype: list[T]
    """
    rset = set()
    rlst = []
    for seq in seqs:
        for item in seq:
            if item not in rset:
                rset.add(item)
                rlst.append(item)
    return rlst


def partition(pred, elms):
    """
    Partition a sequence based on a predicate.
    Two partitions will be returned , one for elements that fulfill the
    predicate and one for those that don't

    :param pred: Predicate
    :type pred: (T) -> bool
    :param elms: Elements to partition
    :type elms: Iterable[T]
    :return: The two partitions, one for elements that fulfill the predicate
             and one for those that don't
    :rtype: (list[T], list[T])
    """
    tLst = []
    fLst = []
    for elem in elms:
        (tLst if pred(elem) else fLst).append(elem)
    return tLst, fLst


def isHexa(data):
    """
    Check a string to its integer value string is assumed to be decimal unless
    it starts with the usual '0x' string
    :param data:
    :type dara: str
    :return:
    """
    if not data:
        return False
    if data[:2] == "0x":
        try:
            int(data, 16)
            return True
        except ValueError:
            return False
    return False


def isDec(data):
    """
    convert a string to its integer value string is assumed to be decimal
    :param data:
    :type dara: str
    :return:
    """
    try:
        int(data)
        return True
    except ValueError:
        return False


def isIpAdress(data, returnMessages=False):
    """
    check if data has the correct format to be an Ip adress
    :param data: variable to be checked
    :type data: Str
    :param returnMessages: whether to return or not messages for UI
    :type returnMessages: Bool
    :return:
    """
    isOk = False
    title = None
    message = None
    # if not data:
    #     return "0x00000000"
    # if hexa ok
    if isHexa(data):
        if len(data) == 10:
            isOk = True
        else:
            title = 'Ip address Error'
            message = 'Ip address hexadecimal need to have a size of 8 hexa'
    else:
        # split by . and check if values are decimal int
        adress = data.split(".")
        if len(adress) == 4:
            isOk = True
            # check if values between each '.' char is int
            for ad in adress:
                if not isDec(ad):
                    title = 'Ip address Error'
                    message = 'Ip address can\'t have non-decimal value!'
                    isOk = False
        else:
            title = 'Ip address Error'
            message = 'Ip address mus be an hexadecimal or X.X.X.X !'
    if returnMessages:
        return isOk, title, message
    else:
        return isOk


def isMacAdress(data, returnMessages=False):
    """
    check if data has the correct format to be a mac adress
    :param data: variable to be checked
    :type data: Str
    :param returnMessages: whether to return or not messages for UI
    :type returnMessages: Bool
    :return:
    """
    isOk = False
    title = None
    message = None
    # if not data:
    #     return "0x000000000000"
    # if hexa ok
    if isHexa(data):
        if len(data) == 14:
            isOk = True
        else:
            title = 'Mac address Error'
            message = 'Mac address hexadecimal need to have a size of 12 hexa'
    else:
        # split by . and check if values are decimal int
        adress = data.split(":")
        if len(adress) == 6:
            isOk = True
            # check if values between each '.' char is int
            for ad in adress:
                if re.match("^[0-9A-Fa-f]{2}", ad) is None:
                    title = 'Mac address Error'
                    message = 'Mac address need to be composed of 2 hexadecimal values separated by a dot!'
                    isOk = False
        else:
            title = 'Mac address Error'
            message = 'Mac address can be an hexadecimal or XX:XX:XX:XX:XX:XX'
    if returnMessages:
        return isOk, title, message
    else:
        return isOk


def unique(sequence):
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]