# ----------------------------------------------------------------------------
# - wact-action                                                              -
# - wild cruncher: action base classes                         core / action -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wact_action
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Action base classes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import traceback

from silkan_common.scmn_log import SCMN_LOG_ERROR
from wcr_core.action.wact_action_api import *
from wcr_core.action.wact_action_parameter_api import WactActionParameterP


#: Action exit codes
WACT_ACTION_EXIT_OK = None
WACT_ACTION_EXIT_CANCELLED = 1
WACT_ACTION_EXIT_DEFERRED = 2


# noinspection PyAbstractClass
class WactActionBaseP(WactActionP):
    @abstractmethod
    def _doExecute(self, actx, paramValuePairs):
        """
        Execute the action.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param paramValuePairs: Param-value-pairs
        :type paramValuePairs: dict[str, object]
        :return: Action exit code
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def _notifyObservers(self, actx, event, *args, **kwargs):
        """
        Notify all registered action observers about a change in the action's
        state.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param event: Event
        :type event: int
        """
        raise NotImplementedError()

    @abstractmethod
    def _isProtocolException(self, e):
        """
        Whether the given exception is part of the action's protocol, i.e.
        used to communicate an expected error state of the action.

        :param e: Exception
        :type e: Exception
        :return: True iff the given exception is part of the action's protocol
        :rtype: bool
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WactActionBase(WactActionBaseP):
    """
    Abstract base class for actions.
    """

    def __init__(self, name, description, *parameters):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameters: Action parameters
        :type parameters: list[WactActionParameterP]
        """
        #: Action name
        self.__name = name
        #: Action description
        self.__description = description
        #: Action parameters
        self.__parameters = parameters
        """:type: list[WactActionParameterP]"""

        #: Registered action listeners
        self.__listeners = []
        """:type: list[WactActionObserverP]"""

    def _isProtocolException(self, e):
        return False

    def getName(self):
        return self.__name

    def getDescription(self):
        return self.__description

    def getParameters(self):
        return iter(self.__parameters)

    def isEnabled(self, actx):
        return True

    def _getParamValue(self, param, actx, pvs):
        """
        :type param: WactActionParameterP
        :type actx: WactActionExecutionContextP
        :type pvs: dict[str, object]
        :rtype: object | None
        """
        if param.getCode() in pvs:
            return pvs[param.getCode()]
        return actx.getValue(param)

    def execute(self, actx, **paramValuePairs):
        """
        :type actx: WactActionExecutionContextP
        """
        if not self.isEnabled(actx):
            return

        self._notifyObservers(
            actx, WACT_ACTION_EXECUTING_EVENT, **paramValuePairs)
        try:
            exitCode = self._doExecute(actx, paramValuePairs)

            if exitCode == WACT_ACTION_EXIT_OK:
                self._notifyObservers(
                    actx, WACT_ACTION_EXECUTION_SUCCEEDED_EVENT,
                    **paramValuePairs)
            elif exitCode == WACT_ACTION_EXIT_CANCELLED:
                self._notifyObservers(
                    actx, WACT_ACTION_EXECUTION_CANCELLED_EVENT,
                    **paramValuePairs)
        except BaseException as e:
            if not self._isProtocolException(e):
                traceback.print_exc()
                trce = traceback.format_exc()
                actx.addMessage(
                    'Failed to execute action {}: {}'.format(
                        self.getName(), e),
                    mcmp='cockpit', mcat=SCMN_LOG_ERROR, trce=trce)
            self._notifyObservers(
                actx, WACT_ACTION_EXECUTION_FAILED_EVENT, e, **paramValuePairs)

    def registerActionObserver(self, observer):
        self.__listeners.append(observer)

    def unregisterActionObserver(self, observer):
        self.__listeners.remove(observer)

    def _notifyObservers(self, actx, event, *args, **kwargs):
        for listener in self.__listeners:
            listener.actionNotify(actx, self, event, *args, **kwargs)


# noinspection PyAbstractClass
class WactActionCommandMixin(WactActionBaseP):
    """
    Mixin forr (possibly deferred command execution)
    """

    def _success(self, actx):
        """
        Handler for successful execution.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        """
        self._notifyObservers(actx, WACT_ACTION_EXECUTION_SUCCEEDED_EVENT)

    def _error(self, actx, e):
        """
        Handler for execution failures.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param e: Exception
        :type e: Exception
        :return: True iff the error has been handled
        :rtype: bool
        """
        self._notifyObservers(actx, WACT_ACTION_EXECUTION_FAILED_EVENT, e)

        return self._isProtocolException(e)

    def _runCommand(self, actx, code, title, runF):
        """
        Execute the given command.

        :param actx: Action execution context
        :type actx: WactActionExecutionContextP
        :param code: Command code
        :type code: str
        :param title: Command title
        :type title: str
        :param runF: Command function
        :type runF: (ScmnProcStreamHandlerP) -> _
        """
        def success():
            self._success(actx)

        def error(excp):
            return self._error(actx, excp)

        actx.executeCommand(code, title, runF, successF=success, errorF=error)

        return WACT_ACTION_EXIT_DEFERRED


# noinspection PyAbstractClass
class WactStateSelectionActionBase(WactActionBase, WactStateSelectionActionP):
    """
    Protocol for actions for making a choice of a given set of options.
    """

    def __init__(self, name, description, parameter):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param parameter: Selection parameter
        :type parameter: WactFixedSelectionActionParameterP
        """
        super().__init__(name, description, parameter)

        self.__parameter = parameter
        """:type: WactFixedSelectionActionParameterP"""

    def getPossibleValues(self):
        return self.__parameter.getPossibleValues()


# noinspection PyAbstractClass
class WactStateToggleActionBase(WactStateSelectionActionBase,
                                WactStateToggleActionP):
    """
    Protocol for actions to toggle a boolean value.
    """

    pass


class WactSimpleAction(WactActionBase):
    """
    Simple action implementation.
    """

    def __init__(self, name, description, actionF, *conditions):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param actionF: Function to be executed
        :type actionF: () -> _
        :param conditions: Conditions under which the action can be executed
        :type conditions: list[() -> _]
        """
        super().__init__(name, description)

        self.__actionF = actionF
        self.__conditions = conditions

    def isEnabled(self, actx):
        return all(c() for c in self.__conditions)

    def _doExecute(self, actx, paramValuePairs):
        self.__actionF()


class WactSimpleContextAction(WactActionBase):
    """
    Simple action implementation for context-aware functions.
    """

    def __init__(self, name, description, actionF, *conditions):
        """
        :param name: Action name
        :type name: str
        :param description: Action description
        :type description: str
        :param actionF: Function to be executed
        :type actionF: (WactActionExecutionContextP) -> _
        :param conditions: Conditions under which the action can be executed
        :type conditions: list[(WactActionExecutionContextP) -> _]
        """
        assert all(hasattr(c, '__call__') for c in conditions)

        super().__init__(name, description)

        self.__actionF = actionF
        self.__conditions = conditions

    def isEnabled(self, actx):
        return all(c(actx) for c in self.__conditions)

    def _doExecute(self, actx, paramValuePairs):
        self.__actionF(actx)
