# ----------------------------------------------------------------------------
# - wcmn-parameter-description                                               -
# - wild cruncher: parameter descriptions                             common -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcmn_parameter_description
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Parameter descriptions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod
from wcr_core.db_api.whdb_protocols_api import WhdbHasFlagsP


WCMN_PARAM_FLAGS_COUNT = 1

WCMN_PARAM_NO_FLAG, \
    WCMN_PARAM_FORMAT_STRING_FLAG = \
    ((2 ** i) >> 1 for i in range(WCMN_PARAM_FLAGS_COUNT + 1))


# noinspection PyAbstractClass
class WcmnParameterDescriptionP(WhdbHasFlagsP):
    """
    Parameter description protocol.
    """

    @abstractmethod
    def getName(self):
        """
        Get the parameter name.

        :return: Name of the parameter
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getLabel(self):
        """
        Get the parameter label.

        :return: Label for the parameter
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getInfo(self):
        """
        Get the parameter description.

        :return: Description for the parameter
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getDefaultValue(self, **kwargs):
        """
        Get the parameter value.

        :return: Value of the parameter
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def hasDynamicDefaultValue(self):
        """
        Whether the default value is dynamic.

        :return: True iff the default value is dynamic
        :rtype: bool
        """
        raise NotImplementedError()


class WcmnParameterDescription(WcmnParameterDescriptionP):
    """
    Parameter description.
    """

    def __init__(self, name, label, info, dval=None, flgs=None, **kwargs):
        """
        :param name: Parameter name
        :type name: str
        :param label: Parameter lable
        :type label: str
        :param info: Parameter info
        :type info: str
        :param dval: Default value
        :type dval: str
        :param flgs: Flags
        :type flgs: int
        """
        self.__name = name
        self.__label = label
        self.__info = info
        self.__defaultValue = dval
        self.__flags = flgs or 0
        self.__dict = {}
        for k, v in kwargs.items():
            if hasattr(self, k):
                raise ValueError('Invalid meta-data key: ' + k)
            setattr(self, k, v)
            self.__dict[k] = v

    def getName(self):
        return self.__name

    def getLabel(self):
        return self.__label

    def getInfo(self):
        return self.__info

    def getDefaultValue(self, **kwargs):
        if self.hasFlag(WCMN_PARAM_FORMAT_STRING_FLAG):
            return self.__defaultValue.format(**kwargs)
        return self.__defaultValue

    def hasFlag(self, flag, *_):
        return flag & self.__flags

    def hasDynamicDefaultValue(self):
        return self.hasFlag(WCMN_PARAM_FORMAT_STRING_FLAG)

    def __str__(self):
        return 'WcmnParameterDescription ' \
               '{{name: {n}, default: {d}}}'.format(
                   n=self.getName(), d=self.__defaultValue)

    def getDictValue(self, k):
        return self.__dict[k]


class WcmnRowParameterDescription():
    """
    row Parameter description.
    """

    def __init__(self, row, prms):
        """
        :param kwargs:
        """
        self.__row = row
        self.__prms = prms

    def getRow(self):
        return self.__row

    def getParameters(self):
        return self.__prms