# ----------------------------------------------------------------------------
# - wctl-controller-api                                                      -
# - wild cruncher: controller api                          core / controller -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wctl_controller_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Controller API

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod


class WctlControllerP(metaclass=ABCMeta):
    """
    Controller protocol.

    Controllers are responsible for controlling and coordinating user inter-
    interactions with database entities, such as creating or deleting
    entities.
    """
    pass


class WctlItemControllerP(WctlControllerP):
    @abstractmethod
    def deleteItem(self, item, block):
        """
        Delete the given item from the block.

        :param item: Item to delete
        :type item: WhdbBaseP
        :param block: Block to delete item from
        :type block: WhdbBlockP
        :return: Items deleted and item potentially changed
        :rtype: (Iterable[WhdbBaseP], Iterable[WhdbBaseP])
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WctlInstanceControllerP(WctlItemControllerP):
    """
    Protocol for instance controllers.
    """

    @abstractmethod
    def canCreateConnectionPoint(self, instance, **kwargs):
        """
        Whether a new connection point can be created on the given instance.

        :param instance: Instance to check
        :type instance: WhdbInstanceP
        :param kwargs: Additional keyword-arguments for the new conn. point
        :type kwargs: dict[str, object]
        :return: True iff a connection point can be created
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnectionPoint(self, instance, valuePrv, **kwargs):
        """
        Create a new connection point on the given instance.

        Will raise an exception if the creation of a new connection point
        with the given `kwargs` is not possible.

        :param instance: Instance to create connection point on
        :type instance: WhdbInstanceP
        :param valuePrv: Controller value provider
        :type valuePrv: WctlInstanceControllerValueProviderP
        :param kwargs: Additional keyword-arguments for the new conn. point
        :type kwargs: dict[str, object]
        :return: Newly created connection point
        :rtype: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def canRemoveConnectionPoint(self, instance, xpoint):
        """
        Whether the given connection point can be removed from the given
        instance.

        :param instance: Instance to check
        :type instance: WhdbInstanceP
        :param xpoint: Connection point to check
        :type xpoint: WhdbConnectionPointP
        :return: True iff the given connection point can be removed
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def removeConnectionPoint(self, instance, xpoint, block):
        """
        Remove given connection point  from the given instance.

        Will raise an exception if the given connection point cannot be
        removed.

        :param instance: Instance to remove connection point from
        :type instance: WhdbInstanceP
        :param xpoint: Connection point to be removed
        :type xpoint: WhdbConnectionPointP
        :param block: Block the instance is in
        :type block: WhdbBlockP
        :return: Items deleted and item potentially changed
        :rtype: (Iterable[WhdbBaseP], Iterable[WhdbBaseP])
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WctlConnectionControllerP(WctlItemControllerP):
    """
    Protocol for connection controllers.
    """

    @abstractmethod
    def canStartConnectionAt(self, block, instance, xpoint=None):
        """
        Whether a connection can be started from the given instance.

        If no connection point is given, it may be assumed that a new
        connection point is to be created. If this is not possible, the
        function will return `False`.

        :param block: Block the instance and potential connection belong to
        :type block: WhdbBlockP
        :param instance: Instance to check
        :type instance: WhdbInstanceP
        :param xpoint: Optional connection point the connection starts from
        :type xpoint: WhdbConnectionPointP | None
        :return: True iff a new connection can start from the given instance
                 (and connection point, if given)
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def canCreateConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                            **kwargs):
        """
        Whether a new connection can be created between the two given end
        points in the given block.

        Omitted connection points are assumed to be created.

        :param block: Block the new connection is to be created in
        :type block: WhdbBlockP
        :param sinst: Start instance
        :type sinst: WhdbInstanceP
        :param einst: End instance
        :type einst: WhdbInstanceP
        :param sxpnt: Optional start connection point
        :type sxpnt: WhdbConnectionPointP | None
        :param expnt: Optional end connection point
        :type expnt: WhdbConnectionPointP | None
        :param kwargs: Additional keyword arguments for potential connection
        :type kwargs: dict[str, object]
        :return: True iff new connection can be created
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                         **kwargs):
        """
        Create a new connection between the given end-points.

        Omitted connection points are assumed to be created.

        Will raise an exception if new connection cannot be created.

        :param block: Block the new connection is to be created in
        :type block: WhdbBlockP
        :param sinst: Start instance
        :type sinst: WhdbInstanceP
        :param einst: End instance
        :type einst: WhdbInstanceP
        :param sxpnt: Optional start connection point
        :type sxpnt: WhdbConnectionPointP | None
        :param expnt: Optional end connection point
        :type expnt: WhdbConnectionPointP | None
        :param kwargs: Additional keyword arguments for the new connection
        :type kwargs: dict[str, object]
        :return: Newly created connection, and updated items
        :rtype: (WhdbConnectionP, Iterable[WhdbBaseP])
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WctlShapeControllerP(WctlItemControllerP):
    """
    Protocol for shape controllers.
    """

    @abstractmethod
    def createShape(self, shapeType, block, **kwargs):
        """
        Create a new shape of the given type.

        :param shapeType: Shape type to create
        :type shapeType: str
        :param block: Block to create shape in
        :type block: WhdbBlockP
        :param kwargs: Additional keyword-arguments for the new shape
        :type kwargs: dict[str, object]
        :return: Newly created shape
        :rtype: WhdbShapeP
        """
        raise NotImplementedError()


class WctlControllerFactoryP(metaclass=ABCMeta):
    """
    Protocol for controller factories.
    """

    @abstractmethod
    def getItemController(self, item):
        """
        Get the controller for the given item.

        :param item: Item to retrieve/create controller for
        :type item: WhdbBaseP
        :return: Controller for the given item
        :rtype: WctlItemControllerP
        """
        raise NotImplementedError()

    @abstractmethod
    def getInstanceController(self):
        """
        Get the general instance controller.

        :return: Instance controller
        :rtype: WctlInstanceControllerP
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnectionController(self):
        """
        Get the general connection controller.

        :return: Conection controller
        :rtype: WctlConnectionControllerP
        """
        raise NotImplementedError()

    @abstractmethod
    def getShapeController(self):
        """
        Get the general shape controller.

        :return: Shape controller
        :rtype: WctlShapeControllerP
        """
        raise NotImplementedError()
