# ----------------------------------------------------------------------------
# - whdb-connection-path-builder-manhattan                                   -
# - wild cruncher: manhattan path builder                          core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_builder_manhattan
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher manhattan path builder with simple object
                  avoidance

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.db_api.whdb_connection_point_api import WhdbConnectionPointP
from .whdb_connection_path_builder import WhdbConnectionPathBuilder, \
    WHDB_DEFAULT_MIN_LINE_LENGTH

_C = 0
_N = 1
_S = 2
_W = 4
_E = 8
_X = _W | _E
_Y = _N | _S
_M = _N | _S | _W | _E


def _dirs2name(dirs):
    s = ''
    if dirs & _N: s += 'N'
    if dirs & _E: s += 'E'
    if dirs & _W: s += 'W'
    if dirs & _S: s += 'S'
    return s


# noinspection PyAbstractClass
class WhdbConnectionManhattanPathBuilder(WhdbConnectionPathBuilder):
    """
    Wild Cruncher manhattan path builder.
    Attempts to build path around connections's start- and end-object if
    necessary (and if possible).
    """

    # While building the path, two partial paths originating from either
    # connection point are generated. They are represented as lists of points,
    # namely triples (float, float, int) with their values corresponding to the
    # x- and y-coordinate and the direction of the segment leading to the
    # point.
    #
    # Directions are encoded bitwise as a single int,
    # using the bit masks `_N`, `_E`, `_S` and `_W`.
    # Bounding boxes are encoded as 4-tuples of floats (x1, x2, y1, y2) where
    # x1 < x2 and y1 < y2.

    def __init__(self, minLineLength=None):
        """
        :param minLineLength: Minimum line segment length
        :type minLineLength: float
        """
        self.__minLineLength = minLineLength or WHDB_DEFAULT_MIN_LINE_LENGTH

    def createPath(self, sobj, scon, eobj, econ):
        spl = self._getInitialPointList(sobj, scon)
        epl = self._getInitialPointList(eobj, econ)

        obbx = self._getOuterBoundingBox(sobj, eobj)
        sbbx = self._getBoundingBox(sobj)
        ebbx = self._getBoundingBox(eobj)

        # If either connection point lies within the other object's bounding
        # box, or if the two bounding boxes intersect, short-circuit by
        # connecting the end-points directly.
        if sobj is not eobj and \
                (self._isInBoundingBox(spl[-1], ebbx) or
                     self._isInBoundingBox(epl[-1], sbbx) or
                     self._areBoundingBoxesIntersection(sbbx, ebbx)):
            return self._connect(spl, epl)

        # Start building the path at the end that lies on its object's
        # bounding box (just in case there is exactly one, otherwise the order
        # is irrelevant)
        if self._isOnBoundingBox(spl[-1], obbx):
            return self._buildPath(spl, epl, sbbx, ebbx, obbx)
        else:
            return list(reversed(self._buildPath(epl, spl, ebbx, sbbx, obbx)))

    def _getInitialPointList(self, obj, conn):
        """
        For a given object/connection point pair, generate a list containing
        the first two points of the partial path that originates from the
        connection point.

        :param obj: The object
        :type obj: WhdbInstanceP
        :param conn: The connection point
        :type conn: WhdbConnectionPointP
        :return: List with initial points and directions
        :rtype: list[(float, float, int)]
        """
        cx, cy = obj.getEffectiveConnectionPointPosition(conn)
        p = None
        gvty = obj.getEffectiveConnectiontPointGravity(conn)
        if gvty == WhdbConnectionPointP.North:
            p = cx, cy - self.__minLineLength, _N
        elif gvty == WhdbConnectionPointP.South:
            p = cx, cy + self.__minLineLength, _S
        elif gvty == WhdbConnectionPointP.West:
            p = cx - self.__minLineLength, cy, _W
        elif gvty == WhdbConnectionPointP.East:
            p = cx + self.__minLineLength, cy, _E

        return [(cx, cy, _C), p] if p is not None else [(cx, cy, _C)]

    def _buildPath(self, spls, epls, sbbx, ebbx, obbx):
        """
        Build up the path between two connection point, assuming that the first
        two points of each path half are already provided in the respective
        point lists.
        The points will be added to the two respective point lists.

        :param spls: Point list for the partial path originating from the start
                     object's connection point
        :type spls: list[(float, float, end)]
        :param epls: Point list for the partial path originating from the end
                     object's connection point
        :type epls: list[(float, float, end)]
        :param sbbx: Bounding box of the end object
        :type sbbx: (float, float, float, float)
        :param ebbx: Bounding box of the start object
        :type obbx: (float, float, float, float)
        :param sbbx: Combined/united bounding box of the two objects
        :type obbx: (float, float, float, float)
        """
        if not self._addFirstPathPoint(spls, epls, sbbx, ebbx):
            # First move along the start objects bounding box could not be
            # determined, path might be all along combined bounding box
            sobx = self._isOnBoundingBox(spls[-1], obbx)
            eobx = self._isOnBoundingBox(epls[-1], obbx)
            if sobx and eobx:
                self._addFirstPathPoint(spls, epls, obbx, obbx)
                self._addPathPointsAlongBoundingBox(spls, epls, obbx, True)
                return self._connect(spls, epls)
        # Follow the path part along the start object's bounding box
        self._addPathPointsAlongBoundingBox(spls, epls, sbbx)
        # Follow the path part along the end object's bounding box
        self._addPathPointsAlongBoundingBox(epls, spls, ebbx)
        # Connect the two halfs
        return self._connect(spls, epls)

    def _addFirstPathPoint(self, plst, opls, bbox, obbx):
        """
        Add the first non-xpnt-connected path-segment to a point list
        (i.e. the third point) if possible.

        This first segment is created differently since its direction can often
        be chosen arbitrarily, yet determines the overall layout and length of
        the path.
        The function will attempt to keep the overall path length minimal
        in its choice of the initial direction.

        The new segment (i.e. third point), if any, will be appended to the
        point list.

        :param plst: Point list
        :type plst: list[(float, float, int)]
        :param opls: Other object's point list
        :type opls: list[(float, float, int)]
        :param bbox: Bounding box of the object
        :type bbox: (float, float, float, float)
        :param obbx: Bounding box of the other object
        :type obbx: (float, float, float, float)
        :return: True iff a point was added to the point list
        :rtype: bool
        """
        rdrs = self._getRequiredDirections(plst, opls)
        pdrs, adrs = self._getPossibleDirections(rdrs, plst, bbox)
        x, y, _ = plst[-1]
        ox, oy, _ = opls[-1]
        x1, x2, y1, y2 = bbox
        ox1, ox2, oy1, oy2 = obbx
        p = None
        if adrs & _X == _X and not pdrs & _Y:
            wl = abs(x - x1) + abs(ox - x1)
            el = abs(x2 - x) + abs(ox - x2)
            if wl < el and (bbox is obbx or x1 > ox2):
                p = x1, y, _W
            elif bbox is obbx or x2 < ox1:
                p = x2, y, _E
        elif adrs & _Y == _Y and not pdrs & _X:
            nl = abs(y - y1) + abs(oy - y1)
            sl = abs(y2 - y) + abs(oy - y2)
            if nl < sl and (bbox is obbx or y1 > oy2):
                p = x, y1, _N
            elif bbox is obbx or y2 < oy1:
                p = x, y2, _S
        if p is not None:
            plst.append(p)
        return p is not None

    @staticmethod
    def _getRequiredDirections(opls, tpls):
        """
        Provides the main directions required to get from the last point in
        the origin's point list `opl` to the last point in the target point
        list `tpl`.

        :param opls: Origin's point list
        :type opls: list[(float, float, int)]
        :param tpls: Target's point list
        :type tpls: list[(float, float, int)]
        :return: Set of required directions, bit-encoded
        :rtype: int
        """
        sx, sy, _ = opls[-1]
        ex, ey, _ = tpls[-1]
        dirs = _C
        if sx < ex:
            dirs |= _E
        elif sx > ex:
            dirs |= _W
        if sy < ey:
            dirs |= _S
        elif sy > ey:
            dirs |= _N
        return dirs

    @staticmethod
    def _getPossibleDirections(rdrs, plst, bbox):
        """
        Provides a list of possible directions that a path could continue in
        around the object's bounding box, from the last point in the given
        point list.
        This function is only applicable while the path is following the
        bounding box.

        :param plst: Point list
        :type plst: list[(float, float, int)]
        :param bbox: Bounding box of the path's starting object
        :type bbox: (float, float, float, float)
        :return: Set of possible directions, bit-encoded
        :rtype: int, int
        """
        x1, x2, y1, y2 = bbox
        x, y, d = plst[-1]
        adrs = _C
        if x <= x1: adrs |= _N | _W | _S
        if x >= x2: adrs |= _N | _E | _S
        if y <= y1: adrs |= _N | _W | _E
        if y >= y2: adrs |= _S | _W | _E

        pdrs = rdrs & adrs
        if pdrs == _C:
            if rdrs == _N or rdrs == _S:
                pdrs = _E | _W
            elif rdrs == _E or rdrs == _W:
                pdrs = _N | _S
        return pdrs, adrs

    def _addPathPointsAlongBoundingBox(self, plst, tpls, bbox, sdir=False):
        """
        Build up a part of a path by following around an object's bounding box
        as long as possible.
        The path's points will be added to the given point list.

        :param plst: Point list of the path originating from the object's
                     connection point
        :type plst: list[(float, float, int)]
        :param tpls: Point list of the other part of the path, originating from
                     the connection point at the other end, the last point of
                     that list serves as target of the path being built up
        :type tpls: list[(float, float, int)]
        :param bbox: Bounding box of the object
        :type bbox: (float, float, float, float)
        :param sdir: Whether to stop as soon as only a single direction is left
                     to go towards the target point (usually `True` when only
                     following the outer, combined bounding box)
        :type sdir: bool
        """

        def singleDir(dirs):
            return dirs == _N or dirs == _S or dirs == _W or dirs == _E

        while True:
            rdrs = self._getRequiredDirections(plst, tpls)
            if sdir and singleDir(rdrs): break
            if not self._addNextPathPoint(plst, rdrs, bbox): break

    def _addNextPathPoint(self, plst, rdrs, bbox):
        """
        Appends the next point of the path to the given path list, as long as
        the new segment lies on the gien bounding box; otherwise, does nothing.

        :param plst: Point list of the path so far
        :type plst: list[(float, float, int)]
        :param rdrs: Required directions towards the target
        :type rdrs: int
        :param bbox: Bounding box
        :type bbox: (float, float, float, float)
        :return: True iff a point was added
        :rtype: bool
        """
        pdrs, _ = self._getPossibleDirections(rdrs, plst, bbox)
        x, y, d = plst[-1]
        if pdrs & d: return False
        x1, x2, y1, y2 = bbox
        if pdrs & _N and not d & _Y:
            # Go north
            plst.append((x, y1, _N))
        elif pdrs & _S and not d & _Y:
            # Go south
            plst.append((x, y2, _S))
        elif pdrs & _W and not d & _X:
            # Go west
            plst.append((x1, y, _W))
        elif pdrs & _E and not d & _X:
            # Go east
            plst.append((x2, y, _E))
        else:
            return False
        return True

    def _getBoundingBox(self, obj):
        """
        Return the bounding box of the given object.

        :param obj: The object
        :type obj: WhdbInstanceP
        :return: The bounding box of the object
        :rtype: (float, float, float, float)
        """
        minlen = self.__minLineLength
        w, h = obj.getEffectiveDimension()
        x, y = obj.getPosition()
        return x - w / 2 - minlen, x + w / 2 + minlen, \
               y - h / 2 - minlen, y + h / 2 + minlen

    def _getOuterBoundingBox(self, obj1, obj2):
        """
        Returns the union of the two bounding boxes of the given objects.

        :param obj1: First object
        :type obj1: WhdbInstanceP
        :param obj2: Second object
        :type obj2: WhdbInstanceP
        :return: Bounding box enclosing both objects
        :rtype: (float, float, float, float)
        """
        minlen = self.__minLineLength
        w1, h1 = obj1.getEffectiveDimension()
        w2, h2 = obj2.getEffectiveDimension()
        ox1, oy1 = obj1.getPosition()
        ox2, oy2 = obj2.getPosition()
        x1 = min(ox1 - w1 / 2,
                 ox2 - w2 / 2)
        x2 = max(ox1 + w1 / 2,
                 ox2 + w2 / 2)
        y1 = min(oy1 - h1 / 2,
                 oy2 - h2 / 2)
        y2 = max(oy1 + h1 / 2,
                 oy2 + h2 / 2)
        return x1 - minlen, x2 + minlen, y1 - minlen, y2 + minlen

    @staticmethod
    def _isOnBoundingBox(point, bbox):
        """
        :param point: First bounding box
        :type point: (float, float, int)
        :param bbox: Bounding box
        :type bbox: (float, float, float, float)
        :return: True iff the point lies on the bounding box
        :rtype: bool
        """
        x1, x2, y1, y2 = bbox
        x, y, _ = point
        return x == x1 or x == x2 or y == y1 or y == y2

    @staticmethod
    def _isInBoundingBox(point, bbox):
        """
        :param point: First bounding box
        :type point: (float, float, int)
        :param bbox: Bounding box
        :type bbox: (float, float, float, float)
        :return: True iff the point lies inside or on the bounding box
        :rtype: bool
        """
        x1, x2, y1, y2 = bbox
        x, y, _ = point
        return x1 <= x <= x2 and y1 <= y <= y2

    @staticmethod
    def _areBoundingBoxesIntersection(bbx1, bbx2):
        """
        :param bbx1: First bounding box
        :type bbx1: (float, float, float, float)
        :param bbx2: Second bounding box
        :type bbx2: (float, float, float, float)
        :return: True iff the two boxes intersect
        :rtype: bool
        """
        b1x1, b1x2, b1y1, b1y2 = bbx1
        b2x1, b2x2, b2y1, b2y2 = bbx2
        return not (b1x1 > b2x2 or b1x2 < b2x1 or b1y1 > b2y2 or b1y2 < b2y1)

    @staticmethod
    def _connect(spls, epls):
        """
        Connect the two parts of a path, one originating from the start object
        and the other from the end object.

        :param spls: First (start) point list
        :type spls: list[(float, float, int)]
        :param epls: Second (end) point list
        :type epls: list[(float, float, int)]
        :return: The combination of the two provided point lists
        :rtype: list[(float, float)]
        """
        sx, sy, sd = spls[-1]
        ex, ey, ed = epls[-1]

        p = None
        if sx != ex and sy != ey:
            if sd == ed:
                if sd == _N:
                    p = ex if ey > sy else sx, min(sy, ey), _N
                elif sd == _S:
                    p = ex if ey < sy else sx, max(sy, ey), _S
                elif sd == _W:
                    p = min(sx, ex), ey if ex > sx else sy, _W
                elif sd == _E:
                    p = max(sx, ex), ey if ex < sx else sy, _E
            elif sd & _X and ed & _X:
                x = (sx + ex) / 2
                spls.append((x, sy, sd))
                epls.append((x, ey, ed))
            elif sd & _Y and ed & _Y:
                y = (sy + ey) / 2
                spls.append((sx, y, sd))
                epls.append((ex, y, ed))
            elif sd & _X and ed & _Y:
                spls.append((ex, sy, sd))
            elif sd & _Y and ed & _X:
                spls.append((sx, ey, sd))

        if p is not None: spls.append(p)
        spls.extend(reversed(epls))
        return [(x, y) for x, y, _ in spls]

    def simplifyPath(self, path, remf, pntf=None):
        pdir = None
        ppnt = None
        rlst = []
        for p in path:
            x, y = pntf(p) if pntf else p
            if ppnt is not None:
                px, py = pntf(ppnt) if pntf else ppnt
                xdif = abs(x - px)
                ydif = abs(y - py)

                if xdif == 0 and ydif == 0:
                    # 2 consecutive points at the same location, remove one
                    rlst.append(ppnt)
                else:
                    if xdif != 0 and ydif == 0:
                        drct = 'X'
                    elif xdif == 0 and ydif != 0:
                        drct = 'Y'
                    else:
                        drct = ''
                    if drct != '' and drct == pdir:
                        # Line segment with inner point, remove it
                        rlst.append(ppnt)
                    pdir = drct
            ppnt = p
        for pnt in rlst:
            remf(pnt)
