# ----------------------------------------------------------------------------
# - whdb-connection-path-circular                                            -
# - wild cruncher: circular connection path                        core / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_circular
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher circular connection path

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
import math

from silkan_common.scmn_trigonometry import scmn_getSlope
from wcr_core.db_api.whdb_connection_path_api import WhdbConnectionPathP, \
    WhdbConnectionPathSegmentP


_WHDB_CIRC_MIN_RADIUS = 5
WHDB_CIRC_DEFAULT_ANGLE = math.pi / 2


def _len(a, b):
    """
    :type a: float
    :type b: float
    :rtype: float
    """
    return math.sqrt(pow(a, 2.0) + pow(b, 2.0))


class WhdbCircularConnectionPath(WhdbConnectionPathP,
                                 WhdbConnectionPathSegmentP):
    """
    Circular connection path.
    """

    def __init__(self, cnctn):
        """
        :param cnctn: Connection
        :type cnctn: WhdbConnectionP
        """
        self.__connection = cnctn

        self.__radius = 0
        self.__startAngle = 0
        self.__sweepLength = 0
        self.__center = None

        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        path = relm.getPath()
        pnts = path.getPoints() if path else None
        if not pnts or len(pnts) != 2 or \
                all(x == 0 and y == 0 for x, y in pnts):
            self.resetPath()
        else:
            self._update()

    def _getPath(self):
        """
        Get the raw path.

        :return: Raw path
        :rtype: WdbsPathP
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        path = relm.getPath()
        if not path:
            self.resetPath()
            path = relm.getPath()
        return path

    def _setRawData(self, xdif, ydif, radius):
        """
        Set the raw data describing the circle arc.

        :param xdif: X-component of slope of line between shape center and
                     circle center
        :type xdif: float
        :param xdif: Y-component of slope of line between shape center and
                     circle center
        :type ydif: float
        :param radius: Radius of the circle
        :type radius: float
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        pnts = relm.getPath().getPoints()
        pnts[0].setXY(xdif, ydif)
        pnts[1].setX(radius)

    def _getRawData(self):
        """
        Get the raw data describing the circle arc.

        :return: Slope X-/Y-component, radius, shape-center distance
        :rtype: (float, float, float)
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        path = relm.getPath()
        pnts = path.getPoints() if path else None
        return pnts[0].getX(), pnts[0].getY(), pnts[1].getX()

    def setRadius(self, radius):
        """
        Set the radius of the circle.

        :param radius: Radius
        :type radius: float
        """
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        pnts = relm.getPath().getPoints()
        pnts[1].setX(radius)
        self._update()

    def getCenter(self):
        """
        Get the center of the circle.

        :return: Center
        :rtype: (float, float)
        """
        return self.__center

    def getRadius(self):
        """
        Get the radius of the circle.

        :return: Radius
        :rtype: float
        """
        return self.__radius

    def getArcAngles(self):
        """
        Get the arc start angle and sweep length.

        :return: Start angle and sweep length
        :rtype: (float, float)
        """
        return self.__startAngle, self.__sweepLength

    def resetPathSrcEnd(self):
        self._update()

    def resetPathDestEnd(self):
        self._update()

    def relocatePath(self):
        self._update()

    def resetPath(self):
        w, h = self.__connection.getStartObject().getDimension()
        r = (w + h) / 4.0
        relm = self.__connection.rawElement
        """:type: WdbsConnectionP"""
        relm.setPath([(0.0, 1.0), (r, r / 2.0)])
        self._update()

    def translate(self, dx, dy):
        pass

    def getPathSegment(self, x, y, pad=5, mthr=3):
        return self

    def getAngle(self):
        """
        :return: Angle in radian of the arc (center point),
            default if no point
        :rtype: float
        """
        path = self.__connection.rawElement.getPath()
        pnts = path.getPoints() if path else None
        if pnts:
            x = pnts[0].getX()
            y = pnts[0].getY()
            return math.atan2(y, x)
        else:
            return math.pi / 2

    def setAngle(self, angle):
        """
        Set the position of the path on a unit circle at the passed angle

        :param angle: Angle in radians to which set the path position
        :type angle: float
        """
        path = self.__connection.rawElement.getPath()
        pnts = path.getPoints() if path else None
        if pnts:
            pnts[0].setXY(math.cos(angle), math.sin(angle))

    def _update(self, center=None):
        # Centers of transition start/end state
        objx, objy = self.__connection.getStartObject().getPosition()
        objw, objh = self.__connection.getStartObject().getDimension()
        objr = (objw + objh) / 4.0

        xdif, ydif, radius = self._getRawData()
        rdst = max(radius / 2.0 + objr, radius) - objr

        cdst = objr + rdst
        segl = (pow(objr, 2.0) - pow(radius, 2.0) + pow(cdst, 2.0)) / \
               (2 * cdst)
        fh = math.sqrt(pow(objr, 2.0) - pow(segl, 2.0))

        slop = scmn_getSlope(xdif, ydif)
        angl = math.asin(fh / radius) if radius != 0 else 0

        self.__center = center or (objx + xdif * (objr + rdst),
                                   objy + ydif * (objr + rdst))
        self.__startAngle = -slop + (math.pi - angl)
        self.__sweepLength = 2 * (math.pi - angl)
        self.__radius = radius

    # noinspection PyProtectedMember
    def move(self, x, y):
        # Centers of transition start/end state
        objx, objy = self.__connection.getStartObject().getPosition()
        objw, objh = self.__connection.getStartObject().getDimension()
        objr = (objw + objh) / 4.0

        pdst = _len(x - objx, y - objy)
        if pdst == 0.0:
            return
        xdif = (x - objx) / pdst
        ydif = (y - objy) / pdst

        radius = max(min(2 * (pdst - objr) / 3, pdst / 2),
                     _WHDB_CIRC_MIN_RADIUS)

        self._setRawData(xdif, ydif, radius)
        self._update()
        self.__connection.setKeepPath()

    def finish(self):
        pass

    def __str__(self):
        return 'Circular Path'
