# ----------------------------------------------------------------------------
# - whdb-connection-path-api                                                 -
# - wild cruncher: connection path protocol                    core / db-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whdb_connection_path_api
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher connection path hierarchy

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta


class WhdbConnectionPathP(metaclass=ABCMeta):
    @abstractmethod
    def resetPathSrcEnd(self):
        """
        Reset the first one or two path segments to reflect changes made to
        the starting object/connection point.
        """
        raise NotImplementedError()

    @abstractmethod
    def resetPathDestEnd(self):
        """
        Reset the last one or two path segments to reflect changes made to
        the ending object/connection point.
        """
        raise NotImplementedError()

    @abstractmethod
    def relocatePath(self):
        """
        Relocate the path according to changes made to both start object/
        connection point pair and the end pair.
        This function should only be called if both
        object/connection point-pairs have been moved by equal amounts; if not,
        the path will be completely reset.
        """
        raise NotImplementedError()

    @abstractmethod
    def resetPath(self):
        """
        Reset the connection path as a single line between the starting
        connection point and the ending connection point.
        """
        raise NotImplementedError()

    @abstractmethod
    def translate(self, dx, dy):
        """
        Translate all path-elements with absolute coordinates by the given
        amounts.

        :param dx: X-amount to translate by
        :type dx: float
        :param dy: Y-amount to translate by
        :type dy: float
        """
        raise NotImplementedError()

    @abstractmethod
    def getPathSegment(self, x, y, pad=5, mthr=3):
        """
        Get connection path segment at the given location, +/- the specified
        padding. The returned segment handle is primarily intended for moving
        the associated segment.

        :param x: X coordinate
        :type x: float
        :param y: Y coordinate
        :type y: float
        :param pad: Padding around coordinates to extend search area for
        segment
        :type pad: float
        :param mthr: Distance threshold at which the returned connection handle
                     will actually execute the moving operation
        :type mthr: float
        :return: Handle for the connection segment at the given location, or
                 `None` if no connection segment could be found
        :rtype: WhdbConnectionPathSegmentP
        """
        raise NotImplementedError()


class WhdbConnectionPathSegmentP(metaclass=ABCMeta):
    """
    Handle for path-segments.
    """

    @abstractmethod
    def move(self, x, y):
        """
        Move the line in x- or y-direction according to the given values,
        depending on the direction of the line

        :param x: Potential move in x-direction
        :type x: float
        :param y: Potential move in y-direction
        :type y: float
        """
        raise NotImplementedError()

    @abstractmethod
    def finish(self):
        """
        Indicate that all movement of the line has been completed.
        """
        raise NotImplementedError()
