# ----------------------------------------------------------------------------
# - wdbi-wrapper                                                             -
# - wild cruncher:ish wrappper objects            core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_wrapper
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable database wrapper objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.dbstore.immutable.wdbi_context import WdbiContext
from wcr_core.dbstore.wdbs_entity import WdbsEntityBase, WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_entity_api import WdbsEntityP
from wcr_core.dbstore_api.wdbs_meta_data_api import WdbsStructuralMetaP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_META_EVENT


# noinspection PyAbstractClass
class WdbiEntityP(WdbsEntityP):
    """
    Wrapper base class.
    """
    __slots__ = tuple()

    @property
    @abstractmethod
    def node(self):
        """
        Get the current version of the wrapped node.

        :return: Node
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiEntityMixinP(WdbiEntityP):
    @abstractmethod
    def _current(self):
        """
        Get the current version of the wrapped node.

        :return: Current node version
        :rtype: WdbiDbNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def _update(self, path, updateF):
        """
        Update the (child element of) wrapped node.

        :param path: Relative path to the element to be updated ([] for node)
        :type path: list[object]
        :param updateF: Update function
        :type updateF: (WdbiNodeP) -> WdbiNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def _wrap(self, wpid, func, *args, **kwargs):
        """
        Get the wrapper for the node with the given id, or create a new one.

        :param wpid: Node id
        :type wpid: UUID
        :param func: Wrapper creation function
        :type func: (*_) -> WdbiWrapperP
        :param args: Arguments for the wrapper creation function
        :param kwargs: Keyword-arguments for the wrapper creation function
        :return: Wrapper for the node with the given id
        :rtype: WdbiEntityP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiEntityBase(WdbsEntityBase, WdbiEntityMixinP):
    __slots__ = WdbsEntityBase.__slots__ + \
        ('__ident', '__type', '__uuid', '__context')

    def __init__(self, dctx, uuid):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        """
        assert isinstance(dctx, WdbiContext)

        super().__init__(dctx)

        self.__context = dctx
        self.__uuid = uuid

        self.__ident = None
        self.__type = None

    def getUUID(self):
        return self.__uuid

    @property
    def nodeType(self):
        """
        Get the wrapped node's type.

        :return: Node type
        :rtype: int
        """
        if self.__type is None:
            self.__type = self._current().nodeType
        return self.__type

    @property
    def metaDescription(self):
        return self._current().metaDescription

    def getMeta(self, *keys):
        meta = self._current().getMeta(*keys)
        return meta.asValue() if meta is not None else None

    def getMetas(self, *keys):
        meta = self._current().getMeta(*keys)
        if not keys:
            return meta
        if isinstance(meta, WdbsStructuralMetaP):
            return meta.kvPairs()
        raise KeyError(
            'No structural meta-data found at {}: {} ({})'.format(
                keys, meta, type(meta)))

    def getMetaKeys(self, *keys):
        if not keys:
            return (f.name for f in self._current().metaDescription.fields)
        meta = self._current().getMeta(*keys)
        if isinstance(meta, WdbsStructuralMetaP):
            return (k for k, _ in meta.kvPairs())
        raise KeyError(
            'No structural meta-data found at {}: {} ({})'.format(
                keys, meta, type(meta)))

    def getMetaCount(self, *keys):
        if not keys:
            return len(list(self._current().metaDescription.fields))
        meta = self._current().getMeta(*keys)
        if isinstance(meta, WdbsStructuralMetaP):
            return len(meta)
        raise KeyError(
            'No structural meta-data found at {}: {} ({})'.format(
                keys, meta, type(meta)))

    def getAllMeta(self):
        return self._current().getAllMeta()

    def _wrap(self, wpid, func, *args, **kwargs):
        return self.__context.wrap(wpid, func, *args, **kwargs)


class WdbiNodeEntityBase(WdbiEntityBase):
    """
    Node wrapper that stores the wrapped node itself.
    """
    __slots__ = WdbiEntityBase.__slots__ + ('__node',)

    def __init__(self, dctx, node):
        """
        :param node: Node to be wrapped
        :type node: WdbiDbNodeP
        """
        WdbiEntityBase.__init__(self, dctx, node.getUUID())

        #: Current version of the node
        self.__node = node
        """:type: WdbiDbNodeP"""

    def _current(self):
        return self.__node

    def setMeta(self, val, *keys):
        self.__node = self.__node.setMeta(val, *keys)
        self._notifyObservers(WDBS_UPDATE_META_EVENT, val, *keys)

    def removeMeta(self, *keys):
        self.__node = self.__node.unsetMeta(*keys)
        self._notifyObservers(WDBS_UPDATE_META_EVENT, None, *keys)

    def _update(self, path, updateF):
        self.__node = self.__node.update(path, updateF)

    @property
    def node(self):
        return self.__node

    def __str__(self):
        return self.__class__.__name__ + ' ' + str(self._current())


class WdbiTreepathEntityBase(WdbiEntityBase, WdbsObservableMixinP):
    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param currentF: Function providing the current node
        :type currentF: (int) -> WdbiDbNodeP
        :param updateF: Function to update the current node
        :type updateF: ((int, WdbiDbNodeP) -> WdbiDbNodeP) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiEntityBase.__init__(self, dctx, currentF(ident).getUUID())

        #: Current node version function
        self.__currentF = currentF
        #: Node update function
        self.__updateF = updateF
        #: Node ident
        self.__ident = ident

    def _current(self):
        return self.__currentF(self.__ident)

    def setMeta(self, val, *keys):
        self.__updateF(self.__ident, lambda c: c.setMeta(val, *keys))
        self._notifyObservers(WDBS_UPDATE_META_EVENT, val, *keys)

    def removeMeta(self, *keys):
        self.__updateF(self.__ident, lambda c: c.unsetMeta(*keys))
        self._notifyObservers(WDBS_UPDATE_META_EVENT, None, *keys)

    def _update(self, path, updateF):
        self.__updateF(self.__ident, lambda node: node.update(path, updateF))

    @property
    def node(self):
        return self._current()

    def __str__(self):
        return self.__class__.__name__ + ' ' + str(self._current())


class WdbiListWrapper(object):
    __slots__ = ('__currentF', '__updateF', '__wrapperF')

    """
    Wraps a list node to resemble a python list.
    """

    def __init__(self, currentF, updateF, wrapperF):
        """
        :param currentF: Function providing the current list node
        :type currentF: (int) -> WdbiListNode
        :param updateF: Function to update the current list node
        :type updateF: ((int, WdbiListNodeP) -> WdbiListNode) -> None
        :param wrapperF: Function to wrap list elements
        :type wrapperF: (WdbiNodeP) -> WdbiEntityP
        """
        self.__currentF = currentF
        self.__updateF = updateF
        self.__wrapperF = wrapperF

    def _current(self):
        """
        Get the current version of the list node.

        :return: Current node version
        :rtype: WdbiListNode
        """
        return self.__currentF()

    def __getitem__(self, item):
        return self._current()[item]

    def __len__(self):
        return len(self._current())

    def __iter__(self):
        return map(self.__wrapperF, self._current())

    def __contains__(self, item):
        return item in self._current()

    def __reversed__(self):
        return map(self.__wrapperF, reversed(self._current()))

    def __nonzero__(self):
        return len(self._current()) > 0

    def append(self, item):
        self.__updateF(lambda l: l.add(self._extractItem(item)))

    def insert(self, pos, item):
        self.__updateF(lambda l: l.add(self._extractItem(item), pos))

    def extend(self, itms):
        self.__updateF(lambda l: l.addAll(map(self._extractItem, itms)))

    def remove(self, item):
        self.__updateF(lambda l: l.remove(self._extractItem(item)))

    def __str__(self):
        return '[{}]'.format(', '.join(map(repr, self)))

    def __repr__(self):
        return '[{}]'.format(', '.join(map(repr, self)))

    def _extractItem(self, item):
        return item.node if isinstance(item, WdbiNodeEntityBase) else item
