# ----------------------------------------------------------------------------
# - wdbi-geometry                                                            -
# - wild cruncher:ish geo objects                 core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_geometry
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable geo objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_tree import WdbiListNode, \
    WdbiArrayNode, WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_tree_db import WDBI_POINT_NODE, \
    WDBI_PATH_NODE, WdbiDbListNode
from wcr_core.dbstore.immutable.wdbi_entity import WdbiNodeEntityBase, \
    WdbiTreepathEntityBase, WdbiEntityMixinP
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_geometry_api import WdbsPathP
from silkan_common.scmn_geometry import ScmnPointMP


_WDBI_POINT_X = 0
_WDBI_POINT_Y = 1


def wdbi_createPoint(x, y, uuid=None):
    """
    Create a point node.

    :param x: X-coordinate
    :type x: float
    :param y: Y-coordinate
    :type y: float
    :param uuid: UUID
    :type uuid: UUID | str | None
    :return: Newly created point node
    :rtype: WdbiArrayNode
    """
    return WdbiArrayNode(
        uuid, WDBI_POINT_NODE,
        [WdbiFieldNode(None, 0, 'xpos', x),
         WdbiFieldNode(None, 0, 'ypos', y)])


# noinspection PyAbstractClass,PyUnresolvedReferences
class WdbiPointEntityMixin(WdbiEntityMixinP, ScmnPointMP):
    """
    Wrapper for point nodes.
    """

    def getX(self):
        return self._current().get(_WDBI_POINT_X).getValue()

    def setX(self, x):
        self._update([_WDBI_POINT_X], lambda f: f.setValue(x))

    def getY(self):
        return self._current().get(_WDBI_POINT_Y).getValue()

    def setY(self, y):
        self._update([_WDBI_POINT_Y], lambda f: f.setValue(y))

    def setXY(self, x, y):
        self._update([_WDBI_POINT_X], lambda f: f.setValue(x))
        self._update([_WDBI_POINT_Y], lambda f: f.setValue(y))

    def __iter__(self):
        c = self._current()
        return iter((c[_WDBI_POINT_X].getValue(), c[_WDBI_POINT_Y].getValue()))


class WdbiPointEntityForNode(WdbiNodeEntityBase, WdbiPointEntityMixin):
    """
    Wrapper for point nodes, handles current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiPointForTreepath(WdbiTreepathEntityBase, WdbiPointEntityMixin):
    """
    Wrapper for externally managed point nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current point node
        :type currentF: (int) -> WdbiArrayNode
        :param updateF: Function to update the current point node
        :type updateF: ((int, WdbiArrayNode) -> WdbiArrayNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)


def wdbi_createPath(metaPrv, pnts, uuid=None, meta=None, **kwargs):
    """
    Create a path node.

    :param pnts: List of point coordinates
    :type pnts: list[(float, float)]
    :return: Newly created path node
    :rtype: WdbiDbListNode
    """
    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsPathP)

    items = [wdbi_createPoint(*p) for p in pnts] if pnts else None
    return WdbiDbListNode(uuid, WDBI_PATH_NODE, meta, items, **kwargs)


# noinspection PyAbstractClass
class WdbiPathEntityMixin(WdbiEntityMixinP, WdbsPathP, WdbsObservableMixinP):
    """
    Wrapper for path nodes.
    """

    def _current(self):
        """
        :rtype: WdbiListNode
        """
        raise NotImplementedError()

    def _getPoint(self, ident):
        """
        Get the point with the given ident from the current path instance.

        :param ident: Ident of the point
        :type ident: int
        :return: Point
        :rtype: WdbiArrayNode
        """
        return self._current().get(ident)

    def _updatePoint(self, ident, updateF):
        """
        Update a point.

        :param ident: Ident of the point
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiArrayNode) -> WdbiArrayNode
        """
        self._update([ident], updateF)

    def _mkPointWrapper(self, point):
        """
        Wrap a point.

        :param point: Point to be wrapped
        :type point: WdbiArrayNode
        :return: Point wrapper
        :rtype: WdbiPathForTreepath
        """
        idnt = point.getUUID()
        return self._wrap(
            idnt, WdbiPointForTreepath, self._getPoint, self._updatePoint,
            idnt)

    def getPoints(self):
        return [self._mkPointWrapper(p) for p in self._current()]

    def addPoint(self, x, y, pos=None):
        point = wdbi_createPoint(x, y)
        self._update([], lambda plst: plst.add(point, pos))
        return self._mkPointWrapper(point)

    def removePoint(self, point):
        self._update([], lambda plst: plst.remove(point))

    def __iter__(self):
        return iter(self.getPoints())


class WdbiPathForNode(WdbiNodeEntityBase, WdbiPathEntityMixin):
    """
    Wrapper for path nodes, handles current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbListNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)


class WdbiPathForTreepath(WdbiTreepathEntityBase, WdbiPathEntityMixin):
    """
    Wrapper for externally managed path nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current path node
        :type currentF: (int) -> WdbiDbListNode
        :param updateF: Function to update the current path node
        :type updateF: ((int, WdbiDbListNode) -> WdbiDbListNode) -> None
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
