# ----------------------------------------------------------------------------
# - wdbi-instance                                                            -
# - wild cruncher:ish instance objects            core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher immutable instance objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.immutable.wdbi_parameter import \
    WdbiEntityParametersMixin, \
    wdbi_createParam
from wcr_core.dbstore.immutable.wdbi_protocols import WdbiWrapperMaker
from wcr_core.dbstore.immutable.wdbi_tree import WdbiArrayNode, WdbiListNode, \
    WdbiDictNode, WdbiFieldNode
from wcr_core.dbstore.immutable.wdbi_tree_db import WDBI_CONNECTOR_LIST_NODE, \
    WDBI_BLOCK_INSTANCE_NODE, WDBI_MODEL_INSTANCE_NODE, WDBI_PARAM_DICT_NODE, \
    WDBI_OPERATOR_INSTANCE_NODE, WdbiDbArrayNode
from wcr_core.dbstore.immutable.wdbi_entity import WdbiNodeEntityBase, \
    WdbiTreepathEntityBase, WdbiEntityMixinP
from wcr_core.dbstore.immutable.wdbi_connection_point import \
    wdbi_createConnectionPoint, WdbiConnectionPointForTreepath
from wcr_core.dbstore.wdbs_entity import WdbsObservableMixinP
from wcr_core.dbstore_api.wdbs_instance_api import WdbsInstanceP, \
    WdbsBlockInstanceP, WdbsModelInstanceP, WdbsOperatorInstanceP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_UPDATE_CORE_EVENT, \
    WDBS_CHILD_ADDED_EVENT, WDBS_CHILD_REMOVED_EVENT


# : Array index for instance connectors
_WDBI_INST_CONNECTORS = 0
# : Array index for instance model
_WDBI_INST_ENTITY = 1
_WDBI_INST_MODEL = 1
_WDBI_INST_BLOCK = 1
#: Array index for instance model
_WDBI_INST_PARAMS = 1
_WDBI_INST_NAME = 2
_WDBI_INST_INFO = 3
_WDBI_INST_OPER = 4


def _wdbi_createInstanceConnectors(metaPrv, node, path, clst):
    """
    Create connector nodes for a given node at the given path.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param node: Node to create connector nodes for
    :type node: WdbiArrayNode
    :param path: Path to create connectors at
    :type path: Iterable[object]
    :param clst: Iterable with connector info
    :type clst: Iterable[(str, UUID | None, dict[str, object])]
    :return: Node with connectors
    :rtype: WdbiArrayNode
    """
    if clst is not None:
        def connF(_clst):
            for ctyp, uuid, meta in clst:
                _clst = _clst.add(
                    wdbi_createConnectionPoint(metaPrv, ctyp, uuid, **meta))
            return _clst

        return node.update(path, connF)
    else:
        return node


def _wdbi_createInstance(metaPrv, entity, ntyp, name=None, info=None,
                         clst=None, uuid=None, meta=None, fields=None,
                         **kwargs):
    """
    Create an instance node of the given type.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param entity: Instantiated entity (model, block, operator params, ...)
    :type entity: WdbiNodeP
    :param ntyp: Instance node type
    :type ntyp: int
    :param name: Instance node name
    :type name: str
    :param info: Instance node info string
    :type info: str
    :param clst: Instance node connectors
    :type clst: Iterable[(str, UUID | None, dict[str, object])]
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param fields: Additional array fields
    :type fields: Iterable[WdbiNodeP] | None
    :param kwargs: Additional meta-data
    :return: Newly created instance node
    :rtype: WdbiArrayNode
    """
    items = [WdbiListNode(None, WDBI_CONNECTOR_LIST_NODE),
             entity,
             WdbiFieldNode(None, 0, 'name', name if name is not None else ''),
             WdbiFieldNode(None, 0, 'info', info if info is not None else '')]
    if fields:
        items.extend(fields)
    node = WdbiDbArrayNode(uuid, ntyp, meta, items, **kwargs)

    return _wdbi_createInstanceConnectors(
        metaPrv, node, [_WDBI_INST_CONNECTORS], clst)


def wdbi_createBlockInstance(metaPrv, btyp, name=None, info=None, clst=None,
                             uuid=None, buuid=None, meta=None, bmeta=None,
                             **kwargs):
    """
    Create a new block instance node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param btyp: Block type
    :type btyp: str
    :param name: the instance name
    :type  name: str
    :param info: the instance info
    :type  info: str
    :param clst: Connector list (type, id, meta)
    :type clst: Iterable[(str, UUID | None, dict[str, object])]
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param buuid: Block UUID
    :type buuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param bmeta: Block meta description
    :type bmeta: WdbsMetaDescriptionP | None
    :param kwargs: Additional meta-data
    :return: Newly created mlock instance node
    :rtype: WdbiDbArrayNode
    """
    from .wdbi_block import wdbi_createBlock

    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsBlockInstanceP, btyp)

    block = wdbi_createBlock(
        metaPrv, btyp=btyp, name=name, info=info, uuid=buuid, meta=bmeta)
    return _wdbi_createInstance(
        metaPrv, block, WDBI_BLOCK_INSTANCE_NODE,
        name=name, info=info, clst=clst, uuid=uuid, meta=meta, **kwargs)


def wdbi_createModelInstance(metaPrv, mtyp=None, model=None, name=None,
                             info=None, clst=None, uuid=None, muuid=None,
                             meta=None, mmeta=None, **kwargs):
    """
    Create a new model instance node.

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param mtyp: Model type
    :type mtyp: str | None
    :param model: Instance model
    :type  model: WdbiNodeP | None
    :param name: the instance name
    :type  name: str | None
    :param info: the instance info
    :type  info: str | None
    :param clst: Connector list (type, id, meta)
    :type clst: Iterable[(str, UUID, dict[str, object])]
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param muuid: Model UUID
    :type muuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param mmeta: Model meta description
    :type mmeta: WdbsMetaDescriptionP | None
    :param kwargs: Additional meta-data
    :return: Newly created instance node
    :rtype: WdbiDbArrayNode
    """
    if model is None:
        from .wdbi_model import wdbi_createModel

        model = wdbi_createModel(
            metaPrv, mtyp=mtyp, name=name, info=info, uuid=muuid, meta=mmeta)

    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsBlockInstanceP, mtyp)

    return _wdbi_createInstance(
        metaPrv, model, WDBI_MODEL_INSTANCE_NODE,
        name=name, info=info, clst=clst, uuid=uuid, meta=meta, **kwargs)


def wdbi_createOperatorInstance(metaPrv, oper, name=None, info=None, clst=None,
                                params=None, uuid=None, meta=None, **kwargs):
    """
    Create a new operator instance node (i.e. function node).

    :param metaPrv: Meta-description provider
    :type metaPrv: WdbsMetaProviderP
    :param oper: Operator code
    :type  oper: str
    :param name: Instance name
    :type  name: str
    :param info: Instance info string
    :type  info: str
    :param clst: Connector list (type, id, meta)
    :type clst: Iterable[(str, UUID | None, dict[str, object])]
    :param params: Operator instance parameter values
    :type params: Iterable[(str, object)]
    :param uuid: UUID
    :type uuid: UUID | str | None
    :param meta: Meta description
    :type meta: WdbsMetaDescriptionP | None
    :param kwargs: Additional meta-data
    :return: Newly created operator instance node
    :rtype: WdbiDbArrayNode
    """
    if meta is None:
        meta = metaPrv.getMetaDescription(WdbsOperatorInstanceP, oper)

    pnode = WdbiDictNode(None, WDBI_PARAM_DICT_NODE)
    node = _wdbi_createInstance(
        metaPrv, pnode, WDBI_OPERATOR_INSTANCE_NODE,
        name=name, info=info, clst=clst, uuid=uuid, meta=meta,
        fields=[WdbiFieldNode(None, 0, 'oper', oper)], **kwargs)
    if params:
        for p, v in params:
            node = node.update(
                [_WDBI_INST_PARAMS],
                lambda ps: ps.add(p, wdbi_createParam(metaPrv, p, value=v)))
    return node


# noinspection PyAbstractClass
class WdbiInstanceEntityMixin(WdbiEntityMixinP, WdbiWrapperMaker,
                              WdbsInstanceP, WdbsObservableMixinP):
    """
    Wrapper for instance nodes.
    """

    def __init__(self, connPath):
        """
        :param connPath: Path to connector list node
        :type connPath: list[object]
        """
        self.__connPath = connPath

    def _current(self):
        """
        :rtype: WdbiArrayNode
        """
        raise NotImplementedError()

    def _getConnectors(self):
        """
        Get the connector-list node.

        :return: Connector list node
        :rtype: WdbiListNode
        """
        return self._current().getPath(self.__connPath)

    def _getConnector(self, ident):
        """
        Get the connector node for the given ident.

        :param ident: Ident value
        :type ident: int
        :return: Connector node with the given ident
        :rtype: WdbiDbArrayNode
        """
        return self._current().getPath(self.__connPath + [ident])

    def _updateConnectors(self, updateF):
        """
        Update the connector-list node.

        :param updateF: Update function
        :type updateF: (WdbiListNode) -> WdbiListNode
        """
        self._update(self.__connPath, lambda clst: updateF(clst))

    def _updateConnector(self, ident, updateF):
        """
        Update a connector.

        :param ident: Ident value of the connector to be updated
        :type ident: int
        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        self._update(self.__connPath + [ident],
                     lambda conn: updateF(conn))

    def _mkConnWrapper(self, conn):
        """
        Wrap a connector.

        :param conn: Connector to be wrapped
        :type conn: WdbiDbArrayNode
        :return: Wrapped Connector
        :rtype: WdbiConnectionPointForTreepath
        """
        idnt = conn.getUUID()
        return self._wrap(
            idnt, WdbiConnectionPointForTreepath, self._getConnector,
            self._updateConnector, idnt)

    def createWrapperForPath(self, path):
        ntyp, idnt, *rest = path
        if ntyp == _WDBI_INST_CONNECTORS:
            wrapper = self._wrap(
                idnt, WdbiConnectionPointForTreepath, self._getConnector,
                self._updateConnector, idnt)
        else:
            raise ValueError('Unknown node type ' + ':'.join(path))

        if rest:
            raise ValueError('Path rest ' + ':'.join(rest))
        else:
            return wrapper

    def getName(self):
        return self._current().getPath([_WDBI_INST_NAME]).getValue()

    def setName(self, name):
        self._update([_WDBI_INST_NAME], lambda f: f.setValue(name))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getInfo(self):
        return self._current().getPath([_WDBI_INST_INFO]).getValue()

    def setInfo(self, info):
        self._update([_WDBI_INST_INFO], lambda f: f.setValue(info))
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def getConnectionPoints(self):
        return (self._mkConnWrapper(c) for c in self._getConnectors())

    def getConnectionPointCount(self):
        return len(self._getConnectors())

    def createConnectionPoint(self, ctyp, uuid=None, meta=None, **kwargs):
        conn = wdbi_createConnectionPoint(
            self.context, ctyp, uuid=uuid, meta=meta, **kwargs)
        self._update(self.__connPath, lambda clst: clst.add(conn))
        wrap = self._mkConnWrapper(conn)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, wrap)
        return wrap

    def setConnectionPointOrder(self, orderF, wrapFs):
        def sort(clst):
            """
            :type clst: WdbiListNode
            :rtype: WdbiListNode
            """
            wrap, unwrap = wrapFs
            wrpd = [wrap(self._mkConnWrapper(c)) for c in clst]
            wrpd.sort(key=orderF)
            clst = clst.clear().addAll(unwrap(c).node for c in wrpd)
            return clst

        self._update(self.__connPath, sort)

    def removeConnectionPoint(self, xpoint):
        self._update(self.__connPath,
                     lambda clst: clst.remove(xpoint))
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, xpoint)

    def clearConnectionPoints(self):
        xpnts = list(self.getConnectionPoints())
        self._update(self.__connPath, lambda clst: clst.clear())
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, *xpnts)


# noinspection PyAbstractClass
class WdbiBlockInstanceEntityMixin(WdbiInstanceEntityMixin,
                                   WdbsBlockInstanceP):
    """
    Wrapper for block instance nodes.
    """
    __slots__ = tuple()

    def __init__(self):
        WdbiInstanceEntityMixin.__init__(self, [_WDBI_INST_CONNECTORS])

    def _getBlock(self, *_):
        """
        Get the current block node.

        :return: Model node
        :rtype: WdbiDbArrayNode
        """
        return self._current()[_WDBI_INST_MODEL]

    def _updateBlock(self, _, updateF):
        """
        Update the block node.

        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        return self._update([_WDBI_INST_MODEL], updateF)

    def getBlock(self):
        from .wdbi_block import WdbiBlockForTreepath

        model = self._current()[_WDBI_INST_MODEL]
        ident = model.getUUID()
        return self._wrap(
            ident, WdbiBlockForTreepath, self._getBlock, self._updateBlock,
            ident)


class WdbiBlockInstanceForNode(WdbiNodeEntityBase,
                               WdbiBlockInstanceEntityMixin):
    """
    Wrapper for block instance nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)
        WdbiBlockInstanceEntityMixin.__init__(self)


class WdbiBlockInstanceForTreepath(WdbiTreepathEntityBase,
                                   WdbiBlockInstanceEntityMixin):
    """
    Wrapper for externally managed block instance nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current instance node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current instance node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> _
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
        WdbiBlockInstanceEntityMixin.__init__(self)


# noinspection PyAbstractClass
class WdbiModelInstanceEntityMixin(WdbiInstanceEntityMixin,
                                   WdbsModelInstanceP):
    """
    Wrapper for model instance nodes.
    """

    def __init__(self):
        WdbiInstanceEntityMixin.__init__(self, [_WDBI_INST_CONNECTORS])

    def _getModel(self, *_):
        """
        Get the current model node.

        :return: Model node
        :rtype: WdbiDbArrayNode
        """
        return self._current()[_WDBI_INST_MODEL]

    def _updateModel(self, _, updateF):
        """
        Update the model node.

        :param updateF: Update function
        :type updateF: (WdbiDbArrayNode) -> WdbiDbArrayNode
        """
        return self._update([_WDBI_INST_MODEL], updateF)

    def getModel(self):
        from .wdbi_model import WdbiModelForTreepath

        model = self._current()[_WDBI_INST_MODEL]
        ident = model.getUUID()
        return self._wrap(
            ident, WdbiModelForTreepath, self._getModel, self._updateModel,
            ident)


class WdbiModelInstanceForNode(WdbiNodeEntityBase,
                               WdbiModelInstanceEntityMixin):
    """
    Wrapper for model instance nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)
        WdbiModelInstanceEntityMixin.__init__(self)


class WdbiModelInstanceForTreepath(WdbiTreepathEntityBase,
                                   WdbiModelInstanceEntityMixin):
    """
    Wrapper for externally managed model instance nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current instance node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current instance node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> _
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
        WdbiModelInstanceEntityMixin.__init__(self)


# noinspection PyAbstractClass,PyUnresolvedReferences
class WdbiOperatorInstanceEntityMixin(WdbiInstanceEntityMixin,
                                      WdbiEntityParametersMixin,
                                      WdbsOperatorInstanceP):
    """
    Wrapper for operator instance nodes.
    """

    def __init__(self):
        WdbiEntityParametersMixin.__init__(self, [_WDBI_INST_PARAMS])
        WdbiInstanceEntityMixin.__init__(self, [_WDBI_INST_CONNECTORS])

    def getOperator(self):
        return self._current().get(_WDBI_INST_OPER).getValue()


class WdbiOperatorInstanceForNode(WdbiNodeEntityBase,
                                  WdbiOperatorInstanceEntityMixin):
    """
    Wrapper for operator instance nodes, handling current node internally.
    """

    def __init__(self, dctx, node):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param node: Node to wrap
        :type node: WdbiDbArrayNode
        """
        WdbiNodeEntityBase.__init__(self, dctx, node)
        WdbiOperatorInstanceEntityMixin.__init__(self)


class WdbiOperatorInstanceForTreepath(WdbiTreepathEntityBase,
                                      WdbiOperatorInstanceEntityMixin):
    """
    Wrapper for externally managed operator instance nodes.
    """

    def __init__(self, dctx, currentF, updateF, ident):
        """
        :param dctx: Database context
        :type dctx: WdbiContext
        :param currentF: Function providing the current instance node
        :type currentF: (int) -> WdbiDbArrayNode
        :param updateF: Function to update the current instance node
        :type updateF: ((int, WdbiDbArrayNode) -> WdbiDbArrayNode) -> _
        :param ident: Ident of the wrapped node
        :type ident: int
        """
        WdbiTreepathEntityBase.__init__(self, dctx, currentF, updateF, ident)
        WdbiOperatorInstanceEntityMixin.__init__(self)


def wdbi_getDynamicInstanceWrapper(inst):
    """
    Get a dynamic instance wrapper for the given instance.

    :param inst: Instance to provide wrapper for
    :type inst: WdbiNodeP
    :return: Instance wrapper for the given instance
    :rtype: WdbiInstanceEntityMixin
    """
    if inst.nodeType == WDBI_BLOCK_INSTANCE_NODE:
        return WdbiBlockInstanceForTreepath
    if inst.nodeType == WDBI_MODEL_INSTANCE_NODE:
        return WdbiModelInstanceForTreepath
    if inst.nodeType == WDBI_OPERATOR_INSTANCE_NODE:
        return WdbiOperatorInstanceForTreepath
    else:
        raise ValueError(str(inst))
