# ----------------------------------------------------------------------------
# - wdbi-tree                                                                -
# - wild cruncher: immutable database tree nodes  core / dbstore / immutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbi_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher - Immutable database tree nodes

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta
from collections import OrderedDict
from itertools import count
from wcr_core.dbstore.wdbs_util import wdbs_makeUUID


def wdbi_getattr(obj, key):
    """
    Fail-safe getattr

    :param obj: Object to get attribute for
    :type obj: object
    :param key: Attribute key
    :type key: str
    :return: Attribute, if found, or `None`
    :rtype: object
    """
    try:
        return getattr(obj, key, None)
    except KeyError:
        return None


class WdbiNodeP(metaclass=ABCMeta):
    """
    Abstract node class.
    Elements in the immutable database are represented as nodes in a tree.
    """
    __slots__ = tuple()

    @abstractmethod
    def getUUID(self):
        """
        Get the node's path-persistent UUID.

        :return: UUID
        :rtype: UUID
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def nodeType(self):
        """
        Get the type of the node (see wdbi_protocols).

        :return: Node type
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getPath(self, path):
        """
        Get the child element at the given path, relative to the node.

        :param path: Path elements
        :type path: list[object]
        :return: Element at the given path
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def update(self, path, updateF):
        """
        Update the element at the given path, using the given update function.

        :param path: Path elements
        :type path: list[object]
        :param updateF: Update function
        :type updateF: (WdbiNodeP) -> WdbiNodeP
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def same(self, other):
        """
        Whether this node and another are the same, i.e. have the same ident-
        value.

        :param other: Node to compare to
        :type other: WdbiNodeP
        :return: True iff this node and another are the same
        :rtype: bool
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiNodeMixinP(metaclass=ABCMeta):
    __slots__ = tuple()

    @abstractmethod
    def _clone(self, **kwargs):
        """
        Clone the current node, taking new values from the provided
        keyword-arguments.

        :return: Clone of the node
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiNodeBase(WdbiNodeP, WdbiNodeMixinP):
    __slots__ = ('_WdbiNodeBase__uuid', '_WdbiNodeBase__nodeType')

    #: ID counter
    __ID_COUNTER = count(1)

    def __init__(self, uuid, ntyp):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        """
        #: UUID
        self.__uuid = wdbs_makeUUID(uuid)
        #: Node type
        self.__nodeType = ntyp

    def _cloneNodeInto(self, new):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiNodeBase
        """
        new.__uuid = self.__uuid
        new.__nodeType = self.__nodeType

    def getUUID(self):
        return self.__uuid

    @property
    def nodeType(self):
        return self.__nodeType

    def same(self, other):
        return isinstance(other, WdbiNodeP) and \
            self.getUUID() == other.getUUID()


# noinspection PyAbstractClass
class WdbiCollectionNodeP(WdbiNodeP):
    """
    Abstract base class for collection nodes.
    """
    __slots__ = tuple()

    @abstractmethod
    def updatedItem(self, key, item):
        """
        Replace the item with the given key with an updated version.

        :param key: Key of the item to be replaced
        :type key: object
        :param item: Replacement
        :type item: WdbiNodeP
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self, key):
        """
        Get the item for the given path-key.

        :param key: Key of the item to be retrieved
        :type key: object
        :return: Item for the given key
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()

    @abstractmethod
    def children(self):
        """
        Get all items.

        :return: All items in the collection
        :rtype: list[WdbiNodeP]
        """
        raise NotImplementedError()

    @abstractmethod
    def __getitem__(self, key):
        """
        Get the item for the given value-key.

        :param key: Key of the item to be retrieved
        :type key: object
        :return: Item for the given key
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiVarCollectionNodeP(WdbiCollectionNodeP):
    """
    Abstract base class for extensible collection nodes.
    """
    __slots__ = tuple()

    @abstractmethod
    def clear(self):
        """
        Clear the collection, i.e. remove all items.

        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbiArrayNode(WdbiNodeBase, WdbiCollectionNodeP):
    """
    Array-based collection node of fixed size.
    """
    __slots__ = WdbiNodeBase.__slots__ + ('_WdbiArrayNode__items',)

    def __init__(self, uuid, ntyp, items):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param items: Child items
        :type items: list[WdbiNodeP]
        """
        super().__init__(uuid, ntyp)

        #: Child items
        self.__items = items

    def _cloneArrayInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiArrayNode
        """
        self._cloneNodeInto(new)
        items = kwargs.get('items')
        new.__items = items if items is not None else self.__items

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiArrayNode.__new__(WdbiArrayNode)
        self._cloneArrayInto(new, **kwargs)
        return new

    @property
    def _items(self):
        """
        :return: Shallow copy of the node's items
        :rtype: list[WdbiNodeP]
        """
        return list(self.__items)

    def getPath(self, path):
        if path:
            return self[path[0]].getPath(path[1:])
        else:
            return self

    def update(self, path, updateF):
        if path:
            item = self[path[0]].update(path[1:], updateF)
            return self.updatedItem(path[0], item)
        else:
            return updateF(self)

    def updatedItem(self, idx, item):
        itms = self._items
        itms[idx] = item
        return self._clone(items=itms)

    def get(self, idx):
        if not isinstance(idx, int):
            raise ValueError('Index has to be an integer')
        if not (0 <= idx < len(self.__items)):
            raise IndexError(
                'Index {} out of bounds (< {})'.format(idx, len(self.__items)))
        return self.__items[idx]

    def children(self):
        return iter(self.__items)

    def __len__(self):
        return len(self.__items)

    def __getitem__(self, idx):
        if not isinstance(idx, int):
            raise ValueError('Index has to be an integer')
        if not (0 <= idx < len(self.__items)):
            raise IndexError('Index out of bounds: ' + str(idx) + ' ' +
                             str(self.__items))
        return self.__items[idx]

    def __iter__(self):
        return iter(self.__items)

    def __reversed__(self):
        return reversed(self.__items)

    def __contains__(self, item):
        return item in self.__items

    def __hash__(self):
        return sum(hash(n) for n in self.__items)

    def __eq__(self, other):
        if not isinstance(other, WdbiArrayNode):
            return False
        if len(self) != len(other):
            return False
        return self.__items == other.__items

    def __str__(self):
        return '[{}]'.format(', '.join(map(repr, self.__items)))

    def __repr__(self):
        return 'WdbiArrayNode ({})'.format(
            ', '.join(map(repr, self.__items)))


# noinspection PyAbstractClass
class WdbiListNode(WdbiNodeBase, WdbiVarCollectionNodeP):
    """
    List-based collection node of variable size.
    """
    __slots__ = WdbiNodeBase.__slots__ + ('_WdbiListNode__items',)

    def __init__(self, uuid, ntyp, items=None):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param items: Child items
        :type items: list[WdbiNodeP] | None
        """
        super().__init__(uuid, ntyp)

        #: Child items
        self.__items = self._createContainer(items)

    @classmethod
    def _createContainer(cls, items=None):
        if isinstance(items, OrderedDict):
            return items
        cont = OrderedDict()
        if items:
            for i in items:
                cont[i.getUUID()] = i
        return cont

    def _cloneListInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiListNode
        """
        self._cloneNodeInto(new)

        items = kwargs.get('items')
        new.__items = items if items is not None else self.__items

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiListNode.__new__(WdbiListNode)
        self._cloneListInto(new, **kwargs)
        return new

    @property
    def _items(self):
        """
        :return: Shallow copy of the node's items
        :rtype: list[WdbiNodeP]
        """
        return self.__items.copy()

    def get(self, key):
        return self.__items.get(key)

    def children(self):
        return iter(self.__items.values())

    def getPath(self, path):
        if path:
            isNode = isinstance(path[0], WdbiNodeP)
            idnt = path[0].getUUID() if isNode else path[0]
            return self.__items[idnt].getPath(path[1:])
        else:
            return self

    def update(self, path, updateF):
        if path:
            try:
                frst = path[0].getUUID()
            except AttributeError:
                frst = path[0]
            item = self.__items[frst].update(path[1:], updateF)
            return self.updatedItem(frst, item)
        else:
            return updateF(self)

    def updatedItem(self, itid, item):
        itms = self.__items.copy()
        itms[itid] = item
        return self._clone(items=itms)

    def add(self, item, pos=None):
        """
        Add a child item.

        :param item: Child item to be added
        :type item: WdbiNodeP
        :param pos: Position for the child item in the list
        :type pos: int
        :return: Updated version of the node
        :rtype: WdbiListNode
        """
        if pos is None:
            itms = self.__items.copy()
            itms[item.getUUID()] = item
        else:
            itms = OrderedDict()
            _itms = list(self.__items.values())
            for i in _itms[:pos]: itms[i.getUUID()] = i
            itms[item.getUUID()] = item
            for i in _itms[pos:]: itms[i.getUUID()] = i
        return self._clone(items=itms)

    def addAll(self, items):
        """
        Add multiple child items.

        :param items: Child items to be added
        :type items: Iterable[WdbiNodeP]
        :return: Updated version of the node
        :rtype: WdbiListNode
        """
        if not items: return self
        itms = self.__items.copy()
        for item in items:
            itms[item.getUUID()] = item
        return self._clone(items=itms)

    def map(self, mapF):
        """
        Apply a map-function to all child items.

        :param mapF: Mapping function
        :type mapF: (WdbiNodeP) -> WdbiNodeP
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        itms = map(mapF, self.__items.values())
        return self._clone(
            items=OrderedDict([(i.getUUID(), i) for i in itms]))

    def remove(self, item):
        """
        Remove a child item.

        :param item: Item to be removed
        :type item: WdbiNodeP
        :return: Updated version of the node
        :rtype: WdbiListNode
        """
        itms = self.__items.copy()
        del itms[item.getUUID()]
        return self._clone(items=itms)

    def clear(self):
        return self._clone(items=OrderedDict())

    def __len__(self):
        return len(self.__items)

    def __getitem__(self, item):
        return list(self.__items.values())[item]

    def __iter__(self):
        return iter(self.__items.values())

    def __reversed__(self):
        return reversed(self.__items.values())

    def __contains__(self, item):
        return item in self.__items

    def __add__(self, node):
        return self.add(node)

    def __sub__(self, node):
        return self.remove(node)

    def __hash__(self):
        s = self.getUUID()
        s += sum(hash(n.getUUID()) + hash(n) for n in self.__items.values())
        return s

    def __eq__(self, other):
        if not isinstance(other, WdbiListNode):
            return False
        if len(self) != len(other):
            return False
        return list(self.__items.values()) == list(other.__items.values())

    def __str__(self):
        return '[{}]'.format(
            ', '.join(map(repr, self.__items.values())))

    def __repr__(self):
        return 'WdbiListNode ({})'.format(
            ', '.join(map(repr, self.__items.values())))


# noinspection PyAbstractClass
class WdbiDictNode(WdbiNodeBase, WdbiVarCollectionNodeP):
    """
    Dictionary-based collection node of variable size.
    """
    __slots__ = WdbiNodeBase.__slots__ + ('_WdbiDictNode__items',)

    def __init__(self, uuid, ntyp, items=None):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param items: Child items
        :type items: dict[object, WdbiNodeP] | list[(object, WdbiNodeP)] | None
        """
        super().__init__(uuid, ntyp)

        #: Child items
        self.__items = self._createContainer(items)

    @classmethod
    def _createContainer(cls, items=None):
        """
        Create an internal dictionary storage container.

        :param items: Optional existing container
                      (may be returned if right type)
        :type items: object
        :return: Container
        :rtype: dict[WdbiNodeP]
        """
        return items if isinstance(items, dict) else {}

    def _cloneDictInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiDictNode
        """
        self._cloneNodeInto(new)

        items = kwargs.get('items')
        new.__items = items if items is not None else self.__items

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiDictNode.__new__(WdbiDictNode)
        self._cloneDictInto(new, **kwargs)
        return new

    @property
    def _items(self):
        """
        :return: Shallow copy of the dict's items
        :rtype: dict[WdbiNodeP]
        """
        return self.__items.copy()

    def getPath(self, path):
        if path:
            return self[path[0]].getPath(path[1:])
        else:
            return self

    def update(self, path, updateF):
        if path:
            item = self[path[0]].update(path[1:], updateF)
            return self.updatedItem(path[0], item)
        else:
            return updateF(self)

    def updatedItem(self, key, item):
        itms = self.__items.copy()
        itms[key] = item
        return self._clone(items=itms)

    def add(self, key, item):
        """
        Add an entry to the dictionary.

        :param key: Key for the new entry
        :type key: object
        :param item: Value
        :type item: WdbiNodeP
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        ndct = self.__items.copy()
        ndct[key] = item
        return self._clone(items=ndct)

    def addAll(self, itms):
        """
        Add multiple entries to the dictionary.

        :param itms: Key-/Value-pairs for the new entries
        :type itms: Iterable[(object,WdbiNodeP)] | dict[(object,WdbiNodeP)]
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        ndct = self.__items.copy()
        for key, item in (itms.items() if isinstance(itms, dict) else itms):
            ndct[key] = item
        return self._clone(items=ndct)

    def remove(self, key):
        """
        Remove an entry from the dictionary.

        :param key: Key of the entry to be removed
        :type key: object
        :return: Updated version of the node
        :rtype: WdbiNodeP
        """
        ndct = self.__items.copy()
        del ndct[key]
        return self._clone(items=ndct)

    def clear(self):
        return self._clone(items=self._createContainer())

    def get(self, key):
        return self.__items.get(key)

    def keys(self):
        return list(self.__items.keys())

    def children(self):
        return list(self.__items.values())

    def __len__(self):
        return len(self.__items)

    def __getitem__(self, key):
        return self.__items[key]

    def __iter__(self):
        return iter(self.__items.items())

    def __reversed__(self):
        return reversed(self.__items.items())

    def __contains__(self, key):
        return key in self.__items

    def __add__(self, kvp):
        return self.add(kvp[0], kvp[1])

    def __sub__(self, key):
        return self.remove(key)

    def __hash__(self):
        return sum(hash(k) + hash(v) for k, v in self.__items.items())

    def __eq__(self, other):
        if not isinstance(other, WdbiDictNode):
            return False
        if len(self) != len(other):
            return False
        return self.__items == other.__items

    def __str__(self):
        return '{{{}}}'.format(
            ', '.join(
                '{}:{}'.format(k, repr(v)) for k, v in self.__items.items()))

    def __repr__(self):
        return 'WdbiDictNode ({})'.format(
            ', '.join(
                '{}:{}'.format(k, repr(v)) for k, v in self.__items.items()))


# noinspection PyAbstractClass
class WdbiLeafNode(WdbiNodeBase):
    """
    Leaf-node without any children.
    """
    __slots__ = WdbiNodeBase.__slots__

    def __init__(self, uuid, ntyp):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        """
        super().__init__(uuid, ntyp)

    # noinspection PyUnusedLocal
    def _cloneLeafInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiLeafNode
        """
        self._cloneNodeInto(new)

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiLeafNode.__new__(WdbiLeafNode)
        self._cloneLeafInto(new, **kwargs)
        return new

    def getPath(self, path):
        if path:
            raise ValueError('LeafNodes have no children')
        return self

    def update(self, path, updateF):
        if path:
            from .wdbi_utils import wdbi_path2str

            raise ValueError(
                'LeafNodes have no children: ' + wdbi_path2str(path) +
                ' ' + str(self))
        return updateF(self)

    def __hash__(self):
        return hash(self.getUUID())

    def __eq__(self, other):
        return isinstance(other, WdbiLeafNode)

    def __str__(self):
        return '@' + str(self.getUUID())

    def __repr__(self):
        return 'WdbiLeafNode @{}'.format(self.getUUID())


class WdbiFieldNode(WdbiNodeBase):
    __slots__ = WdbiNodeBase.__slots__ + \
        ('_WdbiFieldNode__field', '_WdbiFieldNode__value')

    def __init__(self, uuid, ntyp, field, value=None):
        """
        :param uuid: UUID, if known
        :type uuid: UUID | str | None
        :param ntyp: Node type
        :type ntyp: int
        :param field: Field name
        :type field: str
        :param value: Initial field value
        :type value: object | None
        """
        super().__init__(uuid, ntyp)

        #: Field name
        self.__field = field
        #: Field value
        self.__value = value

    # noinspection PyUnusedLocal
    def _cloneFieldInto(self, new, **kwargs):
        """
        Clone the node's values into the new node.

        :param new: New node to clone values into
        :type new: WdbiFieldNode
        """
        self._cloneNodeInto(new)
        new.__field = self.__field
        new.__value = kwargs['value'] if 'value' in kwargs else self.__value

    def _clone(self, **kwargs):
        # noinspection PyArgumentList
        new = WdbiFieldNode.__new__(WdbiFieldNode)
        self._cloneFieldInto(new, **kwargs)
        return new

    def getPath(self, path):
        if path:
            raise ValueError('LeafNodes have no children')
        return self

    def update(self, path, updateF):
        if path:
            from .wdbi_utils import wdbi_path2str

            raise ValueError(
                'LeafNodes have no children: ' + wdbi_path2str(path) +
                ' ' + str(self))
        return updateF(self)

    def getFieldName(self):
        """
        Get the field name.

        :return: Field name
        :rtype: str
        """
        return self.__field

    def getValue(self):
        """
        Get the field value.

        :return: Field value
        :rtype: object | None
        """
        return self.__value

    def setValue(self, value):
        """
        Set the value of the filed node.

        :param value: New value
        :type value: object | None
        :return: Updated node
        :rtype: WdbiFieldNode
        """
        return self._clone(value=value)

    def __hash__(self):
        return hash(self.getUUID())

    def __eq__(self, other):
        return isinstance(other, WdbiLeafNode)

    def __str__(self):
        return '@' + str(self.getUUID())

    def __repr__(self):
        return 'WdbiFieldNode @{}'.format(self.getUUID())
