# ----------------------------------------------------------------------------
# - wdbm-instance                                                            -
# - wild cruncher:msh instance object               core / dbstore / mutable -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbm_instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher mutable instance objects

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.dbstore.mutable.wdbm_model import WdbmModel
from wcr_core.dbstore.mutable.wdbm_parameter import WdbmHasParametersMixin
from wcr_core.dbstore.mutable.wdbm_entity import WdbmEntityBase, \
    WdbmNamedEntityMixin
from wcr_core.dbstore.mutable.wdbm_connection_point import WdbmConnectionPoint
from wcr_core.dbstore_api.wdbs_instance_api import WdbsInstanceP, \
    WdbsBlockInstanceP, WdbsModelInstanceP, WdbsOperatorInstanceP
from wcr_core.dbstore_api.wdbs_observer_api import WDBS_CHILD_ADDED_EVENT, \
    WDBS_UPDATE_CORE_EVENT, WDBS_CHILD_REMOVED_EVENT


class WdbmInstanceBase(WdbmEntityBase, WdbmNamedEntityMixin, WdbsInstanceP):
    """
    The WdbmInstance class is the Silkan hierarchy model instance. An
    instance is a binding between a name and a model. The instance name is
    unique and can be optionally associated with an information string. The
    instance connectors are associated with the model connectors by port id.
    """

    def __init__(self, dctx, name=None, info=None, clst=None,
                 uuid=None, meta=None, **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param name: Instance name
        :type  name: str
        :param info: Instance info string
        :type  info: str
        :param clst: Connector list (type, id, meta)
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        """
        WdbmEntityBase.__init__(self, dctx, uuid, meta, **kwargs)
        WdbmNamedEntityMixin.__init__(self, name, info)

        #: the connection points list
        if clst is not None:
            xpnts = [WdbmConnectionPoint(dctx, t, i, **m)
                     for t, i, m in clst]
        else:
            xpnts = []
        self.__xPoints = xpnts
        """:type: list[WdbmConnectionPoint]"""

    def doDelete(self, recurse=True):
        for xpnt in self.__xPoints:
            xpnt.doDelete()
        super().doDelete()

    def getConnectionPoints(self):
        return iter(self.__xPoints)

    def getConnectionPointCount(self):
        return len(self.__xPoints)

    def createConnectionPoint(self, ctyp, uuid=None, meta=None, **kwargs):
        xpnt = WdbmConnectionPoint(
            self.context, ctyp, uuid=uuid, meta=meta, **kwargs)
        self.__xPoints.append(xpnt)
        self._notifyObservers(WDBS_CHILD_ADDED_EVENT, xpnt)
        return xpnt

    def setConnectionPointOrder(self, orderF, wrapFs):
        wrap, unwrap = wrapFs
        wrpd = [wrap(c) for c in self.__xPoints]
        wrpd.sort(key=orderF)
        self.__xPoints = [unwrap(c) for c in wrpd]
        self._notifyObservers(WDBS_UPDATE_CORE_EVENT)

    def removeConnectionPoint(self, xpoint):
        self.__xPoints.remove(xpoint)
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, xpoint)
        xpoint.doDelete()

    def clearConnectionPoints(self):
        xpnts = self.__xPoints
        self.__xPoints = []
        self._notifyObservers(WDBS_CHILD_REMOVED_EVENT, *xpnts)
        for xpnt in xpnts:
            xpnt.doDelete()


class WdbmBlockInstance(WdbmInstanceBase, WdbsBlockInstanceP):
    """
    Block instance.
    """

    def __init__(self, dctx, name=None, info=None, clst=None,
                 uuid=None, btyp=None, buuid=None, meta=None, bmeta=None,
                 **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param block: Instantiated block
        :type block: WdbsBlockP
        :param name: Instance name
        :type  name: str
        :param info: Instance info string
        :type  info: str
        :param clst: Connector list (type, id, meta)
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param btyp: Block type
        :type btyp: str
        :param buuid: Block-UUID
        :type buuid: UUID | str | None
        :param meta: Meta-data description for the instance
        :type meta: WdbsMetaDescriptionP | None
        :param bmeta: Meta-data description for the block
        :type bmeta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        """
        from wcr_core.dbstore.mutable.wdbm_block import WdbmBlock

        WdbmInstanceBase.__init__(
            self, dctx, name, info, clst, uuid,
            meta or dctx.getMetaDescription(self.__class__, btyp),
            **kwargs)

        #: Block
        self.__block = WdbmBlock(dctx, btyp, name, info, buuid, bmeta)

    def doDelete(self, recurse=True):
        if recurse:
            self.__block.doDelete(recurse)
        super().doDelete()

    def getBlock(self):
        return self.__block


class WdbmModelInstance(WdbmInstanceBase, WdbsModelInstanceP):
    """
    Model instance.
    """

    def __init__(self, dctx, model=None, name=None, info=None, clst=None,
                 uuid=None, mtyp=None, muuid=None, meta=None, mmeta=None,
                 **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param model: Instantiated model
        :type model: WdbsModelP
        :param name: Instance name
        :type  name: str
        :param info: Instance info string
        :type  info: str
        :param clst: Connector list (type, id, meta)
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param mtyp: Model type
        :type mtyp: str | None
        :param muuid: Model-UUID
        :type muuid: UUID | str | None
        :param meta: Meta-data description for the instance
        :type meta: WdbsMetaDescriptionP | None
        :param mmeta: Meta-data description for the model
        :type mmeta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        """
        WdbmInstanceBase.__init__(
            self, dctx, name, info, clst, uuid,
            meta or dctx.getMetaDescription(self.__class__, mtyp),
            **kwargs)

        if model and not isinstance(model, WdbmModel):
            raise ValueError(
                'Invalid model class: ' + model.__class__.__name__)

        #: Model
        self.__model = model if model else WdbmModel(
            dctx, mtyp, name=name, info=info, uuid=muuid, meta=mmeta)

    def doDelete(self, recurse=True):
        self.__model.doDelete()
        super().doDelete()

    def getModel(self):
        return self.__model


# noinspection PyAbstractClass
class WdbmOperatorInstance(WdbmInstanceBase, WdbmHasParametersMixin,
                           WdbsOperatorInstanceP):
    """
    Operator instance.
    """

    def __init__(self, dctx, oper, name=None, info=None, clst=None,
                 params=None, uuid=None, meta=None, status=None,
                 cards=None, statusMask=None, channelPrm=None,
                 arm=None, **kwargs):
        """
        :param dctx: Database context
        :type dctx: WdbsContextP
        :param oper: Operator code
        :type oper: str
        :param name: Instance name
        :type  name: str
        :param info: Instance info string
        :type  info: str
        :param clst: Connector list (type, id, meta)
        :type clst: Iterable[(str, UUID | None, dict[str, object])]
        :param params: Operator parameter values
        :type params: Iterable[(str, object)]
        :param uuid: UUID
        :type uuid: UUID | str | None
        :param meta: Meta-data description
        :type meta: WdbsMetaDescriptionP | None
        :param kwargs: Additional meta-data
        """
        WdbmHasParametersMixin.__init__(self, dctx, params)
        WdbmInstanceBase.__init__(
            self, dctx, name, info, clst, uuid,
            meta or dctx.getMetaDescription(self.__class__, oper), **kwargs)
        if cards is None:
            cards = {}
        #: Operator cards
        self.cardsOper = cards
        self.cardsInst = {}
        #: Operator code
        self.__oper = oper
        #: Operator status
        self.__status = status
        self.__statusMask = statusMask
        self.__arm = arm
        if channelPrm is None:
            channelPrm = {}
        self.__channelParameters = channelPrm

    def doDelete(self, recurse=True):
        WdbmHasParametersMixin.doDelete(self)
        super().doDelete()

    def getOperator(self):
        return self.__oper

    @property
    def getStatus(self):
        return self.__status

    def setStatus(self, status):
        self.__status = status

    @property
    def getStatusMask(self):
        return self.__statusMask

    def setStatusMask(self, statusMask):
        self.__statusMask = statusMask

    @property
    def getArm(self):
        return self.__arm

    def setArm(self, arm):
        self.__arm = arm

    def getChannelPrm(self):
        return self.__channelPrm