# ----------------------------------------------------------------------------
# - wdbs-protocols                                                           -
# - wild cruncher:bsh entities                            core / dbstore-api -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs_protocols
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher raw hierarchy entities

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.dbstore_api.wdbs_observer_api import WdbsObservableP


# noinspection PyAbstractClass
class WdbsEntityP(WdbsObservableP):
    """
    Abstract raw hierarchy entity base class.
    """
    __slots__ = tuple()

    @abstractmethod
    def getUUID(self):
        """
        Get the entity's UUID.

        :return: ID for the entity
        :rtype: UUID
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def metaDescription(self):
        """
        Get the entity's meta-data description.

        :return: Meta-data description
        :rtype: WdbsMetaDescriptionP
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def context(self):
        """
        Get the DBS context.

        :return: DBS context
        :rtype: WdbsContextP
        """
        raise NotImplementedError()

    @abstractmethod
    def getMeta(self, *keys):
        """
        Get the meta-data for the entity with the given key.

        :param key: Meta-data key-path
        :type key: list[object]
        :return: Meta-data, if any, or `None`
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def setMeta(self, value, *keys):
        """
        Set the meta-data for the entity for the given key.

        :param keys: Meta-data key-path
        :type keys: list[object]
        :param value: Meta-data value
        :type value: object
        """
        raise NotImplementedError()

    @abstractmethod
    def removeMeta(self, *keys):
        """
        Remove the meta-data for the given key from the entity.

        :param keys: Meta-data key-path
        :type keys: list[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def getMetaCount(self, *keys):
        """
        Get the number of children for the meta-data item at the given path.

        :param keys: Meta-data key-path
        :type keys: list[object]
        :return: Number of children for the meta-data item at the given path
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getMetaKeys(self, *keys):
        """
        Get the keys for the meta-data item at the given path, if it is a
        structural item.

        :param keys: Meta-data key-path
        :type keys: list[object]
        :return: Keys for the meta-data item at the given path
        :rtype: Iterable[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def getMetas(self, *keys):
        """
        Get the meta-data items at the given path as key-value pairs.

        :param keys: Meta-data key-path
        :type keys: list[object]
        :return: Meta-data items at the given path as key-value pairs
        :rtype: list[(object, object)]
        """
        raise NotImplementedError()

    @abstractmethod
    def getAllMeta(self):
        """
        Get all the entity's meta-data as a dictionary.

        :return: All meta-data values as dictionary
        :rtype: dict[str, object]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsTypedEntityP(WdbsEntityP):
    """
    Entity with type.
    """

    @abstractmethod
    def getEntityType(self):
        """
        Get the entity's type.

        :return: Type
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsNamedEntityP(WdbsEntityP):
    """
    Entity with name/info.
    """

    @abstractmethod
    def getName(self):
        """
        Get the entity's name.

        :return: Name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setName(self, name):
        """
        Set the entity's name.

        :param name: New name
        :type name: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getInfo(self):
        """
        Get the entity's info string.

        :return: Info string
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setInfo(self, info):
        """
        Set the entity's info string.

        :param info: New info string
        :type info: str
        """
        raise NotImplementedError()
