# ----------------------------------------------------------------------------
# - wdbs_meta_type                                                           -
# - wild cruncher: meta-data types                        core / dbstore-api -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wdbs-meta-type
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher meta-data types

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod


class WdbsMetaTypeP(metaclass=ABCMeta):
    """
    Protocol for meta-data types.
    """

    @abstractmethod
    def getStructName(self):
        """
        Return the structural name of the type.

        :return: Structural name of the type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getName(self):
        """
        Return the  name of the type.

        :return: Name of the type
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def setFromValue(self, setF, *args, **kwargs):
        """
        Set the value of the given meta-data instance from the provided
        arguments.

        :param setF: Set-function for meta-data (value, keys, create)
        :type setF: (object, *args) -> _
        :param args: Arguments to extract meta-value from
        :param kwargs: Keyword-arguments to extract meta-value from
        """
        raise NotImplementedError()

    @abstractmethod
    def asValue(self, meta):
        """
        Get the value of the given meta-data instance.

        :param meta: Meta-data instance
        :type meta: WdbsMetaP
        :return: The value of the given meta-data instance
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def getEmptyValue(self):
        """
        Provide an empty default value for the meta-data type.

        :return: Empty value
        :rtype: object
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsAtomicMetaTypeP(WdbsMetaTypeP):
    """
    Protocol for meta-data type for atomic values.
    """

    @property
    @abstractmethod
    def defaultValue(self):
        """
        Get the default value.

        :return: Default value
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def toString(self, metaV):
        """
        Get a string representation for the given meta-value.

        :param metaV: Meta-value
        :type metaV: object
        :return: String representation
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def fromString(self, inStr):
        """
        Convert the given string into a value of the meta-type.

        :param inStr: Input string
        :type inStr: str
        :return: Meta-type value
        :rtype: object
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsStructuralMetaTypeP(WdbsMetaTypeP):
    """
    Protocol for structural meta-data types.
    """

    @abstractmethod
    def getKey(self, item, data):
        """
        Get the key of the given item in the given structure.

        :param item: Item to look for
        :type item: object
        :param data: Data structure, corresponding to this type
        :type data: object
        :return: Key of the given item (corresponding to this type), if found
        :rtype: object | None
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsUnionMetaTypeP(WdbsStructuralMetaTypeP):
    """
    Protocol for meta-data type for union values.
    """

    @abstractmethod
    def getDefaultTypeId(self):
        """
        Get the default sub-type identifier, if none has been set.

        :return: Default sub-type identifier, or `None`
        :rtype: str | None
        """
        raise NotImplementedError()

    @abstractmethod
    def getType(self, typeId):
        """
        Get the meta-type for the given identifier.

        :param typeId: Type identifier
        :type typeId: str
        :return: Meta-type for the given identifier
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsRecordMetaTypeP(WdbsStructuralMetaTypeP):
    """
    Protocol for record meta-data types.
    """

    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    @abstractmethod
    def getKeys(self):
        """
        Get the keys for the record child items.

        :return: Keys for the record child items
        :rtype: Iterable[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getValueType(self, key):
        """
        Get the value-type of the child item with the given key.

        :param key: Child item key
        :type key: str
        :return: Value-type of the child item
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()

    @abstractmethod
    def getLabel(self, key):
        """
        Get the value-type of the child item with the given key.

        :param key: Child item key
        :type key: str
        :return: Value-type of the child item
        :rtype: str
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsTupleMetaTypeP(WdbsStructuralMetaTypeP):
    """
    Protocol for tuple meta-data types.
    """

    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    @abstractmethod
    def getValueType(self, index):
        """
        Get the value-type of the child item at the given index.

        :param index: Child item index
        :type index: int
        :return: Value-type of the child item
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsMultiMetaTypeP(WdbsStructuralMetaTypeP):
    """
    Protocol for container meta-data types.
    """

    @abstractmethod
    def getValueType(self):
        """
        Get the value-type of the child items.

        :return: Value-type of the child items
        :rtype: WdbsMetaTypeP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsListMetaTypeP(WdbsMultiMetaTypeP):
    """
    Protocol for list meta-data types.
    """
    pass


# noinspection PyAbstractClass
class WdbsDictionaryMetaTypeP(WdbsMultiMetaTypeP):
    """
    Protocol for dictionary/map meta-data types.
    """

    @abstractmethod
    def getKeyType(self):
        raise NotImplementedError()


# noinspection PyAbstractClass
class WdbsOrderedDictionaryMetaTypeP(WdbsDictionaryMetaTypeP):
    """
    Protocol for dictionary/map meta-data types that maintain insertion order.
    """
    pass
