# ----------------------------------------------------------------------------
# - whhdb-function                                                           -
# - wild cruncher:wsh hardware operator instance object                      -
# -                                                     core / hardware / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whhdb_Operator_Instance
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher hardware operator instance object

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""

from uuid import UUID

from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION
from wcr_core.db.whdb_instance import WhdbOperatorInstance
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.hardware.db.whhdb_Host_instance_meta import \
    WHHDB_HOST_META
from wcr_core.modules.hardware.db.whhdb_Host_instance_meta import \
    WHHDB_META_CTYPE_KEY, WHHDB_CARD_TYPE_KEY, WHHDB_META_BIND_KEY, \
    WHHDB_CARD_PORTS_KEY, WHHDB_META_IOCTYPE_KEY
from wcr_core.modules.hardware.db.whhdb_instance_base import WhhdbInstanceBaseP
from wcr_core.modules.hardware.db.whhdb_operator import WHHDB_OPER
from wcr_core.modules.hardware.db.whhdb_connection_points import \
     WHHDB_MODULES_IO_NAMES
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError

WHHDB_OP_PARAMS_KEY = 'params'
WHHDB_OPER_SWITCH_TYPE = 'switch' #metavalue to distinguish a switch
WHHDB_OPER_TYPE_META_KEY = 'type' #is the key needed in operator library
WHHDB_IO_SWITCH_CARD_KEY = 'IoCard'

class WhhdbOperatorInstance(WhdbOperatorInstance, WhhdbInstanceBaseP):
    META = WHHDB_HOST_META
    IO_TYPES = WHHDB_MODULES_IO_NAMES
    """
    The WhhdbOperatorInstance class is the Silkan hierarchical generic operator
    instance object. It is an operator model associated with a list of local
    parameters. operator type is defined for it. If the operator is not
    defined, it is assumed to be custom.
    """
    def __init__(self, inst, wreg):
        """
        :param inst: Raw operator instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Raw element wrapper registry
        :type wreg: WmhdbWrapperRegistry
        """
        self.IOCARD_TO_REMOVE = []
        WhdbOperatorInstance.__init__(
            self, WHHDB_OPER[inst.getOperator()], inst, wreg)
        WhhdbInstanceBaseP.__init__(self, inst, wreg)

        self.__rawInstance = inst
        self.__wrapperRegistry = wreg

    def getLabelAlignment(self):
        return self.getOperator().getLabelAlignment()

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if self.canCreateConnectionPoint():
            if WHDB_XPOINT_DIRECTION not in kwargs:
                kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
            return super()._createRawConnectionPoint(ctyp, **kwargs)
        return None

    def removeConnectionPoint(self, xpnt):
        if self.canRemoveConnectionPoint(xpnt):
            super().removeConnectionPoint(xpnt)

    def getNonCardConnectionPoints(self):
        points = self.getConnectionPoints()
        nonCardPoints = []
        for point in points:
            if point not in self.getAllIoConnectionPoints():
                nonCardPoints.append(point)
        return nonCardPoints

    def getAllIoConnectionPoints(self):
        ioCards = self.getIoCards()
        IoCardsPoints = []
        for card in ioCards:
            IoCardsPoints.extend(self.getIoCardConnectionPoints(card))

        return IoCardsPoints

    def getCards(self):
        return self.getIoCards()

    def getIoTypes(self):
        return self.IO_TYPES

    def getIoCards(self):
        """
        return every non marked card
        """
        cards = sorted(self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys())
        for rmc in self.IOCARD_TO_REMOVE:
            cards.remove(rmc)
        return cards

    def setIoCard(self, card):
        """
        initialising a new card or remove from deleted if already exist
        """
        if not card in \
                self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            if not isNameValid(card):
                showErrorWindow("Name Error", NameFormatMessageError,
                                card)
            self.rawElement.setMeta({WHHDB_CARD_PORTS_KEY: []},
                                    WHHDB_META_IOCTYPE_KEY, card)
        else:
            if card in self.CARD_TO_REMOVE:
                self.CARD_TO_REMOVE.remove(card)

    def getModuleIoFromCard(self, cardname):
        modules = []
        for module in self.getIoCards():
            cardN = self.getIoCardBind(module)
            if cardN == cardname:
                modules.append(module)
        return modules

    def checkPortNumberConsistency(self):
        portNumbers = []
        for point in self.getNonCardConnectionPoints():
            portNum = point.getPortNum()
            if portNum in portNumbers:
                msg = "Port consistency error, " \
                      "multiple ports have the same port number " \
                      "for the same card.\n" \
                      "Please check port {}.".format(point.getName())
                return False, msg
            portNumbers.append(portNum)
        modulesIo = self.getModuleIoFromCard(WHHDB_IO_SWITCH_CARD_KEY)
        for moduleIo in modulesIo:
            pointsIo = self.getIoCardConnectionPoints(moduleIo)
            for pointIo in pointsIo:
                portIoNum = pointIo.getPortNum()
                if portIoNum in portNumbers:
                    msg = "Port consistency error, " \
                          "multiple ports have the same port number "\
                          "for the same card.\n"\
                          "Please check port {}.".format(point.getName())
                    return False, msg
                portNumbers.append(portIoNum)
        return True, None

    def getIoCardBind(self, cardname):
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return ''
        bind = self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY, cardname,
                                       WHHDB_META_BIND_KEY)
        return bind

    def setIoCardBind(self, card, bind):
        if not card in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        self.rawElement.setMeta(bind,
                                WHHDB_META_IOCTYPE_KEY, card,
                                WHHDB_META_BIND_KEY)

    def setIoCardPtpMaster(self, card, status):
        return super().setCardPtpMaster(card, status)

    def replaceIoCard(self, hcardname, ncard):
        """
        replace the card hcardname by the ncard.
        :param hcardname: name ot the card ho will be replaced
        :type hcardname: str
        :param ncard: description of the full new card
        :type ncard: [cardname, cardtype, [portsbinded]]
        """
        if hcardname is not '' or hcardname is not None:
            self.rawElement.removeMeta(WHHDB_META_IOCTYPE_KEY, hcardname)
        self.setIoCard(ncard[0])
        self.setIoCardBind(ncard[0], ncard[1])
        for puuid in ncard[2]:
            self.addPortUUIDToIoCard(ncard[0], puuid)

    def getBindedIoPorts(self, cardname):
        """
        Return the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return []
        return self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY, cardname,
                                       WHHDB_CARD_PORTS_KEY)

    def addPortUUIDToIoCard(self, cardname, cuuid):
        """
        add cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        buuid = self.getBindedIoPorts(cardname)
        if buuid:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
            buuid.append(cuuid)
        else:
            if cuuid is list:
                for elem in cuuid:
                    if not isinstance(elem, UUID):
                        raise ValueError(
                            "Every element of the list must be UUIDs")
                buuid = cuuid
            else:
                if not isinstance(cuuid, UUID):
                    raise ValueError(
                        "Invalid port reference Cards "
                        "and port are binded by UUID")
                buuid = [cuuid]
        self.rawElement.setMeta(buuid, WHHDB_META_IOCTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def removeIoCardPort(self, cardname, cuuid):
        """
        remove cuuid from the list of binded port to cardname
        """
        if not cardname in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            return None
        buuid = self.getBindedIoPorts(cardname)

        if cuuid in buuid:
            buuid.remove(cuuid)

        self.rawElement.setMeta(buuid, WHHDB_META_IOCTYPE_KEY, cardname,
                                WHHDB_CARD_PORTS_KEY)

    def getIoCardsDescription(self):
        """
        retrn the full description of the card stocked in metadata
        :return : [cardname, cardtyp, [bindedports]]
        """
        fullcards = []

        for card in self.getIoCards():
            structcards = (card,
                           self.getIoCardBind(card),
                           self.getBindedIoPorts(card))
            fullcards.append(structcards)
        return fullcards

    def getIoCardConnectionPoints(self, card):
        """
        git every connectionpoint binded to card. If card is None return
        every connexion point of Host.
        :return : [ConnectionPoint]
        """
        if not isinstance(card, str):
            return []
        cbpl = self.getBindedIoPorts(card)
        cpl = self.getConnectionPoints()
        res = []
        for cp in cpl:
            if cp.getUUID() in cbpl:
                res.append(cp)
        return res

    def bindFreeIoPorts(self, card):
        """
        Bind to card every connection who isn't allready binded.
        """
        if not card:
            return None
        iobcpuuidl = []
        #get port binded in io card
        for c in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            for cpuuid in self.getBindedIoPorts(c):
                iobcpuuidl.append(cpuuid)
        bcpuuidl = []
        cpl = self.getConnectionPoints()
        for cp in cpl:
            if cp.getUUID() not in bcpuuidl and cp.getUUID() not in iobcpuuidl:
                self.addPortUUIDToIoCard(card, cp.getUUID())

    def getIoConnectionDistributionTypeByUUID(self, uuidl):
        """
        return type distribution of ports having UUID in uuidl
        :return : {str: int}
        """
        distrib={}
        if not isinstance(uuidl, list):
            return distrib
        for c in self.getConnectionPoints():
            if c.getUUID() in set(uuidl):
                if c.getCode() in self.IO_TYPES:
                    tn = c.getCode()
                else:
                    raise Exception('Connector type is unknown !')
                if not tn in distrib.keys():
                    distrib[tn] = 1
                else:
                    distrib[tn] += 1
        return distrib

    def unbindInexistentIoPort(self):
        """
        unbind every UUID of cards port list if they don't
        correspond to a connection point
        """
        cpl = self.getConnectionPoints()
        cpuuidl = []
        for cp in cpl:
            cpuuidl.append(cp.getUUID())
        for c in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            for cpuuid in self.getBindedIoPorts(c):
                if not cpuuid in cpuuidl:
                    self.removeIoCardPort(c, cpuuid)

    def markIoCard(self, cardn):
        """
        mark cardn as card to remove
        """
        if cardn in self.rawElement.getMeta(WHHDB_META_IOCTYPE_KEY).keys():
            self.IOCARD_TO_REMOVE.append(cardn)

    def delIoCardandPorts(self):
        """
        Remove card marked to remove and they binded ports
        """
        dcpl = []
        for rmcn in self.IOCARD_TO_REMOVE:
            cpl = self.getIoCardConnectionPoints(rmcn)
            for cp in cpl:
                dcpl.append(cp)
        for crds in self.getIoCards():
            cpl = self.getIoCardConnectionPoints(crds)
            for cp in cpl:
                if cp in dcpl:
                    dcpl.remove(cp)
        for dcp in dcpl:
            self.removeConnectionPoint(dcp)
        self.delIoCards()

    def delIoCards(self):
        """
        cleaning metada from card marked to remove and asociated information
        """
        portl = []
        for c in self.IOCARD_TO_REMOVE:
            getCon = getattr(self, 'getCardConnectionPoints', None)
            if getCon:
                portl += getCon(c)
            self.rawElement.removeMeta(WHHDB_META_IOCTYPE_KEY, c)
        self.IOCARD_TO_REMOVE = []

    def setStatus(self, status):
        self.rawElement.setStatus(status)

    def getStatus(self):
        return self.rawElement.getStatus

    def setArmPort(self, status):
        self.rawElement.setArm(status)

    def getArmPort(self):
        return self.rawElement.getArm

    def setStatusMask(self, statusMask):
        self.rawElement.setStatusMask(statusMask)

    def getStatusMask(self):
        return self.rawElement.getStatusMask

    def resetIcon(self):
        """
        Reset the operator's icon to its original size.
        """
        self.unsetDimension()

    def getOperatorMetaType(self):
        """
        will return a metadata precising the nature or the operator.
        """
        return self.getOperator().getMeta(WHHDB_OPER_TYPE_META_KEY)

    def hasFlag(self, flag, *_):
        return self.getOperator().hasFlag(flag)

    def __str__(self):
        return 'WhhdbOperatorInstance ' \
               '{{oper: {op}, cns: {ic}}}'.format(
            op=self.getOperator().getCode(),
            ic=[c.getName() for c
                in self.getConnectionPoints()])
