# ----------------------------------------------------------------------------
# - wihdb-operator                                                           -
# - wild cruncher:wsh ICD operator object                    core / icd / db -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wihdb_operator
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model function object

.. moduleauthor:: philippe lambert <lambert@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------
from collections import OrderedDict
from wcr_core.db.whdb_operator import WhdbOperatorRegistry, WhdbOperator


# ----------------------------------------------------------------------------
# - base operator class                                                      -
# ----------------------------------------------------------------------------


class WihdbOperator(WhdbOperator):
    """
    The WcrOperator class is the silkan hierarchical operator object. An
    operator is defined by name and info. It is the operator name which
    defines its functionality. The operator contains an icon representation
    as well as a collection of connectors.
    """
    # TODO: Do not extend HasMediaP

    def __init__(self, code, xmngr, name=None, info=None,
                 meta=None, parameters=None, flags=None,
                 icons=None, shape=None, symbol=None, labelAlign=None,):
        """
        Create a new operator by name and info

        :param code: Operator code
        :type code: str
        :param xmngr: Connector manager
        :type xmngr: WhdbConnectionPointManagerP
        :param name: the operator name
        :type name: str
        :param info: the operator info
        :type info: str
        :param meta: Operator meta data
        :type meta: dict[str, object]
        :param parameters: Operator parameters
                           (name, label, info, value, type, flags)
        :type parameters: list[(str, str, str, str|None, str|None, int|None)]
        :param flags: Flags
        :type flags: set[str]
        :param icons: Operator icon paths
        :type icons: dict[str, str]
        :param shape: Operator shape
        :type shape: list[(float, float)]
        :param symbol: Text symbol
        :type symbol: str
        :param labelAlign: Alignment for labels
        :type labelAlign: int
        """
        super().__init__(code, xmngr, name, info, meta, parameters,
                         flags, icons, shape, symbol, labelAlign)

        self.__entries = {}
        self.__moduleType = None

    def setEntries(self, entries):
        self.__entries = entries

    def setMsgEntry(self, msgName, msgData):
        self.__entries[msgName] = msgData

    def getEntries(self):
        return self.__entries

    def getData(self):
        return self.__entries

    def setModuleType(self, moduleType):
        self.__moduleType = moduleType

    def getModuleType(self):
        return self.__moduleType


class WihdbOperatorRegistry(object):
    """
    Registry for operators.
    """

    def __init__(self):
        #: Registered operators
        self.__items = OrderedDict()

    def __getitem__(self, item):
        try:
            return self.__items[item]
        except KeyError:
            raise KeyError('COULD NOT FIND OPERATOR {}'.format(item))

    def __contains__(self, item):
        if not item:
            return False
        return item in self.__items

    def __iter__(self):
        return iter(self.__items.values())

    def getItems(self):
        return self.__items

    def _clear(self):
        self.__items.clear()

    def addOperator(self, instance):
        """
        Add an instance to the registry.

        :param instance: instance to add
        :type instance: WihdbModuleInstance
        """
        code = instance.getCode()
        self.__items[code] = instance

    def deleteOperator(self, instance):
        """
        delete an instance to the registry.

        :param instance: instance to delete
        :type instance: WihdbModuleInstance
        """
        code = instance.getCode()
        del self.__items[code]

    def addOperators(self, operators):
        """
        Add a list of operators to the registry.

        :param operators: Operators to add
        :type operators: Iterable[WhdbOperatorP]
        """
        for operator in operators:
            self.addOperator(operator)

    def deleteOperators(self, operators):
        """
        delete a list of operators to the registry.

        :param operators: Operators to delete
        :type operators: Iterable[WhdbOperatorP]
        """
        for operator in operators:
            self.deleteOperator(operator)

    def get(self, item):
        """
        Get the operator with the given code.

        :param item: Operator code
        :type item: str
        :return: Operator with the given code
        :rtype: WhdbOperatorP
        """
        return self.__items.get(item)

    def __str__(self):
        return str([k for k in self.__items.keys() if '.' in k])


#: Global HDS function operator registry
WIHDB_OPER = WihdbOperatorRegistry()
