# ----------------------------------------------------------------------------
# - wmctl-controllers                                                        -
# - wild cruncher: model controllers               core / model / controller -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmctl_controllers
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Model controllers

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from wcr_core.controller.wctl_controller_mixins import *
from wcr_core.controller.wctl_default_controllers import \
    WctlSimpleControllerFactoryBase
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION, \
    WHDB_XPOINT_FACING
from wcr_core.db.whdb_operator import whdb_makeConnXYPositions
from wcr_core.db.whdb_protocols import WHDB_POSITION_KEY
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_OUT


class WmctlControllerBase(WctlControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WmctlControllerFactory
        """
        self.__factory = fcty

    @property
    def factory(self):
        return self.__factory


class WmctlInstanceController(WmctlControllerBase,
                              WctlInstanceControllerDefaultXPointMixin,
                              WctlInstanceControllerDefaultDeleteMixin,
                              WctlInstanceControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WmctlControllerFactory
        """
        super().__init__(fcty)

    def canCreateConnectionPoint(self, instance, **kwargs):
        """
        :type instance: WhdbInstanceP
        :rtype: bool
        """
        cdir = kwargs.get(WHDB_XPOINT_DIRECTION)
        if not cdir:
            cdir = instance.getNewConnectionPointDirection(**{
                WHDB_POSITION_KEY: kwargs.get('cpos'),
                WHDB_XPOINT_FACING: kwargs.get('fcng')})

        if cdir is None:
            return False

        return instance.canCreateConnectionPoint(**kwargs)

    def createConnectionPoint(self, instance, valuePrv, **kwargs):
        """
        :type instance: WhdbInstanceP
        :type valuePrv: WctlInstanceControllerValueProviderP
        :type kwargs: dict[str, object]
        :rtype: WmhdbConnector
        """
        cpos = kwargs.get('cpos')
        fcng = kwargs.get('fcng')
        cdir = kwargs.get(WHDB_XPOINT_DIRECTION)
        if not cdir:
            cdir = instance.getNewConnectionPointDirection(**{
                WHDB_POSITION_KEY: cpos,
                WHDB_XPOINT_FACING: fcng})
            kwargs[WHDB_XPOINT_DIRECTION] = cdir

        if cpos is None or fcng is None:
            (x, y, f), *_ = whdb_makeConnXYPositions(cdir, 1)
            if cpos is None:
                kwargs['cpos'] = x, y
            if fcng is None:
                kwargs['fcng'] = f

        kwargs['gvty'] = kwargs['fcng']
        assert self.canCreateConnectionPoint(instance, **kwargs)

        return instance.createConnectionPoint(**kwargs)


class WmctlConnectionController(WmctlControllerBase,
                                WctlConnectionControllerDefaultCreateMixin,
                                WctlConnectionControllerDefaultDeleteMixin,
                                WctlConnectionControllerP):
    def __init__(self, fcty):
        """
        :param fcty: Controller factory
        :type fcty: WmctlControllerFactory
        """
        super().__init__(fcty)

    def canCreateConnection(self, block, sinst, einst, sxpnt=None, expnt=None,
                            **kwargs):
        noNones = sinst and einst and sxpnt and expnt
        if not noNones:
            return False
        if sxpnt not in sinst.getConnectionPoints():
            return False
        if expnt not in einst.getConnectionPoints():
            return False

        sdir = sxpnt.getDirection()
        edir = expnt.getDirection()

        return (sdir == WDBS_XPNT_IN and edir == WDBS_XPNT_OUT) or\
            (sdir == WDBS_XPNT_OUT and edir == WDBS_XPNT_IN)


class WmctlControllerFactory(WctlSimpleControllerFactoryBase):
    def __init__(self):
        super().__init__(
            instCtlr=WmctlInstanceController(self),
            connCtlr=WmctlConnectionController(self))


#: Model controller factory
WMCTL_CONTROLLER_FACTORY = WmctlControllerFactory()
