# ----------------------------------------------------------------------------
# - wmmod-application-module                                                 -
# - wild cruncher:wsh application module                        core / model -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wmmod_application_module
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher model application module

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod

from wcr_core.action.wact_action_api import WACT_WCR_ACTION_CODE, \
    WACT_ACTION_W_DB_VERIFY
from wcr_core.db_api.whdb_block_api import WHDB_TOPLEVEL_BLOCK_CODE, \
    WHDB_DEFAULT_BLOCK_CODE
from wcr_core.db.whdb_top_level_factory import whdb_createTopLevelRawBlock
from wcr_core.modules.model.actions.wmact_action_base import \
    WMACT_ACTION_W_TYPE_PROP
from wcr_core.modules.model.actions.wmact_propagate_types_action import \
    WMACT_PROPAGATE_TYPES_ACTION
from wcr_core.modules.model.controller.wmctl_controllers import \
    WMCTL_CONTROLLER_FACTORY
from wcr_core.modules.model.db.wmhdb_block import WmhdbBlock
from wcr_core.modules.model.db.wmhdb_meta_provider import WmhdbMetaProvider
from wcr_core.modules.model.db.wmhdb_top_block import WmhdbTopHierarchy
from wcr_core.modules.model.db.wmhdb_wrapper_registry import \
    WmhdbWrapperRegistry
from wcr_core.modules.wmod_block_type import WmodBlockType, \
    WMOD_SYSTEM_BLOCK_FLAG
from wcr_core.project.wprj_project_context import WprjProjectContext
from wcr_core.modules.wmod_application_module import \
    WmodApplicationModuleBase, WmodApplicationModuleP, WmodApplicationModuleMP
from wcr_core.common.wcmn_mime_type import WCMN_PAYLOAD_BLOCK_MODEL
from wcr_core.modules.model.db.wmhdb_verifier import WmhdbHierarchyVerifier
from wcr_core.modules.model.db.wmhdb_operator import WMHDB_OPER
from wcr_core.modules.model.db.wmhdb_standard_operators import \
    WMHDB_DEFAULT_OPS


WMMOD_MODEL_APP_CODE = 'model'

WMMOD_BLOCK_TYPES = (
    WmodBlockType(WHDB_TOPLEVEL_BLOCK_CODE, 'Top-level',
                  WmhdbTopHierarchy.META, WmhdbTopHierarchy,
                  WMOD_SYSTEM_BLOCK_FLAG),
    WmodBlockType(WHDB_DEFAULT_BLOCK_CODE, 'Block',
                  WmhdbBlock.META, WmhdbBlock)
)
""":type: tuple[WmodBlockTypeP]"""


# noinspection PyAbstractClass
class WmmodModelApplicationModuleP(WmodApplicationModuleP):
    @abstractmethod
    def getExecutionFactories(self):
        """
        Get all scheduling group execution factories.

        :return: Scheduling group execution factories
        :rtype: Iterable[WmmodSgroupExecutionFactoryP]
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WmmodModelApplicationModuleMP(WmodApplicationModuleMP,
                                    WmmodModelApplicationModuleP):
    @abstractmethod
    def addExecutionFactory(self, executor):
        """
        Add a scheduling group execution factory.

        :param executor: Scheduling group execution factory to be added
        :type executor: WmmodSgroupExecutionFactoryP
        """
        raise NotImplementedError()


class _WmmodModelApplicationModule(WmodApplicationModuleBase,
                                   WmmodModelApplicationModuleMP):
    def __init__(self, operRegistry):
        super().__init__(
            code=WMMOD_MODEL_APP_CODE,
            name='Model',
            blockTypes=WMMOD_BLOCK_TYPES,
            defaultOps=WMHDB_DEFAULT_OPS,
            operRegistry=operRegistry)

        self.__sgrpExecFactories = []
        """:type: list[WmmodSgroupExecutionFactoryP]"""

        self._createDefaultActions()

    def _createDefaultActions(self):
        actnSection = '{}.core.{}'.format(
            WACT_WCR_ACTION_CODE, WMMOD_MODEL_APP_CODE)
        self.addAction(
            WMACT_PROPAGATE_TYPES_ACTION, 'proptyps', actnSection,
            WACT_ACTION_W_DB_VERIFY, WMACT_ACTION_W_TYPE_PROP)

    def addExecutionFactory(self, executor):
        self.__sgrpExecFactories.append(executor)

    def getExecutionFactories(self):
        return self.__sgrpExecFactories

    def createHierarchyVerifier(self, pctx):
        return WmhdbHierarchyVerifier(self._getDbVerifierModules())

    def getPayloadType(self):
        return WCMN_PAYLOAD_BLOCK_MODEL

    def createTopLevelBlock(self, name=None, info=None):
        blockMetas = {b.getCode():b.getMeta() for b in self.getBlockTypes()}
        blockClzzs = {b.getCode():b.getClass() for b in self.getBlockTypes()}
        return wmmod_create_top_block(
            blockMetas, blockClzzs, name=name, info=info)

    def createProjectContext(self, gContext, projectDB, logger=None):
        return WprjProjectContext(
            gContext, self, projectDB, logger, WMCTL_CONTROLLER_FACTORY)


def wmmod_create_top_block(blockMetas, blockClzzs, name=None, info=None):
    tblk = whdb_createTopLevelRawBlock(
        WmhdbMetaProvider(blockMetas), name, info)
    rgty = WmhdbWrapperRegistry(blockClzzs)
    return rgty.createWrapper(tblk)


#: HDS Application module singleton
WMMOD_MODEL_APP_MODULE = _WmmodModelApplicationModule(WMHDB_OPER)
