# ----------------------------------------------------------------------------
# - wcact-generate-file-action                                               -
# - wild cruncher: generate file action             core / sys_map / actions -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 Philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcact_generate_file_action
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - file generation action

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

import os
import subprocess
from xml.etree import ElementTree

from PyQt5.QtWidgets import QCheckBox, QVBoxLayout
from wcr_core.action.wact_action_parameter import WactActionParameter
from wcr_core.action.wact_action_parameter_types \
    import WactActionDirectoryParameterType
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_CETRAC_CODE
from wcr_core.modules.sys_map.actions.wcact_action_base import WcactActionBase
from wcr_core.modules.sys_map.plugin.wcplg_file_config_generator import \
    createConfigurationFileText
from wcr_frontend import showErrorWindow
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.modules.sys_map.gui.wcgui_utils import wcgui_ask_source, \
    wcgui_ask_arp


# WCACT_ACTION_BIND_VERIFY='bindverif'
WCACT_TARGET_GENE_DIR_PARAMETER = WactActionParameter(
    'dest', 'Output directory', 'XML output directory',
    WactActionDirectoryParameterType())


def wcld_get_ccr_path():
    """
    Get the default ccr path
    """
    return 'EMPTY'


def wcld_get_ccr_xml_path():
    """
    Get the default ccr xml path
    """
    return 'EMPTY'


def indent(elem, level=0):
    i = "\n" + level*"  "
    j = "\n" + (level-1)*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for subelem in elem:
            indent(subelem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = j
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = j
    return elem


class WchdbBindsException(Exception):
    def __init__(self, elements, message):
        super().__init__(message)

        self.__elements = elements

    @property
    def elements(self):
        return iter(self.__elements)


class ConfiguratorLaunch(WguiPropertiesWidgetP):

    def __init__(self, options):
        super(ConfiguratorLaunch, self).__init__()
        self.setGeometry(30, 30, 400, 200)
        self.setWindowTitle("Cetrac Configurator Launcher")
        self.options = options
        print(self.options)
        self.checkBoxes = []
        self.setLayout(self._createLayout())

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        for option in self.options:
            checkBox = QCheckBox(option, self)
            layout.addWidget(checkBox)
        return layout

    def lauch(self,fpath, sport):
        subprocess.run(["{}/{}".format(fpath, CETRAC_CONFIGURATOR_EXE),
                        '-i', '"{}"'.format(fpath),
                        '-o', '"{}"'.format(fpath),
                        '-e', '"{}"'.format(sport.getName())
                        ], stdout=subprocess.PIPE)

    def updateFields(self):
        pass

    def save(self):
        pass


class _WcactGenerateFileAction(WcactActionBase):
    """
    Project editor action to reload the hardware source.
    """

    def __init__(self):
        super().__init__(
            'Generate config files',
            'Try to Generate config files for cetrac architecture')

    def _isProtocolException(self, e):
        return isinstance(e, WchdbBindsException)

    def _doExecute(self, actx, paramValuePairs):
        pctx = actx.getProjectContext()
        ok = False
        while not ok:
            # ask for folder where to create file
            dirPath = self._getParamValue(
                WCACT_TARGET_GENE_DIR_PARAMETER, actx, paramValuePairs)
            if not dirPath:
                actx.statusMessage("Generation canceled : "
                                   "No target directory selected.")
                return
            name = "cetracConfiguration.xml"
            fpath = os.path.join(dirPath, name)
            if os.path.exists(fpath):
                showErrorWindow('File Duplication',
                                'A configuration file exist in this folder, '
                                'please rename it or delete it to save new file.',
                                None)
            else:
                ok = True

        # create structure for config files
        xmltext, errs = createConfigurationFileText(pctx, mport=None)
        # prettyXml = BeautifulSoup(xmltext, "xml").prettify()
        # create files.
        targfile = open(fpath, 'w')
        targfile.write(xmltext)
        targfile.close()

        root = ElementTree.parse(fpath).getroot()
        indent(root)
        # ElementTree.write(root)
        targfile = open(fpath, 'w')
        targfile.write(ElementTree.tostring(root).decode('utf-8'))
        targfile.close()

        actx.statusMessage("Generation done at : " + fpath)
        if len(errs) >0:
            showErrorWindow('Configuration Warning',
                            'The file has been generated but some warning were found, '
                            'you can look at them in the system logger.', None)
            actx.statusMessage("Warnings:")
            for err in errs:
                actx.statusMessage(err)

        self.call_configurator(pctx, actx, dirPath)

        return None

    def call_configurator(self, pctx, actx, dirPath):
        ccr_exe = pctx.getConfigurationValue('DEZYGNR_KONFYGURATR_EXE')
        if ccr_exe == 'EMPTY':
            return

        (eq_port, sport, eqp_name), cardName = getConfigPort(pctx)
        cardName = '{}.{}'.format(eqp_name, cardName)
        if not sport:
            actx.statusMessage("Konfyguratr canceled : "
                               "No source port selected.")
            return
        try:
            xml_path = os.path.join(dirPath, 'cetracConfiguration.xml')
            ccf_path = os.path.join(dirPath, 'cetracConfiguration.ccf')
            myddl_path = os.path.join(dirPath, 'mydleware')
            subprocess.call(['{}'.format(ccr_exe),
                             "-i", xml_path,
                             "-o", ccf_path,
                             "-m", myddl_path,
                             "-e", '{}'.format(cardName)])
        except Exception as e:
            message = e.strerror
            showErrorWindow('Konfyguratr Error',
                            message, None)
        else:
            message = "Configuration success !"

        actx.statusMessage(message)


def getRackInfo(pctx):
    map = pctx.getMappingManager()
    racks = map.getRacks()
    hasRack = False
    if len(racks):
        hasRack = True
    return hasRack


def getConfigPort(pctx):

    def formatdata(data):
        return data[0]

    map = pctx.getMappingManager()
    possibilities = map.getHosts()

    datas = []
    # get all CETRAC Cards to select entry point for config
    for host in possibilities:
        for card in host.getCards():
            for pp in host.getCardConnectionPoints(card):
                datas.append((map.formatName(pp, prefix=card), pp, host.getName()))

    psrc = wcgui_ask_source(datas, formatdata)
    try:
        entryCard = psrc[0].split('-')[0]
    except:
        return (None, None, None), None
    return psrc, entryCard


#: Logical simulation reload action
WCACT_GENERATE_FILE_ACTION = _WcactGenerateFileAction()
