# ----------------------------------------------------------------------------
# - wphdb-bus-instance                                                     -
# - wild cruncher: bus instances                         core / system / db -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wphdb_bus_instance
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Bus instance

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_DIRECTION,\
    WHDB_XPOINT_NAME
from wcr_core.db.whdb_instance import WhdbOperatorInstance
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_BI
from wcr_core.modules.system.db.wphdb_phybus_meta import \
    WPHDB_BUS_META_DESCRIPTION, WPHDB_BUS_BOUND_LBUS_KEY
from wcr_core.modules.system.db.wphdb_standard_operator import WPHDB_BUS, \
    WPHDB_BUS_CODE

WPHDB_XPNT_NAME_PATTERN = 'C{id}'


class WphdbPhyBusInstance(WhdbOperatorInstance):
    """
    Operator-instance for bus instances.
    """

    META = WPHDB_BUS_META_DESCRIPTION

    def __init__(self, inst, wreg):
        """
        :param inst: Raw instance
        :type inst: WdbsOperatorInstanceP
        :param wreg: Wrapper registry
        :type wreg: WhdbWrapperRegistryP
        """
        assert inst.getOperator() == WPHDB_BUS_CODE
        WhdbOperatorInstance.__init__(self, WPHDB_BUS, inst, wreg)

    def getBoundLogicalBusesUUID(self):
        """
        :return: List of logical buses UUIDs bound to the physical bus
        :rtype: list[UUID]
        """
        return self.rawElement.getMeta(WPHDB_BUS_BOUND_LBUS_KEY)

    def addBoundLogicalBus(self, logBus):
        """
        Add the logical bus to the bound list of the physical bus unless
        it is already bound, in which case nothing is done

        :param logBus: Logical bus to be bound
        :type logBus: WshdbBusInstance
        """
        if self.isBound(logBus):
            return

        self._addBoundLogicalBusUUID(logBus.getUUID())

    def _addBoundLogicalBusUUID(self, lbUUID):
        """
        Add the logical bus (assumed) UUID to the list without testing

        :param lbUUID: UUID to add
        :type lbUUID: UUID
        """
        listBoundLogBuses = self.getBoundLogicalBusesUUID()
        listBoundLogBuses.append(lbUUID)
        self.rawElement.setMeta(listBoundLogBuses, WPHDB_BUS_BOUND_LBUS_KEY)

    def isBound(self, logBus):
        """
        :param logBus: Logical bus to be bound
        :type logBus: WshdbBusInstance
        :return: True if the passed UUID is already present, False otherwise
        :rtype: bool
        """
        lbUUID = logBus.getUUID()
        for lb in self.getBoundLogicalBusesUUID():
            if lb == lbUUID:
                return True
        return False

    def removeBoundLogicalBus(self, logBus):
        """
        Remove the passed logical bus from the bound list
        Raises ValueError if the logical bus is not found to be bound

        :param logBus: Logical bus to be bound
        :type logBus: WshdbBusInstance
        """
        if not self.isBound(logBus):
            raise ValueError('Cannot remove logical bus: '
                             'Logical bus not bound')

        listBoundLogBuses = self.getBoundLogicalBusesUUID()
        listBoundLogBuses.remove(logBus.getUUID())
        self.rawElement.setMeta(listBoundLogBuses, WPHDB_BUS_BOUND_LBUS_KEY)

    def _createRawConnectionPoint(self, ctyp, **kwargs):
        if WHDB_XPOINT_DIRECTION not in kwargs:
            kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
        return super()._createRawConnectionPoint(ctyp, **kwargs)

    def createConnectionPoint(self, ctyp=None, **kwargs):
        return super().createConnectionPoint(
            ctyp, **_wphdb_setConnectorMeta(self, **kwargs))


def _wphdb_setConnectorMeta(inst, **kwargs):
    """
    Ensure that all meta-data for a new connection point is set correctly.

    :param inst: Instance to which the connection point will be added
    :type inst: WhdbInstanceP
    :param kwargs: Original arguments
    :type kwargs: dict[str, object]
    :return: Updated arguments
    :rtype: dict[str, object]
    """
    kwargs[WHDB_XPOINT_DIRECTION] = WDBS_XPNT_BI
    if not kwargs.get(WHDB_XPOINT_NAME):
        kwargs[WHDB_XPOINT_NAME] = _wphdb_getNewXPointName(inst)
    return kwargs


def _wphdb_getNewXPointName(inst):
    """
    Get the name for the connection point added next to the given instance.

    :param inst: Instance to which the connection point will be added
    :type inst: WhdbInstanceP
    :return: Name for the new connection point
    :rtype: str
    """
    names = set(c.getName() for c in inst.getConnectionPoints())
    cnid = 1
    cname = WPHDB_XPNT_NAME_PATTERN.format(id=cnid)
    while cname in names:
        cnid += 1
        cname = WPHDB_XPNT_NAME_PATTERN.format(id=cnid)
    return cname
