# ----------------------------------------------------------------------------
# - wxio-nested-xml-base                                                     -
# - wild cruncher:xml nested serialization base-class           core / xmlio -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wxio_nested_xml_base
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher xml nested serialization base-class

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from xml.dom.minidom import Document

from wcr_core.db.whdb_connection_point import WHDB_DEFAULT_XPOINT_CODE

from wcr_core.dbstore_api.wdbs_instance_api import WdbsOperatorInstanceP, \
    WdbsModelInstanceP, WdbsBlockInstanceP
from wcr_core.xmlio.wxio_xml_base import *
from wcr_core.modules.hardware.db.whhdb_connection_point_base import WHHDB_XPOINT_CONFIG_KEY
from wcr_core.modules.hardware.db.whhdb_standard_operators import IO_CARDS_CODE
from wcr_core.modules.hardware.db.whhdb_standard_operators import \
    DEFAULT_CARDS_CODES


_META_DATA_TAG = 'meta-data'
_FIXED_META_DATA_XPOINT_TAG = ['port', 'gvty', 'name', 'cpos', 'config', 'inv_pol']
CARDS = (IO_CARDS_CODE + ['wcr.cetc.c2p'] + DEFAULT_CARDS_CODES)

# noinspection PyAbstractClass
class WxioNestedXml(WxioXmlBase):
    """
    Base class for nested XML-IO versions.
    """

    @staticmethod
    def getVersion():
        """
        Get the XML-IO version.

        :return: XML-IO version
        :rtype: str
        """
        raise NotImplementedError()

    def __init__(self, ignoreUUID=False):
        """
        :param ignoreUUID: Whether to ignore UUIDs during serialization
                           (and generate new UUIDs for all entities instead)
        :type ignoreUUID: bool
        """
        super().__init__(ignoreUUID)

    def serialize(self, block, xml, topName, module):
        hrchyNode = xml.createElement('hierarchy')
        hrchyNode.setAttribute('module', module)
        hrchyNode.setAttribute('xmlio-version', self.getVersion())

        # Add all models to the XML output as a flat list
        blocks, models = self.collectModels(block)
        if models:
            modelsNode = xml.createElement('models')
            for model in models:
                modelsNode.appendChild(self._createModelElement(model, xml))
            hrchyNode.appendChild(modelsNode)
        if blocks:
            blocksNode = xml.createElement('blocks')
            for _block in blocks:
                blocksNode.appendChild(
                    self._createBlockElement('block', _block, xml))
            hrchyNode.appendChild(blocksNode)

        # Add the top-level model/instance to the XML output
        hrchyNode.appendChild(self._createBlockElement(topName, block, xml))

        return hrchyNode

    @staticmethod
    def fixFieldName(name):
        """
        :type name: str
        :rtype: str
        """
        return name.replace('_', '-')

    def _setNodeValue(self, text, node, xml):
        """
        Set the inner value of a node, either as text or as CDATA.

        :param text: Inner node value
        :type text: str
        :param node: Node
        :type node: xml.dom.minidom.Element
        :param xml: XML Document
        :type xml: Document
        """
        if '<' in text or '>' in text:
            node.appendChild(xml.createCDATASection(text))
        else:
            node.appendChild(xml.createTextNode(text))

    def _createValueNode(self, tag, value, xml):
        """
        Create a node holding an inner value.

        :param tag: Tag for the new node
        :type tag: str
        :param value: Inner value
        :type value: str
        :param xml: XML Document
        :type xml: Document
        :return: Created node, holding the given value
        :rtype: xml.dom.minidom.Element
        """
        node = xml.createElement(tag)
        if value is not None:
            self._setNodeValue(str(value), node, xml)
        return node

    def _writeMetaData(self, item, node, xml):
        """
        Add the meta-data for the given entity to its node.

        :param item: Hierarchy entity
        :type item: WdbsEntityP
        :param node: Node for the given entity
        :type node: xml.dom.minidom.Element
        :param xml: XML Document
        :type xml: Document
        """
        raise NotImplementedError()

    def _createConnectionElement(self, connection, xml):
        """
        Create a connection node.

        :param connection: Connection to serialize
        :type connection: WdbsConnectionP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given connection
        :rtype: xml.dom.minidom.Element
        """
        connNode = xml.createElement('connection')
        connNode.setAttribute('uuid', str(connection.getUUID()))
        connNode.setAttribute('sobj',
                              str(connection.getStartObject().getUUID()))
        connNode.setAttribute('scon',
                              str(connection.getStartXPoint().getUUID()))
        connNode.setAttribute('eobj', str(connection.getEndObject().getUUID()))
        connNode.setAttribute('econ', str(connection.getEndXPoint().getUUID()))

        path = connection.getPath()
        pathNode = xml.createElement('path')
        if path is not None:
            for x, y in path.getPoints():
                pointNode = xml.createElement('point')
                pointNode.setAttribute('x', str(x))
                pointNode.setAttribute('y', str(y))
                pathNode.appendChild(pointNode)
        connNode.appendChild(pathNode)

        self._writeMetaData(connection, connNode, xml)

        return connNode

    def _createXPointsElement(self, instance, xml):
        """
        Create a node for the connection points of the given instance.

        :param instance: Instance to create connection points node for
        :type instance: WdbsInstanceP
        :param xml: XML Document
        :type xml: Document
        :return: Node containing the serialization of the connection points
                 of the given instance
        :rtype: xml.dom.minidom.Element
        """
        xpointsNode = xml.createElement('xpoints')

        # noinspection PyTypeChecker
        for xpoint in instance.getConnectionPoints():
            xpointNode = xml.createElement('xpoint')
            xpointNode.setAttribute('uuid', str(xpoint.getUUID()))
            try:
                pointType = xpoint.getMeta(WHHDB_XPOINT_CONFIG_KEY).get('type')
            except KeyError:
                pointType = xpoint.getEntityType()
            if pointType is None:
                pointType = xpoint.getEntityType()
            xpointNode.setAttribute('type', pointType)
            xpointNode.setAttribute('direction',
                                    WxioXmlBase.XPOINT_DIR_EXT.get(
                                        xpoint.getDirection()))
            for f in xpoint.metaDescription.fields:
                f.set_temporary(False)
            # filter xpoint type to save only data type
            for f in xpoint.metaDescription.fields:
                if f.name not in _FIXED_META_DATA_XPOINT_TAG:
                    if f.name != pointType.lower():
                        f.set_temporary(True)

            self._writeMetaData(xpoint, xpointNode, xml)
            xpointsNode.appendChild(xpointNode)

        return xpointsNode

    def _createBlockInstanceElement(self, inst, xml):
        """
        Create a block instance node.

        :param inst: Block instance to serialize
        :type inst: WdbsBlockInstanceP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given instance
        :rtype: xml.dom.minidom.Element
        """
        instNode = xml.createElement('block-instance')
        instNode.setAttribute('uuid', str(inst.getUUID()))
        instNode.setAttribute('block', str(inst.getBlock().getUUID()))

        instNode.appendChild(self._createXPointsElement(inst, xml))

        self._writeMetaData(inst, instNode, xml)

        return instNode

    def _createModelInstanceElement(self, inst, xml):
        """
        Create a model instance node.

        :param inst: Model instance to serialize
        :type inst: WdbsModelInstanceP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given instance
        :rtype: xml.dom.minidom.Element
        """
        instNode = xml.createElement('model-instance')
        instNode.setAttribute('uuid', str(inst.getUUID()))
        instNode.setAttribute('model', str(inst.getModel().getUUID()))

        instNode.appendChild(self._createXPointsElement(inst, xml))

        self._writeMetaData(inst, instNode, xml)

        return instNode

    def _createOperatorInstanceElement(self, inst, xml):
        """
        Create an operator instance node.

        :param inst: Operator instance to serialize
        :type inst: WdbsOperatorInstanceP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given operator instance
        :rtype: xml.dom.minidom.Element
        """
        instNode = xml.createElement('operator-instance')
        instNode.setAttribute('uuid', str(inst.getUUID()))
        instNode.setAttribute('operator', inst.getOperator())

        instNode.appendChild(self._createXPointsElement(inst, xml))
        cardsInst = getattr(inst, 'cardsInst', None)
        if cardsInst:
            instNode.appendChild(self._createIoCardsInstanceElement(inst, xml))
        instNode.appendChild(self._createParametersElement(inst, xml))

        self._writeMetaData(inst, instNode, xml)

        return instNode

    def _createIoCardsInstanceElement(self, inst, xml):
        """
        Create a node for the cards of the given entity.

        :param entity: Entity with parameters to serialize
        :type entity: WdbsEntityParametersP
        :type xml: Document
        :return: XML node representing the parameters of the given entity
        :rtype: xml.dom.minidom.Element
        """

        cardsNode = xml.createElement('cards')
        for uuid in inst.cardsInst:
            card = inst.cardsInst[uuid]
            cardNode = xml.createElement('card')
            cardNode.setAttribute('oper', card.op)
            cardNode.setAttribute('name', card.getName())
            cardNode.setAttribute('uuid', str(card.getUUID()))
            cardNode.setAttribute('idtstatus', card.getIDTStatus())
            cardNode.appendChild(self._createChannelParametersElement(card, xml))
            cardsNode.appendChild(cardNode)
        return cardsNode

    def _createParametersElement(self, entity, xml):
        """
        Create a node for the parameters of the given entity.

        :param entity: Entity with parameters to serialize
        :type entity: WdbsEntityParametersP
        :type xml: Document
        :return: XML node representing the parameters of the given entity
        :rtype: xml.dom.minidom.Element
        """
        paramsNode = xml.createElement('parameters')

        for param in entity.getParameters():
            paramNode = xml.createElement('parameter')
            paramNode.setAttribute('name', param.getName())
            value = param.getValue()
            if value is not None:
                self._setNodeValue(value, paramNode, xml)
            paramsNode.appendChild(paramNode)
        return paramsNode

    def _createChannelParametersElement(self, entity, xml):
        """
        Create a node for the parameters of the given entity.

        :param entity: Entity with parameters to serialize
        :type entity: WdbsEntityParametersP
        :type xml: Document
        :return: XML node representing the parameters of the given entity
        :rtype: xml.dom.minidom.Element
        """
        paramsNode = xml.createElement('channels')
        for channelName, channel in entity.parameters.items():
            channelNode = xml.createElement('channel')
            channelNode.setAttribute('number', channelName.strip('channel'))

            for prm in channel.values():
                paramNode = xml.createElement('parameter')
                paramNode.setAttribute('name', prm.getName())
                paramNode.setAttribute('value', str(prm.getValue()))
                channelNode.appendChild(paramNode)
            paramsNode.appendChild(channelNode)
        return paramsNode

    def _createBlockElement(self, typeName, block, xml):
        """
        Create a block node.

        :param typeName: Block type name
        :type typeName: str
        :param block: Block to serialize
        :type block: WdbsContainerP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given block
        :rtype: xml.dom.minidom.Element
        """
        blockNode = xml.createElement(typeName)

        blockNode.setAttribute('uuid', str(block.getUUID()))
        if isinstance(block, WdbsBlockP):
            btyp = block.getEntityType()
            if btyp:
                blockNode.setAttribute('type', str(btyp))

        binsts = [i for i in block.getInstances()
                  if isinstance(i, WdbsBlockInstanceP)]
        """:type: list[WdbsBlockInstanceP]"""
        if binsts:
            instsNode = xml.createElement('block-instances')
            for inst in binsts:
                instsNode.appendChild(
                    self._createBlockInstanceElement(inst, xml))
            blockNode.appendChild(instsNode)

        minsts = [i for i in block.getInstances()
                  if isinstance(i, WdbsModelInstanceP)]
        """:type: list[WdbsModelInstanceP]"""
        if minsts:
            instsNode = xml.createElement('model-instances')
            for inst in minsts:
                instsNode.appendChild(
                    self._createModelInstanceElement(inst, xml))
            blockNode.appendChild(instsNode)

        oinsts = [i for i in block.getInstances()
                  if isinstance(i, WdbsOperatorInstanceP)]
        """:type: list[WdbsOperatorInstanceP]"""
        if oinsts:
            instsNode = xml.createElement('operator-instances')
            for inst in oinsts:
                # handle cards differently
                if inst.getOperator() in CARDS:
                    continue
                instsNode.appendChild(
                    self._createOperatorInstanceElement(inst, xml))
            blockNode.appendChild(instsNode)

        if block.getConnectionCount() > 0:
            connsNode = xml.createElement('connections')
            for conn in block.getConnections():
                connsNode.appendChild(self._createConnectionElement(conn, xml))
            blockNode.appendChild(connsNode)

        if block.getShapeCount() > 0:
            shapesNode = xml.createElement('shapes')
            for element in block.getShapes():
                shapesNode.appendChild(
                    self._createShapeElement(element, xml))
            blockNode.appendChild(shapesNode)

        if block.getButtonCount() > 0:
            interactsNode = xml.createElement('buttons')
            for element in block.getButtons():
                interactsNode.appendChild(
                    self._createButtonElement(element, xml))
            blockNode.appendChild(interactsNode)

        self._writeMetaData(block, blockNode, xml)

        return blockNode

    def _createModelElement(self, model, xml):
        """
        Create a model node.

        :param model: Model to serialize
        :type model: WdbsModelP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given model
        :rtype: xml.dom.minidom.Element
        """
        modelNode = xml.createElement('model')

        modelNode.setAttribute('uuid', str(model.getUUID()))
        mtyp = model.getEntityType()
        if mtyp:
            modelNode.setAttribute('type', str(mtyp))
        modelNode.setAttribute('path', model.getPath())
        modelNode.setAttribute('entry', model.getEntryPoint())

        self._writeMetaData(model, modelNode, xml)

        return modelNode

    def _createShapeElement(self, shape, xml):
        """
        Create a shape node.

        :param shape: Shape to serialize
        :type shape: WdbsShapeP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given shape
        :rtype: xml.dom.minidom.Element
        """
        shapeNode = xml.createElement('shape')
        shapeNode.setAttribute('uuid', str(shape.getUUID()))
        shapeNode.setAttribute('shape-type', shape.getEntityType())

        self._writeMetaData(shape, shapeNode, xml)

        return shapeNode

    def _createButtonElement(self, button, xml):
        """
        Create a button node

        :param button: Button to serialize
        :type button: WdbsButtonP
        :param xml: XML document used for serialization
        :type xml: Document
        :return: XML node representing the given button
        :rtype: xml.dom.minidom.Element
        """
        buttonNode = xml.createElement('button')

        buttonNode.setAttribute('uuid', str(button.getUUID()))
        buttonNode.appendChild(self._createValueNode(
            'title', button.getTitle(), xml))
        buttonNode.appendChild(self._createValueNode(
            'command', button.getCommand(), xml))

        buttonNode.appendChild(self._createParametersElement(button, xml))

        self._writeMetaData(button, buttonNode, xml)

        return buttonNode

    @abstractmethod
    def readName(self, node):
        """
        Read the name-field from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def readInfo(self, node):
        """
        Read the info-field from the given node.

        :param node: Node
        :type node: xml.dom.minidom.Element
        :return: Info
        :rtype: str
        """
        raise NotImplementedError()

    def getBlockNodes(self, node):
        blocksNode = node.find('blocks')
        if blocksNode is not None and len(blocksNode) > 0:
            return iter(blocksNode)
        return []

    def readBlockData(self, node):
        return WxioXmlBlock(
            uuid=node.get('uuid') or node.get('sid'),
            type=node.get('type') or node.get('blockType'),
            name=self.readName(node),
            info=self.readInfo(node))

    def getBlockInstanceNodes(self, node, blocks):
        binstNodes = node.find('block-instances')
        if binstNodes is not None and len(binstNodes) > 0:
            return iter(binstNodes)
        return []

    def getModelInstanceNodes(self, node, models):
        minstNodes = node.find('model-instances')
        if minstNodes is not None and len(minstNodes) > 0:
            return iter(minstNodes)
        return []

    def getOperatorInstanceNodes(self, node):
        oinstNodes = node.find('operator-instances')
        if oinstNodes is not None:
            return iter(oinstNodes)
        return []

    def getIoCardInstanceNodes(self, node):
        oinstNodes = node.find('cards')
        if oinstNodes is not None:
            return iter(oinstNodes)
        return []

    def getConnectionNodes(self, node):
        connsNode = node.find('connections')
        if connsNode is not None:
            return iter(connsNode)
        return []

    def getShapeNodes(self, node):
        shapesNode = node.find('shapes')
        if shapesNode is not None:
            return iter(shapesNode)
        return []

    def getButtonNodes(self, node):
        buttonsNode = node.find('buttons')
        if buttonsNode is not None:
            return iter(buttonsNode)
        return []

    def getModelNodes(self, node):
        modelNodes = node.find('models')
        if modelNodes is not None:
            return iter(modelNodes)
        return []

    def readModelData(self, node):
        return WxioXmlModel(
            uuid=node.get('uuid') or node.get('sid'),
            type=node.get('type') or node.get('modelType'),
            name=self.readName(node),
            info=self.readInfo(node),
            path=node.get('path'),
            entry=node.get('entry') or node.get('fnam'))

    def readXPointData(self, node):
        ctyp = WxioXmlBase.XPOINT_DIR_INT[node.get('direction')]

        return WxioXmlXPoint(
            uuid=node.get('uuid') or node.get('sid'),
            type=node.get('type') or WHDB_DEFAULT_XPOINT_CODE,
            direction=ctyp)

    def getXPointNodes(self, node):
        return node.find('xpoints')

    def readBlockInstanceData(self, node):
        return WxioXmlBlockInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=self.readName(node),
            info=self.readInfo(node),
            block=node.get('block'))

    def readModelInstanceData(self, node):
        return WxioXmlModelInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=self.readName(node),
            info=self.readInfo(node),
            model=node.get('model'))

    def readOperatorInstanceData(self, node):
        return WxioXmlOpInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=self.readName(node),
            info=self.readInfo(node),
            status=self.readStatus(node),
            statusMask=self.readStatusMask(node),
            arm=self.readArm(node),
            oper=node.get('operator'))

    def readIoCardInstanceData(self, node):
        return WxioXmlIOCardInst(
            uuid=node.get('uuid') or node.get('sid'),
            name=node.get('name'),
            idtstatus=node.get('idtstatus'),
            channels=self.readChannels(node),
            oper=node.get('oper'))

    def readConnectionData(self, node):
        pathNode = node.find('path')
        path = None
        if pathNode is not None:
            path = [(float(pn.get('x')), float(pn.get('y')))
                    for pn in pathNode.findall('point')]

        return WxioXmlConnection(
            uuid=node.get('uuid') or node.get('sid'),
            sobj=node.get('sobj'), sxpnt=node.get('scon'),
            eobj=node.get('eobj'), expnt=node.get('econ'),
            name=self.readName(node),
            info=self.readInfo(node),
            path=path)

    def readShapeData(self, node):
        return WxioXmlShape(
            uuid=node.get('uuid') or node.get('sid'),
            stype=node.get('shape-type'))

    def readParameters(self, node):
        paramsNode = node.find('parameters')
        return [(node.get('name'), node.text.strip())
                for node in paramsNode if node.text and node.text != 'None']

    def readOperatorInstanceParameters(self, node):
        return self.readParameters(node)

    def readIoCards(self, node):
        cardsNode = node.find('cards')
        if not cardsNode:
            return []
        return [self.readIoCardInstanceData(cardNode)
                for cardNode in cardsNode]

    def readButtonParameters(self, node):
        return self.readParameters(node)
