# ----------------------------------------------------------------------------
# - wgui-editor                                                              -
# - wild cruncher: editor base classes and protocols                         -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_editor
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher editor base classes and protocols

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
from functools import partial

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget, QErrorMessage

from wcr_frontend.gui.wgui_gui_action import WguiActionExecutionContextP
from wcr_frontend.gui.wgui_protocols import QWidget_Meta_M
from wcr_core.action.wact_action import WactSimpleContextAction


def wgui_create_editor_actions():
    """
    Create a set of general editor functions.

    :return: Editor actions
    :rtype: dict[str, WactActionP]
    """
    #: Condition that there is an active editor
    def hasEditor(context):
        return context.getEditor() is not None

    #: Condition that the active editor is editable
    def isEditable(context):
        editor = context.getEditor()
        return editor is not None and not editor.isReadOnly()

    #: Condition that the active editor can save its content
    def editorWithSaveOnly(context):
        return isinstance(context.getEditor(), WguiEditorHasSaveP)

    #: Condition that the active editor can delete from its content
    def editorWithDeleteOnly(context):
        return isinstance(context.getEditor(), WguiEditorHasDeleteP)

    #: Condition that the active editor can search its content
    def editorWithSearchOnly(context):
        return isinstance(context.getEditor(), WguiEditorHasSearchP)

    return {
        'editor.save': WactSimpleContextAction(
            'Save', 'Save the project to disk',
            partial(_wgui_saveActiveEditor), editorWithSaveOnly),
        'editor.saveas': WactSimpleContextAction(
            'Save &As...', 'Save under a new name',
            partial(_wgui_saveActiveEditorCopy), editorWithSaveOnly),
        'editor.selAll': WactSimpleContextAction(
            'Select &All', 'Select All',
            partial(_wgui_selectAll), hasEditor),
        'editor.cut': WactSimpleContextAction(
            'Cu&t', 'Cut the selection',
            partial(_wgui_cut), hasEditor, isEditable),
        'editor.copy': WactSimpleContextAction(
            '&Copy', 'Copy the selection',
            partial(_wgui_copy), hasEditor),
        'editor.paste': WactSimpleContextAction(
            '&Paste', 'Paste the selection',
            partial(_wgui_paste), hasEditor, isEditable),
        'editor.delete': WactSimpleContextAction(
            '&Delete', 'Delete the selection',
            partial(_wgui_delete), hasEditor, editorWithDeleteOnly),
        'editor.print': WactSimpleContextAction(
            '&Print', 'Print',
            partial(_wgui_print), hasEditor),
        'editor.refresh': WactSimpleContextAction(
            '&Refresh', 'Refresh the project',
            partial(_wgui_refresh), hasEditor),
        'editor.find': WactSimpleContextAction(
            '&Find ...', 'Search the current editor',
            partial(_wgui_find), hasEditor, editorWithSearchOnly),
        'editor.findagain': WactSimpleContextAction(
            '&Find again', 'Search the current editor',
            partial(_wgui_findAgain), hasEditor, editorWithSearchOnly)
    }


def _wgui_saveActiveEditor(context):
    """
    Save the contents of the currently active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    # TODO: Error handling
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasSaveP):
        try:
            editor.save()
        except IOError as e:
            QErrorMessage(context.getParent()).showMessage(
                'Could not save file: {0}'.format(e))


def _wgui_saveActiveEditorCopy(context):
    """
    Save a copy of the contents of the currently active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    # TODO: Error handling
    # TODO: Query path and pass to saveCopy function?
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasSaveP):
        path = ''
        try:
            path, mime = editor.saveCopy()
            if path:
                context.statusMessage('Saved to ' + path + ' as ' + mime)
        except IOError as e:
            QErrorMessage(context.getParent()).showMessage(
                'Could not save file at {0}:\n{1}'.format(path or '?', e))


def _wgui_cut(context):
    """
    Remove the selected contents from the active editor, and copy them to the
    clipboard.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if editor is not None:
        editor.cut()


def _wgui_copy(context):
    """
    Copy the selected contents from the active editor to the clipboard.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if editor is not None:
        editor.copy()


def _wgui_paste(context):
    """
    Paste the contents of the clipboard into the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if editor is not None and not editor.isReadOnly():
        editor.paste()


def _wgui_delete(context):
    """
    Delete the selected contents from the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasDeleteP):
        editor.delete()


def _wgui_selectAll(context):
    """
    Select all contents in the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if editor is not None:
        editor.selectAll()


def _wgui_print(context):
    """
    Print the contents of the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if editor is not None:
        editor.print()


def _wgui_refresh(context):
    """
    Refresh the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasRefreshP):
        editor.refresh()


def _wgui_find(context):
    """
    Search in the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasSearchP):
        editor.activateFind()


def _wgui_findAgain(context):
    """
    Search in the active editor.

    :param context: context
    :type context: WguiActionExecutionContextP
    """
    editor = context.getEditor()
    if isinstance(editor, WguiEditorHasSearchP):
        editor.findAgain()


class WguiEditorHasSaveP(metaclass=ABCMeta):
    """
    Protocol for editors that can save their contents to disk.
    """
    # TODO: WguiFileEditor?

    @abstractmethod
    def isDirty(self):
        """
        Whether the editor's content is dirty, i.e. has changed since loading
        or the last save.

        :return: True iff the editor is dirty
        :rtype: bool
        """
        pass

    @abstractmethod
    def save(self):
        """
        Save the contents of the editor at the path they have been loaded from,
        or last saved to.
        If no such path exists, the user will be asked to provide a path.
        """
        pass

    @abstractmethod
    def saveCopy(self):
        """
        Save the contents of the editor at a new path to be provided by the
        user.

        :return: The path and mime-type of the new file location
        :rtype: (str, str)
        """
        # TODO: Handle file dialog here / pass in path?
        pass

        # TODO: getPath(), getFileTypes(), canSave()


class WguiEditorHasDeleteP(metaclass=ABCMeta):
    """
    Protocol for editors that support deletion of the current selection.
    """

    @abstractmethod
    def delete(self):
        """
        Delete the currently selected/active content.
        """
        pass


class WguiEditorHasRefreshP(metaclass=ABCMeta):
    """
    Protocol for editors that can explicitely refresh.
    """

    @abstractmethod
    def refresh(self):
        """
        Refresh the editor.
        """
        pass


class WguiEditorHasSearchP(metaclass=ABCMeta):
    """
    Protocol for editors that can search their content.
    """

    @abstractmethod
    def activateFind(self):
        """
        Activate the search for the editor.
        """
        pass

    @abstractmethod
    def findAgain(self):
        """
        Search the editor again with the previous settings.
        """
        pass


class WguiEditor(metaclass=ABCMeta):
    """
    Base class for editors.
    """

    @abstractmethod
    def getTitle(self):
        """
        Get the current title of the editor.

        :return: Current title of the editor
        :rtype: str
        """
        pass

    @abstractmethod
    def getIcon(self):
        """
        Get the icon for the editor.

        :return: Path to editor icon
        :rtype: str
        """
        pass

    @abstractmethod
    def getEditorIdent(self):
        """
        Get the icon for the editor.

        :return: Path to editor icon
        :rtype: str
        """
        pass

    @abstractmethod
    def isReadOnly(self):
        """
        Whether the editor is read only.

        :return: True iff the editor is read-only
        :rtype: bool
        """
        pass

    @abstractmethod
    def activated(self):
        """
        Callback called whenever this editor is activated.
        """
        pass

    @abstractmethod
    def print(self):
        """
        Print the text contents or current selection.
        """
        pass

    @abstractmethod
    def copy(self):
        """
        Copy the currently selected contents to the clipboard.
        """
        pass

    @abstractmethod
    def cut(self):
        """
        Copy the currently selected contents to the clipboard, and remove them
        from the editor.
        """
        pass

    @abstractmethod
    def paste(self):
        """
        Paste the contents of the clipboard into the editor.
        """
        pass

    @abstractmethod
    def selectAll(self):
        """
        Select all contents of the editor.
        """
        pass

    # noinspection PyPropertyDefinition
    @property
    @abstractmethod
    def nameChanged(self):
        """
        Get the signal that emits whenever the editor's name changes.

        :return: Signal emitting whenever the editor's name changes
        :rtype: pyqtSignal
        """
        pass


# noinspection PyAbstractClass
class WguiEditorBase(QWidget, WguiEditor,
                     metaclass=QWidget_Meta_M):
    """
    Abstract base class for editors.
    """

    #: Signal for editor name changes (access via nameChanged)
    _nameChanged = pyqtSignal(str)

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)

    @property
    def nameChanged(self):
        return self._nameChanged

    def getEditorIdent(self):
        return 'wcr:editor-{}'.format(id(self))
