# ----------------------------------------------------------------------------
# - wgui-editor-side-button-bar                                              -
# - wild cruncher: side button bar                            frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_editor_side_button_bar
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Side button bar

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import ABCMeta, abstractmethod
import sys

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QStyle, QWidget, QVBoxLayout, QToolButton, \
	QSizePolicy, QStylePainter, QStyleOptionButton
from wcr_frontend.gui.wgui_protocols import QWidget_Meta_M


class WguiSidebarButtonP(metaclass=ABCMeta):
    """
    Protocol for side buttons.
    """

    @abstractmethod
    def code(self):
        """
        Get the button's identifying code.

        :return: Button code
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def toggled(self):
        """
        Signal emitting whenever the button has been toggled (if it's a
        toggle button)

        :return: Button toggled signal
        :rtype: pyqtBoundSignal[bool]
        """
        raise NotImplementedError()

    @abstractmethod
    def setChecked(self, checked):
        """
        Mark the button as toggled (if it's a toggle button).

        :param checked: New button toggle state
        :type checked: bool
        """
        raise NotImplementedError()


class _WguiSidebarToolButton(QToolButton, WguiSidebarButtonP,
                             metaclass=QWidget_Meta_M):
    def __init__(self, code, parent=None):
        """
        :param code: Button code
        :type code: str
        :param parent: Parent widget
        """
        QToolButton.__init__(self, parent)

        #: Button code
        self.__code = code

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.setCheckable(True)

    def code(self):
        return self.__code

    def paintEvent(self, event):
        painter = QStylePainter(self)
        painter.setPen(Qt.black)
        painter.setBrush(Qt.Dense1Pattern)
        painter.rotate(90)
        painter.translate(0, -self.width())

        painter.drawControl(QStyle.CE_PushButton, self.getStyleOption())

    def sizeHint(self):
        size = super().sizeHint()
        """:type: QSize"""
        size.transpose()
        if not sys.platform.lower().startswith('win'):
            size.setWidth(size.width() - 3)
        return size

    def getStyleOption(self):
        styleOpt = QStyleOptionButton()
        styleOpt.initFrom(self)

        size = styleOpt.rect.size()
        size.transpose()
        styleOpt.rect.setSize(size)

        styleOpt.features = QStyleOptionButton.Flat

        if self.isDown():
            styleOpt.state |= QStyle.State_Sunken
        if self.isChecked():
            styleOpt.state |= QStyle.State_On
        #if not self.isDown() and not self.isChecked():
        #    styleOpt.state |= QStyle.State_Raised

        styleOpt.text = self.text()
        #styleOpt.icon = self.icon()
        #styleOpt.iconSize = self.iconSize()

        return styleOpt


class WguiSideButtonBar(QWidget):
    def __init__(self, side, parent=None):
        """
        :param side: Side the bar is on (left/right)
        :type side: int
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Side the bar is on (left/right)
        self.__side = side

        self.__layout = self._createLayout()
        self.setLayout(self.__layout)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QBoxLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(3)

        return layout

    def addToggleButton(self, code, text):
        """
        Add a toggle button to the bar.

        :param code: Button code
        :type code: str
        :param text: Button label
        :type text: str
        :return: Newly created button
        :rtype: WguiSidebarButtonP
        """
        button = _WguiSidebarToolButton(code, self)
        button.setText(text)
        self.__layout.addWidget(button)
        return button
