# ----------------------------------------------------------------------------
# - wgui-editor-sidebar                                                      -
# - wild cruncher: editor widget sidebar                      frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_editor_sidebar
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - Editor widget sidebar

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QSplitter, QLabel, QSizePolicy
from PyQt5.QtGui import QFont

from wcr_frontend.gui.wgui_editor_api import WGUI_EDITOR_LEFT_SIDE

from string import Template
#: Sidebar widget code
WGUI_SIDEBAR_CODE = 'wcr.sidebar'


class _WguiSidebarContent(QSplitter):
    """
    Sidebar content widget.

    """
    DEFAULT_COLOR = ('#5F814D', '#C2D6B8')

    def __init__(self, parent=None):
        super().__init__(Qt.Vertical, parent)

        self.setChildrenCollapsible(False)

        #: Widgets by code
        self.__widgets = {}
        """:type: dict[str, QWidget]"""

    def _wrapWidget(self, widget, title, color):
        titleW = QLabel(title)
        titleW.setMargin(3)
        font = QFont(titleW.font())
        font.setBold(True)
        titleW.setFont(font)
        titleW.setAlignment(Qt.AlignRight)
        style = Template('QLabel { background-color: '
                         'qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,'
                         'stop: 0 $col1, stop: 1 $col2); }')
        titleW.setStyleSheet(style.substitute(col1=color[0], col2=color[1]))
        wdgtLayout = QVBoxLayout()
        wdgtLayout.setContentsMargins(0,0,0,0)
        wdgtLayout.addWidget(titleW)
        wdgtLayout.addWidget(widget)
        wrapper = QWidget()
        wrapper.setLayout(wdgtLayout)

        wrapper.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)

        return wrapper

    def addSidebarWidget(self, code, title, widget, color=None):
        """
        Add the widgets to the sidebar.

        :param code: Widget code
        :type code: str
        :param title: Widget title
        :type title: str
        :param widget: Widget
        :type widget: QWidget
        :param color: defining color widget
        :type color: tuple(hex, hex)
        """
        if color is None:
            color = self.DEFAULT_COLOR
        wrapped = self._wrapWidget(widget, title, color)
        self.addWidget(wrapped)
        self.__widgets[code] = wrapped

    def setWidgetVisible(self, code, visible):
        """
        Set the visibility of the widget with the given code.

        :param code: Code of widget to set visibility for
        :type code: str
        :param visible: Visibility
        :type visible: bool
        """
        wdgt = self.__widgets.get(code)
        """:type: QWidget"""
        if wdgt:
            wdgt.setVisible(visible)


class WguiEditorSidebarInstaller(object):
    """
    Sidebar installer.
    """

    def __init__(self, widgets):
        """
        :param widgets: Sidebar widgets [(Code, Name, Widget)]
        :type widgets: Iterable[(str, str, QWidget)]
        """
        #: Widgets
        self.__widgets = []
        for values in widgets:
            if len(values) == 3:
                c, n, w = values
                cr = None
            elif len(values) == 4:
                c, n, w, cr = values
            self.__widgets.append(('{}.{}'.format(WGUI_SIDEBAR_CODE, c), n, w, cr))
        """:type: list[(str, str, QWidget)]"""

        #: Widget visibility states by code
        self.__states = {c: False for c, _, _, _ in self.__widgets}
        """:type: dict[str, bool]"""
        if self.__widgets:
            self.__states[self.__widgets[0][0]] = True

        #: Sidebar contents widget
        self.__contents = _WguiSidebarContent()

    def install(self, target):
        """
        Install the sidebar into the given target.

        :param target: Target to install sidebar into
        :type target: WguiEditorInstTargetP
        """
        for values in self.__widgets:
            if len(values) == 3:
                code, label, widget = values
                color = None
            elif len(values) == 4:
                code, label, widget, color = values
            state = self.__states.get(code)
            self.__contents.addSidebarWidget(code, label, widget, color)
            self.__contents.setWidgetVisible(code, state)
            button = target.installSideButton(
                code, label, WGUI_EDITOR_LEFT_SIDE, toggle=True)
            button.setChecked(state)

        cntn = self.__contents
        """:type: QWidget"""
        target.insertMainWidget(WGUI_SIDEBAR_CODE, cntn)
        target.sideButtonToggled.connect(self._toggled)

    @property
    def modules(self):
        """
        Get the sidebar's additional module widgets.

        :return: List of sidebar widgets and their identifiers and names
        :rtype: Iterable[(str, str, QWidget)]
        """
        return self.__widgets

    def _toggled(self, code, state):
        """
        :param code: Button code
        :type code: str
        :param state: Button toggle state
        :type state: bool
        """
        _state = self.__states.get(code)
        if _state == state:
            return

        self.__states[code] = state

        self.__contents.setWidgetVisible(code, state)

        if any(self.__states.values()):
            self.__contents.setVisible(True)
        else:
            self.__contents.setVisible(False)

    def getStates(self):
        return self.__states

