# ----------------------------------------------------------------------------
# - wgui-form-widget-creator                                                 -
# - wild cruncher: meta-data widget creator                   frontend / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_form_widget_creator
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher meta-data widget creator

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtCore import QAbstractTableModel, Qt, QModelIndex
from PyQt5.QtWidgets import QTextEdit, QCheckBox, QHBoxLayout, QWidget, \
	QVBoxLayout, QTableView, QHeaderView, QPushButton, QFormLayout, \
	QLineEdit, QAbstractItemView

from wcr_core.dbstore.wdbs_meta_type import *


WGUI_META_KEY = 'meta'

WGUI_META_ITEMS_KEY = '__items'

WGUI_META_LABEL_KEY = 'label'

WGUI_META_SHORTDESC_KEY = 'shortdesc'

WGUI_META_LONGDESC_KEY = 'longdesc'

WGUI_META_WARNING_KEY = 'warning'

WGUI_META_WIDGET_KEY = 'widget'

WGUI_META_WIDGET_ARGS_KEY = 'widgetargs'

WGUI_META_SKIP_KEY = 'skip'

WGUI_META_FIELDS_KEY = 'fields'

WGUI_META_HEADERS_KEY = 'headers'

_WGUI_RECORD_DEPTH_KEY = '__recordDepth'


class WguiWidgetCreatorBaseP(metaclass=ABCMeta):
    """
    Protocol for meta-data UI-widget creators.
    """

    @abstractmethod
    def createWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for the meta-data element found at the given keys.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsMetaTypeP
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiWidgetCreatorBaseIP(WguiWidgetCreatorBaseP):
    @abstractmethod
    def createRecordWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a record element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsRecordMetaTypeP
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createListWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget of a list element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsListMetaTypeP
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createTupleWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a tuple element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsTupleMetaTypeP
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createUnionWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a union element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsUnionMetaTypeP
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createStringWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a string element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsAtomicMetaTypeP[str]
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createTextWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a text element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsAtomicMetaTypeP[str]
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createBoolWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a boolean element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsAtomicMetaTypeP[bool]
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def createTimeWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a widget for a time element.

        :param dtyp: Type of the meta-data element
        :type dtyp: WdbsAtomicMetaTypeP[str]
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiWidgetCreatorBase(WguiWidgetCreatorBaseIP):
    """
    Abstract base class for meta-data UI-widget creators.
    """

    def createWidget(self, dtyp, *keys, data=None, **kwargs):
        meta = kwargs.get(WGUI_META_KEY)
        if meta and meta.get(WGUI_META_SKIP_KEY):
            return None

        if isinstance(dtyp, WdbsRecordMetaTypeP):
            return self.createRecordWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsListMetaTypeP):
            return self.createListWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsTupleMetaTypeP):
            return self.createTupleWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsUnionMetaTypeP):
            return self.createUnionWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsStringMetaType):
            return self.createStringWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsTextMetaType):
            return self.createTextWidget(dtyp, *keys, data=data, **kwargs)
        elif isinstance(dtyp, WdbsBoolMetaType):
            return self.createBoolWidget(dtyp, *keys, data=data, **kwargs)
        elif dtyp is WDBS_TIME_META:
            return self.createTimeWidget(dtyp, *keys, data=data, **kwargs)
        else:
            raise ValueError('Unknown type: {}'.format(dtyp))

    # noinspection PyUnusedLocal
    def getData(self, *keys, data=None, **kwargs):
        """
        Get the meta-data for the given keys.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param data: Source data, to be used instead of global data set
        :type data: object | None
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Meta-data
        :rtype: object
        """
        return data

    def createRecordWidget(self, dtyp, *keys, data=None, **kwargs):
        dpth = kwargs.get(_WGUI_RECORD_DEPTH_KEY)
        kwargs[_WGUI_RECORD_DEPTH_KEY] = dpth + 1 if dpth is not None else 0
        rwdg = self.createEmptyRecordWidget(*keys, **kwargs)
        data = self.getData(*keys, data=data, **kwargs)
        """:type: dict[str, object]"""
        meta = kwargs.get(WGUI_META_KEY) or {}
        for k in dtyp.getKeys():
            ctyp = dtyp.getValueType(k)
            ckey = keys + (k,)
            cdat = data.get(k) if data else None
            kwargs[WGUI_META_KEY] = meta.get(k) if meta else None
            cwdg = self.createWidget(ctyp, *ckey, data=cdat, **kwargs)
            if cwdg is not None:
                self.addToRecordWidget(rwdg, cwdg, *ckey, **kwargs)
        return self.finalizeRecordWidget(rwdg, *keys, **kwargs)

    @abstractmethod
    def createEmptyRecordWidget(self, *keys, **kwargs):
        """
        Create an empty widget for a record meta-data element.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def addToRecordWidget(self, recordWidget, childWidget, *keys, **kwargs):
        """
        Add a child widget to a record widget.

        :param recordWidget: Widget created by `createEmptyRecordWidget`
        :type recordWidget: object
        :param childWidget: Child widget to add
        :type childWidget: QWidget
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def finalizeRecordWidget(self, recordWidget, *keys, **kwargs):
        """
        Finalize a widget for a record meta-data element.

        :param recordWidget: Widget created by `createEmptyRecordWidget`
        :type recordWidget: object
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    def createListWidget(self, dtyp, *keys, data=None, **kwargs):
        meta = kwargs.get(WGUI_META_KEY) or {}
        if meta.get(WGUI_META_WIDGET_KEY) == 'table':
            return self.createTableWidget(dtyp, *keys, data=data, **kwargs)
        lwdg = self.createEmptyListWidget(*keys, **kwargs)
        data = self.getData(*keys, data=data, **kwargs)
        """:type: tuple"""
        if data:
            dmet = meta.get(WGUI_META_ITEMS_KEY) if meta else None
            kwargs[WGUI_META_KEY] = dmet
            for i, d in enumerate(data):
                ckey = keys + (i,)
                cwdg = self.createWidget(
                    dtyp.getValueType(), *ckey, data=d, **kwargs)
                if cwdg is not None:
                    self.addToListWidget(lwdg, cwdg, *ckey, **kwargs)
        return self.finalizeListWidget(lwdg, *keys, **kwargs)

    @abstractmethod
    def createEmptyListWidget(self, *keys, **kwargs):
        """
        Create an empty widget for a list meta-data element.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def addToListWidget(self, listWidget, childWidget, *keys, **kwargs):
        """
        Add a child widget to a list widget.

        :param listWidget: Widget created by `createListRecordWidget`
        :type listWidget: object
        :param childWidget: Child widget to add
        :type childWidget: QWidget
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def finalizeListWidget(self, listWidget, *keys, **kwargs):
        """
        Finalize a widget for a list meta-data element.

        :param listWidget: Widget created by `createListRecordWidget`
        :type listWidget: object
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    def createTableWidget(self, dtyp, *keys, data=None, **kwargs):
        """
        Create a tabular widget for a list containing records.

        :param dtyp: Meta-type of record-containing list
        :type dtyp: WdbsListMetaTypeP
        :param keys: Keys for meta-data item
        :type keys: list[object]
        :param data: Top-level data
        :type data: object
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Table widget
        :rtype: QWidget
        """
        data = self.getData(*keys, data=data, **kwargs)
        assert isinstance(data, list)
        meta = kwargs.get(WGUI_META_KEY)
        """:type: dict[str, dict[str, object] | None]"""
        return _WguiTable(
            dtyp, data,
            meta.get(WGUI_META_WIDGET_ARGS_KEY) if meta else None)

    def createTupleWidget(self, dtyp, *keys, data=None, **kwargs):
        twdg = self.createEmptyTupleWidget(*keys, **kwargs)
        data = self.getData(*keys, data=data, **kwargs)
        """:type: list[object]"""
        meta = kwargs.get(WGUI_META_KEY) or {}
        if data:
            for i, d in enumerate(data):
                ckey = keys + (i,)
                kwargs[WGUI_META_KEY] = meta.get(i) if meta else None
                cwdg = self.createWidget(
                    dtyp.getValueType(i), *ckey, data=d, **kwargs)
                if cwdg is not None:
                    self.addToTupleWidget(twdg, cwdg, *ckey, **kwargs)
        return self.finalizeTupleWidget(twdg, *keys, **kwargs)

    @abstractmethod
    def createEmptyTupleWidget(self, *keys, **kwargs):
        """
        Create an empty tuple for a record meta-data element.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def addToTupleWidget(self, tupleWidget, childWidget, *keys, **kwargs):
        """
        Add a child widget to a tuple widget.

        :param tupleWidget: Widget created by `createTupleRecordWidget`
        :type recordWidget: object
        :param tupleWidget: Child widget to add
        :type childWidget: QWidget
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    @abstractmethod
    def finalizeTupleWidget(self, tupleWidget, *keys, **kwargs):
        """
        Finalize a widget for a tuple meta-data element.

        :param tupleWidget: Widget created by `createEmptyTupleWidget`
        :type tupleWidget: object
        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Widget
        :rtype: QWidget
        """
        raise NotImplementedError()

    def createUnionWidget(self, dtyp, *keys, data=None, **kwargs):
        data = self.getData(*keys, data=data, **kwargs)
        meta = kwargs.get(WGUI_META_KEY) or {}
        if data:
            code, data = data
            vtyp = dtyp.getType(code)
            kwargs[WGUI_META_KEY] = meta.get(code) if meta else None
            return self.createWidget(
                vtyp, *(keys + (1,)), data=data, **kwargs)
        else:
            return None


class WguiSimpleWidgetCreator(WguiWidgetCreatorBase):
    """
    Simple meta-data UI-widget creator.
    """

    def __init__(self, data):
        """
        :param data: Meta-data
        :type data: object
        """
        self.__data = data

    def getData(self, *keys, data=None, **kwargs):
        if data:
            return data

        if keys:
            dataDict = self.__data
            for key in keys[:-1]:
                dataDict = dataDict[key]
            return dataDict[keys[-1]]
        else:
            return self.__data

    # noinspection PyUnusedLocal
    def _setDataF(self, *keys, **kwargs):
        """
        Get a data-setter function for the meta-data at the given keys.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        :return: Data-setter
        :rtype: (object) -> _
        """
        if not keys:
            def setData(data):
                self.__data = data
                self._dataChanged(*keys, **kwargs)
        else:
            dref = self.__data
            for key in keys[:-1]:
                dref = dref[key]
            assert not isinstance(dref, list)
            rkey = keys[-1]

            def setData(data):
                dref[rkey] = data
                self._dataChanged(*keys, **kwargs)

        return setData

    def _dataChanged(self, *keys, **kwargs):
        """
        Call-back function for when data has been changed by the user.

        :param keys: Keys of changed meta-data element
        :type keys: tuple
        :param kwargs: Additional arguments
        :type kwargs: dict[str, object]
        """
        pass

    def _lineEditSaveF(self, *keys, **kwargs):
        return self._setDataF(*keys, **kwargs)

    def _textEditSaveF(self, textEdit, *keys, **kwargs):
        def save():
            setDataF(textEdit.toPlainText())

        setDataF = self._setDataF(*keys, **kwargs)
        return save

    def _checkBoxSaveF(self, *keys, **kwargs):
        def save(state):
            setDataF(bool(state))

        setDataF = self._setDataF(*keys, **kwargs)
        return save

    def _getLabel(self, *keys):
        """
        Get the label for the meta-data element at the given keys.

        :param keys: Keys for location of meta-data element
        :type keys: tuple
        :return: Label for the element at the given keys
        :rtype: str
        """
        return str(keys[-1]) if keys else ''

    def createEmptyRecordWidget(self, *keys, **kwargs):
        layout = QFormLayout()
        layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        if kwargs.get(_WGUI_RECORD_DEPTH_KEY):
            layout.setRowWrapPolicy(QFormLayout.WrapAllRows)
        layout.setContentsMargins(0,0,0,0)
        widget = QWidget()
        widget.setLayout(layout)
        widget.setProperty(WGUI_META_KEY, kwargs.get(WGUI_META_KEY))
        return widget

    def addToRecordWidget(self, recordWidget, childWidget, *keys, **kwargs):
        meta = kwargs.get(WGUI_META_KEY)
        label = meta.get(WGUI_META_LABEL_KEY) if meta else None
        layout = recordWidget.layout()
        """:type: QFormLayout"""
        layout.addRow(label or self._getLabel(*keys), childWidget)

    def finalizeRecordWidget(self, recordWidget, *keys, **kwargs):
        return recordWidget

    def createEmptyListWidget(self, *keys, **kwargs):
        return []

    def addToListWidget(self, listWidget, childWidget, *keys, **kwargs):
        listWidget.append(childWidget)

    def finalizeListWidget(self, listWidget, *keys, **kwargs):
        if len(listWidget) == 1:
            return listWidget[0]
        elif len(listWidget) > 1:
            wdgt = QWidget()
            layout = QVBoxLayout()
            layout.setContentsMargins(0,0,0,0)
            for cwdg in listWidget:
                layout.addWidget(cwdg)
            layout.addStretch(1)
            wdgt.setLayout(layout)
            wdgt.setProperty(WGUI_META_KEY, kwargs.get(WGUI_META_KEY))
            return wdgt
        else:
            return None

    def createEmptyTupleWidget(self, *keys, **kwargs):
        return self.createEmptyListWidget(*keys, **kwargs)

    def addToTupleWidget(self, tupleWidget, childWidget, *keys, **kwargs):
        self.addToListWidget(tupleWidget, childWidget, *keys, **kwargs)

    def finalizeTupleWidget(self, tupleWidget, *keys, **kwargs):
        return self.finalizeListWidget(tupleWidget, *keys, **kwargs)

    def createStringWidget(self, dtyp, *keys, data=None, **kwargs):
        meta = kwargs.get(WGUI_META_KEY)
        widget = QLineEdit(self.getData(*keys, data=data, **kwargs))
        widget.textEdited.connect(self._lineEditSaveF(*keys, **kwargs))
        if meta:
            widget.setPlaceholderText(meta.get(WGUI_META_SHORTDESC_KEY))
            widget.setToolTip(meta.get(WGUI_META_LONGDESC_KEY))
        widget.setProperty(WGUI_META_KEY, meta)
        return widget

    def createTextWidget(self, dtyp, *keys, data=None, **kwargs):
        meta = kwargs.get(WGUI_META_KEY)
        widget = QTextEdit()
        widget.setText(self.getData(*keys, data=data, **kwargs))
        widget.textChanged.connect(
            self._textEditSaveF(widget, *keys, **kwargs))
        widget.setProperty(WGUI_META_KEY, meta)
        return widget

    def createBoolWidget(self, dtyp, *keys, data=None, **kwargs):
        meta = kwargs.get(WGUI_META_KEY)
        widget = QCheckBox()
        widget.setChecked(bool(self.getData(*keys, data=data, **kwargs)))
        widget.stateChanged.connect(self._checkBoxSaveF(*keys, **kwargs))
        widget.setProperty(WGUI_META_KEY, meta)
        return widget

    def createTimeWidget(self, dtyp, *keys, data=None):
        return self.createStringWidget(WDBS_STRING_META, *keys, data=None)


class _WguiTable(QWidget):
    """
    Tabular view of a record-containing list.
    """

    def __init__(self, dataType, data, widgetArgs, parent=None):
        """
        :param dataType: Data-type of record-containing list
        :type dataType: WdbsListMetaTypeP
        :param data: List data
        :type data: list[dict[str, object]]
        :param widgetArgs: Additional widget arguments
        :type widgetArgs: dict[str, object]
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Table model
        self.__model = _WguiTableModel(dataType, data, widgetArgs)

        #: Table view
        self.__table = QTableView()
        #: Button for adding new rows / list elements
        self.__addButton = QPushButton('Add')
        #: Button for removing current row / list element
        self.__delButton = QPushButton('Remove')

        self.setLayout(self._createLayout())
        self._initElements()

    def _createLayout(self):
        """
        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(0)

        layout.addWidget(self.__table)

        btnLayout = QHBoxLayout()
        btnLayout.setContentsMargins(0,0,0,0)
        btnLayout.addWidget(self.__addButton)
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__delButton)

        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        table = self.__table
        table.setModel(self.__model)
        table.setSortingEnabled(True)
        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        table.setSelectionMode(QAbstractItemView.SingleSelection)
        header = table.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.Stretch)
        table.setMinimumHeight(120)

        self.__addButton.clicked.connect(self._add)
        self.__delButton.clicked.connect(self._remove)

    def _add(self):
        self.__model.addRow()

    def _remove(self):
        idxs = self.__table.selectedIndexes()
        if not idxs:
            return
        self.__model.removeRow(idxs[0].row())


class _WguiTableModel(QAbstractTableModel):
    """
    Table model for tabular view of a record-containing list.
    """

    def __init__(self, dataType, data, widgetArgs, parent=None):
        """
        :param dataType: Data-type of record-containing list
        :type dataType: WdbsListMetaTypeP
        :param data: List data
        :type data: list[dict[str, object]]
        :param parent: Parent item
        """
        super().__init__(parent)

        #: Data type
        self.__dataType = dataType
        #: List data
        self.__data = data

        #: List of field keys
        self.__fieldKeys = widgetArgs.get(WGUI_META_FIELDS_KEY)
        assert self.__fieldKeys
        #: List of column headers
        self.__headers = widgetArgs.get(WGUI_META_HEADERS_KEY) or \
                         self.__fieldKeys

    def columnCount(self, *_):
        return len(self.__headers)

    def rowCount(self, *_):
        return len(self.__data)

    def data(self, index, role=None):
        if role != Qt.DisplayRole and role != Qt.EditRole:
            return None

        row = index.row()
        col = index.column()

        rdat = self.__data[row]
        """:type: dict[object, object] | list[object] | tuple"""

        return rdat[self.__fieldKeys[col]]

    def setData(self, index, item, role=None):
        if role != Qt.EditRole:
            return False

        row = index.row()
        col = index.column()

        rdat = self.__data[row]
        """:type: dict[object, object] | list[object] | tuple"""

        rdat[self.__fieldKeys[col]] = item

        return True

    def headerData(self, section, orientation, role=None):
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Vertical:
            return str(section)
        else:
            return self.__headers[section]

    def flags(self, index):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable

    def addRow(self):
        row = len(self.__data)
        self.beginInsertRows(QModelIndex(), row, row)
        data = self.__dataType.getValueType().getEmptyValue()
        """:type: dict[str, object]"""
        self.__data.append(data)
        self.endInsertRows()

    def removeRow(self, row, *_):
        self.beginRemoveRows(QModelIndex(), row, row)
        del self.__data[row]
        self.endRemoveRows()
