# ----------------------------------------------------------------------------
# - wgui-gui-action                                                          -
# - wild cruncher - base classes and protocols for actions                   -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_gui_action
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher base classes and protocols for actions

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtWidgets import QAction, QActionGroup

from silkan_common.scmn_utils import first
from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_ENABLED_CHANGE_EVENT, WactStateSelectionActionP, \
    WACT_ACTION_EXECUTION_SUCCEEDED_EVENT
from wcr_core.action.wact_action_parameter_api import \
    WactFixedSelectionActionParameterP
from wcr_core.action.wact_action_parameter_types import *
from wcr_frontend.gui.wgui_action_execution_context import \
    WguiActionExecutionContextMixin
from wcr_frontend.gui.wgui_action_execution_context_api import \
    WguiActionExecutionContextP
from wcr_frontend.gui.wgui_project_editor import WguiProjectEditor
from wcr_frontend.gui.wgui_protocols import QObject_Meta_M
# noinspection PyUnresolvedReferences
from wcr_frontend.gui.wgui_utils import WGUI_LAST_DIR


# noinspection PyAbstractClass
class WguiActionAdapter(QAction, WactActionObserverP,
                        metaclass=QObject_Meta_M):
    def __init__(self, code, action, actx, hideWhenDisabled, parent):
        """
        :param code: Action code
        :type code: str
        :param action: Action delegate
        :type action: WactActionP
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param hideWhenDisabled: Whether to hide the action if disabled
        :type hideWhenDisabled: bool
        :param parent: Parent widget
        """
        # noinspection PyArgumentList
        super().__init__(action.getName(), parent)

        self.__action = action
        self.__code = code
        self.__actx = actx
        self.__hideWhenDisabled = hideWhenDisabled

        self.setStatusTip(action.getDescription())

        action.registerActionObserver(self)
        # Register with itself to handle triggers internally
        self.triggered.connect(self._triggered)

    @property
    def actionContext(self):
        """
        Get the action context.

        :return: Action context
        :rtype: WguiActionExecutionContextP
        """
        return self.__actx

    @property
    def code(self):
        """
        :return: Action code
        :rtype: str
        """
        return self.__code

    @property
    def action(self):
        """
        Get the wrapped action.

        :return: The wrapped action
        :rtype: WactActionP
        """
        return self.__action

    def update(self):
        """
        Update the state of the action.
        """
        if self.__hideWhenDisabled:
            self.setVisible(self.canExecute())
        else:
            self.setEnabled(self.canExecute())

    def canExecute(self):
        """
        Whether the action can currently execute.

        :return: True iff the action can execute
        :rtype: bool
        """
        return self.__action.isEnabled(self.__actx)

    def _triggered(self):
        """
        Slot callback for when the action has been triggered.
        """
        if self.__action.isEnabled(self.__actx):
            self.__action.execute(WguiOneShotExecutionContext(self.__actx))

    def actionNotify(self, actx, action, event, *args, **kwargs):
        if event == WACT_ACTION_ENABLED_CHANGE_EVENT:
            self.update()


class WguiToggleActionAdapter(WguiActionAdapter):
    def __init__(self, code, action, actx, hideWhenDisabled, parent):
        """
        :param code: Action code
        :type code: str
        :param action: Action delegate
        :type action: WactStateSelectionActionP
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param hideWhenDisabled: Whether to hide the action if disabled
        :type hideWhenDisabled: bool
        :param parent: Parent widget
        """
        super().__init__(code, action, actx, hideWhenDisabled, parent)

        self.__action = action
        self.__parameter = first(action.getParameters())

        self.setCheckable(True)
        self.setChecked(self.__action.getCurrentValue(actx))

    def _triggered(self):
        super()._triggered()
        self.setChecked(self.__action.getCurrentValue(self.actionContext))


class WguiParamValueActionAdapter(WguiActionAdapter):
    def __init__(self, code, action, param, value, actx, hideWhenDisabled,
                 parent):
        """
        :param code: Action code
        :type code: str
        :param action: Action delegate
        :type action: WactActionP
        :param param: Action parameter
        :type param: WactActionParameterP
        :param value: Parameter value (val, text, desc)
        :type value: (object, str, str)
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param hideWhenDisabled: Whether to hide the action if disabled
        :type hideWhenDisabled: bool
        :param parent: Parent widget
        """
        super().__init__(
            code + '-' + str(value[0]), action, actx, hideWhenDisabled, parent)

        self.__action = action
        self.__param = param
        self.__value = value

        self.setText(action.getName() + ' (' + value[1] + ')')
        self.setStatusTip(value[2])

    def _triggered(self):
        self.__action.execute(
            WguiOneShotExecutionContext(self.actionContext),
            **{self.__param.getCode(): self.__value[0]})


class WguiGroupSetActionAdapter(WguiActionAdapter):
    def __init__(self, code, action, value, actx, hideWhenDisabled, parent):
        """
        :param code: Action code
        :type code: str
        :param action: Action delegate
        :type action: WactStateSelectionActionP
        :param value: Value represented by group item (val, title, desc)
        :type value: (object, str, str)
        :param actx: Action execution context
        :type actx: WguiActionExecutionContextP
        :param hideWhenDisabled: Whether to hide the action if disabled
        :type hideWhenDisabled: bool
        :param parent: Parent widget
        """
        super().__init__(
            code + '-' + str(value[0]), action, actx, hideWhenDisabled, parent)

        self.__action = action
        self.__parameter = first(action.getParameters())
        self.__value = value
        self.__siblings = []

        self.setCheckable(True)
        self.setText(value[1])
        self.setStatusTip(value[2])
        self.setChecked(self.__action.getCurrentValue(actx) == self.__value[0])

    def update(self):
        super().update()
        cval = self.__action.getCurrentValue(self.actionContext)
        if cval == self.__value[0]:
            self.setChecked(True)

    def _triggered(self):
        self.__action.execute(
            WguiOneShotExecutionContext(self.actionContext),
            **{self.__parameter.getCode(): self.__value[0]})


def wgui_wrap_action(action, code, actx, parent, **kwargs):
    """
    Wrap an action into a Qt action.

    :param action: Action to wrap
    :type action: WactActionP
    :param code: Action code:
    :type code: str
    :param actx: Action execution context
    :type actx: WguiActionExecutionContextP
    :param parent: Parent widget
    :return: Wrapped action
    :rtype: list[WguiActionAdapter]
    """
    params = list(action.getParameters())
    hideWhenDisabled = bool(kwargs.get('module'))

    if len(params) == 1:
        param = params[0]
        if isinstance(action, WactStateSelectionActionP):
            if isinstance(param.getParamType(), WactActionBoolParameterType):
                return [WguiToggleActionAdapter(
                    code, action, actx, hideWhenDisabled, parent)]
            else:
                actions = []
                actionGrp = QActionGroup(parent)
                for pv in action.getPossibleValues():
                    # noinspection PyTypeChecker
                    a = WguiGroupSetActionAdapter(
                        code, action, pv, actx, hideWhenDisabled, parent)
                    actionGrp.addAction(a)
                    actions.append(a)
                return actions
        elif isinstance(param, WactFixedSelectionActionParameterP):
            possVal = list(param.getPossibleValues())
            if len(possVal) <= 2:
                actions = []
                for pv in possVal:
                    a = WguiParamValueActionAdapter(
                        code, action, param, pv, actx, hideWhenDisabled,
                        parent)
                    actions.append(a)
                return actions

    return [WguiActionAdapter(code, action, actx, hideWhenDisabled, parent)]


class WguiOneShotExecutionContext(WguiActionExecutionContextMixin):
    def __init__(self, delegate):
        """
        :param delegate:
        :type delegate: WguiActionExecutionContextP
        """
        self.__delegate = delegate

        self.__projectCtx = delegate.getProjectContext()
        """:type: WcorProjectContextP"""
        self.__editor = delegate.getEditor()

    def getProjectContext(self):
        return self.__projectCtx

    def getEditor(self):
        return self.__editor

    def addMessage(self, mtxt, mcmp=None, mcat=None, mlvl=None, trce=None):
        if self.__projectCtx:
            self.__projectCtx.addMessage(
                mtxt, mcmp=mcmp, mcat=mcat, mlvl=mlvl, trce=trce)
        else:
            self.__delegate.addMessage(
                mtxt, mcmp=mcmp, mcat=mcat, mlvl=mlvl, trce=trce)

    def getValue(self, parameter):
        return self.__delegate.getValue(parameter)

    def getParent(self):
        return self.__delegate.getParent()


class WguiProjectEditorRefreshHandler(WactActionObserverP):
    def actionNotify(self, actx, action, event, *args, **kwargs):
        """
        :type actx: WguiActionExecutionContextP
        """
        if event & WACT_ACTION_EXECUTION_SUCCEEDED_EVENT:
            editor = actx.getEditor()
            if isinstance(editor, WguiProjectEditor):
                print('EDITOR REFRESH')
                editor.refresh()
