# ----------------------------------------------------------------------------
# - wgui-logger                                                              -
# - wild cruncher - cockpit : system logger window                           -
# ----------------------------------------------------------------------------
# - author    (c) 2012-2012 amaury darsch                                    -
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_text_editor
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher text editor

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

# ----------------------------------------------------------------------------
# - import section                                                           -
# ----------------------------------------------------------------------------

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QComboBox, QTextEdit, QHBoxLayout, \
	QVBoxLayout, QPushButton
from PyQt5.QtGui import QFont, QTextCursor

from wcr_frontend.gui.wgui_editor import WguiEditorBase, \
    WguiEditorHasDeleteP, WguiEditorHasRefreshP, WguiEditorHasSearchP
from wcr_frontend.gui.wgui_protocols import WguiChangeNotifyP
from wcr_frontend.gui.wgui_utils import wgui_print_text, wgui_qcb_get_data
from wcr_frontend.gui.wgui_widgets import WguiTextSearchWidget
from silkan_common.scmn_log import *


# ----------------------------------------------------------------------------
# - class section                                                            -
# ----------------------------------------------------------------------------

WGUI_LOGGER_IDENT = 'wcr:logger'


class WguiLoggerWindow(WguiEditorBase, WguiEditorHasDeleteP,
                       WguiEditorHasRefreshP, WguiEditorHasSearchP,
                       WguiChangeNotifyP, ScmnLogListener):
    """
    The WguiLogger class is the class responsible to display the system
    logger messages.
    """

    MESSAGE_COLOURS = dict({
        SCMN_LOG_INFO: Qt.black,
        SCMN_LOG_WARN: Qt.blue,
        SCMN_LOG_ERROR: Qt.red,
    })

    LOG_LEVELS = []
    for l in range(10):
        LOG_LEVELS.append((SCMN_LOG_INFO, l, 'Info (%d)' % l))
    LOG_LEVELS.append((SCMN_LOG_WARN, 0, 'Warnings'))
    LOG_LEVELS.append((SCMN_LOG_ERROR, 0, 'Errors'))

    __contentChanged = pyqtSignal()

    def __init__(self, msgSource, parent=None):
        """
        Create a new logger window.

        :param msgSource: The message source
        :type msgSource: ScmnMessageSourceP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Application context
        self.__msgSource = msgSource

        #: Text display widget
        self.__textEdit = QTextEdit()
        #: Search widget
        self.__searchWidget = WguiTextSearchWidget(self.__textEdit)
        #: Marker button
        self.__markerButton = QPushButton('Add Marker')
        #: Log level combo box
        self.__logLevelSelect = QComboBox()

        #: Minimum message counter (set by 'delete')
        self.__minCounter = 0
        #: Minimum message counter for single-message adding
        self.__addCounter = 0

        self._initComponents()
        self.setLayout(self._createLayout())

        msgSource.addMessageListener(self)
        self.refresh()

    def _initComponents(self):
        """
        Initialize all UI components as needed.
        """
        # attach a courier like font
        font = QFont()
        font.setFamily('Courier')
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.__textEdit.setFont(font)

        self.__logLevelSelect.addItem('Filter Level')
        self.__logLevelSelect.setItemData(0, (0, 0))
        for cat, lvl, txt in self.LOG_LEVELS:
            idx = self.__logLevelSelect.count()
            self.__logLevelSelect.addItem(txt)
            self.__logLevelSelect.setItemData(idx, (cat, lvl))

        self.__markerButton.clicked.connect(self._addMarker)
        self.__logLevelSelect.currentIndexChanged.connect(self.refresh)

    def _createLayout(self):
        """
        Create the layout for the logger widget.

        :return: Layout for the logger widget
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(5,5,5,5)
        layout.setSpacing(2)

        self.__textEdit.setReadOnly(True)
        layout.addWidget(self.__textEdit)
        layout.setStretch(0, 1)

        btmLayout = QHBoxLayout()
        btmLayout.setContentsMargins(0,0,0,0)
        btmLayout.setSpacing(0)
        btmLayout.setAlignment(Qt.AlignVCenter)

        btmLayout.addWidget(self.__searchWidget)

        fltrLayout = QHBoxLayout()
        fltrLayout.setContentsMargins(0,0,0,0)
        fltrLayout.setSpacing(0)
        fltrLayout.addWidget(self.__logLevelSelect)
        btmLayout.addStretch(1)
        btmLayout.addLayout(fltrLayout)

        btnLayout = QHBoxLayout()
        btnLayout.setContentsMargins(0,0,0,0)
        btnLayout.setSpacing(0)
        btnLayout.addWidget(self.__markerButton)
        btmLayout.addStretch(1)
        btmLayout.addLayout(btnLayout)
        layout.addLayout(btmLayout)

        return layout

    def getTitle(self):
        return 'Logger'

    def getIcon(self):
        return None

    def getEditorIdent(self):
        return WGUI_LOGGER_IDENT

    def isReadOnly(self):
        return True

    def isICDProject(self):
        return False

    def delete(self):
        """
        Delete the system logger content.
        """
        lastMessage = self.__msgSource.lastMessage()
        if lastMessage is not None:
            self.__minCounter = lastMessage.counter
        self.refresh()

    def activated(self):
        self._addMessages()

    def _addMessages(self):
        since = max(self.__addCounter, self.__minCounter)
        if self.__msgSource.lastMessageCounter() <= since:
            return

        cat, lvl = wgui_qcb_get_data(self.__logLevelSelect)
        messages = self.__msgSource.messages(category=cat, minLevel=lvl,
                                             since=since)
        if not messages:
            return

        docLen = self.__textEdit.document().characterCount()
        maxCount = since

        for msg in messages:
            self.__textEdit.setTextColor(self.MESSAGE_COLOURS[msg.category])
            self.__textEdit.append(msg.formatted())
            maxCount = msg.counter

        vscb = self.__textEdit.verticalScrollBar()
        vscb.setValue(vscb.maximum())

        cursor = QTextCursor(self.__textEdit.document())
        cursor.setPosition(docLen)
        self.__searchWidget.highlight(cursor)

        self.__addCounter = maxCount
        return maxCount

    def refresh(self):
        self.__addCounter = 0
        self.__textEdit.clear()
        self._addMessages()

    def print(self):
        wgui_print_text(self.__textEdit, parent=self.parentWidget())

    def copy(self):
        self.__textEdit.copy()

    def cut(self):
        self.__textEdit.copy()

    def paste(self):
        pass

    def selectAll(self):
        self.__textEdit.selectAll()

    def messageAdded(self, *_):
        if self.isVisible():
            self._addMessages()
        self.__contentChanged.emit()

    def activateFind(self):
        self.__searchWidget.setFocus(Qt.OtherFocusReason)

    def findAgain(self):
        self.__searchWidget.findAgain()

    def _addMarker(self):
        line = '=' * 40
        # noinspection PyArgumentList
        time = datetime.now().strftime('%x %X')
        self.__textEdit.setTextColor(Qt.darkGreen)
        self.__textEdit.append(line + ' ' + time + ' ' + line)

    @property
    def contentChanged(self):
        return self.__contentChanged
