# ----------------------------------------------------------------------------
# - wgui_object_selection_dialog                                             -
# - wild cruncher: object selection dialog                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 samuel longchamps                                -
# - copyright (c) 2012-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_object_selection_dialog
   :platform:     Unix, Darwin, Windows
   :synopsis:     Wild Cruncher - object selection dialog

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""

from PyQt5.QtWidgets import QComboBox, QDialog, QVBoxLayout, QLabel, QDialogButtonBox
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data

_WGUI_OBJSEL_DLG_SHOW_DESC_DEFAULT = False
_WGUI_OBJSEL_DLG_ACCEPT_LBL_DEFAULT = "&OK"


class WguiObjectSelectionDialog(QDialog):
    """
    Dialog to allow the user to choose between different objects
    """

    def __init__(self, parent, message, options, title=None, iconPrv=None,
                 acceptLbl=None, showDesc=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        :param message: Dialog message label
        :type message: str
        :param options: List of options as tupples that contain:
            [0] The object to be returned for this option
            [1] Name of the option to be shown in the combo-box
            [2] Description of the option
        :type options: list[(object, str, str)]
        :param title: Dialog title label
        :type title: str
        :param iconPrv: Icon provider object
        :type iconPrv: IconPrvP | None
        :param acceptLbl: Label of the validation button
        :type acceptLbl: str
        :param showDesc: Flag to enable the description of the options to be
            shown in the label of the combo-box
        :type showDesc: bool | None
        """
        if not options:
            raise ValueError('Cannot instantiate object selection dialog: '
                             'No options defined')

        super().__init__(parent)

        self.__selectedObj = None
        self.__message = message
        self.__options = options

        self.__iconPrv = iconPrv
        self.__acceptLbl = acceptLbl if acceptLbl is not None else \
            _WGUI_OBJSEL_DLG_ACCEPT_LBL_DEFAULT
        self.__showDesc = showDesc if showDesc is not None else \
            _WGUI_OBJSEL_DLG_SHOW_DESC_DEFAULT
        if title:
            self.setWindowTitle(title)

        self.__comboBox = QComboBox()
        self.setLayout(self._createLayout())
        self._fillComboBox()

    def _createLayout(self):
        """
        Create the layout for the dialog
        """
        dlgLayout = QVBoxLayout()
        dlgLayout.addWidget(QLabel(self.__message))
        dlgLayout.addWidget(self.__comboBox)

        buttons = QDialogButtonBox()
        buttons.addButton(self.__acceptLbl, QDialogButtonBox.AcceptRole)
        buttons.addButton(QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        dlgLayout.addWidget(buttons)

        return dlgLayout

    def _fillComboBox(self):
        """
        Clear the combo-box and fill-in the options, including icons and
        adding a description if the showDesc flag is True
        """
        self.__comboBox.clear()

        for option in self.__options:
            label = option[1]
            if self.__showDesc:
                label += ' - ' + option[2]

            icon = None
            if self.__iconPrv:
                icon = self.__iconPrv.getIcon(str(option[0]))

            cbParams = (icon, label, option[0]) if icon else \
                       (label, option[0])

            self.__comboBox.addItem(*cbParams)

    def accept(self):
        """
        Set the current selection as the selected object and close the dialog
        """
        self.__selectedObj = wgui_qcb_get_data(self.__comboBox)
        super().accept()

    def getChosenOption(self):
        """
        :return: The object that was selected by the user, or None if the
            dialog has not yet been accepted or was cancelled by user
        :rtype: object | None
        """
        return self.__selectedObj
