# ----------------------------------------------------------------------------
# - wgui-props-dialog-widget                                                 -
# - wild cruncher - cockpit: general property dialog widgets                 -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_props_dialog_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher general property dialog widgets

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from abc import abstractmethod, ABCMeta
from collections import OrderedDict, Counter
from functools import partial
import math
import traceback
import ast

from PyQt5 import QtCore
from PyQt5.QtCore import Qt, pyqtSignal, QObject, QSize, QSortFilterProxyModel
from PyQt5.QtWidgets import QComboBox, QDialog, QCheckBox, QHBoxLayout, \
	QWidget, QRadioButton, QVBoxLayout, QPushButton, QLabel, QTableWidget, \
	QTableWidgetItem, QSplitter, QSizePolicy, QGroupBox, QLineEdit, \
	QTabWidget, QDialogButtonBox, QHeaderView, QListView, QAbstractItemView, \
	QGridLayout, QSpacerItem
from PyQt5.QtGui import QBrush, QStandardItemModel, \
    QStandardItem, QPixmap, QIcon
from wcr_core.db_api.whdb_protocols_api import WhdbHasMediaP

from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data, wgui_getOpenFile
from wcr_core.db.whdb_model import whdb_getMimeForModelFile
from wcr_core.db.whdb_utils import whdb_getPortNamesForEntryPoint
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_TYPE, \
    WHDB_XPOINT_DIRECTION, WHDB_XPOINT_NAME
from wcr_core.dbstore_api.wdbs_connection_point_api import WDBS_XPNT_IN, \
    WDBS_XPNT_OUT, WDBS_XPNT_BI
from wcr_core.wcor_context import wcor_getConfigDescription
from wcr_frontend.gui.wgui_protocols import QWidget_Meta_M, QObject_Meta_M
from silkan_common.scmn_file_path import scmn_isfile, scmn_expshenv, \
    scmn_path_dir
from silkan_common.scmn_log import SCMN_LOG_ERROR


WGUI_NAME_PLACEHOLDER = '[New]'

WGUI_XPNT_CODE = '__xpnt'
_WHGUI_CARD_CONFIG = 'Card Properties'

class WguiPropertiesWidgetP(QWidget, metaclass=QWidget_Meta_M):
    """
    Property dialog widget.

    """

    DISABLED_BRUSH = QBrush(Qt.darkGray)

    def _dis(self, item):
        """
        Make the given widget disabled.

        :param item: Widget to disable
        :type item: QWidget | QTableWidgetItem
        :return: The widget given as parameter
        :rtype: QWidget | QTableWidgetItem
        """
        item.setFlags(item.flags() & ~Qt.ItemIsEditable)
        item.setForeground(self.DISABLED_BRUSH)
        return item

    @abstractmethod
    def updateFields(self):
        """
        Update the widget's fields to reflect the current version of the
        associated data item.
        """
        raise NotImplementedError()

    @abstractmethod
    def save(self):
        """
        Save the data currently held by the widget back to the source data
        structure.
        """
        raise NotImplementedError()


class WguiMultiPropertiesWidgetP(WguiPropertiesWidgetP):
    """
    Composes multiple property widgets into one.
    """

    def __init__(self, *widgets):
        """
        :param widgets: Widgets to be composed into one
        :type widgets: Iterable[WguiPropertiesWidgetP]
        """
        super().__init__()

        self.__widgets = widgets

        self.setLayout(self._createLayout())

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        for widget in self.__widgets:
            layout.addWidget(widget)
        layout.addStretch(2)
        return layout

    def updateFields(self):
        for widget in self.__widgets:
            widget.updateFields()

    def save(self):
        hasErr = False
        for widget in self.__widgets:
            widgErr = widget.save()
            if widgErr:
                hasErr = True
        return hasErr


class WguiGroupedMultiPropertiesWidgetP(WguiPropertiesWidgetP):
    """
    Composes multiple property widgets as groups into one.
    """

    def __init__(self, *widgets):
        """
        :param widgets: Widgets to be composed into one
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        """
        super().__init__()

        self.__widgets = widgets

        self.setLayout(self._createLayout())

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        for title, widget in self.__widgets:
            if title:
                group = QGroupBox(title)
                wLayout = QVBoxLayout()
                wLayout.setContentsMargins(0,0,0,0)
                wLayout.addWidget(widget)
                group.setLayout(wLayout)
            else:
                group = widget
            layout.addWidget(group)
        layout.addStretch(2)
        return layout

    def updateFields(self):
        for _, widget in self.__widgets:
            widget.updateFields()

    def save(self):
        for _, widget in self.__widgets:
            widget.save()


class WguiPropertiesDialog(QDialog):
    """
    Base class for property dialogs.
    """

    #: Signal emitting whenever the state of the dialog's Ok-button changes
    okStateChanged = QtCore.pyqtSignal(bool)

    def __init__(self, title, updateF=None, parent=None):
        """
        :param title: Dialog title
        :type title: str
        :param updateF: Callback function to notify about saves
        :type updateF: () -> _
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Callback function to notify about saves
        self.__updateF = updateF

        #: The dialog's property widgets
        self.__widgets = []

        self.setWindowTitle(title)
        self.setMinimumWidth(400)

    def setTabbedPropertyWidgets(self, *widgets, **kwargs):
        """
        Set the dialog's property widgets to the given list.
        The widgets will be added as tabs.

        Note that only one of the functions to add widgets should be used.

        :param widgets: Property widgets to add, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        :param kwargs: Additional arguments (`activate`)
        """
        tabs = QTabWidget()
        for title, widget in widgets:
            tabs.addTab(widget, title)

        if 'activate' in kwargs:
            tabs.setCurrentWidget(kwargs['activate'])

        self._setMainLayout(widgets, tabs)

    def setGroupedPropertyWidgets(self, *widgets):
        """
        Set the dialog's property widgets to the given list.
        The widgets will be added as groups in one pane.

        Note that only one of the functions to add widgets should be used.

        :param widgets: Property widgets to add, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        """
        layout = QVBoxLayout()
        for title, widget in widgets:
            group = QGroupBox(title)
            wLayout = QVBoxLayout()
            wLayout.setContentsMargins(0,0,0,0)
            wLayout.addWidget(widget)
            group.setLayout(wLayout)
            layout.addWidget(group)
        layout.addStretch(1)
        wrap = QWidget()
        wrap.setLayout(layout)

        self._setMainLayout(widgets, wrap)

    def setPropertyWidget(self, widget):
        """
        Set the dialog's single property widget.

        Note that only one of the functions to add widgets should be used.

        :param widget: Property widget
        :type widget: WguiPropertiesWidgetP
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(widget)
        layout.addStretch(1)
        wrap = QWidget()
        wrap.setLayout(layout)

        self._setMainLayout([('', widget)], wrap)

    def _setMainLayout(self, widgets, widget):
        """
        Set the dialog's property widgets.

        :param widgets: Property widgets, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        :param widget: Wrapper widget containing the property widgets
        :type widget: QWidget
        """
        self.__widgets = [widget for _, widget in widgets]

        layout = QVBoxLayout()
        layout.addWidget(widget)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Apply |
                                   QDialogButtonBox.Reset |
                                   QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        buttons.button(QDialogButtonBox.Apply).clicked.connect(self._apply)
        buttons.button(QDialogButtonBox.Reset).clicked.connect(self._reset)
        self.okStateChanged.connect(
            buttons.button(QDialogButtonBox.Ok).setEnabled)
        self.okStateChanged.connect(
            buttons.button(QDialogButtonBox.Apply).setEnabled)
        layout.addWidget(buttons)
        # create the main layout
        self.setLayout(layout)

    def accept(self):
        hasErr = self._save()
        if not hasErr:
            super().accept()

    def _apply(self):
        """
        Apply the current changes, and leave the dialog window open.
        """
        self._save()
        self._updateFields()

    def _reset(self):
        """
        Reset the dialog window.
        """
        self._updateFields()

    def _save(self):
        """
        Save all changes.
        """
        hasErr = False
        for widget in self.__widgets:
            widgErr = widget.save()
            if widgErr:
                hasErr = True

        if self.__updateF is not None:
            self.__updateF()
        return hasErr

    def _updateFields(self):
        """
        Reset all fields to their original state.
        """
        for widget in self.__widgets:
            widget.updateFields()


class WguiConfigTableProperties(WguiPropertiesWidgetP):
    """
    Configuration table widget.

    UI configuration records (keys starting with 'gui.')can be excluded from
    display as they are usually handled by dedicated widgets.
    """

    #: Default key column width
    DEFAULT_COL_WIDTH = 150

    def __init__(self, configTable, showUiEntries=False, defaults=None,
                 parent=None):
        """
        :param configTable: Configuration table
        :type configTable: ScmnConfigTable
        :param showUiEntries: Whether to display UI configuration records
        :type showUiEntries: bool
        :param defaults: Default configuration values
        :type defaults: dict[str, str]
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Config table
        self.__configTable = configTable
        #: Whether to show UI entries
        self.__showUiEntries = showUiEntries
        #: Default values
        self.__defaults = defaults

        self.__table = QTableWidget()

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        mainLayout = QVBoxLayout()

        table = self.__table
        table.setColumnCount(2)
        table.setColumnWidth(0, self.DEFAULT_COL_WIDTH)
        table.verticalHeader().hide()
        table.horizontalHeader().setStretchLastSection(True)
        table.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        table.setSortingEnabled(True)
        table.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        mainLayout.addWidget(table)

        button = QPushButton('Add new')
        button.clicked.connect(self._addRow)
        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(button)
        mainLayout.addLayout(btnLayout)

        return mainLayout

    def updateFields(self):
        table = self.__table
        table.clear()
        table.setHorizontalHeaderLabels(['Key', 'Value'])

        defaults = {}
        if self.__defaults is not None:
            defaults.update(self.__defaults)
        records = {k: (None, wcor_getConfigDescription(k), v)
                   for k, v in defaults.items()}
        records.update({k: (r.getValue(), r.getInfo(), defaults.get(k))
                        for k, r in self.__configTable
                        if self.__showUiEntries or not k.startswith('gui.')})
        records = OrderedDict(sorted(records.items()))
        table.setRowCount(len(records))

        row = 0
        for key, (value, info, default) in records.items():
            kitm = self._dis(QTableWidgetItem(key))
            table.setItem(row, 0, kitm)

            vwgt = QLineEdit()
            vwgt.setText(value)
            vwgt.setToolTip(info)
            vwgt.setPlaceholderText(default)
            table.setCellWidget(row, 1, vwgt)
            row += 1

    def _addRow(self):
        """
        Slot function to create new rows.
        """
        table = self.__table
        table.setRowCount(table.rowCount() + 1)
        row = table.rowCount() - 1
        kwgt = QLineEdit()
        kwgt.setPlaceholderText('Key')
        table.setCellWidget(row, 0, kwgt)
        vwgt = QLineEdit()
        vwgt.setPlaceholderText('Value')
        table.setCellWidget(row, 1, vwgt)
        kwgt.setFocus(Qt.TabFocusReason)
        table.setCurrentCell(row, 0)
        kwgt.setFocus(Qt.TabFocusReason)

    def save(self):
        # noinspection PyShadowingNames
        def getText(row, col):
            if isinstance(tableView.cellWidget(row, col), QLineEdit):
                return tableView.cellWidget(row, col).text()
            elif tableView.item(row, col) is not None:
                return tableView.item(row, col).text()
            else:
                raise Exception('Cannot retrieve text for r%d c%d' %
                                (row, col))

        tableView = self.__table
        for row in range(tableView.rowCount()):
            key = getText(row, 0)

            if key:
                self.__configTable.setRecordValue(key, getText(row, 1) or None)


class WguiFlipRotationProperties(WguiPropertiesWidgetP):
    def __init__(self, inst, parent=None):
        """
        :param inst: the entity object with flip/rotation functionality
        :type inst: wcr_core.db_api.whdb_protocols_api.WhdbCanFlipRotateP
        :param parent: Parent widget
        """
        # base initialization
        super().__init__(parent)

        #: Instance object
        self.__instance = inst

        #: 0 degree rotation radio button
        self.__rotation000dF = QRadioButton('0°')
        #: 90 degree rotation radio button
        self.__rotation090dF = QRadioButton('90°')
        #: 180 degree rotation radio button
        self.__rotation180dF = QRadioButton('180°')
        #: 270 degree rotation radio button
        self.__rotation270dF = QRadioButton('270°')
        #: Vertical flip checkbox
        self.__verticalFlipF = QCheckBox('Vertically')
        #: Horizontal flip checkbox
        self.__horizontalFlipF = QCheckBox('Horizontally')

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        grpLayout = QHBoxLayout()
        grpLayout.addWidget(self.__rotation000dF)
        grpLayout.addWidget(self.__rotation090dF)
        grpLayout.addWidget(self.__rotation180dF)
        grpLayout.addWidget(self.__rotation270dF)
        grpRotation = QGroupBox('Rotation:')
        grpRotation.setFlat(True)
        grpRotation.setLayout(grpLayout)

        grpLayout = QHBoxLayout()
        grpLayout.addWidget(self.__verticalFlipF)
        grpLayout.addWidget(self.__horizontalFlipF)
        grpFlip = QGroupBox('Mirroring:')
        grpFlip.setFlat(True)
        grpFlip.setLayout(grpLayout)

        layout = QVBoxLayout()
        layout.addWidget(grpRotation)
        layout.addWidget(grpFlip)
        layout.addStretch(1)

        return layout

    def updateFields(self):
        inst = self.__instance

        self.__rotation000dF.setChecked(False)
        self.__rotation090dF.setChecked(False)
        self.__rotation180dF.setChecked(False)
        self.__rotation270dF.setChecked(False)
        r = inst.getRotation()
        if r < math.pi / 4:
            self.__rotation000dF.setChecked(True)
        elif r < 3 * math.pi / 4:
            self.__rotation090dF.setChecked(True)
        elif r < 5 * math.pi / 4:
            self.__rotation180dF.setChecked(True)
        else:
            self.__rotation270dF.setChecked(True)

        self.__verticalFlipF.setChecked(inst.isFlippedVertically())
        self.__horizontalFlipF.setChecked(inst.isFlippedHorizontally())

    def save(self):
        inst = self.__instance

        if self.__rotation090dF.isChecked():
            inst.setRotation(math.pi / 2)
        elif self.__rotation180dF.isChecked():
            inst.setRotation(math.pi)
        elif self.__rotation270dF.isChecked():
            inst.setRotation(3 * math.pi / 2)
        else:
            inst.setRotation(0)

        inst.setFlippedVertically(self.__verticalFlipF.isChecked())
        inst.setFlippedHorizontally(self.__horizontalFlipF.isChecked())


class WguiCustomIconProperties(WguiPropertiesWidgetP):
    def __init__(self, inst, tops, parent=None):
        """
        :param inst: the entity object with a custom icon
        :type inst: wcr_core.db_api.whdb_protocols_api.WhdbHasCustomIconP
        :param tops: Project structure
        :type tops: WprjProjectStructure
        :param parent: Parent widget
        """
        # base initialization
        super().__init__(parent)

        #: Instance object
        self.__instance = inst
        #: Project structure
        self.__tops = tops

        #: Custom text
        self.__textW = QLineEdit()
        #: Icon path
        self.__iconPathW = QLineEdit()
        self.__hideBboxW = QCheckBox(
            'Hide bounding box (only if icon has been set)')
        #: Icon preview
        self.__iconPreviewW = QLabel()
        self.__fileButtonW = QPushButton()

        self.__iconPathW.textChanged.connect(self._updatePreview)

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        layout = QGridLayout()
        layout.addWidget(QLabel('Custom Text:'), 0, 0)
        layout.addWidget(self.__textW, 0, 1)

        layout.addWidget(QLabel('Icon:'), 1, 0)
        lineLayout = QHBoxLayout()
        lineLayout.setContentsMargins(0,0,0,0)
        lineLayout.addWidget(self.__iconPathW)
        self.__fileButtonW.setIcon(QIcon(':/img/app_was_open.png'))
        self.__fileButtonW.clicked.connect(self._selectPath)
        lineLayout.addWidget(self.__fileButtonW)
        layout.addLayout(lineLayout, 1, 1)

        layout.addWidget(self.__hideBboxW, 2, 1)

        prvwLayout = QHBoxLayout()
        prvwLayout.setContentsMargins(0,0,0,0)
        prvwLayout.addStretch(1)
        prvwLayout.addWidget(self.__iconPreviewW)
        prvwLayout.addStretch(1)
        layout.addLayout(prvwLayout, 3, 1)

        return layout

    def updateFields(self):
        inst = self.__instance

        self.__textW.setText(inst.getCustomText())
        self.__iconPathW.setText(inst.getCustomIcon())
        self.__hideBboxW.setChecked(inst.isBoundingBoxHidden())
        self.__hideBboxW.setEnabled(bool(inst.getCustomIcon()))

    def save(self):
        inst = self.__instance
        path = self.__iconPathW.text()

        inst.setCustomText(self.__textW.text())
        inst.setCustomIcon(path)
        inst.setBoundingBoxHidden(self.__hideBboxW.isChecked())

        if scmn_isfile(scmn_expshenv(path)):
            with open(self.__iconPathW.text(), 'rb') as f:
                self.__tops.setResource(path, f.read())
        else:
            # noinspection PyTypeChecker
            self.__tops.setResource(path, None)

    def _selectPath(self):
        # noinspection PyTypeChecker
        file = wgui_getOpenFile(
            'Select Icon File', self,
            filterStr='Icon Image Files (*.png *.jpg *.jpeg *.gif *.svg)',
            idir=scmn_path_dir(scmn_expshenv(self.__iconPathW.text())))
        if file:
            self.__iconPathW.setText(file)

    def _updatePreview(self):
        path = self.__iconPathW.text()
        if scmn_isfile(scmn_expshenv(path)):
            pixmap = QPixmap(path)
            if pixmap.width() > 128 or pixmap.height() > 128:
                pixmap = pixmap.scaled(QSize(128, 128), Qt.KeepAspectRatio)
            self.__iconPreviewW.setPixmap(pixmap)
        else:
            self.__iconPreviewW.setPixmap(QPixmap())
        self.__hideBboxW.setEnabled(bool(path))


class WguiInstanceBaseProperties(WguiPropertiesWidgetP):
    """
    Configuration widget for a model-/function-instance's base data.
    """

    def __init__(self, typeName, inst, tophm, parent=None):
        """
        :param typeName: Type name of the displayed element
        :type typeName: str
        :param inst: the instance object
        :type inst: WhdbInstanceP
        :param tophm: Hierarchy top level
        :type tophm: WhdbBlock
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Type name
        self.__typeName = typeName
        #: instance object
        self.__instance = inst
        #: Hierarchy top level
        self.__tophm = tophm

        #: Name field
        self.__nameF = QLineEdit()
        self.__nameF.setFixedWidth(400)
        #: Info field
        self.__infoF = QLineEdit()
        self.__infoF.setFixedWidth(400)

        if isinstance(inst, WhdbHasMediaP):
            #: Media type field
            self.__mediaF = QLineEdit()
            self.__mediaF.setEnabled(False)
        else:
            self.__mediaF = None

        self.setLayout(self._createLayout())
        self.updateFields()

    def _createLayout(self):
        typeName = self.__typeName
        container = QWidget()
        layout = QVBoxLayout()
        if self.__mediaF is not None:
            layout.addWidget(QLabel('Media Type:'))
            layout.addWidget(self.__mediaF)
        layout.addWidget(QLabel(typeName + ' Name:'))
        layout.addWidget(self.__nameF)
        layout.addWidget(container)
        layout.addWidget(QLabel(typeName + ' Information:'))
        layout.addWidget(self.__infoF)
        layout.addStretch(1)
        return layout

    def updateFields(self):
        inst = self.__instance
        if self.__mediaF is not None and isinstance(inst, WhdbHasMediaP):
            self.__mediaF.setText(inst.getMedia())
        self.__nameF.setText(inst.getName())
        self.__infoF.setText(inst.getInfo())

    def save(self):
        inst = self.__instance
        if isNameValid(self.__nameF.text()):
            inst.setName(self.__nameF.text())
            inst.setInfo(self.__infoF.text())
        else:
            showErrorWindow("Name Error", NameFormatMessageError,
                            self.__nameF.text())

        allEqpNames = tuple(e.getName() for e in self.__tophm.getOperatorInstances())
        counter = Counter(allEqpNames)
        commons = counter.most_common()
        if commons[0][1] > 1:
            showErrorWindow("Duplicate Name Error", "The name {} is already used for another equipement!",
                            commons[0][0])
            return True


class WguiXPointPropsModelP(metaclass=ABCMeta):
    """
    Model to represent connection point data.
    Data entries can be for existing connection points or for new, yet to be
    created, connection points.
    Data is referred to by index (identifying connection point entry) and field
    codes (identifying the data item, e.g. 'name')
    """

    @property
    @abstractmethod
    def xpointAdded(self):
        """
        Signal for newly added connection point entries.
        Carries the index of the new connection point entry.

        :rtype: pyqtSignal[int]
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def xpointRemoved(self):
        """
        Signal for removed connection point entries.
        Carries the index of the removed connection point entry,
        or 0 if all connection points have been removed.

        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def modelReset(self):
        """
        Signal for model resets.

        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def valueChanged(self):
        """
        Signal for a changed value for a connection point entry.
        Carries the index of the affected connection point (`int`) and the
        value code (`str`).

        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def addXPoint(self, **kwargs):
        """
        Add an entry for a new connection point to the model.

        :param kwargs: Code-value pairs
        :return: Index of the new connection point entry
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def removeXPoint(self, index):
        """
        Remove the connection point entry with the given index.

        :param index: Index of the connection point to be removed
        :type index: int
        """
        raise NotImplementedError()

    @abstractmethod
    def setXPoints(self, xpnts):
        """
        Set the model's entries to the given set of connection point entries.
        Will remove all existing model entries.

        :param xpnts: New connection point entries
        :type xpnts: Iterable[dict[str, object]]
        """
        raise NotImplementedError()

    @abstractmethod
    def getXPointCount(self):
        """
        Get the number of connnection point entries.

        :return: Number of connnection point entries
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def getIndices(self):
        """
        Get the indices for all connection point entries.

        :return: Indices for all connection point entries
        :rtype: Iterable[int]
        """
        raise NotImplementedError()

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError

    @abstractmethod
    def getValue(self, index, code, default=None):
        """
        Get the value for the given index/code.

        :param index: Entry index to retrieve value for
        :type index: int
        :param code: Field code to retrieve value for
        :type code: str
        :param default: Default value to return if value cannot be found
        :type default: object
        :return: Value for the given index/code, or the given default value
        :rtype: object
        """
        raise NotImplementedError()

    @abstractmethod
    def getValuesForCode(self, code):
        """
        Get the values for all entries for the given code.

        :param code: Field code to retrieve values for
        :type code: str
        :return: Values for all entries for the given code.
        :rtype: Iterable[object]
        """
        raise NotImplementedError()

    @abstractmethod
    def getValuesForIndex(self, index):
        """
        Get the values for all entries for the given index.

        :param index: Index to retrieve values for
        :type index: int
        :return: Values for all entries for the given code.
        :rtype: dict[str, object]
        """
        raise NotImplementedError()

    @abstractmethod
    def getValues(self):
        """
        Get the values for all entries.

        :return: Values for all entries for the given code.
        :rtype: Iterable[dict[str, object]]
        """
        raise NotImplementedError()

    @abstractmethod
    def setValue(self, index, code, value):
        """
        Set the value for the given index/code.

        :param index: Entry index to set value for
        :type index: int
        :param code: Field code to set value for
        :type code: str
        :param value: Value to set
        :type value: object
        """
        raise NotImplementedError()


class WguiXPointPropsModel(QObject, WguiXPointPropsModelP,
                           metaclass=QObject_Meta_M):
    _xpointAdded = pyqtSignal(int)

    _xpointRemoved = pyqtSignal(int)

    _modelReset = pyqtSignal()

    _valueChanged = pyqtSignal(int, str)

    def __init__(self, xpnts=None):
        """
        :param xpnts: Initial connection point entries
        :type xpnts: Iterable[dict[str, object]]
        """
        super().__init__()
        self.__xpnts = OrderedDict()
        if xpnts:
            self.__xpnts.update({i: x for i, x in enumerate(xpnts, start=1)})
        self.__lastIndex = len(self.__xpnts)

    @property
    def xpointAdded(self):
        return self._xpointAdded

    @property
    def xpointRemoved(self):
        return self._xpointRemoved

    @property
    def modelReset(self):
        return self._modelReset

    @property
    def valueChanged(self):
        return self._valueChanged

    def addXPoint(self, **kwargs):
        self.__lastIndex += 1
        self.__xpnts[self.__lastIndex] = kwargs
        self._xpointAdded.emit(self.__lastIndex)
        return self.__lastIndex

    def removeXPoint(self, index):
        del self.__xpnts[index]
        self._xpointRemoved.emit(index)

    def setXPoints(self, xpnts):
        self.__xpnts.clear()
        self.__lastIndex = len(self.__xpnts)
        self._xpointRemoved.emit(0)
        for idx, xpnt in enumerate(xpnts, start=self.__lastIndex + 1):
            self.__xpnts[idx] = xpnt
        self.__lastIndex += len(self.__xpnts)
        self._modelReset.emit()

    def getXPointCount(self):
        return len(self.__xpnts)

    def getIndices(self):
        return self.__xpnts.keys()

    def __contains__(self, item):
        return item in self.__xpnts

    @property
    def lastIndex(self):
        return self.__lastIndex

    def setLastIndex(self, idx):
        self.__lastIndex = idx

    def getValue(self, index, code, default=None):
        xpnts = self.__xpnts
        xpnt = xpnts[index]
        value = xpnt.get(code) if xpnt else None
        return value if value is not None else default

    def getValuesForCode(self, code):
        return (x.get(code) for x in self.__xpnts.values())

    def getValuesForIndex(self, index):
        return dict(self.__xpnts[index])

    def getValues(self):
        return [dict(xpnt) for xpnt in self.__xpnts.values()]

    def setValue(self, index, code, value):
        self.__xpnts[index][code] = value

        self._valueChanged.emit(index, code)


class WguiModuleIOPointPropsModel(WguiXPointPropsModel):
    _xpointAdded = pyqtSignal(int)

    _xpointRemoved = pyqtSignal(int)

    _modelReset = pyqtSignal()

    _valueChanged = pyqtSignal(int, str)

    def __init__(self, xpnts=None):
        """
        :param xpnts: Initial connection point entries
        :type xpnts: Iterable[dict[str, object]]
        """
        super().__init__()
        self.__xpnts = OrderedDict()
        if xpnts:
            self.__xpnts.update({i: x for i, x in enumerate(xpnts, start=1)})
        self.__lastIndex = len(self.__xpnts)


class WguiXPointPropsFieldCtx(metaclass=ABCMeta):
    """
    Context for connection point property fields, providing access to
    functionality related to the fields' instantiation/usage.
    """

    @property
    @abstractmethod
    def model(self):
        """
        Get the data model that will be used to provide the field with values.

        :return: Data model
        :rtype: WguiXPointPropsModelP
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def instance(self):
        """
        Get the instance that the field will be initiated for.

        :return: Instance
        :rtype: WhdbInstanceP
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def activeXPointChanged(self):
        """
        Get the signal for changes in the active/selected connection point
        entry.

        :return: Active connection point signal
        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def getActiveXPoint(self):
        """
        Get the index of the currently active connection point entry.

        :return: Index of the currently active connection point entry
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def setActiveXPoint(self, index):
        """
        Set the active connection point entry by its index.

        :param index: Index of the new active connection point
        :type index: int
        """
        raise NotImplementedError()


class WguiXPointPropsFieldP(metaclass=ABCMeta):
    """
    Field instantiator.
    """

    @property
    @abstractmethod
    def code(self):
        """
        Get the value code of the field.

        :return: Value code
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def label(self):
        """
        Get the label for the field.

        :return: Field label
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def instantiate(self, context):
        """
        Instantiate the field.

        :param context: Field instantiation context
        :type context: WguiXPointPropsFieldCtx
        :return: Field instantiation
        :rtype: WguiXPointPropsFieldInstanceP
        """
        raise NotImplementedError()


class WguiXPointPropsFieldInstanceP(metaclass=ABCMeta):
    """
    Field instance.
    """

    @property
    @abstractmethod
    def code(self):
        """
        Get the value code of the field.

        :return: Value code
        :rtype: str
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def label(self):
        """
        Get the label for the field.

        :return: Field label
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def installWidget(self, index, widgetF, textF):
        """
        Install the field's widget for the given connection point entry index.

        :param index: Index of the connection point entry to install widget for
        :type index: int
        :param widgetF: Function to install widget
        :type widgetF: (QWidget) -> _
        :param textF: Function to install field as text-only. Will return a
                      function to update the displayed text
        :type textF: (str) -> ((str) -> _)
        """
        raise NotImplementedError()

    @abstractmethod
    def getAuxWidget(self):
        """
        Get the auxilliary widget for the field, if any.

        :return: Auxilliary widget, or `None`
        :rtype: QWidget
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiXPointPropsFieldBase(WguiXPointPropsFieldP):
    """
    Abstract base class for field instantiators.
    """

    def __init__(self, code, label):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        """
        self.__code = code
        self.__label = label

    @property
    def code(self):
        return self.__code

    @property
    def label(self):
        return self.__label


# noinspection PyAbstractClass
class WguiXPointPropsField(WguiXPointPropsFieldBase):
    """
    Base class for field instantiators.
    """

    def __init__(self, code, label, instanceF, *args, **kwargs):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param instanceF: Instantiation function (i.e. field instance
                          constructor)
        :type instanceF: (WguiXPointPropsFieldCtx, *object, **{str:object})
                         -> WguiXPointPropsFieldInstanceP
        :param args: Additional arguments for instantiation function
        :param kwargs: Additional keyword arguments for instantiation function
        """
        super().__init__(code, label)

        self.__instanceF = instanceF
        self.__args = args
        self.__kwargs = kwargs

    def instantiate(self, context):
        return self.__instanceF(
            self.code, self.label, context, *self.__args, **self.__kwargs)


# noinspection PyAbstractClass
class WguiXPointPropsFieldInstance(WguiXPointPropsFieldInstanceP):
    """
    Abstract base class for field instances.
    """

    def __init__(self, code, label):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        """
        self.__code = code
        self.__label = label

    @property
    def code(self):
        return self.__code

    @property
    def label(self):
        return self.__label

    def getAuxWidget(self):
        return None


# noinspection PyAbstractClass
class WguiXPointPropsROTextFieldInstance(WguiXPointPropsFieldInstance):
    """
    Read-only text field.
    """

    def __init__(self, code, label, context):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        #: Value model
        self.__model = context.model

    def installWidget(self, index, widgetF, textF):
        textF(self.__model.getValue(index, self.code))


class WguiXPointPropsROTextField(WguiXPointPropsField):
    def __init__(self, code, label):
        super().__init__(code, label, WguiXPointPropsROTextFieldInstance)


# noinspection PyAbstractClass
class WguiXPointPropsROLookupFieldInstance(WguiXPointPropsFieldInstance):
    """
    Read-only text lookup field.
    """

    def __init__(self, code, label, context, source):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param source: Lookup dictionary
        :type source: dict[object, str]
        """
        super().__init__(code, label)

        #: Value model
        self.__model = context.model
        #: Value source
        self.__source = source

    def installWidget(self, index, widgetF, textF):
        textF(self.__source.get(self.__model.getValue(index, self.code)))


class WguiXPointPropsROLookupField(WguiXPointPropsField):
    def __init__(self, code, label, source):
        super().__init__(
            code, label, WguiXPointPropsROLookupFieldInstance, source)


# noinspection PyAbstractClass
class WguiXPointPropsTextFieldInstance(WguiXPointPropsFieldInstance):
    """
    Editable text field.
    """

    def __init__(self, code, label, context, placeHolder):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param placeHolder: Placeholder text
        :type placeHolder: str
        """
        super().__init__(code, label)

        self.__model = context.model
        self.__placeHolder = placeHolder

    def installWidget(self, index, widgetF, textF):
        def fieldChanged(text):
            self.__model.setValue(index, self.code, text)

        value = self.__model.getValue(index, self.code)
        field = QLineEdit(str(value) if value else None)
        if self.code == 'name':
            # Port name has a max size of 12 characters
            field.setMaxLength(12)
        if self.__placeHolder:
            field.setPlaceholderText(self.__placeHolder)
        field.textChanged.connect(fieldChanged)
        widgetF(field)


class WguiXPointPropsTextField(WguiXPointPropsField):
    def __init__(self, code, label, placeHolder=None):
        super().__init__(code, label, WguiXPointPropsTextFieldInstance,
                         placeHolder)


# noinspection PyAbstractClass
class WguiXPointPropsComboFieldInstance(WguiXPointPropsFieldInstance):
    """
    Combo-box field.
    """

    def __init__(self, code, label, context, options):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param options: Options for combo-box: [(code, label)]
        :type options: list[(str, str)]
        """
        super().__init__(code, label)

        self.__model = context.model
        self.__options = options

    def setOptions(self, options):
        self.__options = options

    def installWidget(self, index, widgetF, textF):
        def fieldChanged():
            self.__model.setValue(
                index, self.code, wgui_qcb_get_data(field))

        value = self.__model.getValue(index, self.code)
        field = QComboBox()
        for code, name in self.__options:
            field.addItem(name, code)
            if value == code:
                field.setCurrentIndex(field.count() - 1)
        field.currentIndexChanged.connect(fieldChanged)
        widgetF(field)


class WguiXPointPropsComboField(WguiXPointPropsField):
    def __init__(self, code, label, options):
        super().__init__(
            code, label, WguiXPointPropsComboFieldInstance, options)


# noinspection PyAbstractClass
class WguiXPointPropsDirectionFieldInstance(WguiXPointPropsFieldInstance):
    """
    Connection point direction field, read-only.
    """

    def __init__(self, code, label, context):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        self.__model = context.model

    def installWidget(self, index, widgetF, textF):
        cdir = self.__model.getValue(index, self.code)
        # noinspection PyTypeChecker
        textF(wgui_getXPointDirectionName(cdir) if cdir else '')


class WguiXPointPropsDirectionField(WguiXPointPropsField):
    def __init__(self, code, label):
        super().__init__(code, label, WguiXPointPropsDirectionFieldInstance)


# noinspection PyAbstractClass
class WguiXPointPropsDeleteButtonInstance(WguiXPointPropsFieldInstance):
    """
    Delete button field.
    """

    def __init__(self, code, label, context, **kwargs):
        """
        Known flags for `kwargs`: `disableable`.

        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param kwargs: Additional flags
        """
        super().__init__(code, label)

        self.__model = context.model
        self.__disableable = bool(kwargs.get('disableable'))

    def installWidget(self, index, widgetF, textF):
        if not self.showButton(index):
            return

        btn = QPushButton('Delete')
        btn.clicked.connect(partial(self.__model.removeXPoint, index))
        widgetF(btn)

        if self.__disableable:
            def added(*_):
                self.updateButtonEnabledState(index, btn)

            def removed(idx):
                if idx == index or idx <= 0:
                    # noinspection PyUnresolvedReferences
                    self.__model.xpointAdded.disconnect(added)
                    # noinspection PyUnresolvedReferences
                    self.__model.xpointRemoved.disconnect(removed)
                else:
                    self.updateButtonEnabledState(index, btn)

            self.__model.xpointAdded.connect(added)
            self.__model.xpointRemoved.connect(removed)
            self.updateButtonEnabledState(index, btn)

    def showButton(self, index):
        """
        Whether to show a delete button for the connection point entry with the
        given index.

        :param index: Connection point entry index
        :type index: int
        :return: True iff a delete button is to be shown for the given index
        :rtype: bool
        """
        return True

    def updateButtonEnabledState(self, index, btn):
        """
        Set a delete button's enabled state.

        :param index: Connection point entry index for the given button
        :type index: int
        :param btn: Delete button
        :type btn: QPushButton
        """
        pass


class WguiXPointPropsDeleteButton(WguiXPointPropsField):
    DELETE_CODE = '__delete'

    def __init__(self, label=None):
        super().__init__(self.DELETE_CODE, label or '',
                         WguiXPointPropsDeleteButtonInstance)


# noinspection PyAbstractClass
class WguiXPointPropsSinglePortROInstance(WguiXPointPropsFieldInstance):
    """
    Read-only port name field.
    """
    #: Port value key, used for model access
    PORT_KEY = 'port'

    def __init__(self, code, label, context):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        """
        super().__init__(code, label)

        self.__model = context.model

    def installWidget(self, index, widgetF, textF):
        textF(self.getPortName(index, self.__model))

    def getPortName(self, index, model):
        """
        Get the port name to be displayed for the connection point entry with
        the given index.

        :param index: Connection point entry index to return port name for
        :type index: int
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :return: Port name
        :rtype: str
        """
        return model.getValue(index, self.PORT_KEY)


class WguiXPointPropsPortsProviderP(metaclass=ABCMeta):
    """
    Port-list provider.
    """

    @property
    @abstractmethod
    def portsChanged(self):
        """
        Sifnal for port-list changes.

        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @abstractmethod
    def getPortsForXpnt(self, index, model):
        """
        Get the list of available ports for the connector entry at the given
        index in the given model.

        :param index: Connector value model index
        :type index: int
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :return: List of available ports (ID, name, direction)
        :rtype: Iterable[(str, str, ind)]
        """
        raise NotImplementedError

    @abstractmethod
    def getAllPorts(self):
        """
        Get the list of all available ports.

        :return: List of available ports (ID, name, direction)
        :rtype: Iterable[(str, str, int)]
        """
        raise NotImplementedError


# noinspection PyAbstractClass
class WguiModelPortProvider(QObject, WguiXPointPropsPortsProviderP,
                            metaclass=QObject_Meta_M):
    portsChanged = pyqtSignal()

    def __init__(self, model, epntPrv, msgTarget):
        """
        :param model: Model
        :type model: WhdbModelP
        :param epntPrv: Port-list provider
        :type epntPrv: WguiXPointPropsEntryPointProviderP
        :param msgTarget: Message target
        :type msgTarget: ScmnMessageTargetP
        """
        QObject.__init__(self)

        self.__epntPrv = epntPrv
        self.__model = model
        self.__msgTarget = msgTarget

        self.__ports = []
        self.updatePorts()
        epntPrv.entryPointChanged.connect(self.updatePorts)

    def updatePorts(self, *_):
        model = self.__model
        path, epnt = self.__epntPrv.getPathEntryPoint()

        path = scmn_expshenv(path)
        mpath = scmn_expshenv(model.getPath())

        try:
            if path == mpath and epnt == model.getEntryPoint():
                ports = model.getPortNames()
            elif path and epnt and scmn_isfile(path):
                ports = whdb_getPortNamesForEntryPoint(
                    path, epnt, whdb_getMimeForModelFile(path))
            else:
                self.__ports = []
                self.portsChanged.emit()
                return

            self.__ports = sorted(((i, n, d) for d, ps
                                   in ports.items() for i, n in ps),
                                  key=lambda p: p[1] or p[0])
            self.portsChanged.emit()
        except Exception as e:
            self.__msgTarget.addMessage(
                mtxt='Could not determine ports for {} in {} ({})'.format(
                    epnt, path, e),
                mcmp='cockpit', mcat=SCMN_LOG_ERROR,
                trce=traceback.format_exc())

    def getPortsForXpnt(self, index, xpntModel):
        return self.getAllPorts()

    def getAllPorts(self):
        return self.__ports


# noinspection PyAbstractClass
class WguiXPointPropsSinglePortFieldInstance(WguiXPointPropsFieldInstance):
    """
    Editable port field for single-port connection points.
    """
    #: Port value key, used for model access
    PORT_KEY = 'port'

    #: Display format for port IDs
    PNID_FORMAT = 'PID {pnid}'

    def __init__(self, code, label, context, portsPrv):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param portsPrv: Ports provider
        :type portsPrv: WguiXPointPropsPortsProviderP
        """
        super().__init__(code, label)

        self.__model = context.model
        self.__portsProvider = portsPrv

    def installWidget(self, index, widgetF, textF):
        def portsChanged(*_):
            portBox.clear()
            portBox.addItem('')
            ports = self.__portsProvider.getPortsForXpnt(index, self.__model)
            for p, n, _ in ports:
                portBox.addItem(n if n else self.PNID_FORMAT.format(pnid=p))
                portBox.setItemData(portBox.count() - 1, p, Qt.UserRole)
                if p == port:
                    portBox.setCurrentIndex(portBox.count() - 1)
            if port and portBox.currentIndex() == 0:
                portBox.addItem(
                    self.PNID_FORMAT.format(pnid=port) + ' (not found)')
                portBox.setItemData(portBox.count() - 1, port, Qt.UserRole)
                portBox.setCurrentIndex(portBox.count() - 1)

        def portChanged(*_):
            _port = wgui_qcb_get_data(portBox)
            self.__model.setValue(index, self.code, _port)

        port = self.__model.getValue(index, self.PORT_KEY)
        portBox = QComboBox()
        portsChanged()

        portBox.currentIndexChanged.connect(portChanged)
        psig = self.__portsProvider.portsChanged
        if psig:
            psig.connect(portsChanged)
        widgetF(portBox)


class WguiXPointPropsMultiPortFieldInstance(WguiXPointPropsFieldInstance):
    """
    Editable port field for multi-port connection points, with an auxilliary
    field for port selection.
    """

    #: Port set value key, used for model access
    PORTS_KEY = 'ports'

    def __init__(self, code, label, context, portsPrv):
        """
        :param code: Value code for the field
        :type code: str
        :param label: Label for the field
        :type label: str
        :param context: Field installation context
        :type context: WguiXPointPropsFieldCtx
        :param portsPrv: Entry-point changed signal
        :type portsPrv: WguiXPointPropsPortsProviderP
        """
        super().__init__(code, label)

        self.__context = context
        self.__portsPrv = portsPrv

        self.__portListF = QListView()
        self.__portListM = QStandardItemModel()
        self.__plstFilter = QSortFilterProxyModel(self.__portListM)
        self.__portFilterF = QLineEdit()

        self.__activeXPoint = 0
        self.__portListUpdate = False

        self.__auxWidget = None

    def _createAuxWidget(self):
        self.__plstFilter.setSourceModel(self.__portListM)
        self.__portListF.setModel(self.__plstFilter)
        self.__portListF.setSelectionMode(QAbstractItemView.NoSelection)
        self.__portFilterF.setPlaceholderText('Filter ports')

        self.__context.activeXPointChanged.connect(self._activeXPointChange)
        self.__portListM.itemChanged.connect(self._updateXPointPorts)
        self.__portFilterF.textChanged.connect(
            self.__plstFilter.setFilterFixedString)

        self._resetPortList()
        psig = self.__portsPrv.portsChanged
        if psig:
            psig.connect(self._resetPortList)

        wdgt = QWidget()
        layout = QVBoxLayout()
        layout.addWidget(
            QLabel('Port names associated with selected connection point:'))
        layout.addWidget(self.__portListF)
        layout.addWidget(self.__portFilterF)
        wdgt.setLayout(layout)
        return wdgt

    def _resetPortList(self, *_):
        ports = self.__portsPrv.getAllPorts()
        instance = self.__context.instance
        self.__portListM.clear()

        portNames = {pidn: (pidn, None, None)
                     for xpnt in instance.getConnectionPoints()
                     for pidn in self.getXPointPorts(xpnt=xpnt)}
        portNames.update(
            {pidn: (pidn, pnam, pdir)
             for pidn, pnam, pdir in ports})

        for pidn, pnam, pdir in sorted(portNames.values()):
            item = QStandardItem(self.getPortDescription(pidn, pnam, pdir))
            item.setFlags((item.flags() | Qt.ItemIsUserCheckable) &
                          ~(Qt.ItemIsSelectable | Qt.ItemIsEnabled))
            item.setCheckState(Qt.Unchecked)
            item.setData(pidn, Qt.UserRole)
            item.setData(pnam, Qt.UserRole + 1)
            item.setData(pdir, Qt.UserRole + 2)
            self.__portListM.appendRow(item)

    def installWidget(self, index, widgetF, textF):
        # noinspection PyShadowingNames
        def formatPorts(pset):
            return ','.join(self.getPortName(p) for p in sorted(pset))

        def portsChanged(idx, code):
            if idx == index and code == self.code:
                _pset = self.__context.model.getValue(index, self.PORTS_KEY)
                textUpdateF(formatPorts(_pset))

        pset = self.__context.model.getValue(index, self.PORTS_KEY)
        textUpdateF = textF(formatPorts(pset))
        self.__context.model.valueChanged.connect(portsChanged)

    def getAuxWidget(self):
        if not self.__auxWidget:
            self.__auxWidget = self._createAuxWidget()
        return self.__auxWidget

    def _activeXPointChange(self, index):
        """
        Callback function, invoked whenever the active connection point
        changes.
        """
        self.__activeXPoint = index
        if index <= 0:
            return

        self.__portListUpdate = True
        if index >= 0:
            ports = self.__context.model.getValue(index, self.PORTS_KEY)
            flgf = lambda i: i.setFlags(i.flags() | Qt.ItemIsEnabled)
        else:
            ports = []
            flgf = lambda i: i.setFlags(i.flags() & ~Qt.ItemIsEnabled)

        for l in range(self.__portListM.rowCount()):
            item = self.__portListM.item(l, 0)
            flgf(item)
            if item.data(Qt.UserRole) in ports:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
        self.__portListUpdate = False

    def _updateXPointPorts(self, portItem):
        """
        Callback function, invoked whenever a port name is checked/unchecked.
        """
        if self.__portListUpdate:
            return
        if not (portItem.flags() & Qt.ItemIsEnabled):
            return
        index = self.__activeXPoint
        if index <= 0:
            return

        model = self.__context.model
        pset = model.getValue(index, self.PORTS_KEY)
        port = portItem.data(Qt.UserRole)
        if portItem.checkState() == Qt.Checked:
            # noinspection PyUnresolvedReferences
            pset.add(port)
        elif port in pset:
            # noinspection PyUnresolvedReferences
            pset.remove(port)
        model.setValue(index, self.PORTS_KEY, pset)

    @abstractmethod
    def getXPointPorts(self, xpnt):
        """
        Get the ports for the given connection point entry.

        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        :return: Port set for the given connection point entry
        :rtype: Iterable[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def getPortName(self, pidn):
        """
        Get the port name for the given port ID.

        :param pidn: Port ID
        :type pidn: str
        :return: Port name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def getPortDescription(self, pidn, pname, pdir):
        """
        Get the formatted port description for the given port.

        :param pidn: Port ID
        :type pidn: str
        :param pname: Port name
        :type pname: str
        :param pdir: Direction
        :type pdir: int
        :return: Port description
        :rtype: str
        """
        raise NotImplementedError()


class WguiXPointPropertiesUIBase(QWidget):
    """
    Base UI for connection point properties.
    """

    def __init__(self, model, fields, fieldOrder, actsig,
                 buttons=None, stretched=None, parent=None):
        """
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param actsig: Active connection point signal
        :type actsig: pyqtSignal
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :param stretched: Stretched fields, by code
        :type stretched: set(str)
        :type parent: Parent widget
        """
        super().__init__(parent)

        self.__model = model
        self.__fields = []
        self.__stretched = stretched
        self.__actsig = actsig

        fieldsLeft = dict(fields)
        for code in fieldOrder:
            field = fields.get(code)
            if field:
                del fieldsLeft[code]
                self.__fields.append(field)
        for field in fieldsLeft.values():
            self.__fields.append(field)

        #: Connection points table
        self.__xpntTableF = QTableWidget(self)
        self.__xpntTableF.setMinimumWidth(600)

        self._initElements()
        self._initUI(self.__xpntTableF, self.__fields, buttons)
        self._populateTable()

    @property
    def xpntTableF(self):
        return self.__xpntTableF

    def _initElements(self):
        """
        Initialize graphical elements.
        """

        def selectionChanged(*_):
            # noinspection PyUnresolvedReferences
            self.__actsig.emit(self.getSelectedXPoint())

        self.__model.xpointAdded.connect(self._addRow)
        self.__model.xpointRemoved.connect(self._delRow)
        self.__model.modelReset.connect(self._populateTable)

        table = self.__xpntTableF

        table.setColumnCount(len(self.__fields))
        table.setRowCount(0)
        table.setSortingEnabled(True)
        table.setHorizontalHeaderLabels([f.label for f in self.__fields])
        table.setSelectionMode(QAbstractItemView.SingleSelection)
        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        for col, fld in enumerate(self.__fields):
            if self.__stretched and fld.code in self.__stretched:
                rszm = QHeaderView.Stretch
            else:
                rszm = QHeaderView.ResizeToContents
            table.horizontalHeader().setSectionResizeMode(col, rszm)

        table.itemSelectionChanged.connect(selectionChanged)

    def _initUI(self, table, fields, buttons):
        """
        Init the UI.

        :param table: Connection point table
        :type table: QTableWidget
        :param fields: Fields
        :type fields: Iterable[WguiXPointPropsFieldInstanceP]
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        """
        raise NotImplementedError()

    def _populateTable(self):
        """
        Populate the table.
        """
        table = self.__xpntTableF
        table.clear()
        table.setHorizontalHeaderLabels([f.label for f in self.__fields])
        table.setRowCount(0)
        for idx in self.__model.getIndices():
            self._addRow(idx)

        if table.rowCount():
            table.selectRow(0)
        self.__xpntTableF.setFocus(Qt.OtherFocusReason)

    def _addRow(self, index):
        """
        Add a row to the table for the connection point entry at the given
        index.

        :param index: Connection point entry index
        :type index: int
        """
        table = self.__xpntTableF
        row = table.rowCount()

        def widgetF(wdgt):
            table.setCellWidget(row, col, wdgt)

        def textF(text):
            item = QTableWidgetItem(text)
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            table.setItem(row, col, item)
            return item.setText

        table.setRowCount(row + 1)
        col = 0

        for fld in self.__fields:
            fld.installWidget(index, widgetF, textF)
            col += 1

        titm = table.item(row, 0)
        if not titm:
            titm = QTableWidgetItem()
            table.setItem(row, 0, titm)
        titm.setData(Qt.UserRole, index)

    def _delRow(self, index):
        """
        Delete the row from the table for the connection point entry with the
        given index.

        :param index: Connection point entry index
        :type index: int
        """
        table = self.__xpntTableF
        for row in range(table.rowCount()):
            if table.item(row, 0).data(Qt.UserRole) == index:
                table.removeRow(row)
                break

    def selectXPoint(self, index):
        """
        Select the given connection point entry with the given index.

        :param index: Connection point entry index to select
        :type index: int
        """
        table = self.__xpntTableF
        if not index or index < 0:
            table.clearSelection()
        else:
            for row in range(table.rowCount()):
                if table.item(row, 0).data(Qt.UserRole) == index:
                    table.selectRow(row)
                    table.setFocus(Qt.OtherFocusReason)
                    break

    def getSelectedXPoint(self):
        """
        Get the index of the selected connection point entry.

        :return: Index of selected connection point
        :rtype: int
        """
        itms = self.__xpntTableF.selectedItems()
        if itms:
            item = self.__xpntTableF.item(itms[0].row(), 0)
            if item:
                slctd = item.data(Qt.UserRole)
                if slctd in self.__model:
                    return slctd
        return -1


class WguiXPointPropertiesSimpleUI(WguiXPointPropertiesUIBase):
    """
    Simple connection point UI, table only.
    """

    def __init__(self, model, fields, fieldOrder, actsig, buttons=None,
                 stretched=None, parent=None):
        """
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param actsig: Active connection point signal
        :type actsig: pyqtSignal
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :param stretched: Stretched fields, by code
        :type stretched: set(str)
        :type parent: Parent widget
        """
        super().__init__(
            model, fields, fieldOrder, actsig, buttons, stretched, parent)

    def _initUI(self, table, field, buttons):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(table)
        if buttons:
            btnLayout = QHBoxLayout()
            btnLayout.addStretch(1)
            for btn in buttons:
                btnLayout.addWidget(btn)
            layout.addLayout(btnLayout)

        self.setLayout(layout)


class WguiXCardPropertiesUI(QWidget):
    """
    Card properties UI with table and auxilliary widgets.
    """

    def __init__(self, model, fields, fieldOrder, stretched=None,
                 buttons=None, parent=None):
        """
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :type parent: Parent widget
        """
        super().__init__(parent)
        self.__model = model
        self.__fields = []
        fieldsLeft = dict(fields)
        for code in fieldOrder:
            field = fields.get(code)
            if field:
                del fieldsLeft[code]
                self.__fields.append(field)
        for field in fieldsLeft.values():
            self.__fields.append(field)
        self.__stretched = stretched

        # self.xpntTableF = QTableWidget(self)

        self._initUI(self.__fields, buttons)
        self._initElements()
        self._populateTable()

    def _initUI(self, fields, buttons):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        # get tablewidget
        f = fields[1]
        wdgt = f.getAuxWidget()
        self.xpntTableF = wdgt.layout().itemAt(3).widget()
        layout.addWidget(wdgt)
        # layout.addWidget(self.xpntTableF)
        self.setLayout(layout)

    def _initElements(self):
        """
        Initialize graphical elements.
        """

        # def selectionChanged(*_):
        #     # noinspection PyUnresolvedReferences
        #     self.__actsig.emit(self.getSelectedXPoint())

        self.__model.xpointAdded.connect(self._addRow)
        self.__model.xpointRemoved.connect(self._delRow)
        self.__model.modelReset.connect(self._populateTable)

        table = self.xpntTableF
        table.setColumnCount(len(self.__fields))
        table.setRowCount(0)
        table.setSortingEnabled(True)
        table.setHorizontalHeaderLabels([f.label for f in self.__fields])
        table.setSelectionMode(QAbstractItemView.SingleSelection)
        table.setSelectionBehavior(QAbstractItemView.SelectRows)
        for col, fld in enumerate(self.__fields):
            if self.__stretched and fld.code in self.__stretched:
                rszm = QHeaderView.Stretch
            else:
                rszm = QHeaderView.ResizeToContents
            table.horizontalHeader().setSectionResizeMode(col, rszm)

        # table.itemSelectionChanged.connect(selectionChanged)


    def _populateTable(self):
        """
        Populate the table.
        """
        table = self.xpntTableF
        table.clear()
        table.setHorizontalHeaderLabels([f.label for f in self.__fields])
        table.setRowCount(0)
        for idx in self.__model.getIndices():
            data = self.__model.getValue(idx, _WHGUI_CARD_CONFIG)
            dictData = ast.literal_eval(data)
            if dictData:
                for key in sorted(dictData):
                    # temporary solution
                    # TODO : implement all channel properties
                    if 'channel' in key:
                        continue
                    self._addRow(idx, key)
        if table.rowCount():
            table.selectRow(0)
        self.xpntTableF.setFocus(Qt.OtherFocusReason)

    def _addRow(self, index, key):
        """
        Add a row to the table for the connection point entry at the given
        index.

        :param index: Connection point entry index
        :type index: int
        """
        table = self.xpntTableF
        row = table.rowCount()
        def widgetF(wdgt):
            table.setCellWidget(row, col, wdgt)

        table.setRowCount(row + 1)

        col = 0
        for fld in self.__fields:
            fld.installWidget(col, index, widgetF, key, )
            col += 1

        titm = table.item(row, 0)
        if not titm:
            titm = QTableWidgetItem()
            table.setItem(row, 0, titm)
        titm.setData(Qt.UserRole, index)

    def _delRow(self, index):
        """
        Delete the row from the table for the connection point entry with the
        given index.

        :param index: Connection point entry index
        :type index: int
        """
        table = self.xpntTableF
        for row in range(table.rowCount()):
            if table.item(row, 0).data(Qt.UserRole) == index:
                table.removeRow(row)
                break

    def selectXPoint(self, index):
        """
        Select the given connection point entry with the given index.

        :param index: Connection point entry index to select
        :type index: int
        """
        table = self.xpntTableF
        if not index or index < 0:
            table.clearSelection()
        else:
            for row in range(table.rowCount()):
                if table.item(row, 0).data(Qt.UserRole) == index:
                    table.selectRow(row)
                    table.setFocus(Qt.OtherFocusReason)
                    break

    def getSelectedXPoint(self):
        """
        Get the index of the selected connection point entry.

        :return: Index of selected connection point
        :rtype: int
        """
        itms = self.xpntTableF.selectedItems()
        if itms:
            item = self.xpntTableF.item(itms[0].row(), 0)
            if item:
                slctd = item.data(Qt.UserRole)
                if slctd in self.__model:
                    return slctd
        return -1



class WguiXPointPropertiesSplitUI(QSplitter, WguiXPointPropertiesUIBase):
    """
    Connection point properties UI with table and auxilliary widgets.
    """

    def __init__(self, model, fields, fieldOrder, actsig, buttons=None,
                 stretched=None, parent=None):
        """
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param actsig: Active connection point signal
        :type actsig: pyqtSignal
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :param stretched: Stretched fields, by code
        :type stretched: set(str)
        :type parent: Parent widget
        """
        QSplitter.__init__(self, parent)
        WguiXPointPropertiesUIBase.__init__(
            self, model, fields, fieldOrder, actsig, buttons, stretched)

    def _initUI(self, table, fields, buttons):
        self.setOrientation(Qt.Horizontal)

        if buttons:
            layout = QVBoxLayout()
            layout.addWidget(table)
            if buttons:
                btnLayout = QHBoxLayout()
                btnLayout.addStretch(1)
                for btn in buttons:
                    if isinstance(btn,QSpacerItem):
                        btnLayout.addItem(btn)
                    else:
                        btnLayout.addWidget(btn)
                layout.addLayout(btnLayout)

            wdgt = QWidget()
            wdgt.setLayout(layout)
            self.addWidget(wdgt)
        else:
            self.addWidget(table)

        for wdgt in filter(None, (f.getAuxWidget() for f in fields)):
            self.addWidget(wdgt)


# noinspection PyAbstractClass
class WguiXPointPropertiesBase(WguiPropertiesWidgetP,
                               WguiXPointPropsFieldCtx):
    """
    Abstract base class for connection point properties widgets.
    """
    #: Active connection point signal
    activeXPointChanged = pyqtSignal(int)

    def __init__(self, instance, parent=None):
        """
        :param instance: Instance
        :type instance: WhdbInstanceP
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__instance = instance
        self.__model = WguiXPointPropsModel()
        self.__fields = None
        self.__ui = None

        if self.__class__ is WguiXPointPropertiesBase:
            self._finishInitialization()

    def _finishInitialization(self):
        self.resetModel(self.__model, self.__instance)
        self.__fields = {c: f.instantiate(self) for c, f
                         in self.createFields().items()}
        self.__ui = self.createUI(
            model=self.__model, fields=self.__fields,
            actsig=self.activeXPointChanged,
            fieldOrder=self.getFieldOrder(),
            stretched=self.getStretchFields(),
            buttons=self.getButtons())

        self.setLayout(self._createLayout())

    def createUI(self, model, fields, fieldOrder, actsig,
                 buttons=None, stretched=None):
        """
        Create the properties UI.

        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param actsig: Active connection point signal
        :type actsig: pyqtSignal
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :param stretched: Stretched fields, by code
        :type stretched: set[str]
        :return: Properties UI
        :rtype: WguiXPointPropertiesUIBase
        """
        if any(f.getAuxWidget() for f in fields.values()):
            return WguiXPointPropertiesSplitUI(
                model=model, fields=fields, fieldOrder=fieldOrder,
                actsig=actsig, stretched=stretched, buttons=buttons)
        else:
            return WguiXPointPropertiesSimpleUI(
                model=model, fields=fields, fieldOrder=fieldOrder,
                actsig=actsig, stretched=stretched, buttons=buttons)

    def createFields(self):
        """
        Get the list of connection point table fields.

        :return: Field list
        :rtype: dict[str, WguiXPointPropsFieldP]
        """
        return {}

    def getFieldOrder(self):
        """
        Get the field order by code. May contain non-existing codes.

        :return: Field order
        :rtype: Iterable[str]
        """
        return []

    def getStretchFields(self):
        """
        Get the stretchable fields, by code.

        :return: Stretchable fields
        :rtype: set[str]
        """
        return None

    def getButtons(self):
        """
        Get the general buttons.

        :return: Button list
        :rtype: list[QPushButton]
        """
        addBtn = QPushButton('Add')
        addBtn.clicked.connect(partial(self.__model.addXPoint))
        return [addBtn]

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(self.__ui)
        return layout

    @property
    def model(self):
        return self.__model

    @property
    def ui(self):
        return self.__ui

    @property
    def fields(self):
        return self.__fields

    @property
    def instance(self):
        return self.__instance

    @property
    def getActiveXPoint(self):
        try:
            return self.__ui.getSelectedXPoint()
        except:
            return 0

    def setActiveXPoint(self, index):
        self.__ui.selectXPoint(index)

    def updateFields(self):
        self.resetModel(self.__model, self.__instance)

    def checkPortNumberConsistency(self):
        portNumbers = []
        for idx in self.model.getIndices():
            xpnt = self.model.getValue(idx, WGUI_XPNT_CODE)
            if xpnt:
                portNum = xpnt.getPortNum()
                if portNum not in portNumbers:
                    portNumbers.append(portNum)
                else:
                    return False
        return True

    def save(self):
        model = self.model
        xpnts = set(model.getValuesForCode(WGUI_XPNT_CODE))
        mark=[]
        for xpnt in list(self.__instance.getConnectionPoints()):
            if xpnt not in xpnts:
                mark.append(xpnt)
                # self.removeXPoint(xpnt)
        xpntAdded = False
        for idx in model.getIndices():
            xpnt = model.getValue(idx, WGUI_XPNT_CODE)
            if not xpnt:
                xpnt = self.createXPoint(idx, model, self.__instance)
                xpntAdded = True
            self.updateXPoint(idx, model, xpnt)
        for xpnt in mark:
            self.removeXPoint(xpnt)
        if xpntAdded:
            self.__instance.resetConnectionPointPositions()

    def removeXPoint(self, xpnt):
        self.__instance.removeConnectionPoint(xpnt)

    def resetModel(self, model, instance):
        """
        Reset/reload the value model.

        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param instance: Instance
        :type instance: WhdbInstanceP
        """
        model.setXPoints(
            {WGUI_XPNT_CODE: x} for x in instance.getConnectionPoints())

    @abstractmethod
    def createXPoint(self, idx, model, instance):
        """
        Create a new connection point.

        :param idx: Connection point entry index
        :type idx: int
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param instance: Instance
        :type instance: WhdbInstanceP
        :return: Newly created connection point
        :rtype: WhdbConnectionPointP
        """
        raise NotImplementedError()

    @abstractmethod
    def updateXPoint(self, idx, model, xpnt):
        """
        Update the given connection point with the values from the model.

        :param idx: Connection point entry index
        :type idx: int
        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param xpnt: Connection point
        :type xpnt: WhdbConnectionPointP
        """
        raise NotImplementedError()


# noinspection PyAbstractClass
class WguiModuleIOPointPropertiesBase(WguiXPointPropertiesBase):
    """
    Abstract base class for connection point properties widgets.
    """

    def __init__(self, instance, parent=None):
        super().__init__(instance, parent)

        self.__instance = instance
        self.__model = WguiModuleIOPointPropsModel()
        self.__fields = None
        self.__ui = None

        if self.__class__ is WguiModuleIOPointPropertiesBase:
            self._finishInitialization()

    def save(self):
        model = self.model
        xpnts = set(model.getValuesForCode(WGUI_XPNT_CODE))
        mark = []

        for xpnt in list(self.__instance.getAllIoConnectionPoints()):
            if xpnt not in xpnts:
                mark.append(xpnt)
                # self.removeXPoint(xpnt)
        xpntAdded = False
        for idx in model.getIndices():
            xpnt = model.getValue(idx, WGUI_XPNT_CODE)
            if not xpnt:
                xpnt = self.createXPoint(idx, model, self.__instance)
                xpntAdded = True
            self.updateXPoint(idx, model, xpnt)
        for xpnt in mark:
            self.removeXPoint(xpnt)
        if xpntAdded:
            self.__instance.resetConnectionPointPositions()



# noinspection PyAbstractClass
class WguiXPointProperties(WguiXPointPropertiesBase):
    """
    Simple read-only connection point properties widget.
    """

    def __init__(self, instance, parent=None):
        """
        :param instance: Instance
        :type instance: WhdbInstanceP
        :param parent: Parent widget
        """
        super().__init__(instance, parent)

        if self.__class__ is WguiXPointProperties:
            self._finishInitialization()

    def createFields(self):
        return {
            WHDB_XPOINT_NAME: WguiXPointPropsROTextField(
                WHDB_XPOINT_NAME, 'Conn. Point'),
            WHDB_XPOINT_DIRECTION: WguiXPointPropsROTextField(
                WHDB_XPOINT_DIRECTION, 'Direction')}

    def getFieldOrder(self):
        return [WHDB_XPOINT_NAME, WHDB_XPOINT_DIRECTION]

    def resetModel(self, model, instance):
        model.setXPoints(
            {WGUI_XPNT_CODE: x,
             WHDB_XPOINT_NAME: x.getName(),
             WHDB_XPOINT_TYPE: x.getEntityType(),
             WHDB_XPOINT_DIRECTION: x.getDirection()}
            for x in instance.getConnectionPoints())

    def createXPoint(self, idx, model, instance):
        """
        :type idx: int
        :type model: WguiXPointPropsModelP
        :type instance: WhdbInstanceP
        :rtype: WhdbConnectionPointP
        """
        ctyp = model.getValue(idx, WHDB_XPOINT_TYPE)
        """:type: str | None"""
        cdir = model.getValue(idx, WHDB_XPOINT_DIRECTION)
        """:type: int | None"""
        return instance.createConnectionPoint(ctyp=ctyp, cdir=cdir)

    def updateXPoint(self, idx, model, xpnt):
        pass


def wgui_createMissingXPointsButton(pkey, pfun, model):
    # TODO: Don't use portsPrv
    """
    Create a button that will add all missing conncetion points on click.

    :param pkey: Port value code
    :type pkey: str
    :param pfun: Port-list function
    :type pfun: () -> Iterable[(str, dict[str, object])]
    :param model: Value model
    :type model: WguiXPointPropsModelP
    :return: Created button
    :rtype: QPushButton
    """

    def addMissingXPoints():
        availablePorts = sorted(pfun(), key=lambda x: x[0])
        ports = set(model.getValuesForCode(pkey))
        for port, kwargs in availablePorts:
            if port not in ports:
                kwargs[pkey] = port
                model.addXPoint(**kwargs)

    mbtn = QPushButton('Add missing')
    mbtn.clicked.connect(addMissingXPoints)
    return mbtn


def wgui_getXPointDirectionName(cdir):
    """
    Get the textual name of the given connector type.

    :param cdir: Connector type to provide name for
    :type cdir: int
    :return: Name of the given connector type
    :rtype: str
    """
    if cdir == WDBS_XPNT_IN:
        return 'Input'
    if cdir == WDBS_XPNT_OUT:
        return 'Output'
    if cdir == WDBS_XPNT_BI:
        return 'Bi-directional'
    return ''


def wgui_getXPointIndex(xpnt, model):
    """
    Get the index for the given connection point.

    :param xpnt: Connection point to return index for
    :type xpnt: WhdbConnectionPointP
    :param model: Value model
    :type model: WguiXPointPropsModelP
    :return: Index for the given connection point, or `None`.
    :rtype: int
    """
    for idx in model.getIndices():
        if model.getValue(idx, WGUI_XPNT_CODE) is xpnt:
            return idx


def wgui_setActiveXPoint(xpnt, fctx):
    """
    Set the active connection point to the given one.

    :param xpnt: Connection point to make active
    :type xpnt: WhdbConnectionPointP
    :param fctx: Context
    :type fctx: WguiXPointPropsFieldCtx
    """
    index = wgui_getXPointIndex(xpnt, fctx.model)
    if index is not None:
        fctx.setActiveXPoint(index)


class WguiCenterWrapper(QWidget):
    """
    Wrapper to center widgets in tables.
    """

    def __init__(self, item, parent=None):
        """
        :param item: Item to be wrapped
        :type item: QWidget
        :param parent: Parent widget
        """
        super().__init__(parent)

        self.__item = item

        layout = QVBoxLayout()
        layout.setContentsMargins(5,5,5,5)
        layout.addWidget(item)
        layout.setAlignment(Qt.AlignCenter)
        self.setLayout(layout)

    @property
    def item(self):
        """
        Get the wrapped item.

        :return: Wrapped item
        :rtype: QWidget
        """
        return self.__item
