# ----------------------------------------------------------------------------
# - wgui-protocols                                                           -
# - wild cruncher: graphical interface protocols                             -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 clemens oertel                                   -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wgui_protocols
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher graphical interface protocols

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""

from abc import ABCMeta, abstractmethod

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QGraphicsItem, QWidget, QMenu

class ViewContext(metaclass=ABCMeta):
    """
    Viewing context that graphical elements are displayed in.

    Todo: Add zoom level
    """

    @abstractmethod
    def snapPoint(self, x, y):
        """
        Snap a point to a grid based on the current setting.
        Returns given point is snapping is not enabled.

        :param x: X value of point to apply snapping to
        :param y: Y value of point to apply snapping to
        :return: Point snapped to grid, or x,y if snapping is not enabled
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def topLevelModelBlock(self):
        """
        Get the current top level model block.

        :return: Current top level model block
        :rtype: WhdbBlockP
        """
        raise NotImplementedError()

    @abstractmethod
    def getConnections(self, inst, conn):
        """
        Get the connections originating from a given
        object/connection-point-pair, if any.

        :param inst: Object to find connection for
        :type inst: WhdbInstanceP
        :param conn: Connector to find connection for
        :type conn: WhdbConnectionPointP
        :return: Connections originating from the given object/connector-pair
        :rtype: Iterator[WhdbConnectionP]
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def highlighter(self):
        """
        Get the highlighter for database element rendering.

        :return: Highlighter
        :rtype: WguiItemHighlighter
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def config(self):
        """
        Get the project editor configuration.

        :return: Project editor config
        :rtype: WguiProjectEditorCfg
        """
        raise NotImplementedError()

    @abstractmethod
    def getResource(self, rkey):
        """
        Get the resource data for the given key.

        :param rkey: Resource key
        :type rkey: str
        :return: Resource for the given key
        :rtype: bytes
        """
        raise NotImplementedError()


class WguiContextMenuFactoryP(metaclass=ABCMeta):
    """
    Protocol for context menu generators.
    """

    @abstractmethod
    def createViewContextMenu(self, parent, scenePos=None):
        """
        Create the global context menu a graphics view.

        :param parent: Parent widget
        :param scenePos: Current scene position (used for pasting)
        :type scenePos: QPoint
        :return: Context menu for the given entity
        :rtype: QMenu
        """
        raise NotImplementedError()

    @abstractmethod
    def createSelectionContextMenu(self, parent, scenePos=None):
        """
        Create a context menu for the current scene selection.

        :param parent: Parent widget
        :param scenePos: Current scene position (used for pasting)
        :type scenePos: QPoint
        :return: Context menu for the given entity
        :rtype: QMenu
        """
        raise NotImplementedError()

    @abstractmethod
    def createItemContextMenu(self, entity, parent):
        """
        Create a context menu for the given item.

        :param entity: Database entity to create context menu for
        :type entity: WhdbBaseP
        :param parent: Parent widget
        :return: Context menu for the given entity
        :rtype: QMenu
        """
        raise NotImplementedError()

    @abstractmethod
    def createConnectorContextMenu(self, connector, entity, parent):
        """
        Create a context menu for the given connector.

        :param connector: Connector to create context menu for
        :type connector: WhdbConnectionPointP
        :param entity: Database entity the connector belongs to
        :type entity: WhdbBaseP
        :param parent: Parent widget
        :return: Context menu for the given entity
        :rtype: QMenu
        """
        raise NotImplementedError()


class WguiEmptyContextMenuFactory(WguiContextMenuFactoryP):
    """
    Dummy context-menu factory returning an empty menu for every request.
    """

    def createItemContextMenu(self, entity, parent):
        return QMenu()

    def createViewContextMenu(self, parent, scenePos=None):
        return QMenu()

    def createConnectorContextMenu(self, connector, entity, parent):
        return QMenu()

    def createSelectionContextMenu(self, parent, scenePos=None):
        return QMenu()


# ----------------------------------------------------------------------------
# - Generic GUI Protocols                                                    -
# ----------------------------------------------------------------------------

class WguiZoomAwareP(metaclass=ABCMeta):
    """
    Protocol for zoom-aware elements.

    Implementers of this interface want to be notified about zoom-level
    changes.
    """

    @abstractmethod
    def zoomChanged(self, zoom:float):
        """
        Notification callback function for

        :param zoom: The new zoom level.
        :type zoom: float
        """
        raise NotImplementedError()


class WguiChangeNotifyP(metaclass=ABCMeta):
    """
    Protocol for objects that can notify about changes of their contents.
    """

    @property
    @abstractmethod
    def contentChanged(self):
        """
        Signal emitted whenever the contents have changed.

        :return: Content change notification signal
        :rtype: pyqtSignal
        """
        raise NotImplementedError()


class WguiHasTitleP(metaclass=ABCMeta):
    """
    Protocol for objects that can notify about changes of their title.
    """

    @property
    @abstractmethod
    def titleChanged(self):
        """
        Signal emitted whenever the contents have changed.

        :return: Content change notification signal
        :rtype: pyqtSignal
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def title(self):
        """
        Get the current title.

        :return: The current title
        :rtype: str
        """
        raise NotImplementedError()


# ----------------------------------------------------------------------------
# - PyQt/ABCMeta                                                             -
# ----------------------------------------------------------------------------

#: Type of QObject
QObject_M = type(QObject)


class QObject_Meta_M(QObject_M, ABCMeta):
    """
    Abstract base class for QObjects.
    Intended solely to resolve metaclass-conflict.
    """
    # noinspection PyArgumentList
    def __init__(cls, name, bases, dct):
        QObject_M.__init__(cls, name, bases, dct)
        ABCMeta.__init__(cls, name, bases, dct)


#: Type of QWidget
QWidget_M = type(QWidget)


class QWidget_Meta_M(QWidget_M, ABCMeta):
    """
    Abstract base class for QWidgets.
    Intended solely to resolve metaclass-conflict.
    """
    # noinspection PyArgumentList
    def __init__(cls, name, bases, dct):
        QWidget_M.__init__(cls, name, bases, dct)
        ABCMeta.__init__(cls, name, bases, dct)


#: Type of QGraphicsItem
QGraphicsItem_M = type(QGraphicsItem)


class QGraphicsItem_Meta_M(QGraphicsItem_M, ABCMeta):
    """
    Abstract base class for QGraphicsItems.
    Intended solely to resolve metaclass-conflict.
    """
    # noinspection PyArgumentList
    def __init__(cls, name, bases, dct):
        QGraphicsItem_M.__init__(cls, name, bases, dct)
        ABCMeta.__init__(cls, name, bases, dct)
