# ----------------------------------------------------------------------------
# - whgui-host-connection-points-props-widget                                -
# - wild cruncher - cockpit : connection point properties dialog widgets     -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       whgui_host_connection_points_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher connection point properties dialog widgets

.. moduleauthor:: Philippe lambert <lambert@silkan.com>
"""
import ast
from collections import Counter
from uuid import UUID
from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtWidgets import QMessageBox, QComboBox, QHBoxLayout, QWidget, \
	QVBoxLayout, QPushButton, QLabel, QTableWidget, QTableView, QHeaderView, \
	QLineEdit, QItemDelegate, QSplitter, QAbstractItemView

from wcr_frontend.modules.hardware.gui.whgui_Io_Card_instance_props_dialog \
            import WhguiIoCardPropertiesDialog
from wcr_frontend.gui.wgui_props_dialog_widget import \
    WguiXPointPropertiesBase, WguiXPointPropsTextField, \
    WguiXPointPropsDeleteButton, WGUI_NAME_PLACEHOLDER, WGUI_XPNT_CODE, \
    WguiXPointPropsROLookupField, WguiXPointPropsFieldInstance, \
    WguiXPointPropsField, _WHGUI_CARD_CONFIG
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend.gui.wgui_props_dialog_widget import WguiXCardPropertiesUI
from wcr_core.db.whdb_connection_point_manager import WHDB_XPOINT_NAME, \
    WHDB_XPOINT_TYPE
from wcr_core.modules.hardware.db.whhdb_connection_points import \
    WHHDB_XPOINT_NAMES, WHHDB_XPOINT_DEFAULT_CODE
from wcr_frontend import isNameValid, showErrorWindow, NameFormatMessageError


WHHDB_REPORT_STATUS = 'Status'
_WHGUI_XPNT_CONFIG = '__config'


# noinspection PyAbstractClass
class WhguiRackXPointProperties(WguiXPointPropertiesBase):
    def __init__(self, instance, pctx, scene, parent=None, ConverifF=None,
                 updateF=None):
        """
        :param instance: The instance object
        :type instance: WhhdbHostInstance
        :param pctx: Project context
        :type pctx: WhmodHardwareProjectContextP
        :param parent: Parent widget
        """
        module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntCfgMetas = {}
        self.updateF = updateF
        super(WhguiRackXPointProperties, self).__init__(instance, parent=parent)
        self.__scene = scene
        self.instance.restoreCards()
        #: Configuration table view
        self.__card_tab = _WhguiCardTableInstance(pctx,instance,parent=self)

        if self.__class__ is WhguiRackXPointProperties:
            self._finishInitialization()

    def getCardTab(self):
        return self.__card_tab

    def getScene(self):
        return self.__scene

    def createUI(self, model, fields, fieldOrder, actsig,
                 buttons=None, stretched=None):
        """
        Create the properties UI.

        :param model: Value model
        :type model: WguiXPointPropsModelP
        :param fields: Field list
        :type fields: dict[str, WguiXPointPropsFieldInstanceP]
        :param fieldOrder: Field order, by code
        :type fieldOrder: Iterable[str]
        :param actsig: Active connection point signal
        :type actsig: pyqtSignal
        :param buttons: Buttons
        :type buttons: Iterable[QPushButton]
        :param stretched: Stretched fields, by code
        :type stretched: set[str]
        :return: Properties UI
        :rtype: WguiXPointPropertiesUIBase
        """
        if any(f.code == _WHGUI_CARD_CONFIG for f in fields.values()):
            return WguiXCardPropertiesUI(
                model=model, fields=fields, fieldOrder=fieldOrder,
                buttons=buttons)

    def updateFields(self):
        self.__card_tab.resettab()
        super().updateFields()

    def createFields(self):
        return {}

    def getFieldOrder(self):
        return [WHDB_XPOINT_NAME, _WHGUI_CARD_CONFIG]

    def getStretchFields(self):
        return {}

    def getButtons(self):
        return []

    def resetModel(self, model, instance, cardId=None):
        if isinstance(cardId, UUID):
            meta = instance.getIOCardMeta(cardId)
            cardname = instance.getIOCardName(cardId)
            model.setXPoints((
                {WGUI_XPNT_CODE: 'card',
                 WHDB_XPOINT_NAME: cardname,
                 WHDB_XPOINT_TYPE: WHHDB_XPOINT_DEFAULT_CODE,
                 _WHGUI_CARD_CONFIG: meta},))

    def createXPoint(self, idx, model, instance):
        return None

    def updateXPoint(self, idx, model, xpnt):
        pass

    def _createLayout(self):
        container = QWidget()
        container_Layout = QHBoxLayout()
        splitprinc = QSplitter()
        splitprinc.setOrientation(Qt.Horizontal)

        left = QWidget()
        left.setLayout(self.__card_tab.createLayout())
        splitprinc.addWidget(left)

        right = QWidget()
        right.setLayout(super()._createLayout())
        container_Layout.addWidget(right)
        container.setLayout(container_Layout)
        splitprinc.addWidget(container)
        layout = QHBoxLayout()
        layout.addWidget(splitprinc)
        layout.setContentsMargins(2,2,2,2)

        return layout

    def save(self):
        tabcards = [c for c, t, cid in self.__card_tab.getCards()]
        s_tabcards = set(tabcards)
        if s_tabcards != tabcards:
            cardsCount = Counter(tabcards)
            for card in cardsCount:
                if cardsCount[card] > 1:
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Information)
                    msg.setText("Card name duplication:"
                                " {}".format(card))
                    msg.setWindowTitle("Save Card Warning")
                    msg.exec_()
                    return True
        self.resetModel(self.model, self.instance)
        self.deleteSceneCards()
        self.instance.delCards()
        self.__card_tab.resettab()

    def deleteSceneCards(self):
        """
        delete form the scene the card that are  not in rack anymore
        :return:
        """
        scene = self.getScene()
        sceneCards = scene.getScenCards()
        cards = self.instance.CARD_TO_REMOVE
        for uuid in cards:
            card = self.instance.cards[uuid]
            if card in sceneCards.keys():
                scene.deleteItem(sceneCards[card])


class _WhguiCardTableInstance(QWidget):
    def __init__(self, pctx, instance, parent=None):
        """
        :param pctx: Project context
        :type pctx:
        :param instance: The instance edited
        :param parent: Parent widget
        """
        self.openedCardProperties = {}
        self.dialogs = []
        super().__init__(parent)
        self.__RackInst=instance
        self.__parent=parent
        module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""

        # : Project context
        self.__pctx = pctx

        #: Add entry button
        # self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        self.__pcicTblM = _WhguiIOCardTableInstanceModel(self.formatcards,
                                                         rackInst=parent,
                                                         parent=self)
        self.__pcicTblW = QTableView(self)
        self.__pcictypF = lambda a,b: a
        self.__CardTypesF = self.__RackInst.getTypes()
        self._initElements()
        self._loadFromData()

    def _initElements(self):
        self.__pcicTblW.setModel(self.__pcicTblM)
        self.__pcicTblW.horizontalHeader().setSectionResizeMode(1)
        self.__pcicTblW.setMinimumWidth(500)
        self.__pcicTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)

        self.__deleteButton.clicked.connect(self._deleteSelEntries)

        self.__pcicTblW.doubleClicked.connect(self.selecClicked)

    def resettab(self):
        """
        reset self.__pcicTblM
        """
        self.__pcicTblM.deleteRows(range(0,self.__pcicTblM.rowCount()))
        self.__RackInst.restoreCards()
        self._loadFromData()

    def _loadFromData(self):
        """
        Import information from HostInst to fill pcicTblM
        """
        #need to initialise tab with instanceinfo
        fullinfo = self.__RackInst.getIOCardsDescription()
        for uuid, card, ctype, pos in sorted(fullinfo, key=lambda x: x[-1]):
            self.__pcicTblM.addRow(Cardid=uuid, Pos=pos, Ctyp=ctype)

    def getCards(self):
        """
        return the description of all card describes in the Table
        """
        return self.__pcicTblM.getFullCards()

    def formatcards(self, data, column, colhead):
        """
        :param data: data to format
        :type data: Cardname | Cardtype |
        """
        if data is None:
            return ''
        if column == 0:
            return data
        elif column == 1:
            return data

    def createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.__pcicTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__pcicTblW.selectedIndexes()
        """:type: list[QModelIndex]"""
        self.__pcicTblM.deleteRows([idx.row() for idx in selIdx])
        rl=[]
        for idx in selIdx:
            rl.append(idx.row())
        rowlist = sorted(set(rl))
        if self.__pcicTblM.rowCount() > 0 and rowlist:
            if rowlist[0] >= self.__pcicTblM.rowCount():
                self.__pcicTblW.setCurrentIndex(self.__pcicTblM.
                                                index(self.__pcicTblM.rowCount()-1,0))
            else:
                self.__pcicTblW.setCurrentIndex(self.__pcicTblM.
                                                index(rowlist[0], 0))

    def removecard(self, uuid):
        """
        add card with name cardn sit the card to dellet when save
        """
        self.__RackInst.markIOCard(uuid)

    def selectedCard(self, selectignor = False, role=Qt.UserRole):
        lrow = []
        for index in self.__pcicTblW.selectionModel().selectedIndexes():
            lrow.append(index.row())
        ulrow = set(lrow)
        if len(ulrow) == 1 and not selectignor:
            return self.__pcicTblM.data(self.__pcicTblM.index(lrow[0],0),
                                        role=role)
        else:
            return None

    def selecClicked(self, item):
        """
        Signal handler : when item is selected
        """
        row = item.row()
        uuid = item.sibling(row, 2).data()
        card = self.__parent.instance.cards[uuid]
        updateF = self.__parent.updateF

        diag = WhguiIoCardPropertiesDialog(card, self.__pctx, updateF=updateF)
        diag.exec_()


class _WhguiIOCardTableInstanceModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_NAME = 'id'
    TITLE_TYPE = 'IO Type'
    HEADERS = [TITLE_NAME, TITLE_TYPE]

    def __init__(self, formatF, rackInst, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__formatF = formatF

        #: Entries as list of tuples of
        #  port, message and type
        self.__entries = []
        """:type: list[(CardId, Cardname, Cardtype)]"""
        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= self.columnCount():
            if column < 3:
                info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                if column == 2:
                    # Get UUID
                    return info
                else:
                    return self.__formatF(info, column, self.HEADERS[column])
        return None

    def setData(self, index, data, role=None):
        assert role == Qt.UserRole
        self._setCardEntry(index, data)

    def _setCardEntry(self, index, data):
        pos, ctype, cardid = self.__entries[index.row()]
        if index.column() == 0:
            actualentry = (data, ctype, cardid)
        elif index.column() == 1:
            actualentry = (pos, data, cardid)
        self.__entries[index.row()] = actualentry

    def getFullCards(self):
        return self.__entries

    def addRow(self, Cardid='', Pos=None, Ctyp=None):
        """
        Add a row at the end of the list (append)

        :param Pos: The position of the card (idtStatus
        :type Pos: str | None
        :param Ctyp: Type of the card
        :type Ctyp: str | None
        :param Ports: ports uuid binded to card
        :type Ports: [uuid] | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        self.__entries.append((Pos, Ctyp, Cardid))
        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        card, ctype, cardid = self.__entries[rowIndex]
        if card is not None and ctype is not None:
            try:
                # Unbound function of message
                self.__parent.removecard(cardid)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def flags(self, index):
        flgs = super().flags(index)
        return flgs
