# ----------------------------------------------------------------------------
# - wigui-avail-module-table                                                      -
# - wild cruncher:icd available module widget                                   -
# ----------------------------------------------------------------------------
# - author    (c) 2017-2017 achraf mahti                                     -
# - copyright (c) 2017-2017 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_message_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan icd message table widget

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

import csv, os, re
from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QVBoxLayout, QPushButton, QItemDelegate, \
                            QMessageBox, QLabel, QAbstractItemView
from wcr_frontend.modules.icd.gui.wigui_table import WiguiTable
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_core.modules.icd.db.wihdb_top_block import GENERIC_TYPES
from wcr_core.common.wcmn_file_handler import WcmnFile
from wcr_core.project.wprj_xmlio import WPRJ_ICD_SILKAN_CSV_MIME_TYPE
from wcr_core.modules.icd.db.wihdb_module_instance import \
    WIHDB_MODULE_TYPE_VALUES, WIHDB_MODULE_STORAGE_VALUES


def wigui_check_imported_data(lines):
    errors = []

    def getData(l, index):
        return l[index] if index < len(l) else None

    for id, line in enumerate(lines[1:]):
        modName = getData(line, 0)
        if not modName:
            errors.append('Module name of row {} is empty.'.format(id))

        modType = getData(line, 1)
        if not modType:
            errors.append('Module type of row {} is empty.'.format(id))
        elif modType not in WIHDB_MODULE_TYPE_VALUES:
            errors.append('Module type should have one of the following values {} '
                          'but {} detected.'.format(WIHDB_MODULE_TYPE_VALUES,
                                                          modType))
        localMsgName = getData(line, 2)
        if not localMsgName:
            errors.append('Local message name of row {} is empty.'.format(id))
        localMsgType = getData(line, 3)
        if not localMsgType:
            errors.append('Local message type of row {} is empty.'.format(id))
        elif localMsgType not in GENERIC_TYPES.keys():
            errors.append('Local message type should have one of the following values {} '
                          'but {} detected.'.format(GENERIC_TYPES.keys(),
                                                          localMsgType))
        ConsProd = getData(line, 4)
        consprodValues = ('C', 'P')
        if not ConsProd:
            errors.append('Consumer/Producer of row {} is empty.'.format(id))
        elif ConsProd not in consprodValues:
            errors.append('Consumer/Producer should have one of the following values {} '
                          'but {} detected.'.format(consprodValues,
                                                          ConsProd))
        Storage = getData(line, 5)
        if not Storage:
            errors.append('Storage of row {} is empty.'.format(id))
        elif Storage not in WIHDB_MODULE_STORAGE_VALUES:
            errors.append('Storage should have one of the following values {} '
                          'but {} detected.'.format(WIHDB_MODULE_STORAGE_VALUES,
                                                          Storage))
        Period = getData(line, 6)
        if not Period:
            errors.append('Period of row {} is empty.'.format(id))
        elif Storage == 'cyclic':
            if not re.match("^([0-9]*|\d*\.\d{1}?\d*)$", eval(Period)):
                errors.append("Period value '{}' is incorrect,"
                              " needs to be a decimal value.".format(Period))
        glbMsgName = getData(line, 7)
        if not glbMsgName:
            errors.append('Global message name of row {} is empty.'.format(id))

        payload = getData(line, 8)
        if not payload:
            errors.append('Payload of row {} is empty.'.format(id))
        elif re.match("^[0-9]+$", payload) is None:
            errors.append("Payload value '{}' is incorrect, "
                          "needs to be a whole value.".format(payload))
    return errors


class WiguiAvailModuleTable(WiguiTable):
    """
    available module table widget.
    """

    def __init__(self, actx, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(pctx, scene, parent)
        self.__parent = parent
        self.__actx = actx
        self.scene = scene
        #: Bus binding table model
        self.gblTblM = _WiguiAvailModuleTableModel(parent=self, tableW=self.gblTblW)
        self.gblTblW.setSelectionMode(QAbstractItemView.SingleSelection)

        #: Creating editor for element in table
        self.gblDlgt = _WiguiAvailModuleDelegate()
        #: Add entry button
        self.__importButton = QPushButton('Import')
        #: Delete entry button
        self.__exportButton = QPushButton('Export')
        #: Creation entry button
        self.__createButton = QPushButton('Add')
        #: Creation entry button
        self.__deleteButton = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def _initElements(self):
        super()._initElements()
        self.gblTblW.doubleClicked.connect(self._editModule)
        self.__importButton.clicked.connect(self._importModule)
        self.__exportButton.clicked.connect(self._exportModule)
        self.__createButton.clicked.connect(self._createModule)
        self.__deleteButton.clicked.connect(self._deleteModule)

    def _loadFromData(self):
        for oper in self.scene.projOperators:
            self.gblTblM.addRow(oper)
        super()._loadFromData()

    def _createLayout(self):
        layout = super()._createLayout()
        btnLayout = QHBoxLayout()
        layout.addLayout(btnLayout)
        btnLayout.addWidget(self.__importButton)
        btnLayout.addWidget(self.__exportButton)
        btnLayout.addWidget(self.__createButton)
        btnLayout.addWidget(self.__deleteButton)
        filtrLayout = QVBoxLayout()
        layout.addLayout(filtrLayout)
        return layout

    def getOpenFile(self):
        # paramType = parameter.getParamType()
        """:type: WactActionOpenFileParameterTypeP"""
        path = wgui_getOpenFile('Select Import file',
                                self.parent(), WPRJ_ICD_SILKAN_CSV_MIME_TYPE)
        if not path:
            return None
        return WcmnFile(path)

    def _importModule(self):
        """
        Import model from csv file in icd view
        """
        file = self.getOpenFile()
        if file:
            with open(file.path) as importFile:
                importReader = csv.reader(importFile)
                lines = [line for line in importReader]
                if len(lines) == 0:
                    QMessageBox.warning(self.parent(),
                                        'Import Warning',
                                        'The file is empty.')

                errors = wigui_check_imported_data(lines)
                if errors:
                    self.__actx.statusMessage("Import icd errors:")
                    for err in errors:
                        self.__actx.addMessage(err, 'Error')
                    errLog = QMessageBox.warning(
                             self.parent(), 'Import error',
                             'The import file used contained some errors, '
                             'do you want to see the logs.',
                             buttons=QMessageBox.Yes | QMessageBox.No)
                    if errLog == QMessageBox.Yes:
                        mainWin = self.__actx.getParent()
                        mainWin._activateLoggerWindow()
                else:
                    self.scene.importOperator(lines)

    def _exportModule(self):
        """
        Export model to csv file
        """
        # TODO: call function that open window to export module selected module
        pass

    def _createModule(self):
        """
        create model to csv file
        """
        self.scene.createOperator()

    def _editModule(self, rowData):
        """
        create model to csv file
        """
        moduleName = rowData.data()
        self.scene.showOperator(moduleName)

    def _deleteModule(self):
        """
            delete  existing model
        """
        index = self.gblTblM.getSelectedIndexes()
        for idx in sorted(index, reverse=True):
            row = idx.row()
            entries = self.gblTblM.getEntries()
            opName = entries[row]

            isUsed = self.scene.checkRemoveOperator(opName)
            if isUsed:
                QMessageBox.warning(
                    self.parent(), 'Deletion attempt',
                    'You are trying to remove a module from the library but'
                    ' that module has been instantiated. \n'
                    'First you need to remove it\'s instances. ')
                return
            btn = QMessageBox.question(
                self.parent(), 'Confirm Deletion',
                'Are you sure you want to delete {i}'.format(i=opName),
                QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
            if btn == QMessageBox.Yes:
                self.gblTblM.deleteRow(row)
                self.scene.removeOperator(opName)


class _WiguiAvailModuleTableModel(QAbstractTableModel):
    """
    Available module table model.
    """
    TITLE_NAME = 'Modules'

    HEADERS = [TITLE_NAME]

    def __init__(self, parent=None, tableW=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.tableW = tableW

        #: Entries as list of tuples of
        #  name and message
        self.__entries = []

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def getEntries(self):
        return self.__entries

    def addRow(self, variable=None):
        """
        Add a row at the end of the list (append)

        :param variable: global variable name
        :type variable: Str
        :param varType: variable type
        :type varType: str
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append(variable)

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def getSelectedIndexes(self):
        return self.tableW.selectionModel().selectedIndexes()

    def data(self, index, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            row = index.row()
            column = index.column()
            if 0 <= column < 2:
                info = self.__entries[row]
                return info
        return None

    # def setData(self, index, data, role=None):
    #     """
    #     :type index: QModelIndex
    #     :type data: object
    #     """
    #     assert role == Qt.EditRole or role == Qt.DisplayRole
    #     scene = self.__parent.scene
    #     oper = scene.projOperators[index.data()]
    #     instanciate = WiguiModuleInstanciateDialog(pctx=scene.projectCtx,
    #                                                oper=oper, parent=scene)
    #     instanciate.exec_()
    #     self._setDataEntry(index, instanciate.getName())

    def _setDataEntry(self, index, data):
        nl = None
        if index.column() == 0:
            nl = data
        self.__entries[index.row()] = nl

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        #allowing editing
        flgs |= Qt.ItemIsEditable
        return flgs

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)


class _WiguiAvailModuleDelegate(QItemDelegate):
    """
    Delegate for Message table
    """

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        """
        Create a label containing for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """
        editor = QLabel(parent)
        return editor