# ----------------------------------------------------------------------------
# - wigui-message-binding-props-dialog                                       -
# - wild cruncher - cockpit : graphic scene widget                           -
# ----------------------------------------------------------------------------
# - author    (c) 2018-2018 achraf mahti                                     -
# - copyright (c) 2018-2018 silkan                                           -
# ----------------------------------------------------------------------------

"""
.. module::       wigui_message_binding_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher wigui graphics scene editor

.. moduleauthor:: achraf mahti <mahti@silkan.net>
"""

from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtWidgets import QComboBox, QWidget, QVBoxLayout, QLineEdit, \
	QTableView, QItemDelegate, QLabel, QDialogButtonBox

from wcr_frontend.gui.wgui_operator_instance_props_dialog import _WguiParametersProperties
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP, \
    WguiFlipRotationProperties, \
    WguiPropertiesDialog, WguiXPointProperties, WguiMultiPropertiesWidgetP
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_frontend import showErrorWindow


class WiguiMessageInstanciateDialog(WguiPropertiesDialog):
    """
    The WguiOperatorInstancePropertiesDialog class is a dialog design to
    access and manipulates operator instance data. This include name,
    info and connection points.
    """

    _TYPE_NAME = 'Message Binding'

    _SHOW_OPERATOR_WIDGET = True

    PARAM_TAB = _WguiParametersProperties

    # noinspection PyTypeChecker
    def __init__(self, pctx, updateF=None, selectXpnt=None, instance=None, parent=None):
        """
        :param inst: Instance
        :type inst: WhdbOperatorInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: _ -> _
        :param selectXpnt: Initially selected connection point, if any
        :type selectXpnt: WhdbConnectionPointP
        :param parent: Parent widget
        """
        winName = '{} {}'.format(self._TYPE_NAME, instance.getName())
        super().__init__(winName, updateF, parent=parent)

        tophm = pctx.getProjectDB().getTopHierarchy()
        instWdgt = WguiOperatorBaseProperties(pctx, tophm, instance)
        self.instWdgt = instWdgt
        addiWdgts = self._createAdditionalInstancePropertiesWidgets(pctx)
        if addiWdgts:
            # noinspection PyTypeChecker
            generalWdgts = [instWdgt] + addiWdgts
        else:
            generalWdgts = [instWdgt]

        widgets = [('Binding', WguiMultiPropertiesWidgetP(*generalWdgts))]

        widgets.extend(self._createAdditionalPropertiesWidgets(pctx))

        xpntWdgt = self._createConnectionPointsPropertiesWidget(pctx)
        if xpntWdgt:
            widgets.append(xpntWdgt)


        # noinspection PyUnresolvedReferences
        self.setTabbedPropertyWidgets(
            *widgets,
            activate=xpntWdgt[1] if selectXpnt and xpntWdgt else None)

    def showOperatorWidget(self):
        return True

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalInstancePropertiesWidgets(self, pctx):
        return []

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalPropertiesWidgets(self, pctx):
        return []

    # noinspection PyMethodMayBeStatic
    def _createConnectionPointsPropertiesWidget(self, pctx):
        pass

    def _setMainLayout(self, widgets, widget):
        """
        Set the dialog's property widgets.

        :param widgets: Property widgets, and their titles
        :type widgets: Iterable[(str, WguiPropertiesWidgetP)]
        :param widget: Wrapper widget containing the property widgets
        :type widget: QWidget
        """
        self.__widgets = [widget for _, widget in widgets]

        layout = QVBoxLayout()
        layout.addWidget(widget)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.okStateChanged.connect(
            buttons.button(QDialogButtonBox.Ok).setEnabled)
        layout.addWidget(buttons)
        # create the main layout
        self.setLayout(layout)

    def _apply(self):
        """
        Apply the current changes, and leave the dialog window open.
        """
        self._save()
        self._updateFields()

    def _reset(self):
        """
        Reset the dialog window.
        """
        self._updateFields()

    def _save(self):
        """
        Save all changes.
        """
        hasErr = False
        for widget in self.__widgets:
            widgErr = widget.save()
            if widgErr:
                hasErr = True

        return hasErr

    def _updateFields(self):
        """
        Reset all fields to their original state.
        """
        for widget in self.__widgets:
            widget.updateFields()


class WguiOperatorBaseProperties(WguiPropertiesWidgetP):
    """
    Configuration widget to create an icd operator base data.
    """

    def __init__(self, pctx, tophm, instance, parent=None):
        """
        :param typeName: Type name of the displayed element
        :type typeName: str
        :param inst: the instance object
        :type inst: WhdbInstanceP
        :param tophm: Hierarchy top level
        :type tophm: WhdbBlock
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__module = pctx.getModule()
        """:type: WpmodSystemApplicationModuleP"""
        self.__xpntTypeNames = ()
        self.instance = instance
        self.messages = instance.getData()
        self.__topBlock = tophm
        self.hasError = True
        # init table necessary for
        self.__msgTblW = QTableView(self)
        self._initModels()
        self._initElements()
        self.createLayout()
        self._loadFromData()

    def updateFields(self):
        pass

    @property
    def msgTblM(self):
        return self.__msgTblM

    @property
    def msgTblW(self):
        return self.__msgTblW

    @property
    def msgDlgt(self):
        return self.__msgDlgt

    def _initModels(self):
        self.__msgTblM = _WhguiMessageTableInstanceModel(self.formatMsgs,
                                                         self.__topBlock,
                                                         parent=self)
        self.__msgDlgt = _WhguiMessageBindDelegate(top=self.__topBlock,
                                                   messages=self.messages,
                                                   parent=self)

    def _initElements(self):
        self.__msgTblW.setModel(self.msgTblM)
        self.__msgTblW.horizontalHeader().setSectionResizeMode(1)
        self.__msgTblW.setMaximumWidth(500)
        self.__msgTblW.setItemDelegate(self.__msgDlgt)

    def resettab(self):
        """
        reset self.__pcicTblM
        """
        self.msgTblM.deleteRows(range(0, self.msgTblM.rowCount()))
        self._loadFromData()

    def _loadFromData(self):
        for message in sorted(self.messages):
            self.msgTblM.addRow(message)

    def getMsgs(self):
        """
        return the description of all card describes in the Table
        """
        return self.msgTblM.getFullMessages()

    def formatMsgs(self, data, column):
        """
        :param data: data to format
        :type data: message | messageGlb | [PortUUID]
        """
        if data is None:
            return ''
        if column == 0:
            (msgType, direction, _, _) = self.messages[data]
            return '({}) {} [{}]'.format(direction, data, msgType)
        elif column == 1:
            return data
        else:
            return 'No formating available'

    def createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.__msgTblW)

        self.setLayout(layout)

        return layout

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.msgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.msgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""
        self.msgTblM.deleteRows([idx.row() for idx in selIdx])
        rl = []
        for idx in selIdx:
            rl.append(idx.row())
        rowlist = sorted(set(rl))
        if self.msgTblM.rowCount() > 0 and rowlist:
            if rowlist[0] >= self.pcicTblM.rowCount():
                self.msgTblW.setCurrentIndex(self.msgTblM.
                                             index(self.msgTblM.rowCount() - 1, 0))
            else:
                self.msgTblW.setCurrentIndex(self.msgTblM.
                                             index(rowlist[0], 0))

    def syncMsgs(self):
        pass

    def removeMsg(self, glbMsg):
        """
        remove message and variables associated to it
        """
        if glbMsg in self.__parent.msg2variable:
            del self.__parent.msg2variable[glbMsg]
        self.__parent.resetModel(self.__parent.model, glbMsg)

    def selectedMsg(self, selectignor=False, role=Qt.UserRole):
        lrow = []
        for index in self.getSelectedIndexes():
            lrow.append(index.row())
        ulrow = set(lrow)
        if len(ulrow) == 1 and not selectignor:
            return self.msgTblM.data(self.msgTblM.index(lrow[0], 0),
                                     role=role)
        else:
            return None

    def getSelectedIndexes(self):
        return self.msgTblW.selectionModel().selectedIndexes()

    def save(self):
        self.check_before_saving()
        if self.hasError:
            return self.hasError
        for modMsg, glbMsg in self.getMsgs():
            msgType = self.messages[modMsg][0]
            state = self.messages[modMsg][1]
            self.__topBlock.bindMsg(modMsg, glbMsg, self.instance.getName(),
                                    state, msgType)

    def check_before_saving(self):
        self.hasError = False
        # logical checks
        for modMsg, glbMsg in self.getMsgs():
            # check that all messages have valid name
            if glbMsg == '':
                showErrorWindow('Message Error',
                                'Message {} is not assigned to a global variable', modMsg)
                self.hasError = True
                return

        allBindings = self.__topBlock.getAllMessageDictBindings()
        for modMsg, glbMsg in self.getMsgs():
            lclMsgType = self.messages[modMsg][0]
            state = self.messages[modMsg][1]
            isUnique = self.__topBlock.checkMsgProducerUnicity(glbMsg, modMsg,
                                                    state, allBindings)
            glbMsgType, size = self.__topBlock.getAllMessages()[glbMsg]
            if not isUnique:
                showErrorWindow('Message Error',
                                'Message {} is already assigned'
                                ' to a producer.', glbMsg)
                self.hasError = True
                return
            elif lclMsgType != glbMsgType:
                showErrorWindow('Message Error',
                                'Message type are not the same.', glbMsg)
                self.hasError = True
                return


class _WhguiMessageTableInstanceModel(QAbstractTableModel):
    """
    icd Binding table model.
    """

    TITLE_MODULE_MESSAGE = 'Module message'
    TITLE_GLOBAL_MESSAGE = 'Global message'
    HEADERS = [TITLE_MODULE_MESSAGE, TITLE_GLOBAL_MESSAGE]

    def __init__(self, formatF, top, parent=None):
        """
        :param XpntTypeNames: The Names of Xpoint of the mule
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent = parent
        self.__formatF = formatF
        self.__top = top

        #: Entries as list of tuples of
        #  port, message and type
        self.__entries = []
        self._update()

    @property
    def entries(self):
        return self.__entries

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
                                    0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.HEADERS)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= self.columnCount():
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
        return None

    def setData(self, index, data, role=None):
        assert role == Qt.UserRole
        self._setMessageEntry(index, data)

    def _setMessageEntry(self, index, data):
        msgGbl, msgLcl = self.__entries[index.row()]
        if index.column() == 0:
            actualentry = (data, msgLcl)
        elif index.column() == 1:
            actualentry = (msgGbl, data)
        self.__entries[index.row()] = actualentry

    def checkUnicityMsg(self, msg, col, row):
        for i, (entry) in enumerate(self.__entries):
            if i != row:
                if msg == entry[col]:
                    return False
        return True

    def getFullMessages(self):
        return self.__entries

    def addRow(self, moduleMsg='', glbMsg=''):
        """
        Add a row at the end of the list (append)

        :param glbMsg: The message name global
        :type glbMsg: str | None
        :param moduleMsg: The message name local
        :type moduleMsg: str | None
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())
        self.__entries.append((moduleMsg, glbMsg))
        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        msgGbl, msgLcl = self.__entries[rowIndex]
        if msgGbl is not None or msgLcl is not None:
            try:
                # Unbound function of message
                self.__parent.removeMsg(msgGbl)
                pass
            except ValueError:
                pass
        self.__entries.pop(rowIndex)
        self.endRemoveRows()
        self._update()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def flags(self, index):
        flgs = super().flags(index)
        flgs |= Qt.ItemIsEditable
        return flgs


class _WhguiMessageBindDelegate(QItemDelegate):
    """
    Delegate for message items
    """

    def __init__(self, top, messages, parent=None):
        """
        :type CardTyp: () -> list[CardTyp]
        :type formatF: (elem,column) -> str
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)
        self.__topBlock = top
        self.__messages = messages

    def createEditor(self, parent, option, index):
        """
        Create a editors for message table the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        col = index.column()
        if col == 0:
            editor = QLabel(parent)
        elif col == 1:
            editor = QComboBox(parent)
            editor.addItem('', '')
            for glbMsg in self.__topBlock.getAllMessages():
                editor.addItem(glbMsg, glbMsg)
            editor.currentIndexChanged.connect(save)

        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox | QLineEdit
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.EditRole)
        if isinstance(editor, QComboBox):
            for index in range(editor.count()):
                if editor.itemData(index, Qt.UserRole) == data:
                    editor.setCurrentIndex(index)
                    break
            else:
                editor.setCurrentIndex(0)
        elif isinstance(editor, QLineEdit):
            editor.setText(data)

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox | QLineEdit
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        if isinstance(editor, QComboBox):
            model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)
        elif isinstance(editor, QLineEdit):
            if editor.text():
                model.setData(index, editor.text(), Qt.UserRole)
