# ----------------------------------------------------------------------------
# - wmegui-data-library-tree                                                 -
# - wild cruncher: data library database tree browser                        -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                   -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_logical_simulation_tree
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher logical simulation database tree browser

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""
from wcr_core.action.wact_action_api import WactActionObserverP, \
    WACT_ACTION_EXECUTED_EVENT
from wcr_core.common.wcmn_protocols import WcmnReleasableP
from wcr_core.db_api.whdb_instance_api import WhdbInstanceP
from wcr_frontend.gui.wgui_db_tree import WguiDbTreeItemFactory, \
    WguiDbTreeModelManager, WguiDbTreeSortFilter
from wcr_frontend.gui.wgui_tree import WguiTree, WguiTreeModelManager
from wcr_frontend.gui.wgui_tree_model import WguiFilterModel, \
    WguiSimpleTreeItemFactory, WguiTreeItemFactoryP, WguiTreeItemBase, \
    WguiTreeModel
from wcr_frontend.gui.wgui_protocols import WguiEmptyContextMenuFactory, \
    QWidget_Meta_M
from wcr_core.db.whdb_block_proxy import WhdbEntityContainerProxy

import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QPushButton

from wcr_frontend.modules.message.gui.wmegui_import_icd_file_handler import \
    WmeguiICDImportFileHandler
from wcr_core.project.wprj_xmlio import WPRJ_ICD_SILKAN_XML_MIME_TYPE
from wcr_frontend.gui.wgui_utils import wgui_getOpenFile
from wcr_core.modules.icd.db.wihdb_module_instance import WihdbModuleInstance



class WmeguiDataLibraryTree(WguiTree, WactActionObserverP,
                                 WcmnReleasableP, metaclass=QWidget_Meta_M):
    def __init__(self, pctx, cctx, actions=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param cctx: Cockpit system context
        :type cctx: WguiCockpitSystemContextP
        :param actions: Action to observe for update
        :type actions: Iterable[WactActionP] | None
        """
        self.__topLevel = WhdbEntityContainerProxy()
        self.__pctx = pctx
        self.__actions = list(actions) if actions else []

        # item factory
        self.__itemFc = _WmeguiModuleTreeItemFactory()

        # WguiTreeModelManager
        modelMngr = WmeguiModuleTreeModelManager(
            self.__pctx.getProjectDB().getTopHierarchy().getModules,
            self.__itemFc)

        #QAbstractItemModel
        model = WguiFilterModel(modelMngr.model)


        #: Collapse-all button
        self.__impButton = QPushButton('Import')
        self.__updateButton = QPushButton('Update')

        #tree init
        super().__init__(model, modelMngr, WguiEmptyContextMenuFactory())

        for action in self.__actions:
            if action:
                action.registerActionObserver(self)


    def _createLayout(self):
        """
        Create the layout.

        :return: Layout
        :rtype: QLayout
        """
        layout=super()._createLayout()

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__impButton)
        btnLayout.addWidget(self.__updateButton)
        layout.addLayout(btnLayout)

        return layout

    def _initElements(self):
        """
        Initialize the UI elements as needed.
        """
        super()._initElements()
        #link to import action
        self.__impButton.clicked.connect(self._importModules)
        # link to update action
        self.__updateButton.clicked.connect(lambda :
                                            self.updateTree(forceupdate=True))

    def release(self):
        for action in self.__actions:
            if action:
                action.unregisterActionObserver(self)
        self.__actions = []
        self.__topLevel.release()
        self.__pctx = None
        super().release()

    def actionNotify(self, actx, action, event, *args, **kwargs):
        if event & WACT_ACTION_EXECUTED_EVENT:
            self.updateTree()

    def updateTree(self, forceupdate=False):
        pctx = self.__pctx
        if pctx is None:
            return
        if self.__pctx.getProjectDB().getTopHierarchy().getModules() == [] \
                or forceupdate:
            for f in self.__pctx.getProjectDB().getTopHierarchy().\
                    getICDFile():
                self._updatemodules(f)
        super().updateTree()

    def _updatemodules(self, path):
        # Recuperation des messages
        handler = WmeguiICDImportFileHandler(
            *self.__pctx.getProjectDB().getModule())
        icdtblk, name = handler.handleFile(path, self.__pctx)
        if not icdtblk:
            raise Exception("No module to be found in {}".format(path))
        if name == '':
            name = 'file:' + os.path.basename(path)
        fmsgs = []
        for m in icdtblk.getOperatorInstances():
            fmsgs.append(m)

        # saving modules in the topblock
        self.__pctx.getProjectDB().getTopHierarchy().setModules(fmsgs)
        return name

    def _importModules(self):
        #Selection d'un fichier
        selected_path=wgui_getOpenFile("Import ICD Module",self,
                                       WPRJ_ICD_SILKAN_XML_MIME_TYPE)
        if selected_path:
            name = self._updatemodules(selected_path)
            self.__pctx.getProjectDB().getTopHierarchy().\
                setICDFile(selected_path, name)
            self.updateTree()
            #update tree to make new messages apear

class WbguiTreeModel(WguiTreeModel):

    def __init__(self, itemFc):
        super().__init__(itemFc)

    def setSourceModel(self, model):
        None

    def setDynamicSortFilter(self, stat):
        None

    def invalidate(self):
        None


class WmeguiModuleTreeModelManager(WguiTreeModelManager):
    """
    Tree-model manager for database trees.
    """

    def __init__(self, mdll, itemFy):
        """
        :param mdll: Database top-level block
        :type mdll: WhdbEntityContainerP
        :param itemFy: Tree-item factory
        :type itemFy: WguiTreeItemFactoryP
        """
        def itemIdF(i):
            if i is None:
                return None
            if isinstance(i, WihdbModuleInstance):
                return i.getUUID()
            return i

        super(WmeguiModuleTreeModelManager, self).__init__(itemIdF, itemFy)

        self.__modlst = mdll

    def populateModel(self, fromScratch = True):
        """
        :type fromScratch: bool
        """
        def addFiles(moduleliste, path):
            for module in moduleliste():
                self.itemAdded(module, path)
                path1=list(path)
                path1.append(module)
                for dat in module.getDatas():
                    self.itemAdded(dat[0]+" -> "+dat[1],path1)

        if fromScratch:
            self.itemClear()
        addFiles(self.__modlst, [])

    def itemClear(self):
        """
        Remove all the items. respecting leaforder
        """
        for itemKey, itemVal in self.getmap().items():
            #we delet only "_WbguiModuleTreeItem" constituing the first
            # level of the tree.
            if isinstance(itemVal, _WbguiModuleTreeItem):
                self.model.removeItem(itemVal)
                del itemVal
        self.getmap().clear()


# noinspection PyAbstractClass
class _WbguiTreeItemBase(WguiTreeItemBase):
    """
    State-machine Datamodel tree item.
    """

    def __init__(self, parentItem=None):
        super().__init__(parentItem)

    def setData(self, data, column, role=None):
        pass

    def getColumnCount(self):
        return 1

    @property
    def flags(self):
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    # @abstractmethod
    def createDetailUI(self):
        """
        Create the detail-view UI.

        :return: Detail-view UI
        :rtype: QWidget
        """
        raise NotImplementedError()


class _WbguiMessageRootTreeItem(_WbguiTreeItemBase):
    def __init__(self):
        super().__init__()

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return 'Item'
        return None

    def createDetailUI(self):
        return None

class _WbguiModuleTreeItem(_WbguiTreeItemBase):
    def __init__(self, itemData, parentItem=None):

        self.data = itemData

        super().__init__(parentItem)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return self.data.getName()
        return None

    def createDetailUI(self):
        return None

class _WbguiDataTreeItem(_WbguiTreeItemBase):
    def __init__(self, itemData, parentItem=None):

        self.data = itemData

        super().__init__(parentItem)

    def getData(self, column, role=None):
        if role == Qt.DisplayRole:
            if column == 0:
                return self.data#.getName()
        return None

    def createDetailUI(self):
        return None


class _WmeguiModuleTreeItemFactory(WguiTreeItemFactoryP):

    def createTreeItem(self, itemData, parentItem, **kwargs):
        if not parentItem:
            return _WbguiMessageRootTreeItem(), None, None
        if not itemData:
            return None, None, None
        #WARNING whdbBlock are used cause it is the type of topblock of
        # message... it coul'd be more specified.
        if isinstance(itemData, WihdbModuleInstance):
            return _WbguiModuleTreeItem(itemData, parentItem), \
                   parentItem, None
        return _WbguiDataTreeItem(itemData, parentItem), parentItem, None