# ----------------------------------------------------------------------------
# - wfgui-instance-props-dialog                                              -
# - wild cruncher - cockpit : instance properties dialog widgets             -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wfgui_instance_props_dialog
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher state machine instance properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtWidgets import QComboBox, QCheckBox, QVBoxLayout

from wcr_frontend.gui.wgui_instance_props_dialog import \
    WguiInstancePropertiesDialog
from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.modules.statemachine.gui.wfgui_executables_props_widget \
    import WfguiExecutableProperties
from wcr_frontend.modules.statemachine.gui.wfgui_datamodel_props_widget \
    import WfguiDatamodelProperties
from wcr_core.modules.statemachine.db.wfhdb_block import WfhdbParallelBlock
from wcr_core.modules.statemachine.db.wfhdb_instance import \
    WfhdbBlockInstance, WfhdbInstanceP, WfhdbStateInstance
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data, wgui_qcb_set_data
from wcr_core.modules.statemachine.db.wfhdb_protocols import *

_WFGUI_HISTORYTYES = [('None', WFHDB_HISTORY_TYPE_NONE),
                      ('Deep', WFHDB_HISTORY_TYPE_DEEP),
                      ('Shallow', WFHDB_HISTORY_TYPE_SHALLOW)]

class WfguiInstancePropertiesDialog(WguiInstancePropertiesDialog):
    """
    The WfguiInstancePropertiesDialog class is a dialog design to access
    and manipulate instance data. This include name, info, initial/final state
    and on-entry/on-exit executables.
    """

    def __init__(self, inst, pctx, updateF, parent=None):
        """
        :param inst: State/parallal instance
        :type inst: WfhdbInstanceP
        :param pctx: Project context
        :type pctx: WcorProjectContextP
        :param updateF: Callback function to notify about saves
        :type updateF: () -> _
        :param parent: Parent widget
        """
        super().__init__(inst, pctx, updateF, parent=parent)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _createAdditionalInstancePropertiesWidget(self, inst, tophm):
        if not isinstance(inst, WfhdbInstanceP):
            return None
        # noinspection PyUnresolvedReferences
        if isinstance(inst, WfhdbBlockInstance) and \
                isinstance(inst.getBlock(), WfhdbParallelBlock):
            return None

        return [_WfguiStateProperties(inst)]

    # noinspection PyMethodMayBeStatic
    def _createAdditionalPropertiesWidget(self, inst, tophm, pctx):
        if isinstance(inst, WfhdbInstanceP):
            return [('On Entry', _WfhdbOnEntryExecutionProperty(inst)),
                    ('On Exit', _WfhdbOnExitExecutionProperty(inst)),
                    ('Datamodel', _WfhdbDataModelProperty(inst))]
        else:
            return []


class _WfguiStateProperties(WguiPropertiesWidgetP):
    def __init__(self, inst, parent=None):
        """
        :param inst: State instance
        :type inst: WfhdbInstanceP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: State instance
        self.__instance = inst

        #: Initial state checkbox
        self.__initialF = QCheckBox('Initial State')
        #: Final state checkbox
        self.__finalF = QCheckBox('Final State')
        #: History state checkbox
        self.__historyF = QCheckBox('History State')
        #: Combo-box for type-selection for history
        self.__histComboW = QComboBox()

        self.setLayout(self._createLayout())
        self._initElements()
        self.updateFields()

    def _createLayout(self):
        """
        Create a new layout.

        :return: Layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.addWidget(self.__initialF)
        layout.addWidget(self.__finalF)
        if isinstance(self.__instance, WfhdbStateInstance):
            layout.addWidget(self.__historyF)
            layout.addWidget(self.__histComboW)
        return layout

    def updateHistory(self):
        if self.__historyF.isChecked():
            self.__histComboW.setEnabled(True)
            wgui_qcb_set_data(self.__instance.getHistoryType(),
                              self.__histComboW)
        else:
            wgui_qcb_set_data(None, self.__histComboW)
            self.__histComboW.setEnabled(False)


    def _initElements(self):
        for item, code in _WFGUI_HISTORYTYES:
            self.__histComboW.addItem(item, code)
        self.__historyF.clicked.connect(self.updateHistory)
        self.updateHistory()

    def updateFields(self):
        self.__initialF.setChecked(self.__instance.isInitialState())
        self.__finalF.setChecked(self.__instance.isFinalState())
        if isinstance(self.__instance, WfhdbStateInstance):
            self.__historyF.setChecked(self.__instance.isHistoryState())
            if self.__historyF.isChecked():
                self.updateHistory()

    def save(self):
        self.__instance.setInitialState(self.__initialF.isChecked())
        self.__instance.setFinalState(self.__finalF.isChecked())
        if isinstance(self.__instance, WfhdbStateInstance):
            self.__instance.setHistoryState(self.__historyF.isChecked())
            self.__instance.setHistoryType(wgui_qcb_get_data(self.__histComboW))
        else:
            self.__instance.setHistoryState(False)
            self.__instance.setHistoryType(None)


class _WfhdbOnEntryExecutionProperty(WfguiExecutableProperties):
    def __init__(self, inst, parent=None):
        """
        :param inst: State instance
        :type inst: WfhdbInstanceP
        :param parent: Parent widget
        """
        #: State instance
        self.__instance = inst

        super().__init__(parent)

    def _getExecutables(self):
        return self.__instance.getOnEntryExec()

    def _setExecutables(self, execs):
        self.__instance.setOnEntryExec(execs)


class _WfhdbOnExitExecutionProperty(WfguiExecutableProperties):
    def __init__(self, inst, parent=None):
        """
        :param inst: State instance
        :type inst: WfhdbInstanceP
        :param parent: Parent widget
        """
        #: State instance
        self.__instance = inst

        super().__init__(parent)

    def _getExecutables(self):
        return self.__instance.getOnExitExec()

    def _setExecutables(self, execs):
        self.__instance.setOnExitExec(execs)

class _WfhdbDataModelProperty(WfguiDatamodelProperties):
    def __init__(self, inst, parent=None):
        """
        :param inst: State instance
        :type inst: WfhdbInstanceP
        :param parent: Parent widget
        """
        #: State instance
        self.__instance = inst

        super().__init__(parent)

    def _getDatas(self):
        od = self.__instance.getDataModel(self.__instance.rawElement)
        l = []
        for k in od.keys():
            Datdic={WFHDB_DATA_ID_KEY: k}
            Datdic.update(od[k])
            l.append(("Data", Datdic))
        return l

    def _setDatas(self, execs):
        add_data={}
        for dat in execs:
            Datadesc = dat[1]
            datkey = Datadesc.pop(WFHDB_DATA_ID_KEY)
            add_data[datkey] = Datadesc
        old_data = self.__instance.getDataModel(self.__instance.rawElement)
        for oldk in old_data.keys():
            if oldk not in add_data.keys():
                self.__instance.removeFromDataModel(self.__instance.rawElement,
                                                    oldk)
        self.__instance.setDataModel(self.__instance.rawElement, add_data)
