# ----------------------------------------------------------------------------
# - wcgui-vlink-prod-binding-table                                           -
# - wild cruncher:sys_map VLink producer bindings widget                     -
# ----------------------------------------------------------------------------
# - author    (c) 2016-2016 philippe lambert                                 -
# - copyright (c) 2016-2016 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wcgui_vlink_prod_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan sys_map vlink producer bindings widget

.. moduleauthor:: Philippe Lambert <lambert@silkan.com>
"""

from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QTableView, QHeaderView, \
	QVBoxLayout, QPushButton, QItemDelegate, QAbstractItemView
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data
from wcr_core.modules.hardware.db.whhdb_Cetrac_switch_instance import \
WhhdbCetracSwitchInstance
from wcr_core.modules.hardware.db.whhdb_kazymir_switch_instance import \
WhhdbKazymirSwitchInstance
from wcr_core.modules.sys_map.plugin.wcplg_config_generator_params import \
    WcplgGetAllRouteOfSameCodeFromPort, WcplgGetAllPortOfSameCodeFromRoute
from silkan_common.scmn_utils import unique


class WcguiVLProdBindingTable(QWidget):
    """
    Virtual link producer binding to physical Host table widget.
    """

    def __init__(self, pctx, scene, filter, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene
        #: View filter
        self.__filter = filter


        #: formating function for element ot table
        self.__formatF = lambda o, c: self.formatF(o, c)
        #: Bus binding table model
        self.__bndgTblM = _WcguiProdBindingTableModel(self.__formatF,
                                                      parent=self)

        #: Creating function to get all possible host
        self.__pportF = lambda: self.getphisicalports()
        #: Creating function geting all available modules
        self.__modF = lambda: self.getNonProdBindedsVLink()
        #: Creating editor for element in table
        self.__bndgPortProdDlgt = _WcguiPortProdBindDelegate(self.__pportF,
                                                             self.__modF,
                                                             self.__formatF,
                                                             self.getModel())

        #: Bus binding table widget
        self.__bndgTblW = QTableView()

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    @property
    def table(self):
        return self.__bndgTblW

    def getAvailPorts(self):
        return self.__pportF

    def getAvailVls(self):
        return self.__modF

    def getScene(self):
        return self.__scene

    def getModel(self):
        return self.__bndgTblM

    def formatF(self, obj, col):
        if not obj :
            return ''
        elif col == 0:
            #: at col 0 we must have a (str, WhhdbHostInstance)
            return obj[0]
        elif col == 1:
            #: at col 1 we must have ( Name, WvhdbVLinkInstance)
            # (will change at implementing of "normal" module
            return obj[0]
        else:
            #: we asking a name for an object in column normaly not present
            return 'No formating available'

    def _initElements(self):
        self.__bndgTblW.setModel(self.__bndgTblM)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__bndgTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        #: delegate classe for edition in table
        self.__bndgTblW.setItemDelegate(self.__bndgPortProdDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.table.selectionModel().selectionChanged.connect(self.highlightConnectors)

        pass

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.addWidget(self.__bndgTblW)

        btnLayout = QHBoxLayout()
        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        layout.addLayout(btnLayout)

        return layout

    def reloadFromData(self):
        """
        remove all row of table and refill it with element from project data
        """
        self.__bndgTblM.deleteRows(range(0,self.__bndgTblM.rowCount()),
                                   Unbind=False)
        self._loadFromData()

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        map = self.__pctx.getMappingManager()

        pportperml = []
        for h in map.getHosts():
            if self.__filter.uuidpassfilter(h.getUUID()):
                for p in h.getConnectionPoints():
                    pportperml.append(p.getUUID())
        sportperml = []
        for sw in self.__pctx.getMappingManager().getSwitchs():
            if self.__filter.uuidpassfilter(sw.getUUID()):
                if (isinstance(sw, WhhdbCetracSwitchInstance) or
                        isinstance(sw, WhhdbKazymirSwitchInstance)):
                    for conIo in sw.getAllIoConnectionPoints():
                        sportperml.append(conIo.getUUID())
                    if sw.getArmPort() == 'Yes':
                        con = sw.hasArmPort()
                        sportperml.append(con.getUUID())

        for bnd in self.__pctx.getProjectDB().getTopHierarchy().getProdBinds():
            if bnd[0] in pportperml:
                port = map.getPhyPortByUUID(bnd[0])
                cardName = port.getCard()
                host = map.getHostFromPortUUID(bnd[0])
                self.__bndgTblM.addRow((
                    self.formatPhisicalPortName(port, cardName, host),
                    map.getPhyPortByUUID(bnd[0])),
                    (map.getVlinkNameByUUID(bnd[1]),
                     map.getVlinkByUUID(bnd[1])))
            elif bnd[0] in sportperml:
                port = map.getPhyPortByUUID(bnd[0])
                cardName = port.getCard()
                switch = map.getSwitchByPortUUID(bnd[0])
                if switch.getArmPort() == 'Yes':
                    portName = '{}.{}_card.{}'.format(switch.getName(),
                                                      switch.getName(),
                                                      port.getName())
                else:
                    portName = self.formatPhisicalPortName(port,
                                                           cardName,
                                                           switch)
                fullPortName = (portName, map.getPhyPortByUUID(bnd[0]))
                self.__bndgTblM.addRow(fullPortName,
                    (map.getVlinkNameByUUID(bnd[1]),
                     map.getVlinkByUUID(bnd[1])))

    def _addEntry(self, port=None, vl=None):
        """
        Add entry in the model
        """
        self.__bndgTblM.addRow(port, vl)

    def getBindings(self):
        """
            Get entry in the model
        """
        return self.__bndgTblM.getDataRows()

    def formatPhisicalPortName(self, port, cardName, host):
        return host.getName() + "." + cardName + "." + port.getName()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__bndgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__bndgTblM.deleteRows([idx.row() for idx in selIdx])

    def unbindPortProd(self, pport, vl):
        """
        unbind the physical port and virtual link for the topblock
        :param pport: physical port
        :type pport: WhhdbHardwareXPoint
        :param vl: virtual link
        :type vl: WvhdbVLinkInstance
        """
        self.__pctx.getProjectDB().getTopHierarchy().removeProdBind(pport[1].getUUID(),vl[1].getUUID())

    def bindPortProd(self, pport, vlink):
        """
        bind the host and the module for the topblock
        :param pport: host instance
        :type pport: WhhdbHardwareXPoint
        :param vlink: module instance
        :type vlink: WvhdbVLinkInstance
        """
        for p,vl in self.__pctx.getProjectDB().\
                getTopHierarchy().getProdBinds():
            if vl == vlink[1].getUUID():
                return
                # raise ValueError("VLink already have producer")
        self.__pctx.getProjectDB().getTopHierarchy(). \
            addProdBind(pport[1].getUUID(),vlink[1].getUUID())

    def getNonProdBindedsVLink(self):
        """
        return the list of virtual links who are not bound to a physical port
        as producers
        """
        tblk = self.__pctx.getProjectDB().getTopHierarchy()
        mngr = self.__pctx.getMappingManager()
        vlinkl = mngr.getVlList()
        if self.__scene.top.getVLinkBlock():
            autoVls = [(inst.getName(), inst)
                       for inst in self.__scene.top.getVLinkBlock().getInstances()]
            vlinkl.extend(autoVls)
        """:type :[( Name, WvhdbVLinkInstance)]"""
        for pport, vl in tblk.getProdBinds():
            if (mngr.getVlinkNameByUUID(vl),
                    mngr.getVlinkByUUID(vl)) in vlinkl:
                vlinkl.remove(
                    (mngr.getVlinkNameByUUID(vl),
                     mngr.getVlinkByUUID(vl)))
        vlinkl = unique(vlinkl)
        return vlinkl

    def getphisicalports(self):
        """
        return the list of the physical port
        """
        conl = []
        for h in self.__pctx.getMappingManager().getHosts():
            if self.__filter.uuidpassfilter(h.getUUID()):
                for con in h.getConnectionPoints():
                    if con.getActive() == "Off":
                        continue
                    cardName = con.getIoCard()
                    conl.append((self.formatPhisicalPortName(con, cardName, h),
                                 con))
        # get Io connectors from switch
        for sw in self.__pctx.getMappingManager().getSwitchs():
            if self.__filter.uuidpassfilter(sw.getUUID()):
                if isinstance(sw, WhhdbCetracSwitchInstance) or \
                        isinstance(sw, WhhdbKazymirSwitchInstance):
                    for conIo in sw.getAllIoConnectionPoints():
                        if conIo.getActive() == "Off":
                            continue
                        cardName = conIo.getIoCard()
                        conl.append((self.formatPhisicalPortName(conIo, cardName, sw),
                                     conIo))
                    if sw.getArmPort() == 'Yes':
                        con = sw.hasArmPort()
                        fullPortName = '{}.{}_card.portArm'.format(sw.getName(),
                                                                   sw.getName())
                        conl.append((fullPortName, con))
        return conl

    def highlightConnectors(self):
        indexes = self.table.selectedIndexes()

        portNames = set()
        for index in indexes:
            row = index.row()
            portData = self.getBindings()[row][0]
            if portData:
                portName = portData[0]
                portNames.add(portName)
        self.__scene.highlightNamedConnector(portNames)


class _WcguiProdBindingTableModel(QAbstractTableModel):
    """
    VLink producer Binding table model.
    """

    TITLE_VLINK = 'Traffic Route'
    TITLE_PPORT = 'Phys Port'

    HEADERS = [TITLE_PPORT, TITLE_VLINK]

    def __init__(self, formatF, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(parent)
        self.__parent=parent


        #: Entries as list of tuples of
        #  physical ports (pport) and virtual link (vlink)
        self.__entries = []
        """:type: list[]"""
        self.__formatF=formatF
        self.rowsInserted.connect(self.dataHasChanged)
        self.rowsRemoved.connect(self.dataHasChanged)
        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, pport=None, vlink=None):
        """
        Add a row at the end of the list (append)

        :param pport: physical port to be set
        :type pport: WhhdbHardwareXPoint | None
        :param vlink: Module to be set
        :type vlink: ( name , WvhdbVLinkInstance) | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((pport, vlink))

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices, Unbind=True):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row, Unbind=Unbind)

    def deleteRow(self, rowIndex, Unbind=True):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        pport, vlinkd = self.__entries[rowIndex]
        if pport is not None and vlinkd is not None and Unbind:
            try:
                self.__parent.unbindPortProd(pport, vlinkd)
                pass
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def getDataRows(self):
        return self.__entries

    def getVlRowIndex(self, vlUUID):
        for rowIndex in range(0, self.rowCount()):
            pport, vlinkd = self.__entries[rowIndex]
            if vlinkd:
                if vlUUID == vlinkd[1].getUUID():
                    return rowIndex

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column < 2:
            info = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(info, column)
            elif role == Qt.UserRole:
                return info
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole
        self._setDataEntry(index, data)

    def _setDataEntry(self, index, data):
        pport, vld = self.__entries[index.row()]
        if pport and vld :
            #a binding could have been done by that line before
            # it have to be deleted.
            try:
                self.__parent.unbindPortProd(pport,vld)
            except ValueError:
                pass

        #creating the new representation line
        nl=()
        if index.column() == 0:
            nl=(data,vld)
        elif index.column() == 1:
            nl=(pport, data)

        self.__entries[index.row()]=nl

        pport, vld = self.__entries[index.row()]
        if pport and vld :
            #line is full a binding have to be done
            try:
                self.__parent.bindPortProd(pport, vld)
            except ValueError:
                pass

    def flags(self, index):
        """
        set flags fot able
        """
        flgs = super().flags(index)
        #allowing editing
        flgs |= Qt.ItemIsEditable
        return flgs

    def dataHasChanged(self):
        self.__parent.getScene().sceneModified.emit()

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2


class _WcguiPortProdBindDelegate(QItemDelegate):
    """
    Delegate for modules and host
    """

    def __init__(self, hostListF, moduleListF, formatF, model, parent=None):
        """
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.formatF = formatF
        self.__hostListPrv = hostListF
        self.__moduleListPrv = moduleListF
        self.__model = model

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """

        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QComboBox(parent)
        row = index.row()
        col = index.column()

        #start withe empty line to allowing unset
        editor.addItem('', None)

        entries = self.__model._WcguiProdBindingTableModel__entries
        # the available port and route depends partially of selected element
        if col == 0:
            routeData = entries[row][1]
            if not routeData:
                busPrvF = self.__hostListPrv
            else:
                busPrvF = lambda: WcplgGetAllPortOfSameCodeFromRoute(routeData[1], self.__hostListPrv())
        elif col == 1:
            # column 0
            portData = entries[row][0]
            if not portData:
                busPrvF = self.__moduleListPrv
            else:
                busPrvF = lambda: WcplgGetAllRouteOfSameCodeFromPort(portData[1], self.__moduleListPrv())
            # add the curent module to the list to conserve it
            # if a binding have been done
            cdata=index.model().data(index.model().index(index.row(), 1),
                                    role=Qt.UserRole)
            if not cdata is None:
                editor.addItem(self.formatF(cdata,1),
                               cdata)
        for elem in busPrvF():
            editor.addItem(self.formatF(elem, col), elem)
        editor.currentIndexChanged.connect(save)

        return editor

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)

    def setEditorData(self, editor, index):
        """
        Initialising editor data with most
        :type editor: QComboBox
        :type index: QModelIndex
        """
        data = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == data:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)
