# ----------------------------------------------------------------------------
# - wpgui-bus-binding-table                                                  -
# - wild cruncher:system bus bindings widget                                 -
# ----------------------------------------------------------------------------
# - author    (c) 2013-2014 samuel longchamps                                -
# - copyright (c) 2013-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_bus_binding_table
   :platform:     Unix, Darwing, Windows
   :synopsis:     Wild Cruncher silkan system bus bindings widget

.. moduleauthor:: Samuel Longchamps <samuel.longchamps@silkan.com>
"""
from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QVBoxLayout, \
	QPushButton, QItemDelegate, QTableView, QHeaderView, QAbstractItemView
from wcr_core.modules.system.wpmod_project_context import \
    wpmod_retrieve_lbus_obj
from wcr_frontend.gui.wgui_utils import wgui_qcb_get_data


class WpguiBusBindingTable(QWidget):
    """
    Logical bus binding to physical buses table widget.
    """

    BTN_PRINT_BINDING_VISIBLE = False

    def __init__(self, pctx, scene, parent=None):
        """
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param scene: Graphics scene (for current block and change signals)
        :type scene: WguiGraphicsScene
        :param parent: Parent widget
        """
        super().__init__(parent)

        # : Project context
        self.__pctx = pctx
        #: Graphical scene
        self.__scene = scene

        # !!! Will need update for handling of multiple logical simulations !!!
        self.__logSimLstPrvF = lambda: [pctx.getLogicalSimulation()] \
            if pctx.getLogicalSimulation() is not None else []
        """:type: () -> list[WshdbTopBlock]"""

        self.__busFormatF = lambda b: wsgui_formatBus(b, self.__logSimLstPrvF)

        #: Bus binding table model
        self.__bndgTblM = _WpguiBusBindingTableModel(self.__busFormatF)

        self.__pbusListPrvF = lambda: [pb for pb in
                                       self.__pctx.getPhysicalBuses()]
        self.__lbusListPrvF = lambda: [lb for lsim in self.__logSimLstPrvF()
                                       for lb in lsim.getBusInstances()]

        """:type: (WphdbPhyBusInstance) -> list[WshdbBusInstance]"""

        #: Bud binding table item delegate
        self.__bndgItmDlgt = _WpguiBusBindDelegate(
            self.__pbusListPrvF, self.__lbusListPrvF,
            self.__busFormatF)

        #: Bus binding table widget
        self.__bndgTblW = QTableView()

        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        #: Clean invalid
        self.__cleanInvalidButton = QPushButton('Clean Invalid')
        #: Print bindings button
        self.__printBindingButton = QPushButton('Print Binding')

        self._initElements()
        self.setLayout(self._createLayout())
        self._loadFromData()

    def _initElements(self):
        self.__bndgTblW.setModel(self.__bndgTblM)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            0, QHeaderView.Stretch)
        self.__bndgTblW.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch)
        self.__bndgTblW.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.__bndgTblW.setItemDelegate(self.__bndgItmDlgt)

        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.__cleanInvalidButton.clicked.connect(self._deleteInvalidEntries)
        self.__printBindingButton.clicked.connect(self._printBinding)

        self.__scene.sceneItemRemoved.connect(self._updateDeletedBus)

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(self.__bndgTblW)

        btnLayout = QHBoxLayout()

        btnLayout.addWidget(self.__printBindingButton)
        btnLayout.addWidget(self.__cleanInvalidButton)
        btnLayout.addStretch(1)
        btnLayout.addWidget(self.__addButton)
        btnLayout.addWidget(self.__deleteButton)
        self.__printBindingButton.setVisible(self.BTN_PRINT_BINDING_VISIBLE)
        layout.addLayout(btnLayout)

        return layout

    def _loadFromData(self):
        """
        Fill the table from the project's data
        """
        lbusDict = wpmod_retrieve_lbus_obj(
            self.__pbusListPrvF(), self.__logSimLstPrvF())

        for pbus in self.__pbusListPrvF():
            lbusList = lbusDict.get(pbus.getUUID(), [])
            for lbus in lbusList:
                self.__bndgTblM.addRow(pbus, lbus)

    def _deleteInvalidEntries(self, validDlgtF=None):
        """
        Delete invalid entries using a validation function.
        If no function is provided, a default one is applied which is mostly
        exhaustive.

        :param validDlgtF: Validity function delegate
        :type validDlgtF: function
        """

        def isEntryValid(pbus, lbus):
            """
            :type pbus: WphdbPhyBusInstance
            :type lbus: WshdbBusInstance
            :rtype: bool
            """
            return pbus is not None and lbus is not None and \
                lbus.getUUID() in pbus.getBoundLogicalBusesUUID() and \
                lbus.getUUID() in [lbusInst.getUUID()
                                   for lbusInst in self.__lbusListPrvF()] and \
                pbus.getUUID() in [pbusInst.getUUID()
                                   for pbusInst in self.__pbusListPrvF()]

        self.__bndgTblM.deleteInvalidEntries(validDlgtF or isEntryValid)

    def _addEntry(self):
        """
        Add entry in the model
        """
        self.__bndgTblM.addRow()

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        selIdx = self.__bndgTblW.selectedIndexes()
        """:type: list[QModelIndex]"""

        self.__bndgTblM.deleteRows([idx.row() for idx in selIdx])

    def _printBinding(self):
        """
        Print the binding for debug purposes
        """
        lbusNameDict = {}
        """:type: dict[UUID, str]"""

        for lbus in self.__lbusListPrvF():
            lbusNameDict[lbus.getUUID()] = self.__busFormatF(lbus)

        print('== List of bound logical buses ==')
        for pbus in self.__pbusListPrvF():
            print('PBUS: {}'.format(self.__busFormatF(pbus)))
            boundList = pbus.getBoundLogicalBusesUUID()
            if len(boundList) == 0:
                print('(No logical bus bound)')
            else:
                for boundLbusUUID in pbus.getBoundLogicalBusesUUID():
                    print('-> LBUS: {}'.format(
                        lbusNameDict.get(boundLbusUUID, 'Not loaded')))


class _WpguiBusBindingTableModel(QAbstractTableModel):
    """
    Bus Binding table model.
    """

    TITLE_PHYBUS = 'Physical Bus'
    TITLE_LOGBUS = 'Logical Bus'

    HEADERS = [TITLE_PHYBUS, TITLE_LOGBUS]

    ERR_WRONG_BUS_TYPE = 'Value newBus must be either of type physical or ' \
                         'logical bus'

    def __init__(self, formatF, parent=None):
        """
        :param formatF: Formatting function
        :type formatF: WshdbBusInstance -> str
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: Buses formatting function
        self.__formatF = formatF

        #: Entries as list of tuples of
        #  physical buses (pbus) and logical buses (lbus)
        self.__entries = []
        """:type: list[(WphdbPhyBusInstance, WshdbBusInstance)]"""

        self._update()

    def _update(self):
        self.beginResetModel()
        self.endResetModel()

    def _doUpdate(self):
        pass

    def beginResetModel(self):
        super().beginResetModel()

    def endResetModel(self):
        self._doUpdate()
        super().endResetModel()

    def headerData(self, section, orientation, role=None):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal and \
               0 <= section < len(self.HEADERS):
                return self.HEADERS[section]
            else:
                return str(section)
        return None

    def addRow(self, pbus=None, lbus=None):
        """
        Add a row at the end of the list (append)

        :param pbus: Physical bus to be set
        :type pbus: WphdbPhyBusInstance | None
        :param lbus: Logical bus to be set
        :type lbus: WshdbBusInstance | None
        :return: The appended row index
        :rtype: int
        """
        index = self.index(self.rowCount(), 0)
        self.beginInsertRows(index, self.rowCount(), self.rowCount())

        self.__entries.append((pbus, lbus))

        if pbus is not None and lbus is not None:
            pbus.addBoundLogicalBus(lbus)

        self.endInsertRows()
        self._update()

    def deleteRows(self, rowIndices):
        """
        :param rowIndices: Iterable of indices
        :type rowIndices: Iterable[int]
        """
        # Make sure no duplicate exist
        rowList = list(set(rowIndices))
        rowList.sort(reverse=True)

        for row in rowList:
            self.deleteRow(row)

    def deleteRow(self, rowIndex):
        """
        Delete a row entry

        :param rowIndex: Row to the delete
        :type rowIndex: int
        """
        if not (0 <= rowIndex < self.rowCount()):
            raise ValueError('Index of row is out of bounds')

        index = self.index(rowIndex, 0)
        self.beginRemoveRows(index, rowIndex, rowIndex)

        pbus, lbus = self.__entries[rowIndex]
        if pbus is not None and lbus is not None:
            try:
                pbus.removeBoundLogicalBus(lbus)
            except ValueError:
                pass

        self.__entries.pop(rowIndex)

        self.endRemoveRows()
        self._update()

    def deleteInvalidEntries(self, validEntryF):
        """
        Delete invalid entries based on a passed function to determine if they
        are valid

        :param validEntryF: Function used to validate the entries
        :type validEntryF: (WphdbPhyBusInstance, WshdbBusInstance) -> bool
        """
        rowsToDelete = []
        """:type: list[int]"""

        for row in range(self.rowCount()):
            pbus, lbus = self.__entries[row]

            if not validEntryF(pbus, lbus):
                rowsToDelete.append(row)

        self.deleteRows(rowsToDelete)

    def index(self, row, column, parentIndex=None, *args, **kwargs):
        return self.createIndex(row, column)

    def data(self, index, role=None):
        row = index.row()
        column = index.column()
        if 0 <= column <= 1:
            bus = self.__entries[row][column]
            if role in [Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole]:
                return self.__formatF(bus)
            elif role == Qt.UserRole:
                return bus
        return None

    def setData(self, index, data, role=None):
        """
        :type index: QModelIndex
        :type data: object
        """
        assert role == Qt.UserRole

    def _setBusEntry(self, index, newBus):
        """
        :param index: Index of the physical bus in the table
        :type index: QModelIndex
        :param newBus: New bus object
        :type newBus: WphdbPhyBusInstance | WshdbBusInstance
        """

        pbus, lbus = self.__entries[index.row()]
        if pbus is not None and lbus is not None:
            try:
                pbus.removeBoundLogicalBus(lbus)
            except ValueError:
                pass

        pbus, lbus = self.__entries[index.row()]
        if pbus is not None and lbus is not None:
            pbus.addBoundLogicalBus(lbus)

    def flags(self, index):
        flgs = super().flags(index)
        flgs |= Qt.ItemIsEditable
        return flgs

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.__entries)

    def columnCount(self, parent=None, *args, **kwargs):
        return 2


class _WpguiBusBindDelegate(QItemDelegate):
    """
    Delegate for bus items (both physical and logical)
    """

    def __init__(self, pbusListPrv, lbusListPrv, formatF, parent=None):
        """
        :type pbusListPrv: () -> list[WphdbPhyBusInstance]
        :type lbusListPrv: () -> list[WshdbBusInstance]
        :type formatF: (WphdbPhyBusInstance | WshdbBusInstance) -> str
        :param parent: Parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.formatF = formatF
        self.__pbusLstF = pbusListPrv
        self.__lbusLstF = lbusListPrv

    def createEditor(self, parent, option, index):
        """
        Create a combobox containing adequate choices for the selected item

        :type parent: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """
        def save(*_):
            self.setModelData(editor, index.model(), index)

        editor = QComboBox(parent)

        mdl = index.model()
        # Get the bus object at the column index of the current row
        getBusColF = lambda colIdx:\
            mdl.data(mdl.index(index.row(), colIdx), Qt.UserRole)

        # Filter function to check if a binding already exist between two buses
        filterF = lambda _: True
        filerBaseF = lambda pb, lb:\
            not (lb.getUUID() in pb.getBoundLogicalBusesUUID())

        # Get the other bus and assign the adequate filter function using it
        # as filling for the base filter function
        col = index.column()
        curBus = getBusColF(col) if (col == 0 or col == 1) else None
        otherBus = getBusColF(1) if col == 0 else \
            getBusColF(0) if col == 1 else None

        if otherBus is not None:
            curChoiceF = lambda _: False
            if curBus is not None:
                curChoiceF = lambda bs: bs.getUUID() == curBus.getUUID()
            filterF = \
                (lambda pb: filerBaseF(pb, otherBus) or curChoiceF(pb))\
                if col == 0 else \
                (lambda lb: filerBaseF(otherBus, lb) or curChoiceF(lb))

        # Iterate through the buses using the adequate provider for the column
        # and add items that pass the filter to the combo box
        busPrvF = self.__pbusLstF if col == 0 else self.__lbusLstF
        for bus in busPrvF():
            if filterF(bus):
                editor.addItem(self.formatF(bus), bus)
        editor.currentIndexChanged.connect(save)

        return editor

    def setEditorData(self, editor, index):
        """
        :type editor: QComboBox
        :type index: QModelIndex
        """
        bus = index.model().data(index, Qt.UserRole)
        for index in range(editor.count()):
            if editor.itemData(index, Qt.UserRole) == bus:
                editor.setCurrentIndex(index)
                break
        else:
            editor.setCurrentIndex(0)

    def setModelData(self, editor, model, index):
        """
        :type editor: QComboBox
        :type model: QAbstractItemModel
        :type index: QModelIndex
        """
        model.setData(index, wgui_qcb_get_data(editor), Qt.UserRole)

    def updateEditorGeometry(self, editor, option, index):
        """
        :type editor: QWidget
        :type option: QStyleOptionViewItem
        :type index: QModelIndex
        """
        editor.setGeometry(option.rect)


def wsgui_formatBus(bus, simTopBlkListPrv):
    """
    Format a buse for textual representation.

    :param bus: Either physical of logical bus
    :type bus: WphdbPhyBusInstance | WshdbBusInstance | None
    :param simTopBlkListPrv: Simulation project top block list provider
    :type simTopBlkListPrv: () -> list[WshdbTopBlock]
    :return: Formatted bus representation
        (with simulation project alias as prefix for logical buses)
    :rtype: str
    """
    if bus is None:
        return ''

    return bus.getName()
