# ----------------------------------------------------------------------------
# - wpgui-host-cards-props-widget                                            -
# - wild cruncher: cards properties dialog           frontend / system / gui -
# ----------------------------------------------------------------------------
# - author    (c) 2014-2014 clemens oertel                                   -
# - copyright (c) 2014-2014 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wpgui_host_cards_props_widget
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher cards properties dialog

.. moduleauthor:: Clemens Oertel <clemens.oertel@silkan.com>
"""
from PyQt5.QtWidgets import QVBoxLayout, QLabel

from wcr_frontend.gui.wgui_props_dialog_widget import WguiPropertiesWidgetP
from wcr_frontend.gui.wgui_record_table_widgets import \
    WguiRecordTableModel, WguiRecordTableWidget
from wcr_core.modules.system.db.wphdb_host_instance import *


class WpguiHostCardsProperties(WguiPropertiesWidgetP):
    """
    Properties widget for host cards.
    """

    def __init__(self, inst, pctx, parent=None):
        """
        :param inst: the instance object
        :type inst: WphdbHostInstance
        :param pctx: Project context
        :type pctx: WpmodSystemProjectContextP
        :param parent: Parent widget
        """
        super().__init__(parent)

        #: instance object
        self.__instance = inst
        #: Project context
        self.__pctx = pctx

        #: Cards table model
        self.__cardsM = _WpguiCardTableModel()
        #: Channels table model
        self.__channelsM = _WpguiChannelTableModel()

        #: Cards table view
        self.__cardsW = WguiRecordTableWidget(self.__cardsM, itemName='Card')
        #: Card channels table view
        self.__channelsW = WguiRecordTableWidget(
            self.__channelsM, itemName='Channel')

        self.setLayout(self._createLayout())
        self._initElements()
        self.updateFields()

    def _createLayout(self):
        """
        Create the layout.

        :return: Newly created layout
        :rtype: QLayout
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)

        layout.addWidget(QLabel('Cards on Host:'))
        layout.addWidget(self.__cardsW)

        layout.addSpacing(10)

        layout.addWidget(QLabel('Channels of selected Card:'))
        layout.addWidget(self.__channelsW)

        return layout

    def _initElements(self):
        """
        Initialize UI elements (run after layout creation, prior to initial
        field update).
        """
        self.__cardsW.currentRecordChanged.connect(
            self.__channelsM.setModelData)
        self.__channelsM.anyDataChanged.connect(
            self.__cardsM.externalDataChange)

    def updateFields(self):
        relm = self.__instance.rawElement
        cards = relm.getMeta(WPHDB_HOST_CARDS_KEY)
        self.__cardsM.setModelData(cards or {})

    def save(self):
        relm = self.__instance.rawElement
        relm.setMeta(self.__cardsM.getModelData(), WPHDB_HOST_CARDS_KEY)


class _WpguiCardTableModel(WguiRecordTableModel):
    """
    Property widget for a card.
    """

    #: Card fields
    _FIELDS = [
        ('Name', WPHDB_HOST_CARD_NAME_KEY),
        ('Type', WPHDB_HOST_CARD_TYPE_KEY)]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(_WpguiCardTableModel._FIELDS, parent)

    def _createEmptyEntry(self):
        rkey = wdbs_makeUUID()
        entry = {
            WPHDB_HOST_CARD_NAME_KEY: 'Card {}'.format(rkey),
            WPHDB_HOST_CARD_TYPE_KEY: '',
            WPHDB_HOST_CARD_CHANNELS_KEY: {}}
        return rkey, entry


class _WpguiChannelTableModel(WguiRecordTableModel):
    """
    Property widget for the channels of a card.
    """

    #: Channel fields
    _FIELDS = [
        ('Name', WPHDB_HOST_CHANNEL_NAME_KEY),
        ('Type', WPHDB_HOST_CHANNEL_TYPE_KEY)]

    def __init__(self, parent=None):
        """
        :param parent: Parent widget
        """
        super().__init__(self._FIELDS, WPHDB_HOST_CARD_CHANNELS_KEY, parent)

    def _createEmptyEntry(self):
        rkey = wdbs_makeUUID()
        entry = {
            WPHDB_HOST_CHANNEL_NAME_KEY: 'Channel {}'.format(rkey),
            WPHDB_HOST_CHANNEL_TYPE_KEY: ''}
        return rkey, entry
