# ----------------------------------------------------------------------------
# - whgui-scene-protocol-table                                               -
# - wild cruncher - cockpit : protocol table widget                          -
# ----------------------------------------------------------------------------
# - author    (c) 2018-2018 achraf mahti                                     -
# - copyright (c) 2017-2018 silkan / wild systems                            -
# ----------------------------------------------------------------------------

"""
.. module::       wvgui_scene_protocol_table
   :platform:     Unix, Windows
   :synopsis:     Wild Cruncher traffic route protocol table scene editor

.. moduleauthor:: Achraf Mahti <mahti@silkan.net>
"""
import re
from PyQt5 import QtCore
from functools import partial

from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget, QTableWidget, QPushButton,\
    QVBoxLayout, QTableWidgetItem, QHeaderView, QLineEdit
from silkan_common.scmn_utils import isHexa, isDec, isIpAdress, isMacAdress
from wcr_core.db.whdb_utils import whdb_getNextName
from wcr_core.modules.vlink.db.wvhdb_Vlink_instance import *
from wcr_core.modules.vlink.db.wvhdb_standard_operators import WVHDB_OPER, \
    WVHDB_VLINK_CODE
from wcr_core.modules.sys_map.plugin.wcplg_config_generator_params import *
from wcr_frontend.gui.wgui_scene_item_ids import *
from wcr_frontend.gui.wgui_scene_item import *
from wcr_frontend.modules.sys_map.gui.wcgui_instance_props_dialog import WCGUI_APP_PERIOD
from wcr_frontend import showErrorWindow


class WvguiSceneProtocolTable(QWidget):
    """
    The WvguiSceneProtocolTable class is the class responsible to manage
    the protocol table scene action in traffic route view
    that WguiGraphicsScene don't handle
    """
    HEADER = []
    PROTOCOL = None
    COMBO_PARAMS = (LOG_TR_AFDX_REDNT_MGNT_TAG, LOG_TR_AFDX_REDNT_VL_TAG,
                    LOG_TR_STORAGE_TAG, LOG_TR_ETH_ARP_TAG, LOG_TR_ETH_DETER_TAG,
                    LOG_TR_ETH_PROTOCOL_TAG, LOG_TR_AFDX_SN_CHECK_TAG,
                    LOG_TR_AFDX_PRIO_TAG, WVHDB_VLINK_TYPE_HYBRID_INPUT[0],
                    WVHDB_VLINK_TYPE_HYBRID_OUTPUT[0], LOG_TR_BAG_TAG)
    #: Signal emitting whenever an editor changes its name
    nameChanged = QtCore.pyqtSignal()
    signalMapper = QtCore.QSignalMapper()

    def __init__(self, scene, parent=None):
        super().__init__(parent)
        # the project database
        self.__scene = scene
        #: Scene items database
        self.sceneItems = {}
        #: table widget
        self.table = QTableWidget()
        #: Add entry button
        self.__addButton = QPushButton('Add')
        #: Delete entry button
        self.__deleteButton = QPushButton('Delete')
        # Flag to allow the class to return signal when modification
        self.triggerScene = False
        self._initElements()
        self.setLayout(self._createLayout())
        self.currentRow = 0
        self.updateTable()

    def getTableIndex(self):
        return self.__scene.orderedProtocolNames.index(self.PROTOCOL)

    def getScene(self):
        return self.__scene

    # Added a reimplementation of the resize event
    def resizeEvent(self, event):
        super(WvguiSceneProtocolTable, self).resizeEvent(event)
        tableSize = self.table.width()
        sideHeaderWidth = self.table.verticalHeader().width()
        tableSize -= sideHeaderWidth
        numberOfColumns = len(self.header)

        remainingWidth = tableSize % numberOfColumns
        for columnNum in range(numberOfColumns):
            if remainingWidth > 0:
                self.table.setColumnWidth(columnNum, int(tableSize / numberOfColumns) + 1)
                remainingWidth -= 1
            else:
                self.table.setColumnWidth(columnNum, int(tableSize / numberOfColumns))

    def getTitle(self):
        return self.PROTOCOL

    def getIcon(self):
        pixmap = QPixmap(100, 100)
        pixmap.fill(QColor(self.TABCOLOR))
        return pixmap

    def isICDProject(self):
        return False

    def activated(self):
        return None

    def getEditorIdent(self):
        return self.PROTOCOL

    def _initElements(self):
        self.header = ['Name']
        self.header.extend([prm[1] for prm in self.HEADER])
        self.header.append('uuid')
        self.table.setColumnCount(len(self.header))
        self.uuidCol = len(self.header)-1
        # uuid col is used for tracking Vl instance only so it is invisible
        self.table.setColumnHidden(self.uuidCol, True)
        self.table.setHorizontalHeaderLabels(self.header)
        header = self.table.horizontalHeader()
        header.setSectionResizeMode(QHeaderView.Stretch)
        for column in range(header.count()):
            width = header.sectionSize(column)
            header.setSectionResizeMode(column, QHeaderView.Interactive)
            header.resizeSection(column, width)
        header.setStretchLastSection(True)
        self.__addButton.clicked.connect(self._addEntry)
        self.__deleteButton.clicked.connect(self._deleteSelEntries)
        self.table.itemChanged.connect(self.itemChanged)

    def updateTable(self):
        for item in self.__scene.items():
            self.addRow(item.getDataObject())

    def _createLayout(self):
        layout = QVBoxLayout()
        layout.addWidget(self.table)
        Hlayout = QHBoxLayout()
        Hlayout.addWidget(self.__addButton)
        Hlayout.addWidget(self.__deleteButton)
        layout.addLayout(Hlayout)
        return layout

    def addRow(self, itemScene):
        item = itemScene.getDataObject()
        self.table.setRowCount(self.currentRow + 1)
        name = QTableWidgetItem(str(item.getName()) if
                                item.getName() is not None else '')
        vlType = item.getVLType()
        if vlType:
            self.table.setItem(self.currentRow, 0, name)
            for i, (name, value, key) in enumerate(item.getVLParameters(vlType)):
                self._createColumn(i, value, key)
            # add uuid as last row
            prmItem = QTableWidgetItem(str(item.getUUID()))
            self.table.setItem(self.currentRow, self.uuidCol, prmItem)
            self.currentRow += 1
            self.sceneItems[str(item.getUUID())] = itemScene
            tableIdx = self.getTableIndex()
            self.__scene.sceneItemAdded.emit(self.currentRow, tableIdx, self.triggerScene)

    def _createTrafficRoute(self):
        oper = WVHDB_OPER[WVHDB_VLINK_CODE]
        return self.__scene.createFunctionItem(oper)

    def _createColumn(self, i, value, key):
        if self.dataCheck(key, value):
            if key in self.COMBO_PARAMS:
                values = []
                if key == LOG_TR_STORAGE_TAG:
                    values = ('Queue', 'Sample')
                elif key in (LOG_TR_ETH_ARP_TAG, LOG_TR_AFDX_REDNT_MGNT_TAG,
                             LOG_TR_AFDX_REDNT_VL_TAG, LOG_TR_AFDX_SN_CHECK_TAG,
                             LOG_TR_ETH_DETER_TAG):
                    values = ('Yes', 'No')
                elif key == LOG_TR_ETH_PROTOCOL_TAG:
                    values = ('TCP', 'UDP')
                elif key == LOG_TR_AFDX_PRIO_TAG:
                    values = ('High', 'Low')
                elif key == WVHDB_VLINK_TYPE_HYBRID_INPUT[0]:
                    values = self.updateComboForHybrid()
                elif key == WVHDB_VLINK_TYPE_HYBRID_OUTPUT[0]:
                    values = self.updateComboForHybrid()
                elif key == LOG_TR_BAG_TAG:
                    values = WCGUI_APP_PERIOD
                prmItem = QComboBox()
                for name in sorted(values):
                    prmItem.addItem(name, name)
                idxCombo = prmItem.findText(value)
                prmItem.setCurrentIndex(idxCombo)
                func = self.itemChanged
                prmItem.currentIndexChanged.connect(partial(
                    func, (self.currentRow, i + 1)))

                self.table.setCellWidget(self.currentRow, i + 1, prmItem)

            else:
                prmItem = QTableWidgetItem(str(value)
                                           if value is not None else '')
                self.table.setItem(self.currentRow, i + 1, prmItem)

    # overide for Hybrid route
    def updateAllHybrid(self):
        pass

    def _addEntry(self):
        itemScene = self._createTrafficRoute()
        item = itemScene.getDataObject()
        name = item.getName()
        protName = name.replace('Traffic', self.PROTOCOL)
        newName = whdb_getNextName(protName, [sc.getDataObject()
                                              for sc in self.sceneItems.values()])
        item.setName(newName)
        item.setVLType(self.PROTOCOL)
        self.addRow(itemScene)

    def _deleteSelEntries(self):
        """
        Delete selected entries from the model
        """
        """:type: list[QModelIndex]"""
        selIdx = self.table.selectedIndexes()
        deleteRows = []
        for idx in selIdx:
            row = idx.row()
            # Get name of the TrafficRoute wherever it is clicked on the row
            uuid = idx.sibling(row, self.uuidCol).data()
            sceneItem = self.sceneItems.get(uuid)
            if sceneItem is None:
                continue
            deleteRows.append((row, sceneItem))
        uniqueRows = set(deleteRows)
        # iter the item in reverse remove them without position issue
        for row, sceneItem in sorted(uniqueRows,
                                     key=lambda row: row[0],
                                     reverse=True):
            self.__scene.deleteItem(sceneItem)
            uuid = sceneItem.getDataObject().getUUID()
            del self.sceneItems[str(uuid)]
            self.table.removeRow(row)
            self.currentRow -= 1
            self.table.selectRow(min(selIdx).row())
            tableIdx = self.getTableIndex()
            self.__scene.sceneItemRemoved.emit(self.currentRow, tableIdx, self.triggerScene)

    def deleteAll(self):
        tableIdx = self.getTableIndex()
        for row in range(0, self.table.rowCount()):
            self.table.removeRow(row)
            self.currentRow -= 1
            self.__scene.sceneItemRemoved.emit(self.currentRow, tableIdx, self.triggerScene)

    def itemChanged(self, selec):
        """
        Signal handler : when item is change update the traffic route
        parameter associated
        """
        if type(selec) == tuple:
            row, col = selec
        else:
            row = selec.row()
            col = selec.column()
        itemUuid = self.table.item(row, self.uuidCol)

        if itemUuid is None:
            return
        uuid = itemUuid.text()
        item = self.sceneItems.get(uuid)
        if not item:
            return
        vlInst = item.getDataObject()
        if type(selec) == tuple:
            data = self.table.cellWidget(row, col).currentText()
        else:
            data = selec.text()

        if col == 0:
            vlInst.setName(data)
        elif type(selec) == tuple:
            # don't check on combobox
            paramkey = self.HEADER[col - 1][0]
            vlInst.setVLParameter(self.PROTOCOL, paramkey, data)
        else:
            # sub 1 because we added Name col
            try:
                paramkey = self.HEADER[col-1][0]
            except IndexError:
                return
            isOk = self.dataCheck(paramkey, data)
            if isOk:
                vlInst.setVLParameter(self.PROTOCOL, paramkey, data)
            else:
                selec.setText(data)
        if self.triggerScene:
            self.__scene.sceneModified.emit()

    def dataCheck(self, paramkey, data):
        """
        Function that will check the data Type depending on the column
        :param paramkey: the name of the column
        :type paramkey: Str
        :param data: data to be checked
        :type data: Str
        :return: Bool
        """
        if not data:
            return True
        isOk = False
        if paramkey in (LOG_TR_ETH_IP_SRC_TAG, LOG_TR_ETH_IP_DEST_TAG):
            isOk, title, message = isIpAdress(data, returnMessages=True)
        elif paramkey in (LOG_TR_ETH_MAC_DEST_TAG, LOG_TR_ETH_MAC_SRC_TAG):
            isOk, title, message = isMacAdress(data, returnMessages=True)
        elif paramkey in self.COMBO_PARAMS:
            return True
        elif paramkey == LOG_TR_ARINC429_ID_TAG:
            res = re.match("((?:\d+;)|(?:0x[0-9A-F]+;))*(0x[0-9A-Fa-f]+|\d+)", data)
            rgroup = res.group()
            lres = rgroup.split(';')
            sres = set(lres)
            if data != rgroup:
                title = 'Data Type Error'
                message = 'Field "{}" must have a list of integer or hexadecimal separated by \';\' !'
            elif len(lres) != len(sres):
                title = 'Data Type Error'
                message = 'ids must be unique !'
            elif len(lres) > 256:
                title = 'Data Type Error'
                message = 'There can\'t be more than 256 ids, {} found !'.format(len(lres))
            else:
                isOk = True
        else:
            if isHexa(data) or isDec(data):
                isOk = True
            else:
                title = 'Data Type Error'
                message = 'Field "{}" must have integer or hexadecimal data!'
        if not isOk:
            showErrorWindow(title, message, paramkey)
        return isOk


class WvguiSceneAFDXTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_AFDX_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_AFDX
    TABCOLOR = "black"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneCetracTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_CETRAC_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_CETRAC
    TABCOLOR = "green"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneEthernetTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_ETNERNET_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_ETNERNET
    TABCOLOR = "blue"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneDiscreteTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_DISCRETE_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_DISCRETE
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiSceneFiberChannelTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_FIBER_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_FIBER
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)


class WvguiScene1553Table(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_1553_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_1553
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)

class WvguiSceneARINC825Table(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_ARINC825_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_ARINC825
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)

class WvguiSceneARINC429Table(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_ARINC429_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_ARINC429
    TABCOLOR = "orange"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)

class WvguiSceneHybridTable(WvguiSceneProtocolTable):
    HEADER = WVHDB_VLINK_TYPE_HYBRID_PARAMETERS
    PROTOCOL = WVHDB_VLINK_TYPE_HYBRID
    TABCOLOR = "red"

    def __init__(self, scene, parent):
        super().__init__(scene, parent)

    def updateAllHybrid(self):
        for rowId in range(self.table.rowCount()):
            prmItemIn = self.table.cellWidget(rowId, 1)
            self.updateComboForHybrid(prmItemIn)
            prmItemOut = self.table.cellWidget(rowId, 2)
            self.updateComboForHybrid(prmItemOut)

    def updateComboForHybrid(self, prmItem=None):
        values = []
        for sceneItem in self.getScene().getSceneItems():
            vl = sceneItem.getDataObject()
            if vl.getVLType() != WVHDB_VLINK_TYPE_HYBRID:
                values.append(vl.getName())
        if prmItem is not None:
            value = prmItem.currentText()
            prmItem.clear()
            for name in sorted(values):
                prmItem.addItem(name, name)
            idxCombo = prmItem.findText(value)
            prmItem.setCurrentIndex(idxCombo)
            prmItem.currentText()

        return values
