##############################################################################################################################
# 											Satusryprt util functions
##############################################################################################################################
import ctypes as ct
from PyCRC.CRC32 import CRC32
import configparser as cp

from constants import CETRAC_Names


def isFrameCetrac(pkt):
    """
        Check if the frame is a CetraC frame
    :param pkt: libpcap packet
    :return: cetracId or None
    """
    if ((pkt[0] == 0x03) and (pkt[1] == 0x00) and (pkt[2] == 0x00)
            and (pkt[3] == 0x01) and (pkt[4] == 0x00) and (pkt[5] == 0x00)):
        cetracId = ((pkt[30] << 8) & 0x0000FF00) + ((pkt[31] << 0) & 0x000000FF)
        return cetracId
    else:
        return None


def isMMResponse(pkt):
    """
        Check if the frame is a MMResponse frame
    :param pkt: libpcap packet
    :return: bool
    """
    if ((pkt[88] == 0xDD) and (pkt[89] == 0xDD) and (pkt[90] == 0xDD) and
            (pkt[91] == 0xDD) and (pkt[92] == 0xDD) and (pkt[93] == 0xDD) and
            (pkt[94] == 0xDD) and (pkt[95] == 0xDD)):
        return True
    else:
        return False


def computeCetracFrame(pkt):
    """
        return the device, port and data info from the cetrac frame
    :param pkt: libpcap packet
    :return: err_device, err_port, err_data
    """
    err_device = pkt[33]
    err_port = pkt[36]
    err_data = ((pkt[112] << 56) + (pkt[113] << 48) + (pkt[114] << 40) +
               (pkt[115] << 32) + (pkt[116] << 24) + (pkt[117] << 16) +
               (pkt[118] << 8) + pkt[119])
    return err_device, err_port, err_data


def computeMMResponse(pkt):
    """
        return the device, board, component and data info from the MMResponse frame
    :param pkt: libpcap packet
    :return: err_device, err_port, err_data
    """
    resp_data = [0]*4
    resp_data[0] = (pkt[100] << 24) + (pkt[101] << 16) + (pkt[102] << 8) + pkt[103]
    resp_data[1] = (pkt[108] << 24) + (pkt[109] << 16) + (pkt[110] << 8) + pkt[111]
    resp_data[2] = (pkt[116] << 24) + (pkt[117] << 16) + (pkt[118] << 8) + pkt[119]
    resp_data[3] = (pkt[124] << 24) + (pkt[125] << 16) + (pkt[126] << 8) + pkt[127]
    resp_equip = pkt[105]
    resp_board = pkt[113]
    resp_comp = pkt[121]
    return resp_data, resp_equip, resp_board, resp_comp


# -----------------------------------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------------------------------
def createCetraCMMread(Device, Adresse):
    """
	Send a MMread command for an address range of 4 words starting at Adresse in the selected Device
    :param Device: interface to send comand
    :return:
    """
    MMBuffer = [0] * 188

    #	CetraC header
    MMBuffer[0] = 0x03
    MMBuffer[3] = 0x01
    MMBuffer[12] = 0x08
    MMBuffer[30] = 0x05
    MMBuffer[31] = Device & 0xFF

    MMBuffer[44] = 0x08
    MMBuffer[46] = 0x04

    #	MMRead payload
    MMBuffer[88] = 0xCC
    MMBuffer[89] = 0xCC
    MMBuffer[90] = 0xCC
    MMBuffer[91] = 0xCC
    MMBuffer[92] = 0xCC
    MMBuffer[93] = 0xCC
    MMBuffer[94] = 0xCC
    MMBuffer[95] = 0xCC
    MMBuffer[96] = 0x00  # ID Reg
    MMBuffer[97] = 0x07  # ID Reg
    MMBuffer[100] = 0x04  # Port = 4 for servitude
    MMBuffer[104] = (Adresse & 0xFFFF) >> 8  # Read Address
    MMBuffer[105] = Adresse & 0xFF  # Read Address

    CRCBuffer = bytes(MMBuffer[32:156])
    crc = CRC32().calculate(CRCBuffer)

    MMBuffer[156] = crc & 0xFF  # CRC
    MMBuffer[157] = (crc & 0xFFFF) >> 8  # CRC
    MMBuffer[158] = (crc & 0xFFFFFF) >> 16  # CRC
    MMBuffer[159] = crc >> 24  # CRC

    #	CetraC trailer
    MMBuffer[160] = 0x00  # Prod stamp
    MMBuffer[161] = 0x01  # Prod stamp
    MMBuffer[162] = 0x00  # Packet size
    MMBuffer[163] = 0x80  # Packet size

    return (ct.c_ubyte * 188)(*MMBuffer)


# ------------------------------------------------------------------------------------------
#										GetMessages
#
#  Load the message table from the external message file
# ------------------------------------------------------------------------------------------
def readMessages(tabMessages, pathfilename=None):
    messages = cp.ConfigParser()
    if pathfilename:
        messages.read(pathfilename)
    else:
        messages.read("Statusryprt.msg")
    i = 0
    for PortType in CETRAC_Names:
        if PortType in messages:
            for index in messages.options(PortType):
                try:
                    flag = int(index)
                    tabMessages[i][flag] = messages[PortType][index]
                except:
                    flag = 0
        i += 1
    return


def getErrorMessage(tabMessages, portType, Err):
    MsgErr = ""
    for index in range(64):
        if Err & 0x01:
            MsgErr = tabMessages[portType][index]
            if MsgErr == "":
                MsgErr = "Unknowned error"
        Err = Err >> 1
    return MsgErr