import sys, os
import unittest
import ctypes as ct

import libpcap as pcap
from PyCRC.CRC32 import CRC32

tstdir = os.path.dirname(os.path.abspath(__file__))
srcdir = os.path.dirname(tstdir)
print('-----------')
print((__file__, srcdir, tstdir))
sys.path.append(srcdir)

from status_utils import isFrameCetrac, isMMResponse, computeCetracFrame, \
						 computeMMResponse, createCetraCMMread,\
                         getErrorMessage, readMessages

from constants import CETRAC_MAXTYPES

########################### TODO ####################################
# change the value depending on the machine
DEVICE = b'\\Device\\NPF_{5ED8EECE-E052-4F7B-86D9-3DE2C525A151}'
#####################################################################

TRAPID = 0x401


@pcap.pcap_handler
def test_is_frame_cetrac(readData, pkthdr, pkt):
    cetracId = isFrameCetrac(pkt)
    unittest.TestCase().assertEqual(cetracId, TRAPID)


@pcap.pcap_handler
def test_check_frame_cetrac_data(readData, pkthdr, pkt):
    err_device, err_port, err_data = computeCetracFrame(pkt)
    # device 2
    unittest.TestCase().assertEqual(err_device, 2)
    # comset 11 (port 6)
    unittest.TestCase().assertEqual(err_port, 11)

    if err_data != 1:
        # err code 0x80
        unittest.TestCase().assertEqual(err_data, 128)


@pcap.pcap_handler
def test_is_frame_MMResponse(readData, pkthdr, pkt):
    mmresp = isMMResponse(pkt)
    unittest.TestCase().assertEqual(mmresp, True)


@pcap.pcap_handler
def test_frame_MMResponse_data(readData, pkthdr, pkt):
    resp_data, resp_equip, resp_board, resp_comp = computeMMResponse(pkt)
    # device 3
    unittest.TestCase().assertEqual(resp_equip, 3)
    # card 1
    unittest.TestCase().assertEqual(resp_board, 1)
    # port PCI
    unittest.TestCase().assertEqual(resp_comp, 0)
    unittest.TestCase().assertEqual(resp_data, [241, 6, 0, 0])

@pcap.pcap_handler
def test_send_cetrac_MMRead(readData, pkthdr, pkt):
    device = 1
    adresse = 1
    MMBuffer = createCetraCMMread(1, 1)
    unittest.TestCase().assertEqual(MMBuffer[0], 0x03)
    unittest.TestCase().assertEqual(MMBuffer[3], 0x01)
    unittest.TestCase().assertEqual(MMBuffer[12], 0x08)
    unittest.TestCase().assertEqual(MMBuffer[30], 0x05)
    unittest.TestCase().assertEqual(MMBuffer[31], device & 0xFF)

    unittest.TestCase().assertEqual(MMBuffer[44], 0x08)
    unittest.TestCase().assertEqual(MMBuffer[46], 0x04)

    # MMRead payload
    unittest.TestCase().assertEqual(MMBuffer[88], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[89], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[90], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[91], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[92], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[93], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[94], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[95], 0xCC)
    unittest.TestCase().assertEqual(MMBuffer[96], 0x00)
    unittest.TestCase().assertEqual(MMBuffer[97], 0x07)
    unittest.TestCase().assertEqual(MMBuffer[100], 0x04)
    unittest.TestCase().assertEqual(MMBuffer[104], (adresse & 0xFFFF) >> 8)
    unittest.TestCase().assertEqual(MMBuffer[105], adresse & 0xFF)

    CRCBuffer = bytes(MMBuffer[32:156])
    crc = CRC32().calculate(CRCBuffer)

    unittest.TestCase().assertEqual(MMBuffer[156], crc & 0xFF)
    unittest.TestCase().assertEqual(MMBuffer[157], (crc & 0xFFFF) >> 8)
    unittest.TestCase().assertEqual(MMBuffer[158], (crc & 0xFFFFFF) >> 16)
    unittest.TestCase().assertEqual(MMBuffer[159], crc >> 24)

    # CetraC trailer
    unittest.TestCase().assertEqual(MMBuffer[160], 0x00)
    unittest.TestCase().assertEqual(MMBuffer[161], 0x01)
    unittest.TestCase().assertEqual(MMBuffer[162], 0x00)
    unittest.TestCase().assertEqual(MMBuffer[163], 0x80)


class TestWcorWorkspace(unittest.TestCase):
    def setUp(self):
        ebuf = ct.create_string_buffer(pcap.PCAP_ERRBUF_SIZE)
        # while 1:
        pathfile = os.path.join(b'data', b'Deconnexion_Link_1G.pcap')
        self.data = pcap.open_offline(ct.c_char_p(pathfile),
                               ebuf)
        if not self.data:
            print("{!s}", ebuf.value.decode("utf-8", "ignore"))
            exit()

    def testIsFrameCetrac(self):
        if pcap.loop(self.data, 0, test_is_frame_cetrac, None) < 0:
            print("pcap_loop() failed: ")
            exit()

    def testIsFrameMMResponse(self):
        ebuf = ct.create_string_buffer(pcap.PCAP_ERRBUF_SIZE)
        pathfile = os.path.join(b'data', b'Response.pcap')
        data = pcap.open_offline(ct.c_char_p(pathfile),
                                      ebuf)
        if pcap.loop(data, 0, test_is_frame_MMResponse, None) < 0:
            print("pcap_loop() failed: ")
            exit()

    def testCetracFrameData(self):
        if pcap.loop(self.data, 0, test_check_frame_cetrac_data, None) < 0:
            print("pcap_loop() failed: ")
            exit()

    def testMMResponseFrameData(self):
        ebuf = ct.create_string_buffer(pcap.PCAP_ERRBUF_SIZE)
        pathfile = os.path.join(b'data', b'Response.pcap')
        data = pcap.open_offline(ct.c_char_p(pathfile),
                                 ebuf)
        if pcap.loop(data, 0, test_frame_MMResponse_data, None) < 0:
            print("pcap_loop() failed: ")
            exit()

    def testSendCetracMMRead(self):
        if pcap.loop(self.data, 0, test_send_cetrac_MMRead, None) < 0:
            print("pcap_loop() failed: ")
            exit()

    def testErrMsg(self):
        tabMessages = [[""]*100 for _ in range(CETRAC_MAXTYPES)]
        pathfilename = os.path.join(srcdir, "Statusryprt.msg")
        readMessages(tabMessages, pathfilename)
        portType = 2 # 1G
        Err = 0x1
        errMsg = getErrorMessage(tabMessages, portType, Err)
        self.assertEqual(errMsg, "Link Loss")
        portType = 2  # 1G
        Err = 0x80
        errMsg = getErrorMessage(tabMessages, portType, Err)
        self.assertEqual(errMsg, "Transmission error (link down while transmitting a packet)")
        portType = 2  # 1G
        Err = 0x1300000000
        errMsg = getErrorMessage(tabMessages, portType, Err)
        self.assertEqual(errMsg, "Cycle / Event scheduler error detected")
        portType = 2  # 1G
        Err = 0x800000
        errMsg = getErrorMessage(tabMessages, portType, Err)
        self.assertEqual(errMsg, "Error in Object Header (Header CRC, Object not Valid, "
                                 "Buffer not Valid, local_ptr mismatch)")


if __name__ == '__main__':
    unittest.main()
