#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/idr.h>
#include <linux/version.h>

#include "cetracFonctions.h"

#define DRV_NAME	CETRAC_MODULE_NAME

#if defined(CONFIG_32BIT)
MODULE_DESCRIPTION("Cetrac 100 kernel 32 bits");
#elif defined(CONFIG_64BIT)
MODULE_DESCRIPTION("Cetrac 100 kernel 64 bits");
#else
MODULE_DESCRIPTION("Cetrac 100 kernel undefined");
#endif

MODULE_AUTHOR("RAMIA Yves (yves.ramia@silkan.com)");

// MODULE_LICENSE("Proprietary");
MODULE_LICENSE("GPL");

MODULE_VERSION("Silkan 3.4.5 / kernel 3.11.0");

/* liste des device geres par le driver */
static struct pci_device_id cetrac_pci_table[] = {
    { PCI_DEVICE(CETRAC_VENDOR_ID, CETRAC_DEVICE_ID), 0, 0, CETRAC_BOARD_V1},
    {0,}
};
MODULE_DEVICE_TABLE(pci,cetrac_pci_table);

/* definition de la structure pci_driver d'cetrac */
static struct pci_driver cetrac_pci_driver =
{
    .name     = DRV_NAME,
    .id_table = cetrac_pci_table,
    .probe    = cetrac_pci_probe,
    .remove   = cetrac_pci_remove,
};

/*
 *  Variables globales du driver
 */

/* variable contenant le nombre de cartes presentes et correctement initialisees */
static DEFINE_MUTEX(cetrac_minors_lock);
static struct idr cetrac_minors;

static struct class *cetrac_class;
static dev_t cetrac_devt;

static void cetrac_dev_release(struct device *dev)
{
	struct cetrac_board *brd = container_of(dev, struct cetrac_board, filedev);

        mutex_lock(&cetrac_minors_lock);
        idr_remove(&cetrac_minors, brd->dev_num);
        mutex_unlock(&cetrac_minors_lock);
        kfree(brd);
}

int cetrac_board_register(struct cetrac_board *brd)
{
	int ret;

	if (!brd)
		return -EINVAL;

	ret = cdev_add(&brd->cdev, brd->filedev.devt, 1);
	if (ret) {
		dev_err(&brd->filedev,
			"cdev_add(%d:%d) failed (err %d)\n",
			MAJOR(brd->filedev.devt),
                        MINOR(brd->filedev.devt), ret);
		return ret;
	}

	ret = device_add(&brd->filedev);
	if (ret) {
		dev_err(&brd->filedev,
			"device_add(%d:%d) failed (err %d)\n",
			MAJOR(brd->filedev.devt),
                        MINOR(brd->filedev.devt), ret);
		cdev_del(&brd->cdev);
		return ret;
	}

	mutex_lock(&cetrac_minors_lock);
	idr_replace(&cetrac_minors, brd, brd->dev_num);
	mutex_unlock(&cetrac_minors_lock);

	return ret;
}

void cetrac_board_unregister(struct cetrac_board *brd)
{
	if (!brd)
		return;

	cdev_del(&brd->cdev);
	device_del(&brd->filedev);

	mutex_lock(&cetrac_minors_lock);
	idr_replace(&cetrac_minors, NULL, brd->dev_num);
	mutex_unlock(&cetrac_minors_lock);
	put_device(&brd->filedev);
}

struct cetrac_board *cetrac_board_alloc(struct device *dev, const struct file_operations *cetrac_fops)
{
	int boardnr, ret;
	struct cetrac_board *brd;

	brd = kzalloc(sizeof(struct cetrac_board), GFP_KERNEL);
	if (!brd)
		return ERR_PTR(-ENOMEM);

	mutex_lock(&cetrac_minors_lock);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
	do {
		ret = idr_pre_get(&cetrac_minors, GFP_KERNEL);
		if (!ret) {
			mutex_unlock(&cetrac_minors_lock);
			dev_err(dev, "idr_pre_get() failed\n");
			kfree(brd);
			return ERR_PTR(-EAGAIN);
		}
		ret = idr_get_new(&cetrac_minors, NULL, &boardnr);
	} while (ret == -EAGAIN);
#else
	ret = idr_alloc(&cetrac_minors, NULL, 0, 0, GFP_KERNEL);
	boardnr = ret;
#endif
	if (ret < 0) {
		mutex_unlock(&cetrac_minors_lock);
		dev_err(dev, "Failed to get a device number\n");
		kfree(brd);
		return ERR_PTR(ret);
	}
	brd->dev_num = boardnr;
	mutex_unlock(&cetrac_minors_lock);

	device_initialize(&brd->filedev);
	brd->filedev.class = cetrac_class;
	brd->filedev.release = cetrac_dev_release;
	brd->filedev.parent = dev;
	brd->filedev.devt = MKDEV(MAJOR(cetrac_devt), brd->dev_num);

	ret = dev_set_name(&brd->filedev, "%s%d", DRV_NAME, brd->dev_num);
        if (ret) {
		dev_err(dev, "Failed to set dev name\n");
		put_device(&brd->filedev);
		return ERR_PTR(ret);
	}

	cdev_init(&brd->cdev, cetrac_fops);
	brd->cdev.owner = THIS_MODULE;
        brd->cdev.kobj.parent = &brd->filedev.kobj;

	return brd;
}

/*
 * Fonction d'initialisation du module.
 * Elle est appelee e l'installation du driver ("insmod ledriver")
 *
 */
static int __init cetrac_init_module(void)
{
    int     result;

    pr_debug("%s\n", __func__);

    idr_init(&cetrac_minors);

    /* creation d'une classe de driver */
    cetrac_class = class_create(THIS_MODULE, DRV_NAME);
    if (IS_ERR(cetrac_class)) {
        pr_err("erreur de creation cetrac_class\n");
	result = PTR_ERR(cetrac_class);
	goto error_idr;
    }

    /* allocation d'un major et d'une plage de minor pour le device */
    result = alloc_chrdev_region(&cetrac_devt, 0, CETRAC_MAX_INSTANCES, DRV_NAME);
    if (result < 0) {
        pr_err("echec de alloc_chrdev_region\n");
	goto error_class;
    }

    pr_info("%s: major: %d, base minor: %d\n",
		DRV_NAME,
		MAJOR(cetrac_devt), MINOR(cetrac_devt));
    pr_info("%s: driver version : %d.%d.%d\n",
	    DRV_NAME, V_PCIDRIVER, V_MAJORDRIVER, V_MINORDRIVER);

    result =  pci_register_driver(&cetrac_pci_driver);
    if (result < 0) {
	pr_err("pci register failed\n");
	goto error_chrdev;
    }

    return 0;

error_chrdev:
	unregister_chrdev_region(cetrac_devt, CETRAC_MAX_INSTANCES);
error_class:
	class_destroy(cetrac_class);
error_idr:
	idr_destroy(&cetrac_minors);
	return result;
}

/*
 * Fonction de de retrait du module.
 * Elle est appelee au dechargement du module ( "rmmod ledriver")
 *
 */
static void cetrac_cleanup_module(void)
{
	pci_unregister_driver(&cetrac_pci_driver);
	unregister_chrdev_region(cetrac_devt, CETRAC_MAX_INSTANCES);
	class_destroy(cetrac_class);
	idr_destroy(&cetrac_minors);
}


module_init(cetrac_init_module);
module_exit(cetrac_cleanup_module);
