#ifndef __CETRACDRIVER_H__
#define __CETRACDRIVER_H__

#include <linux/module.h>       /* pour l'ecriture d'un module */
#include <linux/init.h>         /* pour module_init,module_exit */
#include <linux/fs.h>           /* pour kobj dans cdev.h */
#include <linux/cdev.h>         /* pour struct cdev */
#include <linux/device.h>
#include <linux/interrupt.h>    /* pour tasklet_struct */
#include <linux/sched.h>        /* pour TASK_INTERRUPTIBLE */

#include "cetracTypes.h"
#include "cetracIoctl.h"
#include "cetracLists.h"
#include "cetracHardwareErrorCodes.h"
#include "cetracGeneric.h"

/* Version du driver ( 3.4.5) */
#define V_PCIDRIVER     3
#define V_MAJORDRIVER   4
#define V_MINORDRIVER   5
#define V_CETRACDRIVER   ( (V_PCIDRIVER << 24) + (V_MAJORDRIVER<<16) + (V_MINORDRIVER<<8) )

/* definition du nom du module driver */
#define CETRAC_MODULE_NAME "cetrac"

/* definition du nombre maximum de priorite sur les objets */
#define CETRAC_MAX_PRIORITY 16

/* definition du masque pour recuperer la priorite de l'objet */
#define CETRAC_PRIORITY_MASK (0xF0000000)

/* decalage de bits pour obtenir la priorite */
#define CETRAC_PRIORITY_SHIFT 28

#define CETRAC_GET_OBJ_PRIORITY(idt_obj) ((idt_obj&CETRAC_PRIORITY_MASK)>>CETRAC_PRIORITY_SHIFT)


/************ definition des etats de la carte ******************/

#define CETRAC_BOARD_MEMORY_GRANTED 1

#define CETRAC_DEVICE_ENABLED (CETRAC_BOARD_MEMORY_GRANTED+1)

#define CETRAC_IO_MEMORY_GRANTED (CETRAC_BOARD_MEMORY_GRANTED+2)

#define CETRAC_MEMORY_MAPPED (CETRAC_BOARD_MEMORY_GRANTED+3)

#define CETRAC_HW_INITIALIZED (CETRAC_BOARD_MEMORY_GRANTED+4)

#define CETRAC_FULLY_INITIALISED (CETRAC_BOARD_MEMORY_GRANTED+5)


/***************** definition des bits utiles ****************/

/* bit indiquant si la carte est en cours de traitement d'une requete */
#define CETRAC_BOARD_IN_PROGRESS_BIT 31

/**************** fin definition des bits utiles ******************/

/***************** definition des etats du driver **********************/

#define CETRAC_DRIVER_CLASS_REGISTERED 1

#define CETRAC_DRIVER_MAJOR_ALLOCATED (CETRAC_DRIVER_CLASS_REGISTERED+1)

#define CETRAC_DRIVER_CDEV_REGISTERED (CETRAC_DRIVER_CLASS_REGISTERED+2)

#define CETRAC_DRIVER_PCI_REGISTERED (CETRAC_DRIVER_CLASS_REGISTERED+3)

/***************** fin definition des etats du driver ********************/

/* Constantes generales */
#define CETRAC_DATABUF_SIZE      8192    /* Taille maximale d'un fenetre (allocation du buffer DMA) */
#define CETRAC_NB_PRIORITIES     16      /* Nombre de priorites gerees */
#define CETRAC_REQUEST_SIZE      4       /* Nombre de mots de 32bits formant un requete */
#define CETRAC_REQUEST_TIMEOUT   200     /* TimeOut des attentes dans le driver (200 ms) */

/* Offsets dans le BAR de contrele */
#define CETRAC_OFFSET_RESET      0       /* W : reset de la carte */
#define CETRAC_OFFSET_MASTADD    4       /* R/W : Adresse physique de la memoire du buffer DMA */
#define CETRAC_OFFSET_INT        8       /* R : Acquitement de l'IT - W : Deblocage pour l'IT suivante */
#define CETRAC_OFFSET_STATUS     12       /* R : Statut de la carte */
#define CETRAC_OFFSET_MASTCTL    16       /* W : Desactivation du master PCI (bit0 - Read / bit1 - Write), 1 = desactive */
#define CETRAC_OFFSET_EVENT      24       /* W : Acquittement de l'IT */

/* Types des evenement generes par la carte */
#define CETRAC_TYPE_RETURN       0       /* Type "retour de requete" */
#define CETRAC_TYPE_EVENT        1       /* Type "evenement" */

/* Etats du Driver */
#define CETRAC_STATE_NORMAL              0x0                             /* Etat normal */
#define CETRAC_STATE_HDRW_NOT_INIT       CETRAC_DRIVER_HRDW_NOT_INIT      /* La carte n'est pas en etat de repondre au */
                                                                        /* demarrage du driver (statut de la carte = 0) */
#define CETRAC_STATE_HDRW_NO_RESPONSE    CETRAC_DRIVER_HRDW_NO_RESPONSE   /* La carte n'a pas repondu lors d'un acces */
#define CETRAC_STATE_UNSTABLE            CETRAC_DRIVER_INTERNAL_ERROR     /* Une erreur interne au driver s'est produite */
                                                                        /* rendant le driver instable */

struct cetrac_board {
	int		dev_num;                    /* /dev/cetracX */
	struct device	filedev;
	struct cdev	cdev;
};

/******** definition de la structure representant une carte CETRAC ************/

typedef struct _DEVICE_EXTENSION
{
  struct tasklet_struct tasklet_cetrac;              /* tasklet utilise lors de la prise en compte de l'interruption */
  struct pci_dev        *dev;                       /* pointeur sur la structure pci representant la carte */
  cetracS32              compteur_utilisation;       /* compteur d'utilisation de la carte */

  struct cetrac_board	*brd;
  /* Mappage PCI */
  void __iomem         *Control;                     /* BAR0 : Contrele de la carte (voir offsets ci-dessus) */
  void __iomem         *Events;                      /* BAR1 : Adresse de lecture des evenements et retours de requetes */
  void __iomem         *Requests;                    /* BAR2 : Adresse d'ecriture des requetes */

  /* DMA */
  cetracU32             *VirtualAddress;              /* Buffer Master DMA : Adresse virtuelle */
  dma_addr_t            PhysicalAddress;             /* Buffer Master DMA : Adresse physique */

  /* Systeme */
  cetracU32             DriverState;                  /* Etat courant du driver */

  /* Variables liees aux acces RW */
  int                   bCurrentRequest;                        /* Requete en cours */
  int                   bCurrentRequestWaiting;                 /* Attente de retour en cours (ou e venir) */
  /* int                   bCurrentReset;                        Reset arrive pendant l'attente de retour */
  spinlock_t            RequestCS;                              /* Section critiques d'acces e ces variables */

  cetracU32              RequestQueue[CETRAC_NB_PRIORITIES];      /* Tableau du nombre de requetes en attente par priorite */

  wait_queue_head_t     RequestEvents[CETRAC_NB_PRIORITIES];     /* Tableau des evenements de synchronisation */
                                                                /* des acces (1 par priorite) */
  cetracU32              flagRequestEvents[CETRAC_NB_PRIORITIES]; /* tableau des status associes */
  cetracU32              countWaitRequest[CETRAC_NB_PRIORITIES];  /* tableau compteurs de condition d'attente) */
  cetracU32              countTrigRequest[CETRAC_NB_PRIORITIES];  /* tableau compteurs de condition de reveil */

  wait_queue_head_t     CurrentRequestEvent;                    /* Evenement de synchronisation de l'acces courant */
  cetracU32              statusCurrentRequestEvent;              /* status associe */

  EVENT                 CurrentRequestReturn;                   /* Retour de l'acces courant */

  /* Variables liees aux evenements */
  struct list_head      WaitingAnchor;              /* Ancre de la liste des evenements en attente */
  spinlock_t            WaitingCS;                  /* Section critique de la liste des evenements en attente */
  struct list_head      KnownListAnchor;            /* Ancre de la liste des listes d'attentes connues */
  cetrac_semaphore       KnownListCS;                /* Section critique de la liste des listes d'attentes connues */

  /* pour debug */

  /* nothing ... */

} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

/*
**  Variables globales du driver
*/



typedef  struct echange_s
{
    int     cmd;
    char    *txt;
} echange_t;

#endif /* fin de __CETRAC_H__ */















