// #include <linux/devfs_fs_kernel.h> /* pour devfs_XXX */
#include <linux/pci.h>
#include <linux/dma-mapping.h>      // for pci_alloc_consistent()
#include "cetracFonctions.h"
#include "cetracDriver.h"
#include <linux/version.h>

/*
** Initialisation logicielle du driver.
** C'est du specifique Cetrac sauf la derniere instruction qui lie
** la stucture du driver au systeme.
*/
static inline void cetrac_init_sw(DEVICE_EXTENSION *Current_Board, struct pci_dev *p_dev)
{
    cetracU32 i;

    /* lie la structure pci_dev e la structure interne du driver */
    Current_Board -> dev = p_dev;

    /* raz de la structure tasklet_cetrac */
    memset(&Current_Board->tasklet_cetrac,0,sizeof(struct tasklet_struct));

    /* Initialisation de la section critiques des acces */
    spin_lock_init(&Current_Board->RequestCS);

    /* Initialisation de l'evenement de synchronisation de l'acces courant */
    Current_Board->statusCurrentRequestEvent = 0;
    init_waitqueue_head(&Current_Board->CurrentRequestEvent);

    /* Initialisation de la section critiques des acces list */
    spin_lock_init(&Current_Board->WaitingCS);

    /* Initialisation de la section critiques des acces list */
    sema_init(&Current_Board->KnownListCS, 1);

    /* Initialisation du tableau des evenements de synchronisation des acces (1 par priorite) */
    for(i=0; i<CETRAC_NB_PRIORITIES; i++)
    {
        init_waitqueue_head(&Current_Board->RequestEvents[i]);

        Current_Board->flagRequestEvents[i] = 0x5555;
        Current_Board->countWaitRequest[i]  = 0;
        Current_Board->countTrigRequest[i]  = 0;
    }

    /* Initialisation : pas de requete en cours */
    Current_Board->bCurrentRequest = FALSE;

    INIT_LIST_HEAD(&Current_Board->KnownListAnchor);
    INIT_LIST_HEAD(&Current_Board->WaitingAnchor);

    /* lie la structure interne du driver e la structure pci de la carte */
    pci_set_drvdata(p_dev, Current_Board);

}

/*
**  Initialisation materielle.
*/
static int cetrac_init_hw(DEVICE_EXTENSION *Current_Board)
{
    struct pci_dev *pdev;
    cetracS32 result;
    u32 status;

    pdev = Current_Board->dev;

    /* derniere etape d'initialisation logicielle */
    result = pci_enable_device(pdev);
    if (result) {
        dev_err(&pdev->dev, "Unable to initialise PCI_DEVICE\n");
        return result;
    }

    /* La carte sera utilisee en mode bus-MASTER */
    pci_set_master(pdev);

    /* Set DMA mask */
    if (sizeof(dma_addr_t) > sizeof(u32) &&
	!pci_set_dma_mask(pdev, DMA_BIT_MASK(64))) {
	dev_info(&pdev->dev, "Using 64 bit DMA\n");
    } else {
        result = pci_set_dma_mask(pdev, DMA_BIT_MASK(32));
        if (result) {
            dev_err(&pdev->dev, "Unable to get 32 bit DMA\n");
            goto err_pci_enable;
        }
	dev_info(&pdev->dev, "Using 32 bit DMA\n");
    }

    result = pci_request_regions(pdev, CETRAC_MODULE_NAME);
    if (result < 0) {
        dev_err(&pdev->dev, "Unable to allocate io regions\n");
	goto err_pci_enable;
    }

    /* mapping BAR 0 : control */
    Current_Board->Control = ioremap_nocache(pci_resource_start(pdev, 0),
                                 pci_resource_len(pdev, 0) );
    if (!Current_Board->Control) {
        dev_err(&pdev->dev, "ioremap Bar 0 (control) failed..\n");
	goto err_request_regions;
    }

    /* mapping BAR 1 : events */
    Current_Board->Events = ioremap_nocache(pci_resource_start(pdev, 1),
                                    pci_resource_len(pdev, 1) );
    if (!Current_Board->Events) {
        dev_err(&pdev->dev, "ioremap Bar 1 (events) failed..\n");
	goto err_map_0;
    }

    /* mapping BAR 3 : requests */
    Current_Board->Requests = ioremap_nocache(
                                    pci_resource_start(pdev, 2),
                                    pci_resource_len(pdev, 2) );
    if (!Current_Board->Requests) {
        dev_err(&pdev->dev, "ioremap Bar 2 (requests) failed..\n");
	goto err_map_1;
    }

    /* Mapping du master buffer */
    Current_Board->VirtualAddress = pci_alloc_consistent(   pdev,
                                                            CETRAC_DATABUF_SIZE,
                                                            &Current_Board->PhysicalAddress);
    if(!Current_Board->VirtualAddress) {
        dev_err(&pdev->dev, "Can't allocate master buffer\n");
        result = -ENOMEM;
	goto err_map_2;
    }

    /* Ecriture de l'adresse du buffer DMA dans la carte */
    cetrac_ctl_write32(Current_Board, CETRAC_OFFSET_MASTADD, Current_Board->PhysicalAddress);

    /* Si la carte est prete, changement d'etat du driver (statut de la carte = CETRAC_STATUS_ON) */
    status = cetrac_ctl_read32(Current_Board, CETRAC_OFFSET_STATUS);
    dev_dbg(&pdev->dev, "Status : %08x\n", status);

    if ((status & CETRAC_STATE_ON) == CETRAC_STATE_ON) {
        Current_Board->DriverState = CETRAC_STATE_NORMAL;
    } else {
        Current_Board->DriverState = CETRAC_STATE_HDRW_NOT_INIT;
    }

    /* Reset de la FIFO d'evenements */
    cetrac_evt_write32(Current_Board, 0, 0);

    /* allocation de l'irq pour le module */
    result = request_irq(Current_Board->dev->irq,cetrac_interrupt,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
                           IRQF_SHARED,
#else
                           IRQF_DISABLED | IRQF_SHARED,
#endif
                           CETRAC_MODULE_NAME, Current_Board);
    if ( result ) {
	dev_err(&pdev->dev, "Failed to get irq\n");
	goto err_alloc_consistent;
    }

      /* vider les evenements */
      //PLANTE ???? *(cetracU32*)Current_Board->Events = 0;

      /* initialisation de la tasklet liee e la carte */
      tasklet_init(&Current_Board->tasklet_cetrac,
                     tasklet_cetrac,(unsigned long)Current_Board);

    return 0;

err_alloc_consistent:
    pci_free_consistent(pdev, CETRAC_DATABUF_SIZE,
	Current_Board->VirtualAddress, Current_Board->PhysicalAddress);
err_map_2:
    iounmap(Current_Board->Requests);
err_map_1:
    iounmap(Current_Board->Events);
err_map_0:
    iounmap(Current_Board->Control);
err_request_regions:
    pci_release_regions(pdev);
err_pci_enable:
    pci_disable_device(pdev);
    return result;
}

static void cetrac_deinit_hw(DEVICE_EXTENSION *Current_Board)
{
    struct pci_dev *pdev;

    pdev = Current_Board->dev;

    free_irq(pdev->irq, Current_Board);
    cetrac_ctl_write32(Current_Board, CETRAC_OFFSET_MASTADD, 0);

    pci_free_consistent(pdev, CETRAC_DATABUF_SIZE,
                        Current_Board->VirtualAddress,
                        Current_Board->PhysicalAddress);
    iounmap(Current_Board->Requests);
    iounmap(Current_Board->Events);
    iounmap(Current_Board->Control);
    pci_release_regions(pdev);
    pci_disable_device(pdev);
}

/* liste des operations implementes par le driver de type caractere */
static const struct file_operations cetrac_fops = {
    .open               = cetrac_open,
    .release            = cetrac_release,
    .unlocked_ioctl     = cetrac_ioctl,
    .flush              = cetrac_flush,
    .owner              = THIS_MODULE,
};


/*
**  Fonction lancee e la detection de la carte
*/
int cetrac_pci_probe(struct pci_dev *p_dev,const struct pci_device_id *id)
{
    int result;
    DEVICE_EXTENSION	*priv;
    struct cetrac_board  *brd;

    priv = kzalloc(sizeof(DEVICE_EXTENSION), GFP_KERNEL);
    if (!priv) {
	dev_err(&p_dev->dev, "kzalloc() failed\n");
	return -ENOMEM;
    }

    brd = cetrac_board_alloc(&p_dev->dev, &cetrac_fops);
    if (IS_ERR(brd)) {
       kfree(priv);
       return PTR_ERR(brd);
    }

    priv->brd = brd;

    cetrac_init_sw(priv, p_dev);

    result = cetrac_init_hw(priv);
    if (result < 0)
	goto err_board;

    /*
     *  Enregistrement du device
     */
    result = cetrac_board_register(priv->brd);
    if (result < 0)
	goto err_init_hw;

    dev_info(&p_dev->dev, "irq: %d instance: %d\n", p_dev->irq, brd->dev_num);
    return 0;

err_init_hw:
    cetrac_deinit_hw(priv);
err_board:
    put_device(&brd->filedev);
    kfree(priv);
    return result;
}


/*
**  Fonction lancee au dechargement du driver
*/
void cetrac_pci_remove(struct pci_dev *p_dev)
{
    PDEVICE_EXTENSION Current_Board;
    struct cetrac_board *brd;

    Current_Board = pci_get_drvdata(p_dev);
    brd = Current_Board->brd;

    cetrac_board_unregister(brd);

    cetrac_deinit_hw(Current_Board);
    kfree(Current_Board);
}

