// ------------------------------------------------------------------------------------------------
/// @file cut_cetrac.h
/// @author PiV
/// @date 2015-2016
/// @brief The functions to forge cetrac frames
///
/// The cetrac frame can be forged in two ways called full frame and inner frame. Basically the
/// full frame encapsulate the inner frame. Thus two diffrent headers and two two different tails
/// are defined here.
/// The full frame tail and header are automatically generated by the cetrac PCI cards whereas
/// the inner frame tail and header are software dependant.
// ------------------------------------------------------------------------------------------------
// - CETRAC Project - cetrac util                                                                 -
// - copyright (c) 2016-2016 Silkan                                                               -
// ------------------------------------------------------------------------------------------------

#ifndef CUT_FRAME_CONVERT_H
#define CUT_FRAME_CONVERT_H

#include "cut_types.h"
#include "cut_include.h"
#include "cut_ethernet.h"

/// the cet header1 size
#define CET_FFH_SIZE 18
/// the cet header1 size
#define CET_IFH_SIZE 64
/// the cet tail size
#define CET_IFT_SIZE 32
/// the cet tail size
#define CET_FFT_SIZE 28

/// A type to list the cetrac frame types
typedef enum {
    CETRAC_MM_WRITE,
    CETRAC_MM_RESET,
    CETRAC_MM_READ,
    CETRAC_MM_HARD_RESET,
    CETRAC_MM_AUTH_REQ,
    CETRAC_MM_FIRST_FRAME
} t_cut_frt;

/// The struct to hold the cetrac header of a full frame
typedef struct {
    /// reserved field
    uint8_t p_rsv[2];
    /// the system id
    uint8_t d_syst;
    /// the equipement id
    uint8_t d_equp;
    /// the card id
    uint8_t d_card;
    /// the component id
    uint8_t d_comp;
    /// the entity id
    uint8_t d_enty;
    /// the cetrac b type
    uint8_t d_typb;
    /// the cetrac s type
    uint8_t d_styp;
    /// the cetrac num b
    uint8_t d_numb;
    /// B C D fields
    uint8_t p_bcd[4];
    /// the cetrc object ID
    uint8_t p_idt[4];
} t_cut_cet_ffh;

/// The struct to hold the the cetrac header of a inner frame
typedef struct {
    /// reserved field
    uint8_t p_rsv[56];
    /// the format type
    uint8_t p_frtp[8];
} t_cut_cet_ifh;

/// The struct to hold the cetrac tail of a inner frame
typedef struct {
    /// reserved field
    uint8_t p_rsv[28];
    /// the cetrac objet CRC
    uint8_t p_ccrc[4];
} t_cut_cet_ift;

/// The struct to hold the cetrac tail of a full frame
typedef struct {
    /// the stamp
    uint8_t p_stamp[2];
    /// the message size
    uint8_t p_msize[2];
    /// the VD, Freq, State, Ext values
    uint8_t p_vfse[4];
    /// the system time
    uint8_t p_syst[8];
    /// the status
    uint8_t p_stat[8];
    /// reserved field
    uint8_t p_rsv[4];
} t_cut_cet_fft;

/// The struct to hold a cetrac inner frame. The tail and CRC are integrated
/// into the payload to manage variable frame size.
typedef struct {
    /// the cetrac inner frame header
    t_cut_cet_ifh d_cifh;
    /// the payload and the CRC and the cetrac tails
    uint8_t p_pld_crc[PLD_ETH_SIZE + CRC_SIZE - CET_FFH_SIZE - CET_IFH_SIZE];
} t_cut_cet_iframe;

/// The struct to hold a cetrac full frame. The tails and CRC are integrated
/// into the payload to manage variable frame size.
typedef struct {
    /// the ethernet header
    t_cut_eth_hdr d_ehdr;
    /// the cetrac full frame header
    t_cut_cet_ffh d_cffh;
    /// the cetrac inner frame header
    t_cut_cet_ifh d_cifh;
    /// the payload and the CRC and the cetrac tails
    uint8_t p_pld_crc[PLD_ETH_SIZE + CRC_SIZE - CET_FFH_SIZE - CET_IFH_SIZE];
} t_cut_cet_fframe;

/// @brief return a cetrac inner frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
int cut_cet_inner_get_net_rich(uint8_t *dst, size_t dsz, t_cut_frt typ,
                               uint8_t *const src, size_t ssz,
                               uint16_t frameNum, uint16_t frameNumTotal);

/// @brief return a cetrac full frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
/// @param id  the cetrac object id
/// @param stp the frame stamp
/// @param crc the crc flag, set to true to compute te CRC
int cut_cet_full_get_net_rich(uint8_t *dst, size_t dsz, t_cut_frt typ,
                              uint8_t *const src, size_t ssz,
                              uint32_t id, uint16_t stp, int crc,
                              uint16_t frameNum, uint16_t frameNumTotal);

/// @brief return a cetrac inner frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
int cut_cet_inner_get_net_rich2(uint8_t *dst, size_t dsz, t_cut_frt typ,
                                uint8_t *const src, size_t ssz,
                                uint16_t frameNum, uint16_t frameNumTotal, int isSesBlock);

int cut_cet_inner_get_net_rich3(uint8_t *dst, size_t dsz, t_cut_frt typ,
                                uint8_t *const src, size_t ssz,
                                uint16_t frameNum, uint16_t frameNumTotal, int isEncrypted,
                                uint8_t *tag);

/// @brief return a cetrac full frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
/// @param id  the cetrac object id
/// @param stp the frame stamp
/// @param crc the crc flag, set to true to compute te CRC
int cut_cet_full_get_net_rich2(uint8_t *dst, size_t dsz, t_cut_frt typ,
                               uint8_t *const src, size_t ssz,
                               uint32_t id, uint16_t stp, int crc,
                               uint16_t frameNum, uint16_t frameNumTotal, int isSesBlock);

int cut_cet_full_get_net_rich3(uint8_t *dst, size_t dsz, t_cut_frt typ,
                               uint8_t *const src, size_t ssz,
                               uint32_t id, uint16_t stp, int crc,
                               uint16_t frameNum, uint16_t frameNumTotal, int isEncrypted,
                               uint8_t *tag);

/// @brief return a cetrac inner frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
int cut_cet_inner_get_net(uint8_t *dst, size_t dsz, t_cut_frt typ,
                          uint8_t *const src, size_t ssz);

/// @brief return a cetrac full frame in network order
/// @return frame size in bytes or -1 in case of error
/// @param dst the buffer to get host the data
/// @param dsz the destination size
/// @param typ the cetrac frame type
/// @param src the source buffer (in network order)
/// @param ssz the source size
/// @param id  the cetrac object id
/// @param stp the frame stamp
/// @param crc the crc flag, set to true to compute te CRC
int cut_cet_full_get_net(uint8_t *dst, size_t dsz, t_cut_frt typ,
                         uint8_t *const src, size_t ssz,
                         uint32_t id, uint16_t stp, int crc);

#endif
