// ------------------------------------------------------------------------------------------------
/// @file cut_ethernet.h
/// @author PiV
/// @date 2015-2016
/// @brief The function to forge ethernet frames
// ------------------------------------------------------------------------------------------------
// - CETRAC Project - cetrac util                                                                 -
// - copyright (c) 2016-2016 Silkan                                                               -
// ------------------------------------------------------------------------------------------------

#ifndef CUT_ETHERNET_H
#define CUT_ETHERNET_H

#include "cut_types.h"
#include "cut_include.h"

/// the CRC size
#define CRC_SIZE     4
/// the Ethernet header size
#define ETH_HDR_SIZE  14
/// the IP header size
#define IP_HDR_SIZE  20
/// the udp header size
#define UDP_HDR_SIZE 8
/// the ethernet payload SIZE
#define PLD_ETH_SIZE 1500

// the default vesion and lenght value for an IPV4 frame
static const uint8_t DEF_VER_HLEN = 0x45;

/// the struct to hold an ethernet header
typedef struct {
  /// the mac destination address
  uint8_t p_mac_dst[6];
  /// the mac source address
  uint8_t p_mac_src[6];
  /// the ethernet type
  uint8_t p_eth_typ[2];
} t_cut_eth_hdr;

/// the struct to hold an ip header
typedef struct {
  /// the ip version and length
  uint8_t d_ver_hlen;
  /// the ip type of service
  uint8_t d_tos;
  /// the ip frame length
  uint8_t p_length[2];
  /// the fragment number of a same paquet
  uint8_t p_id[2];
  /// the flags and offset of the fragment in the paquet
  uint8_t p_flags_offset[2];
  /// the time to live in number of hops
  uint8_t d_ttl;
  /// the protocol identifier
  uint8_t d_protocol;
  /// the ip checksum
  uint8_t p_checksum[2];
  /// the ip source address
  uint8_t p_ip_src[4];
  /// the ip destination address
  uint8_t p_ip_dst[4];
} t_cut_ip_hdr;

/// the struct to hold an udp header
typedef struct {
  /// the udp source port
  uint8_t p_port_src[2];
  /// the udp destination port
  uint8_t p_port_dst[2];
  /// the frame length
  uint8_t p_length[2];
  /// the udp checksum
  uint8_t p_crcUDP[2];
} t_cut_udp_hdr;

/// the struct to hold an ip ethernet frame. The CRC is integrated into the payload to
/// manage variable frame size.
typedef struct {
  /// the ethernet header
  t_cut_eth_hdr d_eth;
  /// the ip header
  t_cut_ip_hdr  d_ip;
  /// the payload and CRC
  uint8_t p_payload_crc[PLD_ETH_SIZE + CRC_SIZE - IP_HDR_SIZE];
} t_cut_ip_frame;

/// the struct to hold an udp ethernet frame. The CRC is integrated into the payload to
/// manage variable frame size.
typedef struct {
  /// the ethernet header
  t_cut_eth_hdr d_eth;
  /// the ip header
  t_cut_ip_hdr  d_ip;
  /// the udp header
  t_cut_udp_hdr d_udp;
  /// the payload and CRC
  uint8_t p_payload_crc[PLD_ETH_SIZE + CRC_SIZE - UDP_HDR_SIZE - IP_HDR_SIZE];
} t_cut_udp_frame;

/// the struct to hold an ethernet frame. The CRC is integrated with the payload to
/// manage variable frame size.
typedef struct {
  /// the ethernet header
  t_cut_eth_hdr d_eth;
  /// the payload and CRC
  uint8_t p_payload_crc[PLD_ETH_SIZE + CRC_SIZE];
} t_cut_eth_frame;

/// @brief CRC computation
/// @param[in] dst the pointer to store the CRC in network order
/// @param[in] src the data to compute the CRC
/// @param[in] size the source size
/// @param[in] type the CRC type (8, 16, 32)
void cut_crc_gen(uint8_t* dst, uint8_t* src, unsigned int size, unsigned int type);

/// @brief compute the ip checksum, the field checksum value should be set to O
/// @param ptr the ip header to be used for checksum computation
void cut_ip_checksum (t_cut_ip_hdr* ptr);

/// @brief print a frame into a string
/// @return 0 on success
/// @param dst the string to print in
/// @param dsz the destination size
/// @param src the source to be printed
/// @param ssz the source size
int cut_eth_to_str (char* dst, size_t dsz, uint8_t* const src, size_t ssz);

#endif
