//-------------------------------------------------------------------------------------------------
/// @file mitracErrorCodes.h
/// @author Alexandre.L
/// @date 2015-2017
/// @brief MITRAC errors management.
// ------------------------------------------------------------------------------------------------
// - CETRAC Project - MITRAC
// - Copyright (c) - SILKAN
//-------------------------------------------------------------------------------------------------

#ifndef MITRAC_ERROR_CODES_H_
#define MITRAC_ERROR_CODES_H_

#include <stdio.h>

/// Define MITRAC errors.
typedef enum {
  /// No error.
  MITRAC_NO_ERROR,
  /// Table's index not found.
  MITRAC_INDEX_NOT_FOUND,
  /// Incorrect payload size.
  MITRAC_INCORRECT_SIZE_PAYLOAD,
  /// Incorrect IP frame size.
  MITRAC_INCORRECT_SIZE_IP_FRAME,
  /// MEF's state unknown.
  MITRAC_UNKNOWN_STATE,
  /// CETRAC reading error - no frame available
  MITRAC_NO_FRAME,
  /// CETRAC AFDX reading error - missed sequence number
  MITRAC_AFDX_SN_ERROR,
  /// CETRAC reading error.
  MITRAC_ERROR_CETRAC_READ,
  /// EMPTY FIFO error after reading several data
  MITRAC_EMPTY_FIFO,
  /// CETRAC writing error.
  MITRAC_ERROR_CETRAC_WRITE,
  /// CETRAC swap writing buffer
  MITRAC_ERROR_SWAP_WRITE,
  /// CETRAC swap reading buffer
  MITRAC_ERROR_SWAP_READ,
  /// CETRAC initialization error.
  MITRAC_ERROR_CETRAC_INIT,
  /// CETRAC reset error.
  MITRAC_ERROR_CETRAC_RESET,
  /// CETRAC event error.
  MITRAC_ERROR_CETRAC_EVENT,
  /// Order already executed.
  MITRAC_NO_NEW_ORDER,
  /// MEF in operational mode.
  MITRAC_ORDER_OPE,
  /// MITRAC invalid argument
  MITRAC_INVALID_ARG,
  /// MITRAC memory error.
  MITRAC_MEMORY_ERROR,
  /// MITRAC OPEN SETUP LIB error
  MITRAC_ERROR_OPEN_SETUP_LIB,
  /// MITRAC INIT SETUP LIB error
  MITRAC_ERROR_INIT_SETUP_LIB,
  /// MITRAC CLOSE SETUP LIB error
  MITRAC_ERROR_CLOSE_SETUP_LIB,
  /// MITRAC OPEN APP LIB error
  MITRAC_ERROR_OPEN_APP_LIB,
  /// MITRAC INIT APP LIB error
  MITRAC_ERROR_INIT_APP_LIB,
  /// MITRAC CLOSE APP LIB error
  MITRAC_ERROR_CLOSE_APP_LIB,
  /// MITRAC ERROR MONITORING FILE
  MITRAC_ERROR_MONITOR_FILE,
  /// MITRAC PERIOD ERROR
  MITRAC_ERROR_PERIOD,
  /// MITRAC STOP I/O CARD ERROR
  MITRAC_ERROR_IO_CARD,
  /// MITRAC ID CETRAC ERROR
  MITRAC_TIMEOUT_CETRAC_TIMER,
  /// overflow CetraC Timer
  MITRAC_CLK_OVERFLOW,
  /// interrupt Cetrac Timer
  MITRAC_INTERRUPT_CETRAC_TIMER,
  /// ERROR Cetrac TIMER
  MITRAC_TIMER_CETRAC_ERROR,
  /// MITRAC the application missed at least one data
  MITRAC_ERROR_DATA_LOST,
  /// MITRAC there is no new message for the application
  MITRAC_ERROR_NO_MSG,
  /// The fifo of a message is full
  MITRAC_FIFO_FULL
} t_error;

/// Define MITRAC return codes.
#define MITRAC_TRUE       1
#define MITRAC_FALSE      0

#define MITRAC_ERROR     -1
#define MITRAC_SUCCESS    0

/// @brief Set error function.
/// @param[in] p the pointer to be checked and set.
#define MITRAC_SET_ERROR(p, code) if((p) != NULL) *(p) = code

/// @brief Check pointer function (no error updated).
/// @return MITRAC_TRUE if the pointer is NULL otherwise MITRAC_FALSE.
/// @param[in] p The pointer to be checked.
#define MITRAC_CHK_NULL(p) ((p == NULL) ? MITRAC_TRUE : MITRAC_FALSE)

/// @brief Check pointer function.
/// @return MITRAC_TRUE if the pointer is NULL otherwise MITRAC_FALSE.
/// @param[in]  p The pointer to be checked.
/// @param[out] e The error to be updated.
#define MITRAC_CHK_NULL_PTR(p, e) ((p == NULL) ?                                               \
                                   ((e != NULL) ? *e = MITRAC_INVALID_ARG : 0, MITRAC_TRUE ) : \
                                   MITRAC_FALSE)

/// @brief Check memory function.
/// @return MITRAC_TRUE if allocated memory is NULL otherwise MITRAC_FALSE.
/// @param[in]  p The pointer to be checked.
/// @param[out] e The error to be updated.
#define MITRAC_CHK_NULL_MEM(p, e) ((p == NULL) ?                                                \
                                   ((e != NULL) ? *e = MITRAC_MEMORY_ERROR : 0, MITRAC_TRUE ) : \
                                   MITRAC_FALSE)

/// @brief Display error function.
/// @param[in] d_err  The error to display
void mitracDispErr (t_error d_err);

/// Drivers & library errors mitracErrorCodes
// Definition des provenances des erreurs
#define CETRAC_MASK_TYPE_ERROR    0xF0000000          // Masque du pr�fixe d'erreur
#define CETRAC_BOARD_ERROR        0x00000000          // Pr�fixe : Erreur g�n�r�e par la carte
#define CETRAC_DRIVER_ERROR       0x10000000          // Pr�fixe : Erreur g�n�r�e par le driver

// Definition des erreurs generees par la carte
#define CETRAC_BOARD_UNKNOWN_OBJECT    (CETRAC_BOARD_ERROR | 0x1)  // L'objet concern� par l'acc�s n'est pas connu de la carte
#define CETRAC_BOARD_LIFETIME_OVER     (CETRAC_BOARD_ERROR | 0x2)  // L'horizon d'action de l'objet est d�pass�
#define CETRAC_BOARD_CANCEL            (CETRAC_BOARD_ERROR | 0x4)  // La carte n'est pas en mesure de traiter la requ�te, la requ�te a �t� annul�e
#define CETRAC_BOARD_OVER_SIZE         (CETRAC_BOARD_ERROR | 0x8)  // La taille demand�e lors de l'acc�s est sup�rieure � celle definie dans les propri�t�s
#define CETRAC_BOARD_UNKNOWN_METHOD    (CETRAC_BOARD_ERROR | 0x10)  // La m�thode d'acces de l'objet est inconnue
#define CETRAC_BOARD_UNINITILIZED_DATA (CETRAC_BOARD_ERROR | 0x20)  // Acc�s en lecture � une donn�e non initialis�e (non �crite par le r�seau ou l'application)

// Definition des erreurs generees par le driver
#define CETRAC_DRIVER_HRDW_NOT_INIT     (CETRAC_DRIVER_ERROR | 0x1)  // La carte n'a pas d�marr�
#define CETRAC_DRIVER_HRDW_NO_RESPONSE  (CETRAC_DRIVER_ERROR | 0x2)  // La carte ne r�pond pas lors d'un acc�s
#define CETRAC_DRIVER_WAIT_ABANDONNED   (CETRAC_DRIVER_ERROR | 0x3)  // Une fonction d'attente a �t� interrompue par une APC
#define CETRAC_DRIVER_TIMEOUT           (CETRAC_DRIVER_ERROR | 0x4)  // Une requ�te empil�e a atteint le temps maximal d'attente
#define CETRAC_DRIVER_ACCESS_VIOLATION  (CETRAC_DRIVER_ERROR | 0x5)  // L'acc�s au pointeur de donn�e fournit par l'utilisateur a lev� un execption
#define CETRAC_DRIVER_INTERNAL_ERROR    (CETRAC_DRIVER_ERROR | 0x6)  // Une fonction interne au driver a �chou�
#define CETRAC_DRIVER_QUERY_REMOVE      (CETRAC_DRIVER_ERROR | 0x7)  // Une demande de supression a �t� faite au driver par le PNP
#define CETRAC_DRIVER_CANT_ALLOCATE_MEM (CETRAC_DRIVER_ERROR | 0x8)  // Une allocation a �chou�e dans le driver
#define CETRAC_DRIVER_ALREADY_IN_LIST   (CETRAC_DRIVER_ERROR | 0x9)  // L'identifiant est d�j� dans la liste d'�v�nements
#define CETRAC_DRIVER_NOT_IN_LIST       (CETRAC_DRIVER_ERROR | 0x10)  // L'identifiant n'est pas dans la liste d'�v�nements
#define CETRAC_DRIVER_NO_RECORD         (CETRAC_DRIVER_ERROR | 0x11)  // Aucun identifiant n'est enregistr� sur la liste
#define CETRAC_DRIVER_FIFO_FULL         (CETRAC_DRIVER_ERROR | 0x12)  // La Fifo de la liste d'�v�nement est pleine, au moins un �v�nement a �t� perdu
#define CETRAC_DRIVER_UNKOWN_LIST       (CETRAC_DRIVER_ERROR | 0x13)  // La liste d'�v�nement n'est pas connue
#define CETRAC_DRIVER_SEM_ERROR         (CETRAC_DRIVER_ERROR | 0x14)  // La liste d'�v�nement n'est pas connue
#define CETRAC_DRIVER_OVER_SIZE         (CETRAC_DRIVER_ERROR | 0x15)  // Buffer trop petit pour recuperer toute la trame

#endif // MITRAC_ERROR_CODES_H_11
